/*
 * Copyright 2015 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.unitesk.requality.core;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStreamWriter;
import java.io.UnsupportedEncodingException;
import java.io.Writer;
import java.net.MalformedURLException;
import java.net.URI;
import java.net.URISyntaxException;
import java.nio.file.Path;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Scanner;
import java.util.Set;
import java.util.UUID;
import org.apache.commons.httpclient.URIException;
import org.apache.commons.httpclient.util.URIUtil;
import org.json.JSONException;
import com.unitesk.requality.core.attribute.Attribute;
import com.unitesk.requality.core.attribute.AttributeDefinition;

public class FSStorage extends AbstractTreeStorage {
  private UUID root = null;
  private static final String ROOT_NODE_FILENAME = "root.json";
  private Map<UUID, File> nodeToFile = new HashMap<>();
  private Map<UUID,Map<String,AttributeDefinition>> activeAttributes =
      new HashMap<>();
  private String path;

  public FSStorage(String path) {
    if (!path.endsWith("/")) {
      path = path + "/";
    }
    this.path = path;
    cs = new CacheStorage(this,"");
  }

  private UUID getUUId(File file) {
    try {
      return NodeMapper.getUUId(readFile(file));
    } catch (JSONException e) {
      e.printStackTrace();
      return null;
    }
  }

  public File getFile(UUID uuid) {
    return nodeToFile.get(uuid);
  }

  private String readFile(File file) {
    if (file == null) return null;
    Scanner s = null;
    try {
      StringBuilder jsonsb = new StringBuilder();
      s = new Scanner(new FileInputStream(file), "UTF-8");
      while (s.hasNextLine()) {
        jsonsb.append(s.nextLine().trim());
      }
      return jsonsb.toString();
    } catch (IOException e) {
      e.printStackTrace();
    }
    finally{
      if(s!=null)
        s.close();
    }
    return null;
  }

  @Override
  public UUID getRootNode() {
    if (root != null) return root;
    File file = new File(URI.create(path + "/" + ROOT_NODE_FILENAME));
    if (file.exists()) {
      UUID uuid = getUUId(file);
      nodeToFile.put(uuid, file);
      root = uuid;
      return uuid;
    }
    return null;
  }

  public File getResourceFolder(UUID uuid) {
    File fl = getFile(uuid);
    if (fl == null || fl.getParent() == null) return null;
    String id = getNodeId(uuid);
    File resFolder = new File(fl.getParent() + File.separator + id + "_resources");
    return resFolder;
  }

  @Override
  public String getNodeId(UUID uuid) {
    if (uuid.equals(getRootNode())) {
      return "/";
    }
    if (getFile(uuid) == null) {
      return null; // node not found
    }
    String jext = "json";
    String fileName = getFile(uuid).getName();
    return fileName.substring(0, fileName.length() - jext.length() - 1);
  }

  @Override
  public UUID getParent(UUID uuid) {
    if (getFile(uuid) == null) {
      return null; // node not found
    }
    return getParent(getFile(uuid));
  }

  public UUID getParent(File file) {
    if (file.getAbsoluteFile().equals(path)) {
      return null;
    }

    String folderName = file.getParentFile().getName();
    String parentId = folderName.substring(0, folderName.length());

    File parentFolder = file.getParentFile().getParentFile();
    File parentResource =
        new File(parentFolder.getAbsoluteFile() + File.separator + parentId + ".json");

    if (!parentResource.exists()) {
      return null;
    }
    return getUUId(parentResource);
  }

  @Override
  public Collection<UUID> getChildren(UUID uuid) {
    Collection<UUID> result = new HashSet<>();
    if (getFile(uuid) == null) {
      return result; // node not found
    }

    String nodeId = getNodeId(uuid);
    if (nodeId.equals("/")) {
      nodeId = "root";
    }
    File nodeFolder = getFile(uuid).getParentFile();
    File childrenFolder = new File(nodeFolder.getAbsoluteFile() + File.separator + nodeId);

    if (!childrenFolder.exists()) {
      return result;
    }

    for (File res : childrenFolder.listFiles()) {
      if (res.isFile() && res.getName().endsWith(".json")) {
        UUID childUUID = getUUId(res);
        result.add(getUUId(res));
        nodeToFile.put(childUUID, res);
      }
    }
    return result;
  }

  @Override
  public void createStorageNode(UUID parentUUId, UUID uuid, String id, Map<String, AttributeDefinition> attributes) {
    try {
      File file;
      if (parentUUId == null) {
        file = new File(URI.create(path + ROOT_NODE_FILENAME));
      } else {
        String parentId = getNodeId(parentUUId);
        if (parentId.equals("/")) {
          parentId = "root";
        }
        File childFolder =
            new File(
                getFile(parentUUId).getParentFile().getAbsoluteFile() + File.separator + parentId);

        if (!childFolder.exists()) {
          childFolder.mkdirs();
        }
        file = new File(childFolder.getAbsolutePath() + File.separator + id + ".json");
      }

      if (getDeltaTarget() != null)
        getDeltaTarget().beforeCreated(uuid, parentUUId, attributes, true);
      activeAttributes.put(uuid, new HashMap<>(attributes));
      if (!file.getParentFile().exists()) {
        file.getParentFile().mkdirs();
      }
      if (!file.exists()) {
        file.createNewFile();
      }
      FileOutputStream out = new FileOutputStream(file);
      Writer w = new OutputStreamWriter(out, "UTF-8");

      String jsonString = NodeMapper.toJSON(uuid, attributes,true).toString(4);
      w.append(jsonString);
      w.flush();
      w.close();

      // update model
      nodeToFile.put(uuid, file);
      getDeltaTarget().created(uuid, true);

    } catch (UnsupportedEncodingException e) {
      e.printStackTrace();
    } catch (JSONException e) {
      e.printStackTrace();
    } catch (IOException e) {
      e.printStackTrace();
    }
  }

  @Override
  public void remove(UUID uuid) {
    // TODO: implement me
    String id = getNodeId(uuid);
    File res = getFile(uuid);
    activeAttributes.remove(uuid);
    TreeDB.clearAllCaches(null, uuid);
  //  new File(res.getParentFile() + id);
    res.delete();
  }

  @Override
  public void changeNodeId(UUID node, String id) {
    // TODO Auto-generated method stub

  }

  @Override
  public Map<String, AttributeDefinition> getStorageAttributes(UUID uuid) {
    try {
      if(activeAttributes.containsKey(uuid))
        return activeAttributes.get(uuid);
      Map<String,AttributeDefinition>val=NodeMapper.getAttributes(uuid, readFile(getFile(uuid)));
      activeAttributes.put(uuid, val);
      return val;
    } catch (JSONException e) {
      e.printStackTrace();
      return null;
    }
  }

  @Override
  public void saveStorageAttributes(UUID node, ChangesSet<Attribute> changed) {
    try {
//      getDeltaTarget().changed(node, changed, attributes, Moment.BEFORE, true);
      //Map<String, Attribute> old = oldattributes;//getAttributes(node);
      OutputStreamWriter out = new OutputStreamWriter(new FileOutputStream(getFile(node)), "UTF-8");
      Map<String, AttributeDefinition> oldMap = getAttributes(node);
      Map<String, Attribute> newA = changed.getNewMap();
      for(Entry<String, Attribute> en:newA.entrySet()){
        Attribute a = en.getValue();
        if(a == null)
          oldMap.remove(en.getKey());
        else
          oldMap.put(en.getKey(),en.getValue().getDefinition());
      }
      // for(String k:attributes.keySet())
      activeAttributes.put(node, oldMap);
      String jsonStrring = NodeMapper.toJSON(node, oldMap, true).toString(4);
      out.append(jsonStrring);
      out.close();

//      Map<String,Version> nva = new HashMap<String,Version>();
//      for(String s:changed.getRemoved())
//        nva.put(s, null);
//      for(String s:changed.getAdded())
//        nva.put(s, attributes.get(s).getVersion());
//      for(String s:changed.getChanged())
//        nva.put(s, attributes.get(s).getVersion());
//      saveVersions(node, changed, nva);
//      getDeltaTarget().changed(node, changed,attributes, Moment.AFTER, true);
    } catch (UnsupportedEncodingException e) {
      e.printStackTrace();
    } catch (FileNotFoundException e) {
      e.printStackTrace();
    } catch (IOException e) {
      e.printStackTrace();
    } catch (JSONException e) {
      e.printStackTrace();
    }
  }

  @Override
  public void moveNodeTo(UUID node, UUID parent) {
    // TODO Auto-generated method stub

  }

  @Override
  public boolean canLoad(UUID uuid) {
    File f = nodeToFile.get(uuid);
    if (f != null) {
      return f.exists();
    }
    // TODO Check json
    return false;
  }

  @Override
  public URI getNodeResourcePath(UUID uuId) {
    File f = getResourceFolder(uuId);
    if (f != null) return f.toURI();
    return null;
  }

  @Override
  public InputStream getResourceContents(UUID uuid, String resName) {
    // TODO: stream reuse check
    try {
      URI path = getNodeResourcePath(uuid);
      if (path == null) return null;
      return new FileInputStream(
          new File(new URI(path.toString() + "/" + URIUtil.encodeQuery(resName))));
    } catch (FileNotFoundException e) {
      e.printStackTrace();
    } catch (URISyntaxException e) {
      // TODO Auto-generated catch block
      e.printStackTrace();
    } catch (URIException e) { // TODO Auto-generated catch block
      e.printStackTrace();
    }
    return null;
  }

  @Override
  public void setResourceContents(UUID uuid, String resName, InputStream in, Object data) {
    try {
      if (getDeltaTarget() != null)
        getDeltaTarget().resourceChanged(uuid, resName, Moment.BEFORE, true);
      URI path = getNodeResourcePath(uuid);
      if (path == null) return;
      File fol = new File(path);
      if (!fol.exists()) fol.mkdirs();
      FileOutputStream out = new FileOutputStream(fol + File.separator + resName);
      byte[] buffer = new byte[1024];
      int len;
      while ((len = in.read(buffer)) != -1) {
        out.write(buffer, 0, len);
      }
      out.close();
      if (getDeltaTarget() != null)
        getDeltaTarget().resourceChanged(uuid, resName, Moment.AFTER, true);
    } catch (IOException e) {
      e.printStackTrace();
    }
  }

  @Override
  public boolean isResourceExists(UUID uuid, String resName) {
    URI path = getNodeResourcePath(uuid);
    if (path == null) return false;
    try {
      return new File(new URI(path.toString() + "/" + URIUtil.encodeQuery(resName))).exists();
    } catch (URISyntaxException e) {
      // TODO Auto-generated catch block
      e.printStackTrace();
    } catch (URIException e) { // TODO Auto-generated catch block
      e.printStackTrace();
    }
    return false;
  }

  @Override
  public void deleteResource(UUID uuid, String resName) {
    URI path = getNodeResourcePath(uuid);
    if (path == null) return;
    try {
      new File(new URI(path.toURL() + "/" + URIUtil.encodeQuery(resName))).delete();
    } catch (MalformedURLException e) {
      // TODO Auto-generated catch block
      e.printStackTrace();
    } catch (URISyntaxException e) {
      // TODO Auto-generated catch block
      e.printStackTrace();
    } catch (URIException e) { // TODO Auto-generated catch block
      e.printStackTrace();
    }
  }

  private void recursiveResourceList(File folder, String path, Set<String> result) {
    if (folder.exists())
      for (File f : folder.listFiles()) {
        // use unix path separator
        String curPath = path + "/" + f.getName();
        if (f.isDirectory()) {
          recursiveResourceList(f, curPath, result);
        } else {
          result.add(curPath);
        }
      }
  }

  @Override
  public Set<String> listResources(UUID uuid) {
    Set<String> result = new HashSet<>();
    File f = getResourceFolder(uuid);
    if (f != null) recursiveResourceList(f, "", result);
    return result;
  }

  @Override
  public String getName() {
    return new File(URI.create(path)).getName();
  }

  @Override
  public void createStorage(String name) {
    new File(URI.create(path)).getParentFile().mkdirs();
  }

  @Override
  public void deleteStorage() {
    // TODO Implement deleteStorage in FSStorage
    throw new UnsupportedOperationException("not implemented");
  }

  @Override
  public ITreeStorage getBaseStorage() {
    // TODO Auto-generated method stub
    return null;
  }

  @Override
  public File getResource(UUID uuid, String resName) {
    URI path = getNodeResourcePath(uuid);
    if (path == null) return null;
    try {
      return new File(new URI(path.toString() + "/" + URIUtil.encodeQuery(resName)));
    } catch (URISyntaxException e) {
      // TODO Auto-generated catch block
      e.printStackTrace();
    } catch (URIException e) { // TODO Auto-generated catch block
      e.printStackTrace();
    }
    return null;
  }
  @Override
  public void clearCachedCache(UUID uid) {
    super.clearCachedCache(uid);
    activeAttributes.remove(uid);
  }
  @Override
  public void clearCaches() {
    super.clearCaches();
    activeAttributes.clear();
  }

  @Override
  public Path getLocation() {
    return new File(URI.create(path)).toPath();
  }
}
