/*
 * Copyright 2015 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.unitesk.requality.core;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.net.URI;
import java.nio.file.Path;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedHashMap;
import java.util.Map;
import java.util.Set;
import java.util.UUID;
import com.unitesk.requality.core.attribute.Attribute;
import com.unitesk.requality.core.attribute.AttributeDefinition;
import com.unitesk.requality.core.attribute.AttributeStorage;

public class MemoryStorage extends AbstractTreeStorage {
  private UUID root = null;
  private Map<UUID, String> toId = new HashMap<>();
  private Map<UUID, Collection<UUID>> toChildren = new HashMap<>();
  private Map<UUID, Map<String, AttributeDefinition>> toAttributes =
      new HashMap<>();
  private Map<UUID, HashMap<String, ByteArrayOutputStream>> toResourсes =
      new LinkedHashMap<>();

  public MemoryStorage() {
    root = UUID.randomUUID();
    toId.put(root, "/");
    toChildren.put(root, Collections.emptyList());
    toAttributes.put(root, AttributeStorage.formTypeMap(root, TreeNode.ATTR_TYPE));
    init(root);
  }

  public MemoryStorage(UUID root) {
    init(root);
  }
  private void init(UUID root){
    this.root = root;
    cs = new CacheStorage(this,"");
  }

  public void emulateDeltaCreate(
      UUID parentUUId, UUID uuid, String type, String id, Map<String, AttributeDefinition> attributes) {
    getDeltaTarget().beforeCreated(uuid, parentUUId, attributes, false);
    toId.put(uuid, id);
    toChildren.put(uuid, new HashSet<UUID>());
    toAttributes.put(uuid, attributes);
    if (parentUUId == null) {
      root = uuid;
    } else {
      toChildren.get(parentUUId).add(uuid);
    }
    getDeltaTarget().created(uuid, false);
  }

  public void emulateDeltaDelete(UUID uuid) {
    getDeltaTarget().deleted(uuid, Moment.BEFORE, false);
    toId.remove(uuid);
    toAttributes.remove(uuid);
    Collection<UUID> childrens = toChildren.remove(uuid);
    for (UUID puuid : toChildren.keySet()) {
      if (toChildren.get(puuid).contains(uuid)) {
        toChildren.get(puuid).remove(uuid);
      }
    }
    for (UUID cuuid : childrens) {
      emulateDeltaDelete(cuuid);
    }
    getDeltaTarget().deleted(uuid, Moment.AFTER, false);
  }

  public void emulateDeltaMove(UUID uuid, UUID newParent) {
    UUID oldParent = null;
    for (UUID puuid : toChildren.keySet()) {
      if (toChildren.get(puuid).contains(uuid)) {
        oldParent = puuid;
      }
    }
    assert oldParent != null;
    getDeltaTarget().moved(uuid, oldParent, newParent, Moment.BEFORE, false);
    toChildren.get(oldParent).remove(uuid);
    toChildren.get(newParent).add(uuid);
    getDeltaTarget().moved(uuid, oldParent, newParent, Moment.AFTER, false);
  }

  public void emulateDeltaChanged(UUID uuid, ChangesSet<Attribute> what, Map<String, AttributeDefinition> attributes) {
//    getDeltaTarget().changed(uuid, what, attributes, Moment.BEFORE, false);
    toAttributes.put(uuid, attributes);
//    getDeltaTarget().changed(uuid, what, attributes, Moment.AFTER, false);
  }

  @Override
  public UUID getRootNode() {
    return root;
  }

  @Override
  public String getNodeId(UUID uuid) {
    return toId.get(uuid);
  }

  @Override
  public void createStorageNode(UUID parentUUId, UUID uuid, String id, Map<String, AttributeDefinition> attrs) {
    if (getDeltaTarget() != null) getDeltaTarget().beforeCreated(uuid, parentUUId, attrs, true);
    toId.put(uuid, id);
    if (!toChildren.containsKey(uuid)) toChildren.put(uuid, new HashSet<UUID>());
    if (attrs == null) {
      attrs = new HashMap<>();
    }
    toAttributes.put(uuid, new HashMap<>(attrs));
    if (parentUUId == null) {
      root = uuid;
    } else {
      toChildren.get(parentUUId).add(uuid);
    }
    getDeltaTarget().created(uuid, true);
  }

  @Override
  public void remove(UUID uuid) {
    getDeltaTarget().deleted(uuid, Moment.BEFORE, true);
    toId.remove(uuid);
    toAttributes.remove(uuid);
    Collection<UUID> toDelete = toChildren.get(uuid);
    if (toDelete != null) {
      toDelete = new ArrayList<>(toChildren.get(uuid));
      for (UUID duuid : toDelete) {
        remove(duuid);
      }
    }
    // remove from parent
    for (UUID puid : toChildren.keySet()) {
      if (toChildren.get(puid).contains(uuid)) {
        toChildren.get(puid).remove(uuid);
      }
    }
    toChildren.remove(uuid);
    getDeltaTarget().deleted(uuid, Moment.AFTER, true);
  }

  @Override
  public Map<String, AttributeDefinition> getStorageAttributes(UUID uuid) {
    Map<String, AttributeDefinition> m = toAttributes.get(uuid);
    return m;
  }

  @Override
  public Collection<UUID> getChildren(UUID uuid) {
    if (toChildren.get(uuid) == null) {
      return null;
    }
    return Collections.unmodifiableCollection(toChildren.get(uuid));
  }

  @Override
  public UUID getParent(UUID uuid) {
    for (UUID i : toChildren.keySet()) {
      if (toChildren.get(i).contains(uuid)) {
        return i;
      }
    }
    return null;
  }

  @Override
  public boolean canLoad(UUID uuid) {
    return toId.containsKey(uuid);
  }

  @Override
  public void changeNodeId(UUID uuid, String id) {
    String old = getNodeId(uuid);
//    getDeltaTarget()
//        .changed(
//            uuid,
//            new ChangesSet<>(Collections.emptySet(),AttributeStorage.formSet(TreeNode.ATTR_ID),Collections.emptySet(),
//                AttributeStorage.formAttributeDefMap(TreeNode.formIdAttribute(uuid, old))),
//            AttributeStorage.formAttributeDefMap(TreeNode.formIdAttribute(uuid, id)),
//            Moment.BEFORE,
//            true);
//    getDeltaTarget().moved(uuid, getParent(uuid), getParent(uuid), Moment.BEFORE, true);
    toId.put(uuid, id);
//    getDeltaTarget()
//        .changed(
//            uuid,
//            new ChangesSet<>(Collections.emptySet(),AttributeStorage.formSet(TreeNode.ATTR_ID),Collections.emptySet(),
//                AttributeStorage.formAttributeDefMap(TreeNode.formIdAttribute(uuid, old))),
//            AttributeStorage.formAttributeDefMap(TreeNode.formIdAttribute(uuid, id)),
//            Moment.AFTER,
//            true);
//    getDeltaTarget().moved(uuid, getParent(uuid), getParent(uuid), Moment.AFTER, true);
  }

  @Override
  public void moveNodeTo(UUID node, UUID parent) {
    UUID oldParent = getParent(node);
    getDeltaTarget().moved(node, oldParent, parent, Moment.BEFORE, true);
    if (oldParent != null) toChildren.get(oldParent).remove(node);
    toChildren.get(parent).add(node);
    getDeltaTarget().moved(node, oldParent, parent, Moment.AFTER, true);
  }

  @Override
  public void saveStorageAttributes(UUID uuid, ChangesSet<Attribute> changes) {
//    getDeltaTarget().changed(uuid, changes, attributes, Moment.BEFORE, true);
    //Map<String, Attribute> old = getAttributes(uuid);

    //Map<String, Attribute> oldMap = getAttributes(uuid);
    if(!toAttributes.containsKey(uuid))
      toAttributes.put(uuid, new HashMap<String,AttributeDefinition>());
    AttributeStorage.updateMap(toAttributes.get(uuid),changes);
    //toAttributes.put(uuid, new HashMap(attributes));
//    for(String s:oldMap.keySet())
//      if(!changes.contains(s))
//        toAttributes.get(uuid).put(s,oldMap.get(s));
//    getDeltaTarget().changed(uuid, changes, attributes, Moment.AFTER, true);
  }

  @Override
  public URI getNodeResourcePath(UUID uuId) {
    return null;
    // throw new UnsupportedOperationException();
  }

  @Override
  public InputStream getResourceContents(UUID uuid, String resName) {
    // TODO check stream
    byte[] data = getResourceData(uuid, resName);
    if (data == null) {
      return null;
    }
    return new ByteArrayInputStream(data);
  }

  @Override
  public void setResourceContents(UUID uuid, String resName, InputStream in, Object data) {
    if (!toResourсes.containsKey(uuid)) {
      toResourсes.put(uuid, new HashMap<String, ByteArrayOutputStream>());
    }
    getDeltaTarget().resourceChanged(uuid, resName, Moment.BEFORE, true);
    Map<String, ByteArrayOutputStream> r = toResourсes.get(uuid);
    ByteArrayOutputStream out = new ByteArrayOutputStream();
    r.put(resName, out);
    byte[] buffer = new byte[1024];
    int len;
    try {
      while ((len = in.read(buffer)) != -1) {
        out.write(buffer, 0, len);
      }
      getDeltaTarget().resourceChanged(uuid, resName, Moment.AFTER, true);
    } catch (IOException e) {
      e.printStackTrace();
    }
  }

  private byte[] getResourceData(UUID uuid, String resName) {
    if (!isResourceExists(uuid, resName)) {
      return null;
    }
    Map<String, ByteArrayOutputStream> nodeResources = toResourсes.get(uuid);
    ByteArrayOutputStream resData = nodeResources.get(resName);
    return resData.toByteArray();
  }

  @Override
  public boolean isResourceExists(UUID uuid, String resName) {
    Map<String, ByteArrayOutputStream> nodeResources = toResourсes.get(uuid);
    if (nodeResources == null) {
      return false;
    }
    ByteArrayOutputStream resData = nodeResources.get(resName);
    if (resData == null) {
      return false;
    }
    return true;
  }

  @Override
  public void deleteResource(UUID uuid, String resName) {
    getDeltaTarget().resourceChanged(uuid, resName, Moment.BEFORE, true);
    toResourсes.get(uuid).remove(resName);
    getDeltaTarget().resourceChanged(uuid, resName, Moment.AFTER, true);
  }

  @Override
  public Set<String> listResources(UUID uuid) {
    if (!toResourсes.containsKey(uuid)) return new HashSet<>();
    return new HashSet<>(toResourсes.get(uuid).keySet());
  }

  @Override
  public void createStorage(String name) {}

  @Override
  public void deleteStorage() {}

  @Override
  public String getName() {
    return toString();
  }

  @Override
  public ITreeStorage getBaseStorage() {
    // TODO Auto-generated method stub
    return null;
  }

  @Override
  public File getResource(UUID uuid, String name) {
    throw new UnsupportedOperationException();
  }

  @Override
  public Path getLocation() {
    return CoreUtils.getTempDir().resolve("CacheStorage"+root);
    //throw new UnsupportedOperationException();
  }
}
