/*
 * Copyright 2015 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.unitesk.requality.core;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TreeMap;
import java.util.UUID;
import com.unitesk.requality.core.attribute.Attribute;
import com.unitesk.requality.core.attribute.AttributeDefinition;
import com.unitesk.requality.core.attribute.AttributeStorage;
import com.unitesk.requality.core.attribute.AttributeType;
import com.unitesk.requality.core.attribute.ListValue;
import com.unitesk.requality.core.attribute.TypedValue;
import com.unitesk.requality.core.attribute.Version;
import com.unitesk.requality.core.attribute.valueproviders.ConstValueProvider;
import com.unitesk.requality.nodetypes.VirtualNode;

public class NodeDesc {
  // taken from transaction storage
  private static final String TMP_RESOURCE_PREFIX = "requality";
  private UUID uuid;
  private Map<String, AttributeDefinition> attributes;
  private Map<String, File> resources;
  private Map<UUID, NodeDesc> children;
  private UUID parentUUId;
  private UUID showparentUUId;
  private TreeDB db;
  private Version version;
  private UUID cloneParent;

  private static Map<UUID, NodeDesc> listToMap(List<UUID> uids) {
    Map<UUID, NodeDesc> map = new TreeMap<>();
    for (UUID uid : uids) map.put(uid, null);
    return map;
  }

  public NodeDesc(
      TreeDB db,
      UUID uuid,
      UUID cloneParent,
      UUID parentUUId,
      UUID showparentUUId,
      String id,
      Map<String, AttributeDefinition> map,
      List<UUID> uids,
      boolean renew) {
    this(db, uuid, cloneParent, parentUUId, showparentUUId, id, map, listToMap(uids), renew);
  }

  public NodeDesc(
      TreeDB db,
      UUID uuid,
      UUID cloneParent,
      UUID parentUUId,
      UUID showparentUUID,
      String id,
      Map<String, AttributeDefinition> attributes,
      Map<UUID, NodeDesc> map,
      boolean renew) {
    this.uuid = uuid;
    this.db = db;
    this.attributes = attributes;
    setId(id);
    this.cloneParent = cloneParent;
    this.children = map;
    this.parentUUId = parentUUId;
    this.showparentUUId = showparentUUID;
  }

  @SuppressWarnings("unchecked")
  static Object renewList(ListValue value, List<Object> newValues) {
    if (value.getType() == AttributeType.LIST) {
      List<Object> k = new ArrayList<>();
      Iterator<?> i = newValues.iterator();
      for (Object o : value.getValues()) {
        Object o2 = i.next();
        k.add(renewList((ListValue) o, (List<Object>) o2));
      }
      return new ListValue(value.getType(), k);
    } else {
      return new ListValue(value.getType(), new ArrayList<>(newValues));
    }
  }

  public void setId(String id) {
    this.attributes.put(TreeNode.ATTR_ID, TreeNode.formIdAttribute(uuid, id));
  }

  @Override
  public boolean equals(Object obj) {
    if (!(obj instanceof NodeDesc)) {
      return false;
    }
    NodeDesc nodeDesc = (NodeDesc) obj;
    if (!nodeDesc.getVersion().equals(getVersion())) {
      return false;
    }
    if (!nodeDesc.getId().equals(getId())) {
      return false;
    }
    if (!nodeDesc.getUUId().equals(getUUId())) {
      return false;
    }
    if (!attrubuteEquals(nodeDesc)) {
      return false;
    }
    if (!childrenEquals(nodeDesc)) {
      return false;
    }
    return true;
  }

  public boolean equalsButUUID(Object obj) {
    if (!(obj instanceof NodeDesc)) {
      return false;
    }
    NodeDesc nodeDesc = (NodeDesc) obj;
    if (!nodeDesc.getId().equals(getId())) {
      return false;
    }
    if (!attrubuteEquals(nodeDesc)) {
      return false;
    }
    if (!childrenEquals(nodeDesc)) {
      return false;
    }
    return true;
  }

  private boolean childrenEquals(NodeDesc nodeDesc) {
    return getChildrenUUId().equals(nodeDesc.getChildrenUUId());
  }

  private boolean attrubuteEquals(NodeDesc nodeDesc) {
    return AttributeStorage.attributesDefsEquals(nodeDesc.getAttributes(), getAttributes());
  }

  public Version getVersion() {
    return version == null ? Version.getUnset() : version;
  }

  public String getId() {
    return attributes.get(TreeNode.ATTR_ID).getValue(null).toString();
  }

  public UUID getUUId() {
    return uuid;
  }

  public List<UUID> getChildrenUUId() {
    return Collections.unmodifiableList(new ArrayList<>(children.keySet()));
  }

  public Map<String, AttributeDefinition> getAttributes() {
    return attributes;
  }

  public void setAttribute(AttributeDefinition attr) {
    // List<Attribute> list = attributes.get(attr.getKey());
    // if(list==null)
    // list = new LinkedList<Attribute>();
    // list = AttributeStorage.addToList(attributes,list, attr, uuid);
    attributes.put(attr.getKey(), attr);
  }

  public AttributeDefinition removeAttribute(String key) {
    return attributes.remove(key);
    // List<Attribute> clist = attributes.get(key);
    // if(clist != null && !clist.isEmpty()){
    // Iterator<Attribute> i = clist.iterator();
    // Attribute a = i.next();
    // if(a.getOrigin().equals(uuid))
    // i.remove();
    // return a;
    // }
    // return null;
  }

  public UUID getWorkParentUUID() {
    return parentUUId;
  }

  public UUID getShownParentUUID() {
    return showparentUUId;
  }

  @Override
  public String toString() {
    return getId() + "[" + getType() + "]";
  };

  public String getType() {
    if (getAttributes().containsKey(TreeNode.ATTR_TYPE)) {
      // Iterator<Attribute> it = getAttributes().get(TreeNode.ATTR_TYPE);
      // if(it.hasNext()){
      AttributeDefinition a = getAttributes().get(TreeNode.ATTR_TYPE);
      TypedValue tv = ((ConstValueProvider)a.getValueProvider()).getRawValue();
      if (tv == null)
        return AttributeStorage.getCachedStorage(db, a.getOrigin())
            .getStringAttributeValue(a.getKey());
      return tv.getValue().toString();
      // }
    }
    return "TreeNode";
  }

  public void setChildrenDesc(UUID chUUId, NodeDesc chDesc) {
    assert children.containsKey(chUUId);
    chDesc.setParentUUId(uuid);
    children.put(chUUId, chDesc);
  }

  public NodeDesc getChildDesc(UUID chUUId) {
    return children.get(chUUId);
  }

  public TreeNode createNode(TreeDB db, UUID pUUId, Boolean deep) {
    return createNode(db, pUUId, deep, false);
  }

  public TreeNode createNode(TreeDB db, UUID pUUId, Boolean deep, Boolean saveUUID) {
    // new node - new uuid
    UUID uid = uuid;
    if (!saveUUID) uid = db.getUuidGenerator().nextUUId();
    TreeNode result = db.createChildNode(uid, pUUId, getId(), getType(), true);
    Map<String, AttributeDefinition> attributes = getAttributes();
    for (String s : attributes.keySet()) {
      AttributeDefinition a = attributes.get(s);
      if (!uid.equals(uuid) && a.getOrigin().equals(uuid))
        a = new AttributeDefinition(uid, a.getKey(), a.getAvailability(), a.getValueProvider(), a.getFlags());
      result.attrs().setAttribute(a);
    }
    result.cloneParent = getCloneParent();
    result.commitAttributes();
    if (deep) {
      for (UUID chUUId : getChildrenUUId()) {
        children.get(chUUId).createNode(db, result.getUUId(), deep, saveUUID);
      }
    }
    if (resources != null)
      for (String s : resources.keySet())
        try {
          File f = resources.get(s);
          if (f.exists()) {
            FileInputStream fis = new FileInputStream(f);
            result.setResourceContent(s, fis);
            fis.close();
          }
        } catch (FileNotFoundException e) {
          e.printStackTrace();
        } catch (IOException e) {
          // TODO Auto-generated catch block
          e.printStackTrace();
        }
    return result;
  }

  private UUID getCloneParent() {
    return cloneParent;
  }

  public TreeNode createNode(TreeDB db) {
    return createNode(db, false);
  }

  public TreeNode createNode(TreeDB db, Boolean saveLink) {
    return createNode(db, parentUUId, true, saveLink);
  }

  public void setParentUUId(UUID pUUId) {
    this.parentUUId = pUUId;
    showparentUUId = pUUId; // TODO
  }

  public static NodeDesc fromNode(TreeNode node) {
    return fromNode(node, true, true, false);
  }

  public static NodeDesc fromNode(final TreeNode node, Boolean deep, boolean keepattributesorigin) {
    return fromNode(node, deep, keepattributesorigin, false);
  }

  public static NodeDesc fromNode(
      final TreeNode node, Boolean deep, boolean keepattributesorigin, boolean onlyloaded) {
    List<UUID> uids = node.getWorkChildrenUUIDs(onlyloaded);
    if (VirtualNode.TYPE_NAME.equals(node.getType())) uids = Collections.emptyList();
    Map<String, Attribute> attrs = node.attrs().getActiveAttributes();
    // if(node.getCloneParentUUId()!=null){
    // attrs.put("_cloneparent", new LinkedList<Attribute>(){{{
    // add(new Attribute(Version.getZero(), node.getUUId(), "_cloneparent",
    // node.getCloneParentUUId().toString()));
    // }}});
    // attrs.put("_cloneparentversion", new LinkedList<Attribute>(){{{
    // add(new Attribute(Version.getZero(), node.getUUId(),
    // "_cloneparentversion", node.getCPVersion().toString()));
    // }}});
    // }
    // attrs.put("_version", new LinkedList<Attribute>(){{{
    // add(new Attribute(Version.getZero(), node.getUUId(), "_version",
    // node.getVersion().toString()));
    // }}});
    // tends to be loading by vattrs alg
    Map<String, Attribute> attributes = new HashMap<>(attrs);
    Map<String, AttributeDefinition> natr = new HashMap<>();
    for (Map.Entry<String, Attribute> entry : attributes.entrySet()) {
      String key = entry.getKey();
      if (TreeNode.isVirtName(node,key))
        // if(key.equals("_%id"))
        continue; // TODO look into
      Attribute newa = null;
      Attribute attr = entry.getValue();
      if(attr!=null)
      {
        UUID origin = attr.getOrigin();
        if (!attr.isLocal(node.getUUId()) && !keepattributesorigin)
          attr = node.attrs().getAttribute(attr.getKey());
        if (attr == null || !attr.getOrigin().equals(origin)) continue;
        
        //if (!attr.getDefinition().isGenerated()) {
//          Object value = attr.getRawValue();
//          if (value != null) {
//            AttributeType at = attr.getResultType();
//            if (attr.getRawType() == AttributeType.LIST) {
//
//              value = renewList((ListValue) value, (List<Object>) attr.getNewValue(node.attrs()));
//            }
//            if (attr.getRawType() == AttributeType.ENUM) {
//              try {
//                value = ((AppliedEnumValue) value).clone();
//                at = AttributeType.ENUM;
//              } catch (CloneNotSupportedException e) {
//                e.printStackTrace();
//              }
//            }
//            if (attr.getRawType() == AttributeType.ENUM_DEFINITION) {
//              try {
//                value = ((EnumDefinitionValue) value).clone();
//                at = AttributeType.ENUM_DEFINITION;
//              } catch (CloneNotSupportedException e) {
//                e.printStackTrace();
//              }
//            }
//            newa =
//                (new Attribute(
//                    attr.getVersion(),
//                    !keepattributesorigin ? node.getUUId() : attr.getOrigin(),
//                    at,
//                    attr.getKey(),
//                    value,
//                    attr.getGenerator()));
//          }
//        } else
//          newa =
//              (new Attribute(
//                  attr.getVersion(),
//                  !keepattributesorigin ? node.getUUId() : attr.getOrigin(),
//                  attr.getRawType(),
//                  attr.getKey(),
//                  attr.getRawValue(),
//                  attr.getGenerator()));
        if(keepattributesorigin)
          natr.put(key, attr.getDefinition());
        else
          natr.put(key, new AttributeDefinition(node.getUUId(), attr.getKey(), attr.getAvailability(),
                attr.getDefinition().getValueProvider(),
                attr.getDefinition().getFlags()));
      }
      
    }
    NodeDesc result =
        new NodeDesc(
            node.getTreeDB(),
            node.getUUId(),
            node.getCloneParentUUId(),
            node.getWorkParentUUID(),
            node.getShownParentUUID(),
            node.getId(),
            natr,
            uids,
            !keepattributesorigin);
    Set<String> ress = node.listResources();
    result.resources = new HashMap<>();

    for (String s : ress) {
      InputStream is = node.getResourceContent(s);
      try{
        if (is == null) continue;
        byte[] bt;
        try {
          File tmpNewFile =
              File.createTempFile(
                  TMP_RESOURCE_PREFIX, node.getUUId().toString() + result.resources.size());
          bt = new byte[is.available()];
          is.read(bt);
          FileOutputStream fos = new FileOutputStream(tmpNewFile);
          fos.write(bt);
          fos.close();
          result.resources.put(s, tmpNewFile);
          tmpNewFile.deleteOnExit();
        } catch (IOException e) {
          // TODO Auto-generated catch block
          e.printStackTrace();
        }
      }
      finally{
        try {
          is.close();
        } catch (IOException e) {
          // TODO Auto-generated catch block
          e.printStackTrace();
        }
      }
    }
    // node.getResourceContent(name)
    if (deep) {
      int num = node.sizeWorkChildren(onlyloaded);
      for (int i = 0; i < num; i++) {
        TreeNode child = node.getWorkChild(i, false, onlyloaded);
        if (child.isVirtual() && !node.isVirtual()) RLogger.logInfo("not add");
        else result.setChildrenDesc(child.getUUId(), child.toNodeDesc(deep, keepattributesorigin));
      }
    }
    return result;
  }

  public void setChildrenUUIDs(List<UUID> childs) {
    for (UUID uid : childs)
      if (!children.containsKey(uid)) children.put(uid, null); // may cause error
  }

  public Map<UUID, NodeDesc> getChildren() {
    return children;
  }

  public String getIndex() {
    if (getAttributes().containsKey(TreeDB.INDEX_ATTRIBUTE)) {
      AttributeDefinition a = getAttributes().get(TreeDB.INDEX_ATTRIBUTE);
      ;
      if (a.getRawValue()==null || a.getRawValue().getValue() == null)
        return AttributeStorage.getCachedStorage(db, a.getOrigin())
            .getStringAttributeValue(a.getKey());
      return a.getRawValue().getValue().toString();
    }
    return "0";
  }
}
