package com.unitesk.requality.core;

import java.io.IOException;
import java.net.URI;
import java.util.Map;
import com.unitesk.requality.core.instancer.Instantiator;
import com.unitesk.requality.nodetypes.DocFolder;
import com.unitesk.requality.nodetypes.InstancesFolder;
import com.unitesk.requality.nodetypes.NodeType;
import com.unitesk.requality.nodetypes.ReportFolder;
import com.unitesk.requality.nodetypes.Requirement;
import com.unitesk.requality.nodetypes.RestrictedFolder;
import com.unitesk.requality.report.Template;

public class ResManager {
  //Static tracker to be inited for new resource manager instance
  private static IResTracker gtt;
  //Singleton exemplar
  private static ResManager tm = null;
  private static String[] lastParams = new String[0];
  //Tracker of single instance
  private IResTracker tt;
  /**
   * * Default constructor. May have tracker-specific init params.
   *
   * @see FilesTracker
   * @param params tracker-specific init params
   */
  private ResManager(String... params) {
    this(false, params);
  }
  /**
   * * Constructor that allows to init Files Tracker even when eclipse environment is active.
   *
   * @param forceFiles if true then ResManager will use Files Tracker instead of Eclipse Tracker
   * @param params tracker-specific init params
   */
  private ResManager(boolean forceFiles, String... params) {
    if (params.length > 0) lastParams = params;
    else params = lastParams;
    if (gtt == null || forceFiles) tt = new FilesTracker();
    else tt = gtt;
  }
  /** * Get list of available TreeDBs from active tracker. */
  public TreeDB[] getTreeDBs() {
    return tt.getTreeDBs();
  }
  /**
   * * Method to init instance of Resource Manager. If forceFiles is enabled then the instance will
   * be initiated with Files Tracker instance. Otherwise it will use available tracker. In a case
   * eclipse is active it will use Resource Tracker.
   *
   * @param forceFiles if true then ResManager will use Files Tracker instead of Eclipse Tracker
   * @param params tracker-specific init params
   * @return Resource Manager instance with available tracker.
   */
  public static ResManager getManager(boolean forceFiles, String... params) {
    boolean ch = false;
    if (tm == null || forceFiles && !(tm.tt instanceof FilesTracker)) {
      tm = new ResManager(params);
      ch = true;
    }
    if (ch) tm.tt.init(params);
    return tm;
  }
  /**
   * * Method to init instance of Resource Manager. It will use available tracker. In a case eclipse
   * is active it will use Resource Tracker.
   *
   * @param params tracker-specific init params
   * @return Resource Manager instance with available tracker.
   */
  public static ResManager getManager(String... params) {
    return getManager(false, params);
  }
  /**
   * * Check that given TreeDB is tracking by active tracker.
   *
   * @param db TreeDB to be checked
   * @return true if db is tracked
   */
  public boolean isTracking(TreeDB db) {
    for (TreeDB db2 : tt.getTreeDBs()) if (db.equals(db2)) return true;
    return false;
  }
  /**
   * * Get name of given TreeDB. Names will be taken from trees tracker and in most cases are equal
   * to name of db-containing folder.
   *
   * @param db TreeDB for which name will be given
   * @return name of TreeDB if it is tracked, null in other cases
   */
  public String getName(TreeDB db) {
    if (db == null) return null;
    return tt.getName(db);
  }
  /**
   * * Get TreeDB by name if have one with given name in trees tracker.
   *
   * @param name name of TreeDB to be searched
   * @return TreeDB instance if tracker have one with given name, null in other cases
   */
  public TreeDB getTreeDB(String name) {
    return tt.getTreeDB(name);
  }
  /**
   * * Remove TreeDB with given name if have one tracked.
   *
   * @param name name of TreeDB to be deleted
   */
  public void remove(String name) {
    tt.remove(name);
  }
  /**
   * * Returns string representation of path to TreeDB instance if it is tracked.
   *
   * @param treeDB TreeDB to get path
   * @return path to TreeDB if it is tracked
   */
  public String getLocation(TreeDB treeDB) {
    return tt.getLocation(treeDB);
  }
  /**
   * * Get report templates from tracker. TODO move it to API
   *
   * @return report templates in form of map id->template
   */
  public Map<String, Template> getTemplates() {
    return tt.getReportTemplates();
  }
  /**
   * * Method to create TreeDB with given name and add to active tracker. May throw runtime
   * exception if TreeDB with given name was already exist.
   *
   * @param string name of TreeDB
   * @return TreeDB instance if no TreeDBs with given name was tracked before, null if creation was
   *     failed
   */
  public TreeDB createProject(String string) {
    if (getTreeDB(string) != null)
      throw new RuntimeException("Project with given name was already exist");
    TreeDB db = tt.createProject(string);
    db.setIsLoaded(false);
    // Create nodes
    RLogger.logInfo(string + " was created on " + tt.getClass().getName());
    TreeNode node = db.getRootNode();
    if (node.findWorkChild("Requirements") == null
        || node.findWorkChild("Types") == null
        || node.findWorkChild("Documents") == null
        || node.findWorkChild("Reports") == null
        || node.findWorkChild("Instances") == null) {
      if (db.startTransaction("Create Project Structure", true)) {
        if (node.findWorkChild("Requirements") == null) {
          node.createChild("Requirements", Requirement.TYPE_NAME);
        }
        if (node.findWorkChild("Types") == null) {
          node.createChild("Types", NodeType.TYPE_NAME);
        }
        if (node.findWorkChild("Documents") == null) {
          node.createChild("Documents", DocFolder.TYPE_NAME);
        }
        if (node.findWorkChild("Reports") == null) {
          node.createChild("Reports", ReportFolder.TYPE_NAME);
        }
        if (node.findWorkChild("Templates") == null) {
          node.createChild("Templates", RestrictedFolder.TYPE_NAME, null);
        }
        if (node.findWorkChild("Instances") == null) {
          node.createChild("Instances", InstancesFolder.TYPE_NAME, null);
          // Init vtree
          // Instantiator.getVirtualTreeDB(node.getTreeDB(),Instantiator.DEFAULT_INSTANCE_NAME,true);
        }
        db.commit();
      }
    }
    db.setIsLoaded(true);
    db.getLinkResolver().initCaches();
    db.getTermsStorage().initCaches();
    Instantiator.loadDBs(db);
    tt.trackProject(string, db);
    return db;
  }
  /** * Method to clear caches for given TreeDB. */
  public void clearCache(TreeDB p) {
    tt.clearCache(p);
  }
  /**
   * * Method to set default resource tracker. It will be used for resource manager instance by
   * default.
   */
  public static void setTreesTracker(IResTracker tracker) {
    if (tm != null) {
      tm = null;
    }
    gtt = tracker;
  }
  /**
   * * Method to close and open TreeDB. Now used for compatibility with eclipse project close-open
   * mechanics and corresponding Requality operations like storage version conversion
   *
   * @param treeDB
   */
  public void reOpen(TreeDB treeDB) {
    tt.reOpen(treeDB);
  }

  public static void resetOne() {
    tm = null;
  }

  public IResTracker getTracker() {
    return tt;
  }
  public TreeDB trackTreeDB(URI path) throws IOException{
    return tt.trackTreeDB(path);
  }
  public TreeDB trackTreeDB(URI path, String name) throws IOException{
    return tt.trackTreeDB(path,name);
  }
  public java.nio.file.Path getRootLocation() {
    return tt.getRootLocation();
  }
  public java.nio.file.Path getGlobalCacheLocation() {
    return tt.getGlobalCacheLocation();
  }
}
