/*
 * Copyright 2015 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * limitations under the License.
 */
package com.unitesk.requality.core;

import java.io.InputStream;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.Stack;
import java.util.UUID;
import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.concurrent.CopyOnWriteArrayList;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.locks.ReentrantLock;
import java.util.concurrent.locks.ReentrantReadWriteLock;
import java.util.concurrent.locks.ReentrantReadWriteLock.ReadLock;
import java.util.concurrent.locks.ReentrantReadWriteLock.WriteLock;
import java.util.regex.Pattern;
import com.unitesk.requality.core.SortingSettings.SortingMethod;
import com.unitesk.requality.core.TreeNodeChangeEvent.TreeNodeChangeEventType;
import com.unitesk.requality.core.attribute.Attribute;
import com.unitesk.requality.core.attribute.AttributeAvailability;
import com.unitesk.requality.core.attribute.AttributeDefToMerge;
import com.unitesk.requality.core.attribute.AttributeDefinition;
import com.unitesk.requality.core.attribute.AttributeStorage;
import com.unitesk.requality.core.attribute.AttributeToMerge;
import com.unitesk.requality.core.attribute.AttributeType;
import com.unitesk.requality.core.attribute.Flag;
import com.unitesk.requality.core.attribute.ListValue;
import com.unitesk.requality.core.attribute.MergeParams;
import com.unitesk.requality.core.attribute.Version;
import com.unitesk.requality.core.attribute.VersionsStorage;
import com.unitesk.requality.core.attribute.VersionsStorage.Type;
import com.unitesk.requality.core.attribute.valueproviders.ConstValueProvider;
import com.unitesk.requality.core.attribute.valueproviders.ReferenceValueProvider;
import com.unitesk.requality.core.checkers.CheckerContainer;
import com.unitesk.requality.core.checkers.IMark;
import com.unitesk.requality.core.instancer.Instantiator;
import com.unitesk.requality.core.instancer.TreeInstancer;
import com.unitesk.requality.core.operations.CommonRemoveLocation;
import com.unitesk.requality.core.propagators.AttributePropagatorComparator;
import com.unitesk.requality.core.propagators.IAttributeChangePropagator;
import com.unitesk.requality.core.transaction.ITransactionListener;
import com.unitesk.requality.core.transaction.TransactionStorage;
import com.unitesk.requality.nodetypes.Document;
import com.unitesk.requality.nodetypes.Location;
import com.unitesk.requality.nodetypes.NodeType;
import com.unitesk.requality.nodetypes.Requirement;
import com.unitesk.requality.nodetypes.RestrictedFolder;
import com.unitesk.requality.nodetypes.TestPurpose;
import com.unitesk.requality.nodetypes.VirtualNode;
import com.unitesk.requality.tools.DeepFirstTreeWalker;
import com.unitesk.requality.tools.RequalityCLI;

public class TreeDB {
  public static final int DEFAULT_DEPTH = 4;
  public static final String MAX_FID = "_maxForeignId";
  public static final String ORDER_ATTRIBUTE = "_order";
  public static final String INDEX_ATTRIBUTE = "_index";
  public static final String TEST_GENERATORS_FILTERS = "_testGenFilters";
  public static final String DEFAULT_REFERENCE_NAME = "_defRefName";
  public static final String ACTIVE_TEMPLATES_ATTRIBUTE = "_active_templates";
  public static final String EMPTY_TEMPLATE = "&empty";
  public static final String SORT_BY_INDEX = "_sort_by_index";
  public static final String SORT_BY_LOCATIONS = "_sort_by_locations";
  public static final String SORT_SEP_TYPES = "_separate_types_in_sort";
  public static final String SORT_METHOD = "_sorting_method";
  private static final String PREFERENCE_FNMODE = "_com.unitesk.requality.full_name_mode";

  public boolean textIsSimple = true;

  private int loa = 0;
  private List<ITreeChangeListener> listeners = new CopyOnWriteArrayList<>();
  private List<ITreeChangeListener> guilisteners = new CopyOnWriteArrayList<>();
  private Map<UUID, IMark[]> chprob = new HashMap<>();
  private TreeNode root;
  private Map<UUID, TreeNode> cache = new HashMap<>();
  private List<String> testgenfilters;
  private ITreeStorage storage;
  private IUUIDGenerator uuidGenerator;
  private TreeLogic tlogic;
  private static Stack<TreeNode> rempool = new Stack<>();
  public static char DATA_SEPARATOR = ((char) 0xA6);
  private static SortingSettings defaultSortSettings =
      SortingSettings.init(SortingMethod.BY_INDEX, false);

  private Map<String, NodeType> defaultTypes = new HashMap<>();

  private boolean isVirtual;

  private ConcurrentLinkedQueue<TreeNodeChangeEvent> guichanges =
      new ConcurrentLinkedQueue<>();
  private int activeChangesRowsCount = 0;

  private Map<String, String> activeTemplates = null;

  protected CheckerContainer container = null;
  
  private Map<UUID, INodeChangeListener> latechangelisteners;
  Map<UUID, ReentrantLock> cacheLocks = new HashMap<>();
  Map<UUID, Thread> cacheThreads = new HashMap<>();
  private Set<UUID> loading = new HashSet<>();
  private ArrayList<ITransactionListener> transactionListeners =
      new ArrayList<>();
  private static List<String> typeFilter = new ArrayList<>();
  static final Pattern new_linkto_pattern =
      Pattern.compile("\\[linkto: ([^\\]]*)\\]", Pattern.MULTILINE & Pattern.DOTALL);
  static final Pattern linkto_pattern =
      Pattern.compile(
          "\\[<span id=\"([a-zA-Z0-9-]+)\">.{0,10}</span>linkto: ([^\\]]*)\\]",
          Pattern.MULTILINE & Pattern.DOTALL);

  private LinkResolver lr = null;
  private TermsStorage ts = null;
  private static LinkResolver glr = null;
  private static CacheStorage gcs = null;

  private TreeNode refTarget;
  private int refTargetState = 0;
  boolean loaded = false;
  public Map<UUID, Set<String>> attrRootRevCache = new HashMap<>();
  public Map<String, Set<UUID>> attrRootCache = new HashMap<>();

  private Map<String, TreeNode> btg = null;
  
  private AttributesRulesHolder vrh;
  
  public AttributesRulesHolder getAttributeRules(){
    if(vrh == null)
      vrh = new AttributesRulesHolder(this);
    return vrh;
  }
  
  private String getMaxFid(){
    String mfid = root.attrs().getStringAttributeValue(MAX_FID);
    if(mfid!=null){
      
    }else{
      
    }
    return mfid;
  }
  
  public void clearTagsCache() {
    btg = null;
  }

  public TreeNode getByTags(TreeNode node, String link) {
    if (btg == null) {
      btg = new HashMap<>();
      TreeNode rt = node.getTreeDB().getNode(Requirement.getTypeRootQId());
      DeepFirstTreeWalker dftw = new DeepFirstTreeWalker(rt);
      Iterator<TreeNode> it = dftw.iterator();
      while (it.hasNext()) {
        TreeNode nd = it.next();
        Attribute a = nd.attrs().getAttribute("_tags");
        if (a != null && !nd.isDisposed()) {
          List<Object> list = ((ListValue)a.getGValue()).getValues();
          for (Object o : list) btg.put(o.toString(), nd);
        }
      }
    }
    return btg != null ? btg.get(link) : null;
  }

  public String[] getNodesByUFName(String s, List<String> filter) {
    TreeNode[] nodes = getLinkResolver().getNodesByUFName(this, s, filter);
    String[] arr = new String[nodes.length];
    for (int i = 0; i < nodes.length; i++) arr[i] = nodes[i].getUUId().toString();
    return arr;
  }

  public CheckerContainer getCheckerContainer() {
    if (container == null) container = new CheckerContainer();
    return container;
  }

  public boolean getFullNameMode() {
    if (root == null) return false;
    Attribute a = root.attrs().getAttribute(PREFERENCE_FNMODE);
    return a != null && a.getValue()!=null && a.getValue().toString().toLowerCase().equals("true");
  }

  public void setFullNameMode(boolean useUVName) {
    if (root == null) return;
    root.attrs().setConstAttribute(PREFERENCE_FNMODE, useUVName, Flag.HIDDEN);
    root.commitAttributes();
  }

  public void loadActiveTemplates() {
    activeTemplates = new HashMap<>();
    if (getTemplatesFolder() == null) return;
    Attribute a = getTemplatesFolder().attrs().getAttribute(ACTIVE_TEMPLATES_ATTRIBUTE);
    if (a != null) {
      ListValue info = (ListValue) a.getValue();
      for (Object o : info.getValues()) {
        String s = o.toString();
        String type = s.substring(0, s.indexOf("|"));
        String id = s.substring(s.indexOf("|") + 1);
        if (id.isEmpty()) activeTemplates.put(type, null);
        else {
          TreeNode node = getTemplate(id);
          if (node != null) activeTemplates.put(type, id);
          else activeTemplates.put(type, null);
        }
      }
    }
  }

  public Map<String, String> getActiveTemplates() {
    if (activeTemplates == null) loadActiveTemplates();
    return activeTemplates;
  }

  public void saveActiveTemplates() {
    List<Object> infos = new ArrayList<>();
    for (String s : getActiveTemplates().keySet()) {
      String name = getActiveTemplates().get(s);
      if (name != null) {
        TreeNode tmpl = getTemplate(name);
        if (tmpl != null) infos.add(s + "|" + tmpl.getId());
        else infos.add(s + "|");
      } else infos.add(s + "|");
    }
    TreeNode tf = getOrCreateTemplatesFolder();
    tf.attrs().setConstAttribute(ACTIVE_TEMPLATES_ATTRIBUTE, new ListValue(AttributeType.STRING, infos),
        Flag.HIDDEN);
    
    getOrCreateTemplatesFolder().commitAttributes();
  }

  public TreeNode getTemplatesFolder() {
    try {
      return getRootNode().findWorkChild("Templates");
    } catch (RuntimeException re) {
    }
    return null;
  }

  INodeChangeListener templch =
      new INodeChangeListener() {
        @Override
        public void moved(TreeNode node, TreeNode oldParent, boolean internal) {}

        @Override
        public void deleted(TreeNode source) {}

        @Override
        public void attributeChange(TreeNode source, String what, Attribute old) {
          if (what.equals(ACTIVE_TEMPLATES_ATTRIBUTE)) {
            loadActiveTemplates();
          }
        }

        @Override
        public void attributesChange(TreeNode source, ChangesSet<Attribute> what) {
          if (what.contains(ACTIVE_TEMPLATES_ATTRIBUTE))
            attributeChange(source, ACTIVE_TEMPLATES_ATTRIBUTE, what.getOld(ACTIVE_TEMPLATES_ATTRIBUTE));
        }

        @Override
        public void fireNodeResourceChange(
            TreeNode treeNode, String resName, Moment m, boolean inner) {
          // TODO Auto-generated method stub

        }
      };

  public CacheStorage getCacheStorage() {
    //if(isVirtual())
      //return Instantiator.getBaseTree(this).getCacheStorage();
    return getStorage().getCacheStorage();
  }

  public TreeNode getOrCreateTemplatesFolder() {
    TreeNode node = getTemplatesFolder();
    if (node == null) 
      node = root.createChild("Templates", RestrictedFolder.TYPE_NAME);
    node.addChangeListener(templch);
    return node;
  }

  public TreeNode createTemplate(String type) {
    TreeNode fol = getOrCreateTemplatesFolder();
    String id = RequalityCLI.getNextId(fol, type, "Template_");
    return fol.createChild(id, type, null);
  }

  public boolean extractTemplate(TreeNode node, String name) {
    TreeNode fold = getOrCreateTemplatesFolder();
    if (getTemplate(name) != null) return false;
    Map<String, AttributeDefinition> map = new HashMap<>();
    UUID newuid = UUID.randomUUID();
    Map<String, Attribute> omap = node.attrs().getMap();
    for (String s : new HashSet<>(omap.keySet())) {
      if(!omap.get(s).isVirtual())
        map.put(s, omap.get(s).getDefinition().copy(newuid));
      // map.put(a.getKey(), list);
      // }
      // else
      // map.put(a.getKey(), a);
    }
    int num = 0;
    String retname = name;
    while (fold.findWorkChild(retname) != null) {
      retname = name + "_" + num;
      num++;
    }
    Map<String, AttributeDefinition> attributes = map;
    NodeDesc nd =
        new NodeDesc(
            this,
            newuid,
            null,
            fold.getUUId(),
            fold.getUUId(),
            retname,
            map,
            new ArrayList<UUID>(),
            true);
    nd.createNode(this);
    return true;
  }

  public List<TreeNode> getTemplates() {
    List<TreeNode> templates = new ArrayList<>();
    TreeNode tfolder = getTemplatesFolder();
    if (tfolder == null) return Collections.emptyList();
    for (TreeNode templ : tfolder.getWorkChildren()) // TODO sort
    templates.add(templ);
    return templates;
  }

  public TreeNode getTemplate(String name) {
    if (name == null || name.equals("&empty")) return null;
    if (getTemplatesFolder() == null) return null;
    return getTemplatesFolder().findWorkChild(name);
  }

  public void setActiveTemplate(String type, String templ) {
    getOrCreateTemplatesFolder();
    getActiveTemplates().put(type, templ);
    saveActiveTemplates();
  }

  public TreeNode getActiveTemplate(String type) {
    if (!getActiveTemplates().containsKey(type)) return null;
    return getTemplate(getActiveTemplates().get(type));
  }

  private ConcurrentLinkedQueue<TreeNodeChangeEvent> getGuiChanges() {
    if (isVirtual) { // Test here
      TreeDB db = Instantiator.getBaseTree(this);
      if (db != null && db != this) return db.getGuiChanges();
    }
    return guichanges; // TODO check
  }

  @SuppressWarnings("unchecked")
  public List<String> getTestGeneratorsFilters() {
    if (testgenfilters == null)
      testgenfilters =
          (List<String>)
              getRootNode()
                  .attrs()
                  .getListAttributeValue(TEST_GENERATORS_FILTERS, AttributeType.STRING);
    return testgenfilters;
  }

  /**
   * Returns Type of element with given uid if it exists.
   *
   * @param uid uuid of element
   * @return type of element or null
   */
  public String getNodeType(UUID uid) {
    if (isLoaded(uid)) return getNode(uid).getType();
    String tp = storage.getNodeRawType(uid);
    if (tp == null) {
      AttributeStorage as = AttributeStorage.getCachedStorage(this, uid);
      if (as != null){
        Object res = AttributeStorage.getAttributeValue(this, as.getActiveAttributes(),uid,TreeNode.ATTR_TYPE);
        if(res != null)
          return res.toString();
      }
    }
    return tp;
  }
  private ReentrantReadWriteLock rwl = new ReentrantReadWriteLock();
  private ReadLock rl = rwl.readLock();
  private WriteLock wl = rwl.writeLock();
  /** Combine all future events in a row to be handled concurrently. */
  public void startChangeRow() {
//    if (isVirtual) {
//      if(CoreUtils.DEBUG_LEVEL>1)
//        RLogger.logInfo("Start change row on "+ResManager.getManager().getName(TreeDB.this));
//      TreeDB root = Instantiator.getRootTree(this);
//      if (this != root) {
//        root.startChangeRow();
//        return;
//      }
//    }
    wl.lock();
    activeChangesRowsCount++;
    wl.unlock();
  }

  /**
   * Returns 'virtuality' status of catalog. Catalog is virtual if it is linked to virtual storage.
   *
   * @return true if catalog is virtual
   */
  public boolean isVirtual() {
    return isVirtual;
  }

  /**
   * Sets 'virtuality' status of catalog. Catalog is virtual if it is linked to virtual storage.
   * This method is public only to be available for inner mechanics. Should not be called by side
   * plugins.
   *
   * @return true if catalog is virtual
   */
  public void setVirtualState(boolean value) {
    isVirtual = value;
  }

  /** Stop event row started with startChangeRow() method */
  @SuppressWarnings("unchecked")
  public void stopChangeRow() {
//    if (isVirtual) {
//      TreeDB root = Instantiator.getRootTree(this);
//      if (this != root) {
//        root.stopChangeRow();
//        return;
//      }
//    } 
    wl.lock();
    try{
      activeChangesRowsCount--;
      if (activeChangesRowsCount <= 1){
        try{
          clearRemPool();
        }catch(Exception e){
          e.printStackTrace();
        }
      }
      if(CoreUtils.DEBUG_LEVEL>1)
        RLogger.logInfo("Stop change row for "+ResManager.getManager().getName(this)); 
      if (activeChangesRowsCount == 0) {
        if(CoreUtils.DEBUG_LEVEL>1)
          RLogger.logInfo("Row is 0, actions for "+ResManager.getManager().getName(this));  
        // TODO needed?
        HashMap<UUID, Integer> lastchange = new HashMap<>(); // ,
        // lastmove = new HashMap<UUID, Integer>();
        HashSet<UUID> deleted = new HashSet<>(), created = new HashSet<>();
        HashMap<UUID, ChangesSet<Attribute>> changeParams = new HashMap<>();
        int i = 0;
        TreeNodeChangeEvent lastfocus = null;
        List<TreeNodeChangeEvent> handled = new LinkedList<>();
        while (!getGuiChanges().isEmpty()) {
          TreeNodeChangeEvent change = getGuiChanges().poll();
          handled.add(change);
          if (change == null || change.getType() == null) continue;
          if (change.getType() == TreeNodeChangeEventType.DELETE) {
            NodeDesc nd = change.getNodeDescTarget();
            if (nd != null) {
              deleted.add(nd.getUUId());
            } else deleted.add(change.getTreeNodeTarget().getUUId());
          } else if (change.getType() == TreeNodeChangeEventType.CHANGE) {
            TreeNode node = change.getTreeNodeTarget();
            if (node.isDisposed()) {
              if (!deleted.contains(node)) deleted.add(node.getUUId());
            } else {
              lastchange.put(node.getUUId(), i);
              ChangesSet<Attribute>na = (ChangesSet<Attribute>) change.getArguments()[1];
              if(na!=null){
                if (!changeParams.containsKey(node.getUUId()))
                  changeParams.put(node.getUUId(), na);
                else
                  changeParams.put(node.getUUId(), 
                      changeParams.get(node.getUUId()).merge((ChangesSet<Attribute>) change.getArguments()[1]));
              }
            }
          } else if (change.getType() == TreeNodeChangeEventType.MOVE) {
            TreeNode node = change.getTreeNodeTarget();
            if (node.isDisposed()) {
              if (!deleted.contains(node)) deleted.add(node.getUUId());
            } // else
            // lastmove.put(node.getUUId(), i);
          }
          if (change.getType() == TreeNodeChangeEventType.CREATE) {
            TreeNode node = change.getTreeNodeTarget();
            if(node!=null){
              if (node.isDisposed()) {
                if (!deleted.contains(node)) deleted.add(node.getUUId());
              } else created.add(node.getUUId());
            }
          }
          i++;
        }
        for (UUID uid : deleted) {
          lastchange.remove(uid);
          // lastmove.remove(uid);
          created.remove(uid);
        }
        for (UUID uid : created) {
          lastchange.remove(uid);
          // lastmove.remove(uid);
        }
        i = 0;
        Set<Integer> events = new HashSet<>();
        events.addAll(lastchange.values());
        // events.addAll(lastmove.values());
        List<TreeNodeChangeEvent> retevents = new LinkedList<>();
        if(CoreUtils.DEBUG_LEVEL>1)
          RLogger.logInfo("Before change set  on "+ResManager.getManager().getName(TreeDB.this));
        while (!handled.isEmpty()) {
  
          TreeNodeChangeEvent change = handled.remove(0);
          try {
            if (change != null)
              switch (change.getType()) {
                case CHANGE:
                  {
                    TreeNode node = change.getTreeNodeTarget();
                    try {
                      if (events.contains(i)) {
                        ChangesSet<Attribute> params = changeParams.get(node.getUUId());
                        Object[] args = change.getArguments();
                        args[1] = params;
                        retevents.add(
                            new TreeNodeChangeEvent(TreeNodeChangeEventType.CHANGE, node, args));
                        change.getTreeNodeTarget().fireGUINodeAttributeChange(params);
                      }
                    } catch (NullPointerException p) {
                      p.printStackTrace();
                    }
                  }
                  break;
                case CACHE_CHANGE:
                  {
                    TreeNode node = change.getTreeNodeTarget();
                    if (node != null) {
                      retevents.add(change);
                    }
                  }
                  break;
                case CREATE:
                  {
                    TreeNode node = change.getTreeNodeTarget();
                    if (node != null) {
                      retevents.add(change);
                    }
                    // change.getTreeNodeTarget().fireNodeAttributeChange(
                    // change.getArguments().length>1?(String)change.getArguments()[1]:TreeNode.ATTR_UNDEFINED);
                  }
                  break;
                case DELETE:
                  {
                    retevents.add(change);
                    TreeNode targ = change.getTreeNodeTarget();
                    if (targ != null) targ.fireGUINodeDelete();
                  }
                  break;
                case MOVE:
                  {
                    // if(events.contains(i)){
                    retevents.add(change);
                    // }
                    // change.getTreeNodeTarget().fireNodeAttributeChange(
                    // change.getArguments().length>2?(String)change.getArguments()[2]:TreeNode.ATTR_UNDEFINED);
                  }
                  break;
                case SETFOCUS:
                  {
                    TreeNode node = change.getTreeNodeTarget();
                    if (!node.isDisposed()) lastfocus = change;
                  }
                  break;
              }
          } catch (Exception e) {
            RLogger.logInfo("error during event ex");
            e.printStackTrace();
          }
          i++;
        }
        if(CoreUtils.DEBUG_LEVEL>1)
          RLogger.logInfo("After change set  on "+ResManager.getManager().getName(TreeDB.this));
        if (lastfocus != null) retevents.add(lastfocus);
        if (retevents.size() > 0)
          for (ITreeChangeListener l : getGUIListeners()) {
            l.changed(retevents);
          }
      }
    }
    finally{
      wl.unlock();
    }
  }

  private void clearRemPool() {
    while (!rempool.isEmpty()) {
      TreeNode node = rempool.pop();
      if (!node.isDisposed()) node.delete();
    }
  }

  /**
   * Get specific object delegated for different checks like check on parent-child relations and
   * element identifiers check.
   *
   * @return TreeLogic object
   */
  public TreeLogic getTreeLogic() {
    if (tlogic == null) tlogic = new TreeLogic();
    return tlogic;
  }

  private TreeDB(int loa2) {
    latechangelisteners = new HashMap<>();
    this.loa = loa2;
  }
  public void initRootTypes(){
    NodeType dbroot = new NodeType("DBRoot", TreeDB.this);
    dbroot.setClass(com.unitesk.requality.core.TreeNode.class);
    defaultTypes.put("DBRoot", dbroot);
    initNodeAttributes(dbroot,dbroot.attrs().getChanged().getNewMap());
    
    NodeType nodetype = new NodeType("NodeType", TreeDB.this);
    nodetype.setClass(com.unitesk.requality.nodetypes.NodeType.class);
    defaultTypes.put("NodeType", nodetype);
    initNodeAttributes(nodetype,nodetype.attrs().getChanged().getNewMap());
    
    NodeType vnodetype = new NodeType("VirtualNode", TreeDB.this);
    vnodetype.setClass(com.unitesk.requality.nodetypes.VirtualNode.class);
    defaultTypes.put(VirtualNode.TYPE_NAME, vnodetype);
    initNodeAttributes(vnodetype,vnodetype.attrs().getChanged().getNewMap());
    
    NodeType req = new NodeType("Requirement", TreeDB.this);
    req.setClass(com.unitesk.requality.nodetypes.Requirement.class);
    defaultTypes.put(Requirement.TYPE_NAME, req);
    initNodeAttributes(req,req.attrs().getChanged().getNewMap());
    
    NodeType node = new NodeType("Node", TreeDB.this);
    node.setClass(com.unitesk.requality.core.TreeNode.class);
    defaultTypes.put("Node", node);
    initNodeAttributes(node,node.attrs().getChanged().getNewMap());

  }

  public TreeNode createChildNode(UUID pUUId, String id, String type) {
    return createChildNode(pUUId, id, type, EMPTY_TEMPLATE);
  }

  /**
   * Method to create child element for element with pUUId identifier.
   *
   * @param pUUId parent UUID for element
   * @param id identifier of element, needs to comply with TreeLogic.isChildIdValidExc
   * @param type type of element, @see {@link TreeNode.getType}
   * @param templ
   * @return created Element or null
   */
  public TreeNode createChildNode(UUID pUUId, String id, String type, String templ) {
    UUID uuid = uuidGenerator.nextUUId();
    return createChildNode(uuid, pUUId, id, type, templ, false);
  }

  public TreeNode createChildNode(
      UUID uuid, UUID pUUId, String id, String type, boolean ignoreTemplate) {
    return createChildNode(uuid, pUUId, id, type, EMPTY_TEMPLATE, ignoreTemplate);
  }

  /**
   * Method to create child element for element with pUUId identifier.
   *
   * @param uuid UUID of element
   * @param pUUId parent UUID for created element
   * @param id identifier of element, needs to comply with TreeLogic.isChildIdValidExc
   * @param type type of element, @see {@link TreeNode.getType}
   * @return created Element or null
   */
  public TreeNode createChildNode(
      UUID uuid, UUID pUUId, String id, String type, String templ, boolean ignoreTemplate) {
    if (getNode(pUUId).findWorkChild(id) != null) {
      throw new IllegalArgumentException(
          "Child with id: \"" + id + "\" exists in " + getNode(pUUId).getQualifiedId());
    }
    if (type == null) RLogger.logInfo("null type");
    storage.createNode(pUUId, uuid, type, id);
    TreeNode node = getNode(uuid), rnode = getRefTarget();
    Map<String,AttributeDefinition> map = 
        getAttributeRules().getRequiredAttrRule().getMap().get(type);
    boolean changed = false;
    if(map!=null)
      for(Entry<String, AttributeDefinition> ev:map.entrySet()){
        AttributeDefinition ad = ev.getValue();
        node.attrs().setAttribute(new AttributeDefinition(node.getUUId(), 
            ev.getKey(), AttributeDefinition.detectValueProvider(ev.getKey(), 
                ad.getType(), ad.getRawValue().getValue()), ad.getFlags()));
        changed = true;
      }
    if (getRefTargetState(rnode) == 0) rnode = null;
    TreeNode template = null;
    if (templ != null) {
      template = getTemplate(templ);
    } else template = getActiveTemplate(type);
    if (node != null && template != null && !ignoreTemplate) {
      for (Attribute a : template.attrs()) {
        //TODO check textNode types
        if (node.attrs().getLocalAttributeKeys().contains(a.getKey())) continue;
        changed = true;
        //if (a.getOrigin().equals(template.getUUId()))
          node.attrs().setAttribute(new AttributeDefinition(node.getUUId(), a.getKey(),a.getAvailability(),
          a.getDefinition().getValueProvider(),
          a.getDefinition().getFlags()));
          //.storeAttribute(a.copy(node.getUUId(), node.getUUId()));
       // else node.attrs().storeAttribute(a.copy(node.getUUId()));
      }
    }
    if (node != null && Requirement.TYPE_NAME.equals(node.getType()) && rnode != null) {
      changed = true;
      node.attrs().setAttribute(new AttributeDefinition(node.getUUId(), 
          getDefReferenceName(), new ReferenceValueProvider(AttributeType.REFERENCE,
              rnode.getUserVisibleName(true)),AttributeDefinition.detectFlags(getDefReferenceName())));
      //node.attrs().getLinkAttributeValue(getDefReferenceName());
      if (getRefTargetState(rnode) == 1) {
        setRefTarget(null);
        Map<String, Attribute> old = new HashMap<>();
        old.put(getDefReferenceName(), rnode.attrs().getAttribute(getDefReferenceName()));
        ChangesSet<Attribute> cs = new ChangesSet<>();
        cs.addSomeChange(getDefReferenceName(),old.get(getDefReferenceName()),rnode.attrs().getAttribute(getDefReferenceName()));
        fireNodeGUIChange(
            rnode,
            cs,
            true);
      }
    }
    if (node!=null && changed) node.commitAttributes();
    return node;
  }

  public TreeNode createChildNode(UUID uuid, UUID pUUId, String id, Map<String, AttributeDefinition> attrs) {
    if (getNode(pUUId).findWorkChild(id) != null) {
      throw new IllegalArgumentException(
          "Child with id: \"" + id + "\" exists in " + getNode(pUUId).getQualifiedId());
    }
    String type = null;
    try {
      type = attrs.get(TreeNode.ATTR_TYPE).getRawValue().getValue().toString();
    } catch (Exception e) {
    }
    if (type == null) RLogger.logInfo("null type");
    storage.createNode(pUUId, uuid, id, attrs);
    TreeNode node = getNode(uuid), rnode = getRefTarget();
    if (getRefTargetState(rnode) == 0) rnode = null;
    if (node != null && Requirement.TYPE_NAME.equals(node.getType()) && rnode != null) {
      node.attrs().setAttribute(new AttributeDefinition(
          node.getUUId(),getDefReferenceName(),
          new ReferenceValueProvider(AttributeType.REFERENCE,rnode.getUserVisibleName(true)),
          AttributeDefinition.detectFlags(getDefReferenceName())));
      node.attrs().getLinkAttributeValue(getDefReferenceName());
      if (getRefTargetState(rnode) == 1) {
        setRefTarget(null);
        Map<String, Attribute> old = new HashMap<>();
        old.put(getDefReferenceName(), rnode.attrs().getAttribute(getDefReferenceName()));
        ChangesSet<Attribute> cs = new ChangesSet<>();
        cs.addSomeChange(getDefReferenceName(),null,old.get(getDefReferenceName()));
        fireNodeGUIChange(
            rnode,
            cs,
            true);
      }
      node.commitAttributes();
    }
      
    return node;
  }

  /**
   * Recursively remove element and all its descendants. Child elements are removed first.
   *
   * @param node element to remove
   */
  public void removeNode(TreeNode node) {
    if (node.isVirtual() && !isVirtual) {
      TreeDB db = Instantiator.getExistVirtualTreeDB(this, node.getVRoot());
      if (db != this) {
        db.removeNode(node);
        return;
      }
    }
    startChangeRow();
    try{
      storage.remove(node.getUUId());
    }
    catch(Exception e){
      e.printStackTrace();
    }
    finally{
      stopChangeRow();
    }
  }

  /**
   * Move element from one parent to another
   *
   * @param node node to move
   * @param newParent new parent of element
   */
  public void moveNode(TreeNode node, TreeNode newParent) {
    String ret = storage.getCacheStorage().startCacheTransaction("Move node",CacheStorage.COMMIT_SINGLE);
    storage.moveNodeTo(node.getUUId(), newParent.getUUId());
    if(ret == null)
      storage.getCacheStorage().commitCache();
  }

  /**
   * Set identifier of node to id
   *
   * @param node element to set new identifier
   * @param id new identifier
   */
  void setId(TreeNode node, String id) {
    if (!node.getId().equals(id)) {
      storage.changeNodeId(node.getUUId(), id);
      // if(isVirtual){
      // Attribute ca = node.attrs().getLocalAttribute(TreeInstancer.ATTR_CLONE_ID);
      // if(ca.getOrigin().equals(node.getUUId())){
      // if(!ca.getValue().toString().equals(node.getCloneParent().getId())){
      // node.attrs().setAttribute(TreeInstancer.ATTR_CLONE_ID, id);
      // node.attrs().saveAttributes();
      // }
      // }
      // }
      //
    }
  }
  public UUID getRootNodeUUID() {
    return root!=null?root.getUUId():storage!=null?storage.getRootNode():null;
  }
  
  public TreeNode getCachedRootNode() {
    return root;
  }
  /** @return root node of TreeDB */
  public TreeNode getRootNode() {
    if (root == null) {
      UUID rootUUID;
      if ((rootUUID = storage.getRootNode()) == null) { // root doesn't
        // exists
        return null;
      }
      try{
    	  root = getNode(rootUUID, DEFAULT_DEPTH, false);
      }catch(NodeNotFoundException nne){
    	  RLogger.logInfo(CoreUtils.getTrace(nne));
    	  return null;
      }
    }
    return getCachedRootNode();
  }

  /**
   * Method to update cached requirement name.
   *
   * @param node requirement to update name
   */
  public void updateRequirementName(TreeNode node) {
    // if(!(node instanceof Requirement))
    // return;
    // Requirement r = (Requirement)node;
    // if(getLinkResolver().hasRevNameCache(node.getUUId()))
    getLinkResolver().updateName(node.getUUId(), node.getName());
  }

  public TreeNode getNode(UUID uuid){
    try{
      return getNode(uuid, DEFAULT_DEPTH, false);
    }
    catch(NodeNotFoundException nnfe){
      RLogger.logInfo(CoreUtils.getTrace(nnfe));
      return null;
    }
  }
  
  public boolean isNodeExist(UUID uid){
    TreeNode node = getCachedNode(uid);
    if(node!=null)
      return !node.isDisposed();
    return storage.canLoad(uid);
  }
  
  public TreeNode getNodeExc(UUID uuid) throws NodeNotFoundException{
    return getNode(uuid, DEFAULT_DEPTH, false);
  }

  public TreeNode getCachedNodeOrVirt(UUID uuid) {
    return getCachedNodeOrVirt(uuid, DEFAULT_DEPTH);
  }

  public TreeNode getCachedNodeOrVirt(UUID uuid, int deep) {
    TreeNode node = cache.get(uuid);
    if (node != null) {
      return node;
    }
    if (isVirtual && deep > 0)
      return Instantiator.getRootTree(this).getCachedNodeOrVirt(uuid, deep - 1);
    if (!isVirtual && deep > 0) { // TODO another dbs
      for (TreeDB db : Instantiator.getVirtualTreeDBs(this))
        // TreeDB db =
        // Instantiator.getVirtualTreeDB(this,Instantiator.DEFAULT_INSTANCE_NAME);
        if (db != null && db != this) {
          TreeNode nde = db.getCachedNodeOrVirt(uuid, deep - 1);
          if (nde != null) return nde;
        }
    }
    return null;
  }

  public TreeNode getCachedNode(UUID uuid) {
    TreeNode node = cache.get(uuid);
    if (node != null) {
      return node;
    }
    return null;
  }

  // Condition cacheCond = cacheLock.newCondition();
  @SuppressWarnings("unchecked")
  /**
   * Method to get element by UUID
   *
   * @param uuid element identifier
   * @return element with given uuid
   */
  public TreeNode getNode(UUID uuid, int depth, boolean notcheck) throws NodeNotFoundException{

    if (uuid == null) return null;
    TreeNode node = getCachedNodeOrVirt(uuid, DEFAULT_DEPTH);
    if (node != null) {
      return node;
    }
    if (root != null && root.getUUId().equals(uuid)) return root;
    boolean isClone = false, canLoad = false;
    Map<String, AttributeDefinition> attrs = null;
    boolean isroot = false;
    String type = null;
    if (CoreUtils.DEBUG_LEVEL > 1) {
      RLogger.logInfo("Loading " + uuid);
    }
    long l = System.currentTimeMillis();
    TreeDB db = this;
    Object[] tl = testLoad(storage, uuid, depth);
    canLoad = tl != null;
    if (tl != null) {//to avoid warn
      attrs = (Map<String, AttributeDefinition>) tl[0];
      canLoad = (boolean) tl[1];
      type = (String) tl[2];
      isroot = (boolean) tl[3];
      isClone = (boolean) tl[4];
    }
    if (CoreUtils.DEBUG_LEVEL > 1) {
      RLogger.logInfo(uuid + " Determine params " + (System.currentTimeMillis() - l));
      l = System.currentTimeMillis();
    }
    if (!isroot) isroot = db.getRootNodeUUID().equals(uuid);
    // if(isVirtual && isClone && first)
    // return VirtualNode.getBaseTree(this).getNode(uuid,false);//single
    // deep into
    if (!canLoad) {
      db = null;
      if (depth > 0) {
        if (depth > 0 && !isVirtual()) {
          if (uuid.equals(storage.getRootNode())) {
            RLogger.logInfo("Root node damaged! on " + getName());
            return null; // root damaged!
          }
          if (attrs == null
              || attrs.get(TreeNode.ATTR_VIRT_STATUS) == null
              || attrs.get(TreeNode.ATTR_VIRT_STATUS) == null)
            db = this; // Instantiator.getVirtualTreeDB(this,Instantiator.DEFAULT_INSTANCE_NAME);
          else
            db =
                Instantiator.getExistVirtualTreeDB(
                    this,
                    UUID.fromString(attrs.get(TreeNode.ATTR_VIRT_STATUS).getRawValue().getValue().toString()));

          // if(db!=null && db!=this)
          // return db.getNode(uuid,depth-1,notcheck);
        } else if (isVirtual()) {
          db = Instantiator.getBaseTree(TreeDB.this);
          // if(db!=null && db!=this)
          // return db.getNode(uuid,depth-1,notcheck);
        }
      }
    }
    if (db != null && db != this) {
      tl = db.testLoad(storage, uuid, depth);
      if (tl == null) canLoad = false;
      else {
        attrs = (Map<String, AttributeDefinition>) tl[0];
        canLoad = (boolean) tl[1];
        type = (String) tl[2];
        isroot = (boolean) tl[3];
        isClone = (boolean) tl[4];
        if (!isroot) isroot = db.getRootNodeUUID().equals(uuid);
      }
    }
    if (!canLoad) canLoad = isClone || isroot;
    if (CoreUtils.DEBUG_LEVEL > 1) {
      RLogger.logInfo(uuid + " Get proper tree " + (System.currentTimeMillis() - l));
      l = System.currentTimeMillis();
    }
    if (db == null && canLoad) db = this;
    if (db != null && canLoad) {
      if (attrs == null) throw new RuntimeException("Null attributes!");
      node = db.getCache().get(uuid);
      if (node != null) return node;
      // synchronized (cacheLocks){
      ReentrantLock lock = db.cacheLocks.get(uuid);
      if (lock != null) {

        int tl1 = 100;
        while (tl1 > 0) {
          node = db.getCachedNodeOrVirt(uuid);
          if (lock.isLocked())
            try {
              Thread.currentThread().join(10);
            } catch (InterruptedException e) {
              // TODO Auto-generated catch block
              e.printStackTrace();
            }
          tl1--;
          if (node != null) return node;
        }
        return null;
      } else {
        synchronized (db.cacheLocks) {
          if(cacheLocks.containsKey(uuid)){
            try {
              db.cacheLocks.get(uuid).wait();
            } catch (InterruptedException e) {
              // TODO Auto-generated catch block
              e.printStackTrace();
            }
            return db.cache.get(uuid);
          }
          ReentrantLock rl = new ReentrantLock();
          db.cacheLocks.put(uuid, rl);
          db.cacheThreads.put(uuid,Thread.currentThread());
          boolean locked = false;
          while (!locked) {
            try {
              locked = rl.tryLock(100, TimeUnit.MILLISECONDS);
            } catch (InterruptedException e) {
              RLogger.logInfo("interrupted!"+ uuid );
              if(CoreUtils.DEBUG_LEVEL>0)
                e.printStackTrace();
            }
            if (locked) {
              node = db.cache.get(uuid);
              if (node != null) {
                rl.unlock();
                //db.cacheLocks.get(uuid).unlock();
                if(rl == db.cacheLocks.get(uuid)){
                  db.cacheLocks.remove(uuid);
                  db.cacheThreads.remove(uuid);
                }
                return node;
              }
              locked = true;
            }else{
              if(CoreUtils.DEBUG_LEVEL>0){
                System.out.println("T1:" );
                System.out.println(Thread.currentThread().getId()+" "+Thread.currentThread().isAlive()+" "+Thread.currentThread().isInterrupted() );
                print(Thread.currentThread().getStackTrace());
                System.out.println("T2:" );
                System.out.println(db.cacheThreads.get(uuid).getId()+" "+db.cacheThreads.get(uuid).isAlive()+" "+db.cacheThreads.get(uuid).isInterrupted() ); 
                print(db.cacheThreads.get(uuid).getStackTrace());
                if (db.cacheLocks.get(uuid).isHeldByCurrentThread())
                  System.out.println(1 );
                System.out.println(db.cacheThreads.get(uuid).getState());
              }
            }
          }
        }
      }
      // code below was written to deal with the problem of
      // loops in some case and to add change listeners
      // for linked Nodes.
      AttributeStorage as = AttributeStorage.getCachedStorage(db, uuid);
      TreeNode n = null;
      if (CoreUtils.DEBUG_LEVEL > 1) {
        RLogger.logInfo(uuid + " preload " + (System.currentTimeMillis() - l));
        l = System.currentTimeMillis();
      }
      if (as != null && as.getNode() != null) {
        n = as.getNode();
      } else {
        if (db.loading.contains(uuid)) {
          return db.cache.get(uuid);
        }
        db.loading.add(uuid);
        //db.updateParentCloneParentAttributeDefs(uuid, null, null, db.storage.getParent(uuid), attrs);
        if (CoreUtils.DEBUG_LEVEL > 1) {
          RLogger.logInfo(uuid + " attrs upped " + (System.currentTimeMillis() - l));
          l = System.currentTimeMillis();
        }
        n = db.createFromStorage(uuid, type, attrs);
        if (CoreUtils.DEBUG_LEVEL > 1) {
          RLogger.logInfo(uuid + " item created " + (System.currentTimeMillis() - l));
          l = System.currentTimeMillis();
        }
      }
      if ("/".equals(n.getId())) {
        assert n.parent == null;
        db.root = n;
      }
      //n.attrs().incVersions(attrs.keySet(), attrs, n.getVS()); // TODO look
      n.getVS().saveVers();
      // into
      if (CoreUtils.DEBUG_LEVEL > 1) {
        RLogger.logInfo(uuid + " inc versies " + (System.currentTimeMillis() - l));
        l = System.currentTimeMillis();
      }
      // fix for event loop
      if (db.loading.contains(uuid)) db.loading.remove(uuid);
      db.cache.put(n.getUUId(), n);
      // update of name cache for Requirements
      // Using for UFName supporting.
      // if(n instanceof Requirement)
      db.updateRequirementName(n);
      db.cacheLocks.get(uuid).unlock();
      db.cacheLocks.remove(uuid);
      // Get links if they are contained in
      if (CoreUtils.DEBUG_LEVEL > 1) {
        RLogger.logInfo(uuid + " up names " + (System.currentTimeMillis() - l));
        l = System.currentTimeMillis();
      }
      // current node
      // n.updateGenerators();
      Attribute a = n.attrs().getAttribute(TreeNode.ATTR_LINKS);
      if (a != null) {
        ArrayList<Object> links = (ArrayList<Object>) ((ListValue)a.getValue()).getValues();
        for (Object linko : links) {
          String link = linko.toString();
          // Link with * - hidden links
          // we need to deal with them in
          // some situations.
          if (link.startsWith("*")) link = link.substring(1);
          // link contains UUID, so we use it here
          UUID uid = UUID.fromString(link.substring(0, link.indexOf('_')));
          INodeChangeListener ncl = n.getChangeListener();
          if (ncl == null) continue;
          if (!n.getUUId().equals(uid)) {
            try {
              db.getNode(uid).addChangeListener(ncl);
            } catch (RuntimeException re) {
              db.latechangelisteners.put(uid, ncl);
              // Node not found
            }
          } else {
            n.addChangeListener(ncl);
          }
        }
      }
      if (CoreUtils.DEBUG_LEVEL > 1) {
        RLogger.logInfo(uuid + " up links " + (System.currentTimeMillis() - l));
        l = System.currentTimeMillis();
      }
      // updateGenerators(n);
      if (db.latechangelisteners.containsKey(uuid)) {
        n.addChangeListener(db.latechangelisteners.get(uuid));
        db.latechangelisteners.remove(uuid);
      }
      if (!notcheck && !db.getCheckerContainer().isChecked(n)) {
        IMark[] cp = db.getCheckerContainer().check(n);
        db.getChprob().put(uuid, cp);
      }
      if (n instanceof ITreeChangeListener) db.addListener((ITreeChangeListener) n);
      if (n instanceof INodeChangeListener) n.addChangeListener((INodeChangeListener) n);
    // t = System.currentTimeMillis()-t;
      // if(t>300)
      // RLogger.logInfo("anomaly!");
      // RLogger.logInfo("Load node time: "+(t));
      n.loaded = true;

      return n;
      // }
    }
    /*
     * if(getStorage() instanceof TreeInstancer){ TreeDB db =
     * VirtualNode.getBaseTree(this); if(db == null) //TODO ups throw new
     * RuntimeException("node not found: " + uuid);
     * if(db.getRootNodeUUID().equals(uuid)) return
     * db.root!=null?db.getRootNode():db.root;
     * if(!db.isNodeExist(uuid)) throw new
     * RuntimeException("node not found: " + uuid); node = db.getNode(uuid);
     * if (node != null) return node; throw new
     * RuntimeException("node not found: " + uuid); } //one level deeper?
     * skip for now //TreeDB db = VirtualNode.getVirtualTreeDB(this);
     * //if(!db.equals(this)) // return db.getNode(uuid);
     */
    throw new NodeNotFoundException(uuid);
  }

  private void print(StackTraceElement[] trace) {
        System.out.println(this);
        for (StackTraceElement traceElement : trace)
            System.out.println("\tat " + traceElement);
  }

  private Object[] testLoad(ITreeStorage storage2, UUID uuid, int depth) {
    if (storage.canLoad(uuid)) {

      boolean isClone = false, canLoad = false;
      Map<String, AttributeDefinition> attrs = null;
      boolean isroot = false;
      String type = null;
      attrs = storage.getAttributes(uuid);
      Object ta = AttributeStorage.getAttributeDefValue(this, attrs, uuid, TreeNode.ATTR_TYPE);
      if (ta != null) type = ta.toString();
      if (type != null) {
        if (attrs != null) {
          canLoad = true;
          boolean virt = false;
          if (canLoad
              && attrs.containsKey(TreeNode.ATTR_VIRT_STATUS)
              && attrs.get(TreeNode.ATTR_VIRT_STATUS) != null) {
            Object val = attrs.get(TreeNode.ATTR_VIRT_STATUS).getRawValue().getValue();
            // TODO trans storage
            if (val != null) { // can be loaded only by virt. tree
              canLoad = isVirtual && getVStorage().getRootNode().toString().equals(val);
              isroot = uuid.toString().equals(val);
              virt = true;
              if (depth == 0) canLoad = true;
            }
          }
          if (!isroot && Instantiator.isDefVRoot(this, uuid, attrs)) {
            isroot = true;
            virt = true; // hack
          }
          if (!virt && isVirtual()) canLoad = false;
          // if(isroot&&!isVirtual()&&forceload)
          // canLoad=true;
          if (attrs.containsKey(TreeNode.ATTR_CLONE_PARENT)
              || attrs.containsKey("_cloneparent")) isClone = true;
        }
      }
      return new Object[] {attrs, canLoad, type, isroot, isClone};
    }
    return null;
  }

  /**
   * Method to get element by "QualifiedId" or "User-visible name" or "Full name"
   *
   * @param qId path to element
   * @return element if exists, null otherwise, can throw Exception
   */
  public TreeNode getNode(String qId) {
    return getNode(qId, Path.QID | Path.UV | Path.UUID);
  }

  /**
   * Method to get element by "QualifiedId" or "User-visible name" or "Full name". If rawpath
   * parameter is true then search ignores hide mechanics.
   *
   * @param qId path to element
   * @param rawPath if true, ignore hide mechanisms
   * @return element if exists, null otherwise, can throw Exception
   */
  public TreeNode getNode(String qId, int mode) {
    try{
      return getNodeExc(qId,mode);
    }
    catch(NodeNotFoundException nnfe){
      
    }
    return null;
  }
  public TreeNode getNodeExc(String qId, int mode) {
    int pos = qId.indexOf("::");
    if(pos>-1)
    {
      String dbname = qId.substring(0, pos);
      qId = qId.substring(pos+2);
      TreeDB fdb = ResManager.getManager().getTreeDB(dbname);
      if(fdb == null)
        return null;
      return fdb.getNode(qId);
    }
    List<TreeNode> res = getLinkResolver().resolvePath(null, null, this, qId, mode);
    if (res == null || res.size() == 0) return null;
    return res.iterator().next();
  }
  /**
   * Save attributes of element with uuid as attributes
   * @param uuid identifier of element
   */
  public void commitAttributes(TreeNode node, ChangesSet<Attribute> changes, boolean save) {
    // do not need to save if nothing changes
    if (changes.isEmpty())
      return;
    //TreeDB.addAttrsRec(node, changes.getAll(),changes, map);
    //Map<String, AttributeDefinition> old = storage.getAttributes(node.getUUId());
//    Map<String, Version> va = storage.getVersions(uuid);
    List<IAttributeChangePropagator> pps = new ArrayList<>();
    for(IAttributeChangePropagator pch : node.getAPropogators()){
      if(pch.shouldBeApplied(node, changes, Moment.BEFORE))
        pps.add(pch);
    }
    pps.sort(AttributePropagatorComparator.getOne());
    for(IAttributeChangePropagator pch:pps)
        pch.apply(node, changes);
    //Map<String,AttributeDefinition> oldMap = storage.getAttributes(uuid);
    /*boolean save = false;
    for(String s:changes.getAll())
      if((AttributeDefinition.detectFlags(s)&Flag.VIRTUAL)!=Flag.VIRTUAL)
        save = true;
    if(save)*/
    saveAttributes(node,changes,save);
    pps = new ArrayList<>();
    for(IAttributeChangePropagator pch : node.getAPropogators()){
      if(pch.shouldBeApplied(node, changes, Moment.AFTER))
        pps.add(pch);
    }
    
    pps.sort(AttributePropagatorComparator.getOne());
    for(IAttributeChangePropagator pch:pps)
        pch.apply(node, changes);

  }
  public void saveAttributes(TreeNode node, ChangesSet<Attribute> changes, boolean save){
    handleTreeNodeChanges(node, changes, Moment.BEFORE, true);
    if(!changes.isEmpty()&&save){
      if(getAttributeRules().matchAChanges(node.getType(), changes))
      {
        String val = "1*";
        Attribute old = node.attrs().getAttribute(AttributesVersioningRule.ATTR_NODE_VERSION);
        if(old!=null)
        {
          String s = old.getGValue();
          if(s!=null && !s.endsWith("*"))
            val = Long.parseLong(s)+1+"*";
        }
        String key = AttributesVersioningRule.ATTR_NODE_VERSION;
        Attribute newa = new Attribute(
            new AttributeDefinition(node.getUUId(), key, 
                new ConstValueProvider(AttributeType.STRING, val)),node.attrs());
        node.attrs().setAttribute(newa);
        changes.addSomeChange(AttributesVersioningRule.ATTR_NODE_VERSION, 
            old, 
            newa);
      }
      storage.saveAttributes(node.getUUId(), changes);
    }else
      System.out.println( "null ch" ); 
    handleTreeNodeChanges(node, changes, Moment.AFTER, true);
  }
  /**
   * Method to set content of resource with name 'name' on element node
   *
   * @param node element to set resource for
   * @param name name of resource
   * @param in content of resource
   * @param data
   * @deprecated
   */
  public void setResourceContent(TreeNode node, String name, InputStream in, Object data) {
    storage.setResourceContents(node.getUUId(), name, in, data);
    // update Requirement name
    // if(node.getType().equals(Requirement.TYPE_NAME))
    // updateRequirementName(node);
    // TODO possible error
  }

  /**
   * Delete resource on element node with name "name"
   *
   * @param node element from which resource will be deleted
   * @param name name of resource
   */
  public void deleteResource(TreeNode node, String name) {
    storage.deleteResource(node.getUUId(), name);
  }

  /**
   * Method to check existence of resource with name "name" on element node
   *
   * @param node element where resource can be
   * @param name name of resource to check
   * @return true if resource exists, false otherwise
   */
  public boolean isResourceExists(TreeNode node, String name) {
    return storage.isResourceExists(node.getUUId(), name);
  }

  /**
   * Get name of linked to treeDB Storage
   *
   * @return name of storage
   */
  public String getName() {
    // TODO replace by ResManager.getManager().getName(this)
    return storage.getName();
  }

  public String getResName() {
    String rn = ResManager.getManager().getName(this);
    if (rn != null) return rn; 
    return null;
  }

  // //////////////////////////////////////////////////////////////////////////
  //
  // Transaction API methods
  //
  // //////////////////////////////////////////////////////////////////////////
  public boolean startTransaction(String name) {
    return startTransaction(name, false);
  }

  /**
   * Method to start transaction - cancelable set of changes. Data storage will be replaces by
   * Transaction Storage. All changes will be applied on main storage only after commit.
   *
   * @param name name of transaction to show on undo\redo mechanics
   */
  public boolean startTransaction(String name, boolean check) {
    if (isVirtual && Instantiator.getRootTree(this) != this) {
      return Instantiator.getRootTree(this).startTransaction(name, check);
    }
    if (inTransaction()) {
      if (check) return false;
      throw new IllegalStateException("transaction \"" + getName() + "\"is started");
    }
    if(CoreUtils.DEBUG_LEVEL>1)
      System.out.println("\t Start transaction "+ResManager.getManager().getName(this)+" "+name);
    // saveLateChanges(null);
    startChangeRow();
    TreeInstancer vstorage = null;
    if (!isVirtual)
      vstorage = Instantiator.getVirtualStorage(this, Instantiator.DEFAULT_INSTANCE_NAME);
    TransactionStorage tStorage =
        new TransactionStorage(storage, vstorage, getTransactionListeners());
    tStorage.setName(name);
    tStorage.setDeltaTarget(storageDeltaListener);
    storage.setDeltaTarget(nullDeltaListener);
    // if(vstorage!=null)
    // vstorage.setDeltaTarget(nullDeltaListener);
    storage = tStorage;
    // if(vstorage!=null){
    // TreeDB vdb =
    // Instantiator.getVirtualTreeDB(this,Instantiator.DEFAULT_INSTANCE_NAME);
    // vdb.startChangeRow();
    // vdb.storage = tStorage;
    // }
    return true;
  }

  /**
   * Method to end transaction and apply changes on main storage.
   *
   * @return temporaty storage
   */
  
  public TransactionStorage commit() {
    if (isVirtual && Instantiator.getRootTree(this) != this)
      return Instantiator.getRootTree(this).commit();
    try{
      if (!inTransaction()) {
        throw new IllegalStateException("transaction not started");
      }
      if(CoreUtils.DEBUG_LEVEL>1)
        System.out.println("\t Stop transaction "+ResManager.getManager().getName(this));
      // saveLateChanges(null);
      
      
      while(!que.isEmpty()){
        Set<Future<?>> toq = new HashSet<>(que);
        Iterator<Future<?>> it = toq.iterator();
        while(it.hasNext()){
          Future<?> r = it.next();
          try {
            if(!r.isCancelled()&&!r.isDone())
              r.get();
          } catch (InterruptedException e) {
            e.printStackTrace();
          } catch (ExecutionException e) {
            e.printStackTrace();
          }
          it.remove();
          que.remove();
        }
      }
      
      boolean isl = isLoaded();
      TransactionStorage tStorage = (TransactionStorage) storage;
      setIsLoaded(false);
      tStorage.apply();
      // tStorage.apply();
      tStorage.setDeltaTarget(nullDeltaListener);
      // TreeDB vdb =
      // Instantiator.getVirtualTreeDB(this,Instantiator.DEFAULT_INSTANCE_NAME);
      storage = tStorage.getStorage();
      // if(vdb!=null)
      // vdb.storage = tStorage.getVStorage();
      // if(vdb!=null)
      // vdb.stopChangeRow();
      setIsLoaded(isl);
      storage.setDeltaTarget(storageDeltaListener);
      // if(vdb!=null)
      // vdb.storage.setDeltaTarget(vdb.getDeltaTarget());
      return tStorage;
    }
    finally{
      stopChangeRow();
    }
  }

  /**
   * Cancel all changes after startTransaction. If transaction is not started will throw Exception
   */
  public void rollback() {
    if (isVirtual) {
      Instantiator.getRootTree(this).rollback();
      return;
    }
    try{
      while(!que.isEmpty()){
        Set<Future<?>> toq = new HashSet<>(que);
        Iterator<Future<?>> it = toq.iterator();
        while(it.hasNext()){
          Future<?> r = it.next();
          if(!r.isCancelled()&&!r.isDone())
            r.cancel(true);
          it.remove();
          que.remove();
        }
      }
      if (!(storage instanceof TransactionStorage)) {
        throw new IllegalStateException("transaction not started");
      }
  
      if(CoreUtils.DEBUG_LEVEL>1)
        System.out.println("\t Rollback "+ResManager.getManager().getName(this));
      // saveLateChanges(null);
  
      TransactionStorage tStorage = (TransactionStorage) storage;
      tStorage.setDeltaTarget(nullDeltaListener);
      tStorage.rollback(tStorage);
      // TreeDB vdb =
      // Instantiator.getVirtualTreeDB(this,DEFAULT_REFERENCE_NAME);
      storage = tStorage.getStorage();
      // vdb.storage = tStorage.getVStorage();
      // vdb.stopChangeRow();
      storage.setDeltaTarget(storageDeltaListener);
    }finally{
      stopChangeRow();
    }
    // vdb.storage.setDeltaTarget(vdb.getDeltaTarget());
    // for (ITransactionListener tl : transactionListeners) {
    // tl.undo(tStorage);
    // }
  }

  /**
   * Add listener on transactions events
   *
   * @param l listener to add
   */
  public void addTransactionListener(ITransactionListener l) {
    TreeDB root = Instantiator.getRootTree(this);
    if (root.equals(this)) transactionListeners.add(l);
    else root.addTransactionListener(l);
  }

  /**
   * Remove listener from transaction listeners
   *
   * @param l listener to remove
   */
  public void removeTransactionListener(ITransactionListener l) {
    TreeDB root = Instantiator.getRootTree(this);
    if (root.equals(this)) transactionListeners.remove(l);
    else root.removeTransactionListener(l);
  }

  // ////////////////////////////////////////////////////////////////////////
  //
  // Types operations
  //
  // ////////////////////////////////////////////////////////////////////////
  public void addNodeType(String name, Class<? extends TreeNode> clazz) {
    if (isVirtual) {
      // TreeDB db = Instantiator.getRootTree(this);
      // if(db!=null)
      // db.addNodeType(name,clazz);
    } else {
      NodeType nt = new NodeType(name, TreeDB.this);
      nt.setClass(clazz);
      defaultTypes.put(name, nt);
      initNodeAttributes(nt,nt.attrs().getChanged().getNewMap());
    }
  }

  public void addNodeType(Class<? extends TreeNode> clazz) {
    if (isVirtual) {
      TreeDB db = Instantiator.getRootTree(this);
      if (db != null && db != this) db.addNodeType(clazz);
    } else addNodeType(clazz.getSimpleName(), clazz);
  }

  public void addNodeTypes(Map<String, Class<? extends TreeNode>> nts) {
    if (isVirtual) {
      TreeDB db = Instantiator.getRootTree(this);
      if (db != null && db != this) db.addNodeTypes(nts);
    } else
      for (String name : nts.keySet()) {
        addNodeType(name, nts.get(name));
      }
  }

  /**
   * Method to get map of element types in form of pairs String type name -> NodeType
   *
   * @return available element types
   */
  public Map<String, NodeType> getTypes() {
    if (isVirtual) {
      TreeDB db = Instantiator.getRootTree(this);
      if (db != null && db != this) return db.getTypes();
    }
    Map<String, NodeType> result = new HashMap<>(defaultTypes);
    TreeNode typesNode = getNode("/Types", Path.QID);
    if (typesNode != null && typesNode.sizeWorkChildren() > 0) {
      int num = typesNode.sizeWorkChildren();
      for (int i = 0; i < num; i++) {
        NodeType node = (NodeType) typesNode.getWorkChild(i);
        result.put(node.getId(), node);
      }
    }
    return result;
  }

  public void createType(NodeType newType) {

    if (isVirtual) Instantiator.getRootTree(this).createType(newType);
    else {
      TreeNode types = getNode("/Types", Path.QID);
      if (types == null) {
        types = getRootNode().createChild("Types", NodeType.TYPE_NAME, null);
      }
      TreeNode newTypeNode = types.createChild(newType.getName());
      newTypeNode.commitAttributes();
    }
  }

  public NodeType getNodeType(String name) {
    if (isVirtual) {
      TreeDB db = Instantiator.getRootTree(this);
      if (db != null && db != this) return db.getNodeType(name);
    }
    for (NodeType type : getTypes().values()) {
      if (type.getId().equals(name)) {
        return type;
      }
    }
    return null;
  }

  public boolean removeType(String typeName) {
    if (isVirtual) {
      TreeDB db = Instantiator.getRootTree(this);
      if (db != null && db != this) return db.removeType(typeName);
    }
    TreeNode types = getNode("/Types", Path.QID);
    if (types == null) {
      return false;
    }
    TreeNode n = types.findWorkChild(typeName);
    n.delete();
    return true;
  }

  // ////////////////////////////////////////////////////////////////////////
  //
  // TreeDB factories
  //
  // ////////////////////////////////////////////////////////////////////////

  /**
   * Creates db structures and root node
   *
   * @param name
   * @param storage
   * @param uuidGenerator
   * @return
   */
  public static TreeDB create(String name, ITreeStorage storage, IUUIDGenerator uuidGenerator) {
    TreeDB newDB = new TreeDB(0);//, storage);
    newDB.uuidGenerator = uuidGenerator;
    storage.setDeltaTarget(newDB.nullDeltaListener);
    storage.createStorage(name);
    newDB.storage = storage;
    newDB.initRootTypes();
    // create root node
    UUID uuid = uuidGenerator.nextUUId();
    storage.createNode(null, uuid, "DBRoot", "/");
    newDB.tlogic = new TreeLogic();
    storage.setDeltaTarget(newDB.storageDeltaListener);
    newDB.setIsLoaded(true);
    return newDB;
  }

  /**
   * Creates db structures and root node
   *
   * @param name
   * @param storage
   * @return treeDB object
   */
  public static TreeDB create(String name, ITreeStorage storage) {
    return create(name, storage, new DefaultUUIDGenerator());
  }

  public static TreeDB load(ITreeStorage storage, IUUIDGenerator uuidGenerator) {
    return load(storage, uuidGenerator, 0, null);
  }

  public static TreeDB load(
      ITreeStorage storage, IUUIDGenerator uuidGenerator, int loa, TreeDB base) {
    TreeDB newDB = new TreeDB(loa);
    newDB.uuidGenerator = uuidGenerator;
    newDB.storage = storage;
    newDB.initRootTypes();
    // FIXME: for test we can use mock target, so we left old one if it is
    // not null. Its look like bad design.
    if (storage.getDeltaTarget() == null) {
      storage.setDeltaTarget(newDB.storageDeltaListener);
    }
    return newDB;
  }

  /** Update cache of requirements names. Locate requirements and save its names. */
  public void updateRequirementsNames(TreeNode node) {
    //TODO change to subscription 
    DeepFirstTreeWalker tw = new DeepFirstTreeWalker(node, false, false, true);
    Iterator<TreeNode> it = tw.iterator();
    while (it.hasNext()) {
      TreeNode tn = it.next();
      if (tn != null) updateRequirementName(tn);
    }
  }

  /*
   * public void updateRequirementNames(UUID node) { Collection<UUID> children
   * = storage.getChildren(node); for(UUID uid:children){ Collection c =
   * storage.getAttributes(uid); String name = null, type = "";
   * Iterator<Attribute> it = c.iterator(); while(it.hasNext()){ Attribute a =
   * it.next(); if(a.getKey().equals(TreeNode.ATTR_TYPE)) type =
   * (String)a.getValue(); if(a.getKey().equals(TreeNode.ATTR_NAME)) name =
   * (String)a.getValue(); } if(type.equals(Requirement.TYPE_NAME))
   * updateRequirementName(uid, name); updateRequirementNames(uid); } }
   */
  public static TreeDB load(ITreeStorage storage) {
    return load(storage, new DefaultUUIDGenerator());
  }

  // ////////////////////////////////////////////////////////////////////////
  /**
   * Add listener on element changes
   *
   * @param listener listener on element changes
   * @return true if listener is added, false otherwise
   */
  public boolean addListener(ITreeChangeListener listener) {
    if (listener == null) return false;
    if (isVirtual) {
      TreeDB db = Instantiator.getRootTree(this);
      if (db != null && db != this) return db.addListener(listener);
    }
    return listeners.add(listener);
  }

  /**
   * Remove listener on element changes
   *
   * @param listener listener to remove
   * @return true if listener is removed, false otherwise
   */
  public boolean removeListener(ITreeChangeListener listener) {
    if (listener == null) return false;
    if (isVirtual) {
      TreeDB db = Instantiator.getRootTree(this);
      if (db != null && db != this) return db.removeListener(listener);
    }
    return listeners.remove(listener);
  }

  /**
   * Get list of listeners on tree changes.
   *
   * @return list of listeners
   */
  public List<ITreeChangeListener> getListeners() {
    if (isVirtual) {
      TreeDB db = Instantiator.getRootTree(this);
      if (db != null && db != this) return db.getListeners();
    }
    return Collections.unmodifiableList(listeners);
  }

  public List<ITreeChangeListener> getGUIListeners() {
    return Collections.unmodifiableList(guilisteners);
  }

  // ////////////////////////////////////////////////////////////////////////
  public boolean addGUIListener(ITreeChangeListener listener) {
    if (listener == null) return false;
    return guilisteners.add(listener);
  }

  public boolean removeGUIListener(ITreeChangeListener listener) {
    if (listener == null) return false;
    return guilisteners.remove(listener);
  }
  
  private IStorageDeltaListener nullDeltaListener =
      new IStorageDeltaListener() {

        @Override
        public void startChangeRow() {
        }

        @Override
        public void stopChangeRow() {
          // TODO Auto-generated method stub

        }

        @Override
        public void created(UUID uuid, boolean forceInternal) {
          // TODO Auto-generated method stub

        }

        @Override
        public void deleted(UUID uuid, Moment m, boolean forceInternal) {
          // TODO Auto-generated method stub

        }
//
//        @Override
//        public void changed(
//            UUID uuid,
//            ChangesSet<Attribute> what,
//            Map<String, AttributeDefinition> newAttrs,
//            Moment m,
//            boolean forceInternal) {
//          // TODO Auto-generated method stub
//
//        }

        @Override
        public void resourceChanged(UUID uuid, String resName, Moment m, boolean forceInternal) {
          // TODO Auto-generated method stub

        }

        @Override
        public void moved(
            UUID uuid, UUID oldParent, UUID newParent, Moment m, boolean forceInternal) {
          // TODO Auto-generated method stub

        }

        @Override
        public void beforeCreated(
            UUID uuid, UUID par, Map<String,AttributeDefinition> attrs, boolean internal) {
          // TODO Auto-generated method stub

        }

        @Override
        public void versionChanged(
            UUID uuid,
            ChangesSet<Version> changes,
            Moment m, 
            boolean forceInternal) {
          // TODO Auto-generated method stub

        }

        @Override
        public boolean changeAttributes(
            UUID uuid, ChangesSet<Attribute> what) {
          return false;
        }
      };

  private void notifyExternalChange() {
    for (ITransactionListener tl : transactionListeners) {
      tl._dispose();
    }
  }

  private boolean isSameClone(TreeNode node, NodeDesc desc) {
    return isSameClone(node, desc.getAttributes());
  }

  private boolean isSameClone(TreeNode node, TreeNode desc) {
    return isSameClone(node, desc.attrs().getActiveAttributeDefs());
  }

  private boolean isSameClone(TreeNode node, Map<String, AttributeDefinition> atrs) {
    boolean descVirt =
        !atrs.containsKey(TreeNode.ATTR_VIRT_STATUS)
            ? false
            : atrs.get(TreeNode.ATTR_VIRT_STATUS) == null
                ? false
                : atrs.get(TreeNode.ATTR_VIRT_STATUS).getRawValue().getValue() != null;
    // TODO compare trees
    if (!descVirt) return false;
    String cloneid =
        !atrs.containsKey(TreeInstancer.ATTR_CLONE_ID)
            ? ""
            : atrs.get(TreeInstancer.ATTR_CLONE_ID) == null
                ? ""
                : getAttributeValue(node.getUUId(), atrs.get(TreeInstancer.ATTR_CLONE_ID), "", true).toString();
    String clonepar =
        !atrs.containsKey(TreeNode.ATTR_CLONE_PARENT)
            ? ""
            : atrs.get(TreeNode.ATTR_CLONE_PARENT) == null
                ? ""
                : getAttributeValue(node.getUUId(), atrs.get(TreeNode.ATTR_CLONE_PARENT), "", true)
                    .toString();
    // atrs.get(TreeInstancer.ATTR_CLONE_PARENT).iterator().next().getValue().toString();
    return (node.isVirtual()
        && isVirtual
        && cloneid.equals(node.attrs().getAttributeValue(TreeInstancer.ATTR_CLONE_ID, ""))
        && clonepar.equals(node.getCloneParentUUId().toString()));
  }
  private ChangesSet<Attribute> updateParentCloneParentAttributes(
      UUID nodeUUID, TreeNode node, UUID cp, UUID parent, Map<String, Attribute> attrs) {
    UUID cloneP = cp;
    if (isVirtual && cp == null) cloneP = TreeDB.this.getVStorage().getCloneParent(cp);
    Map<String, Attribute> cpa = null, pa = null;
    TreeNode cloneParent = null;
    if (cloneP == null) { // TODO mb remove
      Object cpo =
          AttributeStorage.getAttributeValue(
              TreeDB.this, attrs, nodeUUID, TreeNode.ATTR_CLONE_PARENT);
      if (cpo != null) {
        cloneP = UUID.fromString(cpo.toString());
      }
    }
    if (cloneP != null) {
      try {
        cloneParent = getNode(cloneP);
        if (cloneParent != null) {
          if (cloneParent.getVS().isCloneParentVersionOutdated(Type.ATTRIBUTES))
            cloneParent
                .getTreeDB()
                .updateParentCloneParentAttributes(
                    cloneParent.getUUId(),
                    cloneParent,
                    null,
                    cloneParent.getWorkParentUUID(),
                    cloneParent.attrs().getLocalAttributes());
        }
      } catch (Exception e) {
        e.printStackTrace();
      }
    }
    if (cloneParent != null
        && (node == null
            || VersionsStorage.isCloneParentVersionOutdated(
                Instantiator.getRootTree(this), Type.ATTRIBUTES, nodeUUID, cloneP))) {
      cpa = cloneParent.attrs().getLocalAttributes();
      if (node != null) node.attrs().getVersionsStorage().updateCloneParentVersion(Type.ATTRIBUTES);
    }
    if (attrs == null) RLogger.logInfo("null attributes! " + nodeUUID);
    if (attrs != null && (cpa != null)) {
      // VersionsStorage.initOrLoad(Instantiator.getRootTree(this).getStorage(),
      // node).incVersion(Type.ATTRIBUTES);
      Set<String> keys = new HashSet<>(attrs.keySet());
      //if (pa != null) keys.addAll(pa.keySet());
      keys.addAll(cpa.keySet());
      ChangesSet<Attribute> changes = new ChangesSet<>();
      for (String s : keys) {
        if (!s.startsWith("_%")) {
          Attribute old = attrs.get(s);
          attrs.put(
              s,
              AttributeStorage.mergeAttrs(
                  this,
                  nodeUUID,
                  new AttributeToMerge(
                      cpa.get(s),
                      MergeParams.ALL,
                      true),
//                  new AttributeDefToMerge(
//                      !s.startsWith("_%") && pa != null ? pa.get(s) : null,
//                      MergeParams.NOTLOCAL,
//                      true),
                  new AttributeToMerge(attrs.get(s), MergeParams.ALLHP, false)));
          if (attrs.get(s) == null) attrs.remove(s);
          changes.addSomeChange(s,old,attrs.get(s));
          //TODO change
        }
      }
      return changes;
    }
    return null;
  }
  private void updateParentCloneParentAttributeDefs(
      UUID nodeUUID, AttributeStorage as, UUID cp, UUID parent, Map<String, AttributeDefinition> attrs) {
    UUID cloneP = cp;
    if (isVirtual && cp == null) cloneP = TreeDB.this.getVStorage().getCloneParent(cp);
    Map<String, AttributeDefinition> cpa = null, pa = null;
    TreeNode cloneParent = null;
    if (cloneP == null) { // TODO mb remove
      Object cpo =
          AttributeStorage.getAttributeDefValue(
              TreeDB.this, attrs, nodeUUID, TreeNode.ATTR_CLONE_PARENT);
      if (cpo != null) {
        cloneP = UUID.fromString(cpo.toString());
      }
    }
    if (cloneP != null) {
      try {
        cloneParent = getNode(cloneP);
        if (cloneParent != null) {
          if (cloneParent.getVS().isCloneParentVersionOutdated(Type.ATTRIBUTES))
            cloneParent
                .getTreeDB()
                .updateParentCloneParentAttributeDefs(
                    cloneParent.getUUId(),
                    cloneParent.attrs(),
                    null,
                    cloneParent.getWorkParentUUID(),
                    cloneParent.attrs().getRawDMap());
        }
      } catch (Exception e) {
        e.printStackTrace();
      }
    }
    if (cloneParent != null
        && (as == null
            || VersionsStorage.isCloneParentVersionOutdated(
                Instantiator.getRootTree(this), Type.ATTRIBUTES, nodeUUID, cloneP))) {
      cpa = cloneParent.attrs().getActiveAttributeDefs();
      if (as != null) as.getVersionsStorage().updateCloneParentVersion(Type.ATTRIBUTES);
    }
    if(parent!=null){
      TreeNode par = getCachedNodeOrVirt(parent);
      if(par!=null)
        pa = par.attrs().getPropogatingAttributeDefs();
    }
    if (attrs == null) RLogger.logInfo("null attributes! " + nodeUUID);
    if (attrs != null && (cpa != null || pa != null)) {
      Set<String> keys = new HashSet<>(attrs.keySet());
      if (pa != null) keys.addAll(pa.keySet());
      if (cpa != null) keys.addAll(cpa.keySet());
      for (String s : keys) {
        if (!s.startsWith("_%")) {
          AttributeDefinition old = attrs.get(s);
          attrs.put(
              s,
              AttributeStorage.mergeAttrDefs(
                  this,
                  nodeUUID,
                  new AttributeDefToMerge(
                      !s.startsWith("_%") && cpa != null ? cpa.get(s) : null,
                      MergeParams.ALL,
                      true),
                  new AttributeDefToMerge(
                      !s.startsWith("_%") && pa != null ? pa.get(s) : null,
                      MergeParams.NOTLOCAL,
                      true),
                  new AttributeDefToMerge(attrs.get(s), MergeParams.ALLHP, false)));
          if (attrs.get(s) == null) attrs.remove(s);
        }
      }
    }
  }
//  private ICacheChangeListener cacheDeltaListener =
//    new ICacheChangeListener(){
//     private ICacheChangeListener getProperListener(
//            UUID uuid) {
//          TreeNode curNode = getCachedNodeOrVirt(uuid);
//          if (curNode!=null && !isVirtual&&curNode.isVirtual()) {
//            //if (attrs != null) {
//              Attribute la = curNode.attrs().getAttribute(TreeNode.ATTR_VIRT_STATUS);
//              if (la != null && la.getValue()!=null && la.getValue() != null
//                  || curNode.isVirtual()) {
//                boolean nvirt = false;
//                // if(attrs.containsKey(TreeNode.ATTR_VIRT_STATUS)&&!attrs.get(TreeNode.ATTR_VIRT_STATUS).isEmpty()){
//                Object val = getAttributeValue(uuid, la, null, false);
//                if (val == null) nvirt = true;
//                else {
//                  TreeDB db =
//                      Instantiator.getVirtualTreeDB(TreeDB.this, UUID.fromString(val.toString()));
//                  if (db != null && db != TreeDB.this) { // null if
//                    // root is
//                    // creating
//                    // right now
//                    return db.getCacheDeltaTarget();
//                  }
//                }
//                // }
//                if (nvirt) {
//                  nvirt = true;
//                  return TreeDB.this
//                      .getCacheDeltaTarget(); // Instantiator.getVirtualTreeDB(TreeDB.this,Instantiator.DEFAULT_INSTANCE_NAME).getDeltaTarget();
//                }
//              }
//           // }
//          }
//          return this;
//        }
//        @Override
//        public void cacheChanged(
//          UUID uuid,
//          String section,
//          Set<String> keys,
//          Map<String, Map<String, Object>> oldvs,
//          Map<String, Map<String, Object>> newvs,
//          boolean forceInternal) {
//            ICacheChangeListener lis = getProperListener(uuid);
//            if(lis!=this){
//              lis.cacheChanged(uuid, section, keys, oldvs, newvs, forceInternal);
//              return;
//            }
//            TreeNode node = getNode(uuid);
//            if (node != null) {
//              boolean changes = false;
//              for (String s : keys) {
//                Map<String,Object> newv = newvs != null ? newvs.get(s) : null,
//                    oldv = oldvs != null ? oldvs.get(s) : null;
//                if (newv!=null && !newv.equals(oldv)||newv==null&&oldv==null) {
//                  changes = true;
//                }
//              }
//              if (changes) {
//                startChangeRow();
//                try{
//                  for (ITreeChangeListener li : getListeners())
//                    li.cacheChanged(node, section, keys, oldvs, newvs, forceInternal);
//                  addGuiChange(
//                      new TreeNodeChangeEvent(
//                          TreeNodeChangeEventType.CACHE_CHANGE,
//                          node,
//                          section,
//                          keys,
//                          oldvs,
//                          newvs,
//                          true));
//                }finally{
//                  stopChangeRow();
//                }
//              }
//            }
//        }
//      };
  private IStorageDeltaListener storageDeltaListener =
      new IStorageDeltaListener() {
        private IStorageDeltaListener getProperListener(
            UUID uuid, String newId, Map<String, AttributeDefinition> attrs) {
          TreeNode curNode = getCachedNodeOrVirt(uuid);
          if (attrs == null) attrs = storage.getAttributes(uuid);
          if (!isVirtual) {
            if (attrs == null) attrs = storage.getAttributes(uuid);
            if (attrs != null) {
              AttributeDefinition la = attrs.get(TreeNode.ATTR_VIRT_STATUS);
              if (la != null && la.getRawValue()!=null && la.getRawValue().getValue() != null
                  || curNode != null && curNode.isVirtual()) {
                boolean nvirt = false;
                // if(attrs.containsKey(TreeNode.ATTR_VIRT_STATUS)&&!attrs.get(TreeNode.ATTR_VIRT_STATUS).isEmpty()){
                Object val = getAttributeValue(uuid, la, null, false);
                if (val == null) nvirt = true;
                else {
                  TreeDB db =
                      Instantiator.getExistVirtualTreeDB(TreeDB.this, UUID.fromString(val.toString()));
                  if (db != null && db != TreeDB.this) { // null if
                    // root is
                    // creating
                    // right now
                    return db.getDeltaTarget();
                  }
                }
                // }
                if (nvirt) {
                  nvirt = true;
                  return TreeDB.this
                      .getDeltaTarget(); // Instantiator.getVirtualTreeDB(TreeDB.this,Instantiator.DEFAULT_INSTANCE_NAME).getDeltaTarget();
                }
              }
            }
          }
          return this;
        }

        @Override
        public void beforeCreated(
            UUID uuid, UUID parent, Map<String,AttributeDefinition> attrs, boolean internal) {
          if(CoreUtils.DEBUG_LEVEL>1)
            RLogger.logInfo("Before creation of "+uuid+" on "+ResManager.getManager().getName(TreeDB.this)); 
          TreeNode curNode = getCachedNodeOrVirt(uuid);
          if (curNode != null){
            return;
          }
          String newId = storage.getNodeId(uuid);
          IStorageDeltaListener lis = getProperListener(uuid, newId, attrs);
          if (lis == null) return;
          if (lis != this) {
            lis.beforeCreated(uuid, parent, attrs, internal);
            return;
          }
          // saveLateChanges(getCachedNodeOrVirt(parent));
          startChangeRow();
          try{
            Object cp =
                AttributeStorage.getAttributeDefValue(
                    TreeDB.this, attrs, uuid, TreeNode.ATTR_CLONE_PARENT);
            UUID cpu = null;
            if (cp != null) cpu = UUID.fromString(cp.toString());
            updateParentCloneParentAttributeDefs(uuid, null, cpu, parent, attrs);
          }
          finally{
            stopChangeRow();
          }
          if(CoreUtils.DEBUG_LEVEL>1)
            RLogger.logInfo("Before creation of "+uuid+" on "+ResManager.getManager().getName(TreeDB.this)+" ended"); 

        }

        @Override
        public void created(UUID uuid, boolean forceInternal) {
          TreeNode curNode = getCachedNodeOrVirt(uuid);
          if (curNode != null) {
            curNode.isDisposed();
            return;
          }
          String newId = storage.getNodeId(uuid);

          Map<String, AttributeDefinition> satrs =
              new HashMap<>(storage.getAttributes(uuid));
          IStorageDeltaListener lis = getProperListener(uuid, newId, satrs);
          if (lis == null) return;
          if (lis != this) {
            lis.created(uuid, forceInternal);
            return;
          }
          NodeDesc newNodeDesc = loadNodeDescription(uuid, satrs, forceInternal);
          if (newNodeDesc == null) {
            newNodeDesc = loadNodeDescription(uuid, satrs, forceInternal);
            ;
            return;
          }
          curNode = cache.get(uuid);
          if (curNode != null) {
            return;
          }
          startChangeRow();
          try{
            TreeNode par = getNode(newNodeDesc.getWorkParentUUID());
            if (par != null)
              for (UUID chi : par.getActiveWorkChildrenUUIDs()) {
                TreeNode child = getNode(chi);
                if (child == null) {
                  par.removeChildUUID(chi);
                  if (!par.isVirtual()) RLogger.logInfo("null child in " + uuid);
                } else if (!child.getUUId().equals(newNodeDesc.getUUId())
                    && isSameClone(child, newNodeDesc)) {
                  deleted(chi, Moment.BEFORE, true);
                  deleted(chi, Moment.AFTER, true);
                }
              }
            curNode = getNode(uuid); 
            if (curNode == null) return;
            curNode.parent = newNodeDesc.getWorkParentUUID();
            TreeNode parent = curNode.getWorkParent();
            addGuiChange(
                new TreeNodeChangeEvent(
                    TreeNodeChangeEventType.CREATE, curNode, forceInternal)); // do
            // add
            // before
            // clone
            // upds
            if (parent != null && !parent.containsWorkChildUUID(uuid)) // test
              // here
              // for
              // stranges
              parent.addChildUUID(uuid);
            if (!forceInternal) notifyExternalChange();
            for (ITreeChangeListener l : getListeners()) {
              if (l != null)
                l.created(curNode, forceInternal);
            }

  
            // update Requirement name in cache
            
            if (curNode instanceof ITreeChangeListener) addListener((ITreeChangeListener) curNode);
            if (curNode instanceof INodeChangeListener)
              curNode.addChangeListener((INodeChangeListener) curNode);
            if (Requirement.TYPE_NAME.equals(curNode.getType())) {
              Location.markLocations(curNode, null);
              for (Location l : ((Requirement) curNode).getLocations()) {
                // l.updateRevLocations((Requirement) curNode);
                l.resetLabel(); // TODO
                l.commitAttributes();
              }
            }
  //          if (Location.TYPE_NAME.equals(curNode.getType())) {
  //            Location l = (Location) curNode;
  //            Requirement[] reqs = l.getRequirements();
  //            // l.resetLabel();
  //            if (reqs != null)
  //              for (Requirement r : reqs) for (Location lc : r.getLocations()) lc.resetLabel();
  //            l.resetLabel();
  //          }
            updateRequirementName(curNode);
          }
          finally{
            stopChangeRow();
          }
        }

        private NodeDesc loadNodeDescription(
            UUID uuid, Map<String, AttributeDefinition> attributes, boolean raw) {
          if (!storage.canLoad(uuid)) return null;
          String typeName = getNodeType(uuid);
          UUID parentUUId;
          String id;
          if ("DBRoot".equals(typeName)) {
            id = "/";
            parentUUId = null;
          } else {
            id = storage.getNodeId(uuid);
            parentUUId = storage.getParent(uuid);
          }
          List<UUID> childrenUUId = new ArrayList<>(storage.getChildren(uuid));
          UUID cp = null;
          // Version v = Version.getOne(),cpv=Version.getUnset();
          if (attributes.containsKey(TreeNode.ATTR_CLONE_PARENT))
            cp =
                UUID.fromString(
                    attributes.get(TreeNode.ATTR_CLONE_PARENT).getRawValue().getValue().toString());
          return new NodeDesc(
              TreeDB.this, uuid, cp, parentUUId, parentUUId, id, attributes, childrenUUId, false);
        }

        /** Reaction on element deletion */
        @Override
        public void deleted(UUID uuid, Moment m, boolean internal) {
          TreeNode node = cache.get(uuid);
          if(CoreUtils.DEBUG_LEVEL>1)
            RLogger.logInfo("TreeDB delete of "+uuid+" on "+ResManager.getManager().getName(TreeDB.this)+" "+m.toString()); 
          if (node == null) {
            TreeDB db = TreeDB.this;
            if (db.isVirtual) db = Instantiator.getRootTree(db);
            node = db.getCachedNodeOrVirt(uuid);
            if (node != null) {
              IStorageDeltaListener delta =
                  getProperListener(uuid, null, node.attrs().getActiveAttributeDefs());
              if (delta != null) {
                if (delta != this) {
                  delta.deleted(uuid, m, internal);
                  return;
                }
              } else return;
            }
          }
          if (node == null || node.getType() == null) {
            if (m == Moment.AFTER)
              if (node != null && node.getWorkParentUUID() != null)
                removeCache(uuid, node.getWorkParentUUID(), DEFAULT_DEPTH);
            return;
          }
          // if(m==Moment.BEFORE)
          // saveLateChanges(node);
          startChangeRow();
          try{
            TreeDB root = Instantiator.getRootTree(TreeDB.this);
            if (m == Moment.BEFORE) {
              node.setDisposed();
              TreeNode[] children = node.getWorkChildren(null, true, false, true, false);
              for (TreeNode child : children) child.delete(true);
              List<TreeNode> clones = Instantiator.getClones(TreeDB.this, node);
              for (TreeNode clone : clones) clone.delete(true);
              String tp = node.getType();

              TreeNode fnode = node;
              switch (tp) {
                case Requirement.TYPE_NAME:
                  {

                    List<Requirement> reqs = new ArrayList<>();
                    reqs.add((Requirement) fnode);
                    Map<Location,List<Requirement>> locs = new HashMap<>();
                    List<Location> llist = ((Requirement) fnode).getLocations();
                    for(Location s:llist)
                      locs.put(s, reqs);
                    if(llist.size()>0)
                      CommonRemoveLocation.getOne().removeLocations(locs, true);
                    
                  }
                  break;
                case Location.TYPE_NAME:
                  {
                    Requirement[] rqs = ((Location) fnode).getRequirements();
                    ArrayList<Requirement> reqs = (rqs != null) ?
                        reqs = new ArrayList<>(Arrays.asList(rqs)):null;
                    if(reqs!=null){

                      CommonRemoveLocation.getOne()
                          .removeRequirementLocation(
                              (Location) fnode, null, true, fnode.getUUId().toString());
                      CommonRemoveLocation.getOne()
                          .removeFragment(
                              (Location) fnode, null, true, fnode.getUUId().toString(), TreeDB.this);
                    }
                    
                  }
                  break;
              }

              for (ITreeChangeListener l : getListeners()) {
                l.deleted(node, internal, m);
              }
              // if(node.getCloneParent()!=null)
              // node.getCloneParent().removeClone(nodeUUID);
              // for(TreeNode chi:node.getActiveWorkChildren())
              // chi.delete();//delete childs if there is som
  
            } else {
              if (VirtualNode.TYPE_NAME.equals(node.getType())) {
                // TODO move
                if (Instantiator.revreuses.containsKey(node)) {
                  UUID uid = Instantiator.revreuses.remove(node);
                  Instantiator.reuses.get(uid).remove(node);
                  if (Instantiator.reuses.get(uid).isEmpty()) Instantiator.reuses.remove(uid);
                }
              }
              if (node instanceof ITreeChangeListener) removeListener((ITreeChangeListener) node);
              UUID par = node.getWorkParentUUID();
              node.setNOTDisposed();
              //NodeDesc nodeDesc = NodeDesc.fromNode(node, false, false, true);
              node.setDisposed();
  
              //tends to be done by cache
             // getLinkResolver().updateName(uuid, null);
  
              node.fireNodeDelete();
              if (node.getWorkParentUUID() != null) {
                TreeNode wp = node.getWorkParent();
                if (wp != null) wp.removeChildUUID(uuid);
              }
              for (ITreeChangeListener l : getListeners()) {
                l.deleted(node, internal, m);
              }
              addGuiChange(
                  new TreeNodeChangeEvent(TreeNodeChangeEventType.DELETE, node, internal, node, m));
              if (!internal) {
                notifyExternalChange();
              }
              // if(node instanceof Requirement){
              // Requirement req = (Requirement) node;
              // for(Location s :req.getLocations()){
              // s.updateRevLocations(req);
              // s.resetLabel();
              // }
              // }
              // TODO look
              // node.getTreeDB().getCacheStorage().removeCacheListValue(uuid,"",TreeNode.ATTR_CLONE_PARENT,node.getCloneParentUUId());
              for (TreeNode nd : Instantiator.getClones(root, node)) {
                try {
                  nd.delete();
                  // ti.remove(uid);
                } catch (Exception e) {
                  // clone del error
                }
              }
              removeCache(uuid, par, DEFAULT_DEPTH);
            }
            // if(!isVirtual)//{
            // In.getVirtualTreeDB(TreeDB.this).getDeltaTarget().deleted(uuid,
            // internal);
            // }
            // if(!isVirtual)
            // VirtualNode.getVirtualTreeDB(TreeDB.this).getDeltaTarget().deleted(uuid,
            // internal);
            if(CoreUtils.DEBUG_LEVEL>1)
              RLogger.logInfo("TreeDB delete fin on "+ResManager.getManager().getName(TreeDB.this) );
          }
          finally{
            stopChangeRow();
          }
          
        }
        @Override
        public boolean changeAttributes(UUID uuid, ChangesSet<Attribute> cs){
          TreeNode curNode = getNode(uuid);
          if(curNode == null){
            RLogger.logInfo("null curNode "+uuid);
            return false;//TODO mb true
          }
          curNode.attrs().commitChanges(cs, true);
          return true;
          //TODO call commit?
        }
        /** Reaction on element change */
        @Override
        public void moved(
            UUID nodeUUID, UUID oldParentUUId, UUID newParentUUId, Moment m, boolean internal) {
          TreeNode curNode = cache.get(nodeUUID);
          if (curNode == null) {
            TreeDB db = TreeDB.this;
            if (db.isVirtual) db = Instantiator.getRootTree(db);
            curNode = db.getCachedNodeOrVirt(nodeUUID);
          }
          if (curNode != null) {
            IStorageDeltaListener delta =
                getProperListener(nodeUUID, null, curNode.attrs().getActiveAttributeDefs());
            if (delta != null) {
              if (delta != this) {
                delta.moved(nodeUUID, oldParentUUId, newParentUUId, m, internal);
                return;
              }
            } else return;
          }else
            throw new RuntimeException("Moved node not found! "+nodeUUID);
          startChangeRow();
          try{
            TreeNode oldParent = getNode(oldParentUUId);
            TreeNode newParent = getNode(newParentUUId);
            if (m == Moment.BEFORE) {
              boolean changes = false;
              if (oldParent != null) {
                oldParent.removeChildUUID(nodeUUID);
                curNode.parent = null;//temp
                for (String s : oldParent.attrs().getLocalAttributeKeys()) { 
                  // all
                  Attribute vars = oldParent.attrs().getLocalAttribute(s);
                  if (vars != null) {
                    Attribute a = curNode.attrs().getAttribute(s);
                    if (a != null && a.getOrigin().equals(vars.getOrigin()))
                      if (a.getAvailability() != AttributeAvailability.LOCAL) {
                        curNode.attrs().removeAttribute(s);
                        changes = true;
                      }
                  }
                }
                curNode.getVS().setParentVersion(Type.ATTRIBUTES, Version.getMinusOne());
                curNode.getVS().setParentVersion(Type.STRUCT, Version.getMinusOne());
                changes = true;
              }
              if (changes)
                // curNode.parent = null;
                curNode.commitAttributes();
              // if(AttributeStorage.isMarkedUUIDs(TreeDB.this,nodeUUID)){
              // AttributeStorage.removeMark(TreeDB.this, nodeUUID);
              // curNode.save
              // }
            } else {
              boolean changes = false;
              if (newParent != null) { // TODO change par on null
                // RLogger.logInfo("null par");
                TreeNode par = getNode(oldParentUUId);
                if (par != null)
                  for (UUID chi : par.getActiveWorkChildrenUUIDs()) {
                    TreeNode child = getNode(chi);
                    if (child != null
                        && !child.getUUId().equals(curNode.getUUId())
                        && isSameClone(child, curNode)) {
                      deleted(chi, Moment.BEFORE, true);
                      deleted(chi, Moment.AFTER, true);
                    }
                  }
                try{
                  newParent.addChildUUID(nodeUUID);
                }catch(Exception e){
                  newParentUUId = oldParentUUId;
                  newParent = oldParent;
                  newParent.addChildUUID(nodeUUID);
                  RLogger.logInfo(CoreUtils.getTrace(e));
                }
                curNode.parent = newParentUUId;
                for (String s : newParent.attrs().getLocalAttributeKeys()) // TODO
                // all
                {
                  Attribute list = newParent.attrs().getLocalAttribute(s);
                  if (list != null)
                    if (curNode.attrs().getAttribute(s) == null) {
  
                      Attribute a = list;
                      if (a.getAvailability() != AttributeAvailability.LOCAL) {
                        curNode.attrs().setAttribute(a.getDefinition());
                        changes = true;
                      }
                    }
                }
              }
              curNode.attrs().updateAttribute(TreeNode.ATTR_USERVISIBLEID);
              if (changes || curNode.attrs().hasChanges()) {
                curNode.commitAttributes();
              }
  
              if (!internal) {
                notifyExternalChange();
              }
              if (curNode.getType() == null) RLogger.logInfo("Null type!");
  
              if (VirtualNode.TYPE_NAME.equals(curNode.getType()))
                ((VirtualNode) curNode).markMoved();
              curNode.fireNodeMoved(oldParent == null ? null : oldParent, internal);
              curNode.fireGUINodeMoved(oldParent == null ? null : oldParent, internal);
              for (ITreeChangeListener l : getListeners()) {
                l.moved(
                    curNode,
                    oldParent == null ? curNode.getWorkParent() : oldParent,
                    newParent == null ? curNode.getWorkParent() : newParent,
                    internal);
              }
              fireNodeGUIMove(
                  curNode,
                  oldParent == null ? null : oldParent,
                  newParent == null ? null : newParent,
                  internal);
    //          if (oldParent != null && oldParent.equals(newParent))
                //if (curNode instanceof Requirement) {
//                  ((Requirement) curNode).updateTreeNodes(curNode, TreeNode.ATTR_ID);
  //              }
              clonesReact(nodeUUID, oldParent, newParent);
              if (newParent != null) {
                newParent.getVS().updateVersion(Type.STRUCT);
                VirtualNode.updateClonesChildren(newParent); // TODO check
              }
              // if(oldParent!=null)
              // VirtualNode.updateClonesChildren(oldParent);//TODO check
  
            }
          }
          finally{
            stopChangeRow();
          }
        }

        private void clonesReact(UUID uuid, TreeNode oldParent, TreeNode newParent) {
          TreeDB db = Instantiator.getRootTree(TreeDB.this);
          LinkedList<TreeNode> pars1 = new LinkedList<>(),
              pars2 = new LinkedList<>();
          TreeNode nde = newParent;
          while (nde != null) {
            pars1.addFirst(nde);
            nde = nde.getWorkParent();
          }
          nde = oldParent;
          while (nde != null) {
            pars2.addFirst(nde);
            nde = nde.getWorkParent();
          }
          Iterator<TreeNode> it1 = pars1.iterator(), it2 = pars2.iterator();
          StringBuilder sb = new StringBuilder();
          TreeNode node1 = null, node2 = null;
          while (it1.hasNext() && it2.hasNext()) {
            node1 = it1.next();
            node2 = it2.next();
            if (!node1.getUUId().equals(node2.getUUId())) break;
          }
          if (node1 != null && node2 != null && node1.getUUId().equals(node2.getUUId())) {
            node1 = node2 = null;
            if (it1.hasNext()) node1 = it1.next();
            if (it2.hasNext()) node2 = it2.next();
          }
          while (node2 != null) {
            if (sb.length() > 0) sb.append("/");
            sb.append("..");
            node2 = null;
            if (it2.hasNext()) node2 = it2.next();
          }
          while (node1 != null) {
            if (sb.length() > 0) sb.append("/");
            sb.append(node1.getUUId());
            node1 = null;
            if (it1.hasNext()) node1 = it1.next();
          }
          String sbs = sb.toString();
          // TODO ugly version
          Set<UUID> clones = new HashSet<>(Instantiator.getClones(uuid, db));
          TreeNode node = db.getNode(uuid);
          if (clones.size() > 0) {
            for (UUID uid : clones) {
              try {
                TreeNode clone = db.getNode(uid);
                if (clone == null) continue;
                TreeNode posClone = clone.getWorkParent();
                // on place
                if (sbs.length() > 0) {
                  String[] correction = sbs.split("/");
                  for (int i = 0; i < correction.length; i++) {
                    String action = correction[i];
                    if (action.equals("..")) {
                      if (posClone.getType().equals(VirtualNode.TYPE_NAME)) {
                        posClone = null; // cannot get upper chan
                        // vnode
                        break;
                      }
                      posClone = posClone.getWorkParent();
                      if (!posClone.isVirtual()
                          && !VirtualNode.TYPE_NAME.equals(posClone.getType())) { // TODO
                        // change
                        // for
                        // proper
                        posClone = null;
                        break;
                      }
                    } else {
                      posClone =
                          Instantiator.getChildOrClone(db, posClone, UUID.fromString(action));
                      if (posClone == null) break;
                    }
                  }
                }
                if (posClone == null)
                  try {
                    clone.delete();
                  } catch (Exception e) {
                    deleted(clone.getUUId(), Moment.BEFORE, true);
                    deleted(clone.getUUId(), Moment.AFTER, true);
                  }
                else {
                  // if(posClone.getWorkParent().getType().equals(VirtualNode.TYPE_NAME))
                  // npclones.remove(posClone.getWorkParent().getUUId());
                  // npclones.remove(posClone.getUUId());
                  if (!clone.getUUId().equals(uuid)) {
                    int cnt = 0;
                    for (UUID ndeu : posClone.getActiveWorkChildrenUUIDs()) {
                      Object cpo =
                          AttributeStorage.getAttributeDefValue(
                              db, db.getAttributes(ndeu), ndeu, TreeNode.ATTR_CLONE_PARENT);
                      if (cpo != null
                          && cpo.toString().equals(clone.getCloneParentUUId().toString())) cnt++;
                    }
                    if (cnt <= 1) { // if only one{
                      try {
                        clone.attrs().setConstAttribute(TreeInstancer.ATTR_CLONE_ID, node.getId(),Flag.HIDDEN);
                        clone.setId(node.getId());
                        clone.commitAttributes();
                      } catch (Exception e) {
                        e.printStackTrace();
                      }
                    }
                  }
                  if (!posClone.getUUId().equals(clone.getWorkParent().getUUId()))
                    clone.setParent(posClone);
                  // if(!posClone.getParent(true).getType().equals(VirtualNode.TYPE_NAME)){

                  // }
                  // if(!clone.getId().equals(node.getId()))
                  // clone.setId(node.getId());
                }
              } catch (Exception e) {
                // e.printStackTrace();
              }
            }
            // deleted(node.toNodeDesc(false));
          }
        }

        @Override
        public void resourceChanged(UUID uuid, String resName, Moment m, boolean forceInternal) {
          TreeNode node = getCachedNodeOrVirt(uuid, DEFAULT_DEPTH);
          if (node != null) {
            node.fireNodeResourceChange(resName, m, forceInternal);
            node.fireGUINodeResourceChange(resName, m, forceInternal);
          }
        }

        @Override
        public void startChangeRow() {
          TreeDB.this.startChangeRow();
        }

        @Override
        public void stopChangeRow() {
          //if (TreeDB.this.isVirtual) Instantiator.getRootTree(TreeDB.this).stopChangeRow();
          TreeDB.this.stopChangeRow(); 

        }

        @Override
        public void versionChanged(
            UUID uuid, ChangesSet<Version> changes, Moment m, 
            boolean forceInternal) {
//          if (m == Moment.BEFORE) {
//            TreeNode node = getNode(uuid);
//            if (node != null) {
//              VersionsStorage vs = node.getVS();
//              Version v = vs.getVersion(Type.STRUCT);
//              lastStruct.put(uuid, v);
//            }
//          }
//          if (m == Moment.AFTER) {
//            TreeNode node = getNode(uuid);
//            if (node != null) {
//              VersionsStorage vs = node.getVS();
//              Version v = lastStruct.get(uuid);
//              if (v == null) v = Version.getUnset();
//              if (!v.equals(vs.getVersion(Type.STRUCT))) {
////                 TreeNode node = getCachedNodeOrVirt(uuid);
//                startChangeRow();
//                try{
//                  addGuiChange(
//                    new TreeNodeChangeEvent(
//                        TreeNodeChangeEventType.VERSION_CHANGE, node, true, Type.STRUCT));
//                }
//                finally{
//                  stopChangeRow();
//                }
//              }
//              for (String s : changes.getAddedChanged())
//                if(vers.get(s)==null)
//                  vs.setVersion(s, Version.getZero(), false);
//                else
//                  vs.setVersion(s, vers.get(s).clone(), false);
//              for(String s:changes.getRemoved())
//                vs.setVersion(s, Version.getZero(), false);
//            }
//          }
        }
      };
      public void handleTreeNodeChanges(
          TreeNode curNode, ChangesSet<Attribute> what, Moment m, boolean internal) {

        if(CoreUtils.DEBUG_LEVEL>1)
          RLogger.logInfo("TreeDB changed on "+ResManager.getManager().getName(TreeDB.this) +" "+m.toString()); 
        //TreeNode curNode = getNode(uuid);
        //if (curNode == null) return;
        if(curNode.getTreeDB()!=this && m == Moment.BEFORE){//TODO remove if needed moment check
          curNode.getTreeDB().handleTreeNodeChanges(curNode,what,m,internal);
          return;
        }
        startChangeRow();
        try{
          if (m == Moment.BEFORE) {
            //TODO order?
            Map<String,Attribute> newA = what.getNewMap();
            for(Entry<String, Attribute> s:newA.entrySet()){
              Attribute a = s.getValue();
              curNode.reactBeforeAttributeChange(s.getKey(), a!=null?a.getDefinition():null);
            }
          } else {
            if (what == null) {
              throw new RuntimeException("Empty changes set");
            }// else what = new HashSet<String>(what);
            if (curNode.getId() == null) {
              throw new RuntimeException("null id");
            }
            if (!what.isEmpty()){
              Map<String,Attribute> newA = what.getNewMap();
              for(Entry<String, Attribute> s:newA.entrySet()){
                Attribute a = s.getValue();
                curNode.reactAfterAttributeChange(s.getKey(), a);
              }
//              for (String s : what.getAll()) {
//                curNode.reactAfterAttributeChange(s, curNode.attrs().getAttribute(s));
//              }
              fireNodeChange(curNode, what, internal);
              curNode.fireNodeAttributeChange(what);
            }
            if (what.contains(TreeNode.ATTR_NAME))
              getLinkResolver().updateName(curNode.getUUId(), curNode.getName());
            if (!what.isEmpty()) {
              fireNodeGUIChange(curNode, what, internal);
              curNode.fireGUINodeAttributeChange(what);
              if (!internal) 
                notifyExternalChange();
            }
            if(curNode.getTreeDB()!=this&& m == Moment.AFTER){//TODO remove if needed moment check
              curNode.getTreeDB().handleTreeNodeChanges(curNode,what,m,internal);
              return;
            }
          }
        }
        finally{
          stopChangeRow();
        }

        if(CoreUtils.DEBUG_LEVEL>1)
          RLogger.logInfo("TreeDB changed on "+ResManager.getManager().getName(TreeDB.this) +" "+m.toString()); 
      }
  private ConcurrentLinkedQueue<Future<?>> que = new ConcurrentLinkedQueue<>();
  public void addDelayed(Runnable object) {
    if (isVirtual && Instantiator.getRootTree(this) != this){
      Instantiator.getRootTree(this).addDelayed(object); 
      return;
    }
    ExecutorService executor = Executors.newSingleThreadExecutor();
    Future<?> f = executor.submit(object);
    que.add(f);
  }

  public boolean isHidden(UUID uid) {
    Object attr = getAttributeValue(uid, VirtualNode.ATTR_HIDDEN_STATUS, false, false);
    if (attr != null && attr.toString().toLowerCase().equals("true")) return true;
    return false;
  }

//  public static ChangesSet<Attribute> addAttrsRec(TreeNode node, Set<String> what, ChangesSet<Attribute> chVis) {
//    return addAttrsRec(node, what, chVis, null);
//  }
  /**
   * Method to gather local deps from "what" set, update as and chVis. Adds order of execution to chVis 
   * @param as Attribute storage context
   * @param what set of names to be handled
   * @param chVis changes set to be extended
   * @param now set of new definitions to be used in attribute form mechanics
   */
  public static ChangesSet<Attribute> addAttrsRec(AttributeStorage as) {
//    AttributeStorage as = node.attrs();
    ChangesSet<Attribute> ret = new ChangesSet<>();
    Set<String> handled = new HashSet<>();
    List<String> order = new LinkedList<>();
    Map<String, Set<String>> links = new HashMap<>();
    LinkedList<String>toh = new LinkedList<>(as.getAllAttributeKeys());
    
    //forming a set of definitions to be checked
    ChangesSet<AttributeDefinition> ch = as.getChanged();
    toh.addAll(ch.getAll());
    
    //forms dependencies 
    while (!toh.isEmpty()) {
      String s = toh.removeFirst();
      Set<String> rev = null;
      rev = as.getDependencies(s);
      for (String s2 : rev) {
        if (!links.containsKey(s2)){
          links.put(s2, new HashSet<>());
        }
        links.get(s2).add(s);
      }
    }
    //forms set of attribute names to 
    LinkedList<String> openSet = new LinkedList<>(ch.getAll());
    Set<String> fullSet = new HashSet<>(openSet);
    while (!openSet.isEmpty()) {
      Iterator<String> next = openSet.iterator();
      String cs = next.next();
      next.remove();
      //handled.add(cs);
      Set<String> rel = links.get(cs);
      if (rel != null)
        for(String s:rel) 
          if(!fullSet.contains(s)){
            fullSet.add(s);
            openSet.add(s);
          }
    }
    
    openSet = new LinkedList<>(ch.getAll());
//    if(openSet.contains(TreeNode.ATTR_TYPE)){
//      openSet.remove(openSet.indexOf(TreeNode.ATTR_TYPE));
//      openSet.addFirst(TreeNode.ATTR_TYPE);
//    }
//    
//    if(openSet.contains(TreeNode.ATTR_ID)){
//      openSet.remove(openSet.indexOf(TreeNode.ATTR_ID));
//      openSet.addFirst(TreeNode.ATTR_ID);
//    }
    //while there are some 
    while (!openSet.isEmpty()) {
      Iterator<String> next = openSet.iterator(); 
      String chosen = null;
      while(next.hasNext() && chosen == null){
        String cand = next.next();
        boolean f = true;
        for(String s:as.getDependencies(cand))
          if(fullSet.contains(s)&&!handled.contains(s)){
            f = false;
            break;
          }
        if(f){
          chosen = cand;
          next.remove();
        }
      }
      if(chosen != null){
        String cs = chosen;
        order.add(cs);
        handled.add(cs);
        AttributeDefinition a = as.getWorkAttributeDefinition(cs);
        Attribute nw = null, ow = null;
        if(a == null){
          ow = as.actualRemoveFromAttributeList(cs);
        }else{
          try{
            nw = a.toAttribute(as);
            ow = as.setAttribute(nw);
          }catch(Exception e){
            RLogger.logError(e);
            throw e;
          }
        }
        ret.addSomeChange(cs, ow, nw);
        Set<String> rel = links.get(cs);
        if (rel != null)
          for(String s:rel) 
            if(!handled.contains(s)){
              openSet.add(s);
            }
      }else
        break;
    }
    ret.setOrder(order);
    as.clearChanges();
    if(!openSet.isEmpty())
      RLogger.logInfo("uses cannot be resolved for "+as.getNode().getUUId());
    return ret;
  }

  public Object getAttributeValue(UUID where, String key, Object object, boolean forcelocal) {
    // TODO check
    TreeNode node = getCachedNodeOrVirt(where);
    if (node != null)
      return node.getTreeDB()
          .getAttributeValue(
              where,
              forcelocal ? node.attrs().getAttribute(key) : node.attrs().getAttribute(key),
              object,
              forcelocal);
    Map<String, AttributeDefinition> atrs = getStorage().getAttributes(where);
    return AttributeStorage.getAttributeDefValue(this, atrs, where, key);
    // if(node!=null)
    // return getAttributeValue(forcelocal?
    // nde.attrs().getLocalAttribute(key,where):
    // nde.attrs().getLocalAttribute(key), object,forcelocal);
    // return object;
  }

  protected Object getAttributeValue(
      UUID where, Attribute next, Object object, boolean forcelocal) {
    if (next == null) return object;
    Object val = next.getValue();
    if (val == null)
      if (where.equals(next.getOrigin())) return null;
      else return getAttributeValue(next.getOrigin(), next.getKey(), object, forcelocal);
    return val;
  }

  protected Object getAttributeValue(
      UUID where, AttributeDefinition next, Object object, boolean forcelocal) {
    if (next == null || next.getRawValue() == null) return object;
    Object val = next.getRawValue().getValue();
    if (val == null)
      if (where.equals(next.getOrigin())) return null;
      else return getAttributeValue(next.getOrigin(), next.getKey(), object, forcelocal);
    return val;
  }

  protected Map<String, AttributeDefinition> getAttributes(UUID uid) {
    AttributeStorage as = AttributeStorage.getCachedStorage(this, uid);
    if (as != null) return as.getActiveAttributeDefs();
    TreeNode cache = getCachedNodeOrVirt(uid);
    if (cache != null) return cache.attrs().getActiveAttributeDefs();
    if (isVirtual) return Instantiator.getRootTree(this).getStorage().getAttributes(uid);
    return getStorage().getAttributes(uid);
  }

//  /**
//   * Method to replace node attributes by stored values
//   *
//   * @param node element to replace attributes on
//   */
//  public void updateNodeAttributes(TreeNode node,boolean first) {
//    Map<String, Attribute> stA = storage.getAttributes(node.getUUId());
//    updateNodeAttributes(node, stA,first);
//  }

  public void initNodeAttributes(final TreeNode node, Map<String, AttributeDefinition> newAttributes) {
    Map<String, AttributeDefinition> now = newAttributes; // TODO check
    updateParentCloneParentAttributeDefs(
      node.getUUId(), node.attrs(), null, node.getWorkParentUUID(), now);
    node.attrs().setChanged(new ChangesSet<>(new HashSet<>(now.keySet()),
        null,null,null,now,null));
    addAttrsRec(node.attrs());
    node.attrs().clearChanges();
    //node.attrs().commitChanges(cs, now, false);
  }

  /**
   * Method to get uuid generator
   *
   * @return uuid generator
   */
  public IUUIDGenerator getUuidGenerator() {
    return uuidGenerator;
  }

  /**
   * Get current data storage. It may be related to one of possible storage mechanisms. Most
   * frequently used are Resource storage based on eclipse mechanisms and Transaction storage -
   * temporary storage during transactions. One more variant is TreeInstancer that store data during
   * report generation
   *
   * @return current data storage
   */
  public ITreeStorage getStorage() {
    return storage;
  }

  public TreeInstancer getVStorage() {
    if (!isVirtual) return Instantiator.getVirtualStorage(this, Instantiator.DEFAULT_INSTANCE_NAME);
    if (storage instanceof TreeInstancer) return (TreeInstancer) storage;
    if (storage instanceof TransactionStorage)
      return (TreeInstancer) ((TransactionStorage) storage).getVStorage();
    return null;
  }

  private boolean defaultTypesContainsKey(String typeName) {
    if (isVirtual) {
      TreeDB db = Instantiator.getRootTree(this);
      if (db != null && db != this) return db.defaultTypesContainsKey(typeName);
    }
    return defaultTypes.containsKey(typeName);
  }

  private NodeType defaultTypesGet(String typeName) {
    if (isVirtual) {
      TreeDB db = Instantiator.getRootTree(this);
      if (db != null && db != this) return db.defaultTypesGet(typeName);
    }
    return defaultTypes.get(typeName);
  }

  private TreeNode createModelNode(
      UUID uuid,
      UUID puuid,
      UUID cloneParent,
      String id,
      String typeName,
      Map<String, AttributeDefinition> attributes) {
    attributes = new HashMap<>(attributes);
    TreeNode result = null;
    if (defaultTypesContainsKey(typeName)) {
      NodeType nt = defaultTypesGet(typeName);
      try {
        Class<?>[] treeNodeArgs = new Class[] {TreeDB.class, UUID.class};
        result = nt.getTypeClass().getConstructor(treeNodeArgs).newInstance(TreeDB.this, uuid);
        result.setType(typeName);
      } catch (Exception e) {
        e.printStackTrace();
      }
    }
    if (result == null) {
      result = new TreeNode(this, uuid);
      result.setType(typeName);
    }
    result.setId(id);
    result.cloneParent = cloneParent;
    result.parent = puuid;
    if (cloneParent == null) {

      Object cp = null;
      AttributeDefinition cpa = attributes.get(TreeNode.ATTR_CLONE_PARENT);
      if (cpa != null) cp = cpa.getRawValue().getValue();
      // getAttributeValue(uuid, TreeInstancer.ATTR_CLONE_PARENT, null,
      // true);
      if (cp != null) result.cloneParent = UUID.fromString(cp.toString());
    }
    // TODO check
    cache.put(uuid, result);
    // if(!isVirtual)
    // VirtualNode.getVirtualTreeDB(this).cache.put(uuid, result);
    initNodeAttributes(result, attributes);
    Attribute parentRoot = result.attrs().getAttribute(TreeNode.ATTR_VIRT_STATUS);
    if (parentRoot != null && parentRoot.getValue() != null)
      result.vroot = UUID.fromString(parentRoot.getDefinition().getRawValue().getValue().toString());
    return result;
  }

  public void updateNodeChildren(TreeNode result) {
    updateNodeChildren(result, false);
  }

  public static void updateNodeChildren(TreeNode result, boolean init) {
    if (result == null) return;
    TreeDB db = Instantiator.getRootTree(result.getTreeDB(), 3);
    if ((!db.storage.canLoad(result.getUUId()) || result.isVirtual() && !db.isVirtual)) {
      if (!db.isVirtual && result.isVirtual()) {
        UUID vroot = result.getVRoot();
        if (vroot != null) db = Instantiator.getExistVirtualTreeDB(db, vroot);
      }
    }
    db.updateNodeChildren(result, db.storage.getChildren(result.getUUId()), init);
  }

//  public void updateShownNodeChildren(TreeNode result, boolean init) {
//    // if(result.getTreeDB()!=this)
//    // RLogger.logInfo("wrong db!");
//    // Collection<UUID> uids = null;
//    // if(result.getShownChildrenUUIDs()!=null)
//    // uids = new HashSet<UUID>(result.getShownChildrenUUIDs());
//    // int ocnt = 0;
//    // if(uids!=null)
//    // ocnt = uids.size();
//    // List<UUID> ids = null;
//    // //uids.removeAll(ids);
//    //
//    result.clearShownChildren();
//    if (!init) {
//      //TODO struct update
//      //fireNodeGUIChange(result, TreeNode.formIdEvent(), Collections.emptyMap(), true);
//      //result.fireGUINodeAttributeChange(TreeNode.IdEvent);
//    }
//  }

  private void updateNodeChildren(TreeNode result, Collection<UUID> children, boolean init) {

    result.setChildrenUUIDs(new ArrayList<>(children), init);
    if (!init) {
      // result.getVS().incVersion(Type.STRUCT);
      // fireNodeChange(result,new HashSet<String>(){{add("_%id");}},
      // true);
      // fireNodeGUIChange(result,new HashSet<String>(){{add("_%id");}},
      // true);
    }
  }

  private TreeNode createFromStorage(
      UUID uuid, String typeName, Map<String, AttributeDefinition> attributes) {
    String id;
    UUID puuid;
    long l = System.currentTimeMillis();
    if ("DBRoot".equals(typeName)) {
      id = "/";
      puuid = null;
    } else {
      id = storage.getNodeId(uuid);
      puuid = storage.getParent(uuid);
    }
    if (id == null) {
      throw new RuntimeException("Id is null for" + uuid);
    }
    // if(isVirtual&&storage instanceof TreeInstancer)
    // ((TreeInstancer)storage).createIfNeeded(uuid);
    UUID cloneP = null;
    if (attributes.containsKey(TreeNode.ATTR_CLONE_PARENT)
        && attributes.get(TreeNode.ATTR_CLONE_PARENT) != null) {
      AttributeDefinition o = attributes.get(TreeNode.ATTR_CLONE_PARENT);
      Object val = o.getRawValue().getValue();
      if (val == null){
        val = getAttributeValue(uuid, o, null, false);
      }
      if (val != null) cloneP = UUID.fromString(val.toString());
    }
    if (CoreUtils.DEBUG_LEVEL > 1) {
      RLogger.logInfo("\t get cp " + (System.currentTimeMillis() - l));
      l = System.currentTimeMillis();
    }

    // if(isVirtual)
    // cloneP = VirtualNode.getVirtualStorage(this).getCloneParent(uuid);
    TreeNode result = createModelNode(uuid, puuid, cloneP, id, typeName, attributes);
    if (result!=null && result.getId() == null) {
      throw new RuntimeException("Id is null for" + uuid);
    }
    if (CoreUtils.DEBUG_LEVEL > 1) {
      RLogger.logInfo("\t create model " + (System.currentTimeMillis() - l));
      l = System.currentTimeMillis();
    }
    Collection<UUID> children = storage.getChildren(uuid);
    if (CoreUtils.DEBUG_LEVEL > 1) {
      RLogger.logInfo("\t get children " + (System.currentTimeMillis() - l));
      l = System.currentTimeMillis();
    }
    updateNodeChildren(result, children, true);
    if (CoreUtils.DEBUG_LEVEL > 1) {
      RLogger.logInfo("\t update children " + (System.currentTimeMillis() - l));
      l = System.currentTimeMillis();
    }
    if (result == null) RLogger.logInfo("null res");
    else {
    }
    if (result!=null && result.getId().equals("/")) {
      assert result.parent == null;
      root = result;
    }
    return result;
  }

  /**
   * Method to fire element change event. It will be translated to all subscribed listeners.
   *
   * @param curNode changed element
   * @param internal true if event appears from inner mechanics
   */
  public void fireNodeChange(
      TreeNode curNode, ChangesSet<Attribute> params, boolean internal) {
    if (isVirtual && curNode != null) {
      TreeDB db = Instantiator.getRootTree(this);
      if (db != null && db != this) db.fireNodeChange(curNode, params, internal);
    } else
      for (ITreeChangeListener l : getListeners()) {
        if (l != null) // Oo
        l.changed(curNode, params, internal);
      }
  }

  /**
   * String for representing of Node's types filter in form of list of string names of types. For
   * ex. {TestPurpose.TYPE_NAME}"
   *
   * @return
   */
  public List<String> getLinkToTypeFilter(String linktype) {
    // TODO TreeLogic methods
    if (linktype.equals("TestStepLinks")) {
      ArrayList<String> a = new ArrayList<>();
      a.add(TestPurpose.TYPE_NAME);
      return a;
    }
    return typeFilter;
  }

  public List<String> getRelatedLinkTo() {
    ArrayList<String> a = new ArrayList<>();
    a.add("TestStepLinks");
    return a;
  }

  public List<String> getLinkToTypeFilter() {
    // TODO TreeLogic methods
    List<String> types = new ArrayList<>(), lt = getRelatedLinkTo();
    for (String s : lt) types.addAll(getLinkToTypeFilter(s));
    return types;
  }

  /**
   * Update stored linkto: constructions and attribute where [linkto: ] data stored.
   *
   * @param attrname Name of attribute where links can be placed
   */
  public void updateLinks(TreeNode node, String attrname, boolean removeHiddenLinks) {}

  public SortingSettings getSortingSettings() {
    if (isVirtual) {
      TreeDB db = Instantiator.getBaseTree(this);
      if (db != null && db != this) return db.getSortingSettings();
      return null;
    }
    return loadSettings(getRootNode());
  }

  public static SortingSettings loadSettings(TreeNode node) {
    if (node == null) return defaultSortSettings;
    // old style
    Attribute al = node.attrs().getAttribute(SORT_BY_LOCATIONS);
    SortingMethod sm = SortingMethod.BY_NAMES;
    if (al != null
        && ((al.getValue().equals("true")
            || al.getValue().equals(true)))) sm = SortingMethod.BY_LOCATIONS;
    Attribute ai = node.attrs().getAttribute(SORT_BY_INDEX);
    if (ai != null
        && (ai.getValue().equals("true")
            || ai.getValue().equals(true))) sm = SortingMethod.BY_INDEX;
    // new style
    Attribute a = node.attrs().getAttribute(SORT_METHOD);
    if (a != null) {
      String aval = a.getValue().toString();
      if (aval != null) {
        SortingMethod t = SortingMethod.valueOf(aval);
        sm = t;
      } else sm = defaultSortSettings.getMethod();
    } else {
      if (al == null && ai == null) sm = defaultSortSettings.getMethod(); // buggy?
    }
    a = node.attrs().getAttribute(SORT_SEP_TYPES);
    boolean sepState =
        a != null
            ? (a.getValue().equals("true")
                || a.getValue().equals(true))
            : defaultSortSettings.getSeparateByTypesFlag();
    return SortingSettings.init(sm, sepState);
  }

  public void updateLinks(TreeNode node, String attrname) {
    updateLinks(node, attrname, false);
  }
  public void fireNodeChildrenOrderChange(TreeNode shownParent) {
    fireNodeChildrenOrderChange(shownParent,true);
  }
  public void fireNodeChildrenOrderChange(TreeNode shownParent,boolean clear) {
    if(clear)
      shownParent.clearShownChildren();
    if (activeChangesRowsCount == 0)
      for (ITreeChangeListener l : getGUIListeners()) {
        l.childrenOrderChanged(shownParent);
      }
    else
      addGuiChange(
          new TreeNodeChangeEvent(TreeNodeChangeEventType.CHILDREN_ORDER_CHANGE, shownParent));
  }
  public void fireNodeGUIChange(
      TreeNode curNode, ChangesSet<Attribute> cs, boolean internal) {
    if (activeChangesRowsCount == 0)
      for (ITreeChangeListener l : getGUIListeners()) {
        l.changed(curNode, cs, internal);
      }
    else
      addGuiChange(
          new TreeNodeChangeEvent(TreeNodeChangeEventType.CHANGE, curNode, internal, new ChangesSet<>(cs)));
  }

  public void fireNodeGUIMove(TreeNode curNode, TreeNode from, TreeNode to, boolean internal) {
    if (activeChangesRowsCount == 0)
      for (ITreeChangeListener l : getGUIListeners()) {
        l.moved(curNode, from, to, internal);
        ;
      }
    else {
      addGuiChange(
          new TreeNodeChangeEvent(TreeNodeChangeEventType.MOVE, curNode, internal, from, to));
    }
  }

  private void addGuiChange(TreeNodeChangeEvent treeNodeChangeEvent) {
    wl.lock();
    guichanges.add(treeNodeChangeEvent);
    wl.unlock();
  }

  public IStorageDeltaListener getDeltaTarget() {
    return storageDeltaListener;
  }

  public LinkResolver getLinkResolver() {
    if (lr == null)
      // if(isVirtual&&getLevelOfAbstraction()>1){//reports
      // TreeDB db = Instantiator.getBaseTree(this);
      // if(db!=null && db!=this)
      // return db.getLinkResolver();
      // }else
      lr = new LinkResolver(this);
    return lr;
  }


  public boolean isLoaded(UUID uid) {
    if (cache.containsKey(uid)) return true;
    if (isVirtual) {
      TreeDB base = Instantiator.getBaseTree(this);
      if (base == null)
        return false;
      if (base != this) return base.isLoaded(uid);
      return false;
    } else {
      for (TreeDB vdb : Instantiator.getVirtualTreeDBs(this))
        if (vdb.cache.containsKey(uid)) return true;
    }
    return false;
  }

  public TreeNode getRefTarget() {
    return refTarget;
  }

  public int getRefTargetState(TreeNode node) {
    if (refTarget != null && refTarget.equals(node)) return refTargetState;
    return 0;
  }

  public void setRefTarget(TreeNode node) {
    if (refTarget == null) refTargetState = node != null ? 1 : 0;
    else if (!refTarget.equals(node)) {
      refTargetState = 0;
      refTarget = null;
      return;
    }
    if (refTarget != null && refTarget.equals(node)) refTargetState = (this.refTargetState + 1) % 3;
    refTarget = node;
    getRootNode().attrs().setConstAttribute("_refTarget", 
        node == null ? "" : node.getQualifiedId(),Flag.HIDDEN);
    getRootNode().commitAttributes();
  }

  public String getDefReferenceName() {
    if (isVirtual) {
      TreeDB db = Instantiator.getBaseTree(this);
      if (db != null && db != this) return db.getDefReferenceName();
      return null;
    }
    if (getRootNode() == null || root.attrs() == null) RLogger.logInfo("null root or attrs");
    Attribute a = root.attrs().getAttribute(DEFAULT_REFERENCE_NAME);
    if (a != null) return a.getValue().toString();
    return "elaborates";
  }

  public int getGlobalRefTargetState() {
    return refTargetState;
  }

  public List<ITransactionListener> getTransactionListeners() {
    TreeDB root = Instantiator.getRootTree(this);
    if (root.equals(this)) return transactionListeners;
    else return root.getTransactionListeners();
  }

  public String getQualifiedId(UUID origin) {
    if (origin == null || !getStorage().canLoad(origin)) return "";
    String id = getStorage().getNodeId(origin);
    if (id.equals("/")) return "";
    return getQualifiedId(getStorage().getParent(origin)) + "/" + id;
  }

  public void fireNodeGUIRemoved(TreeNode node, boolean b, Moment m) {
    node.fireGUINodeDelete();
    for (ITreeChangeListener inc : getGUIListeners()) inc.deleted(node, true, m);
  }

  public String getFullName(String targetQID) {
    try {
      return getFullName(getNode(targetQID, Path.QID | Path.UUID));
    } catch (Exception e) {

    }
    return null;
  }

  public String getFullName(TreeNode node) {
    return getFullName(node, false);
  }

  public String getFullName(TreeNode node, boolean inner) {
    if (node == null) return null;
    if (getFullNameMode()) return node.getUserVisibleName(inner);
    return Requirement.getFullName(node, this, inner);
  }

  public IStorageDeltaListener getNullDeltaListener() {
    return nullDeltaListener;
  }

  public Collection<Attribute> getEnumDefinitons() {
    HashSet<Attribute> ret = new HashSet<>();
    for (Attribute a : root.attrs()) {
      if (a != null && a.getDefinition().equals(AttributeType.ENUM_DEFINITION)) ret.add(a);
    }
    return ret;
  }

  public void setCheckers(CheckerContainer checkers) {
    this.container = checkers;
  }

  public IMark[] getCheckerProblems(TreeNode node) {
    return getChprob().remove(node.getUUId());
  }

  public int getLevelOfAbstraction() {
    return loa;
  }

  public void setIsLoaded() {
    setIsLoaded(true);
  }

  public void setIsLoaded(boolean val) {
    loaded = val;
  }

  public boolean isLoaded() {
    return loaded;
  }
  // public TreeNode getDVNode(String qualifiedId, boolean b) {
  // try{
  // TreeNode[]res = getLinkResolver().resolvePath(this, qualifiedId,
  // PathResolveMode.QID_UV_CREATE, b);
  // if(res == null || res.length == 0)
  // return null;
  // return res[0];
  // }
  // catch(RuntimeException e){tiator
  // e.printStackTrace();
  // }
  // return null;
  // }

  public int getChangeRowCount() {
    return activeChangesRowsCount;
  }

  public boolean isBusy() {
    TreeDB db = this;
    if (isVirtual) db = Instantiator.getRootTree(db);
    return !db.isLoaded()
        || db.inTransaction()
        || db.getChangeRowCount() > 0;
  }

  public UUID getCloneParent(UUID node) {
    Object val = getAttributeValue(node, TreeNode.ATTR_CLONE_PARENT, null, false);
    if (val != null) return UUID.fromString(val.toString());
    return null;
  }

  public void setSortingMethod(SortingSettings sortingSettings, boolean save) {
    TreeNode root = Instantiator.getRootTree(this).getRootNode();
    root.attrs().setConstAttribute(TreeDB.SORT_METHOD, sortingSettings.getMethod().toString(),Flag.HIDDEN);
    root.attrs().setConstAttribute(TreeDB.SORT_SEP_TYPES, sortingSettings.getSeparateByTypesFlag(),Flag.HIDDEN);
    if (save) root.commitAttributes();
  }

  public static List<UUID> updateVisablility(TreeDB db, List<UUID> nodes) {
    if (nodes == null || nodes.size() == 0) return nodes;
    List<UUID> rnodes = new ArrayList<>();
    // TODO look into predicate updates mechanics
    // may be incorrect
    for (UUID uid : nodes) {
      try {
        TreeNode node = db.getNode(uid, DEFAULT_DEPTH, true);
        if (node == null
            || node.attrs()
                .getBooleanAttributeValue(TreeNode.ATTR_VIRT_REMOVED, false) || node.isDisposed()) // Virtually
          //
          continue;
        if (!node.isHidden()) rnodes.add(uid);
      } catch (RuntimeException e) {
        e.printStackTrace();
      } catch (Exception e) {
        e.printStackTrace();
      }
    }
    return rnodes;
  }

  public Map<UUID, IMark[]> getChprob() {
    return chprob;
  }

  public void setChprob(Map<UUID, IMark[]> chprob) {
    this.chprob = chprob;
  }

  public void setChprobUnit(UUID uid, IMark[] marks) {
    this.chprob.put(uid, marks);
  }

  public void clearAllCaches(boolean clearOnlyLoaded) {
    //TODO handle lates
    Instantiator.clearCache(this);
    clearTagsCache();
    DeepFirstTreeWalker dftw =
        new DeepFirstTreeWalker(getRootNode(), false, false, true, true, clearOnlyLoaded);
    Iterator<TreeNode> it = dftw.iterator();
    CheckerContainer cc = getCheckerContainer();
    if (cc != null) cc.clearChecked();
    Set<UUID> uids = new HashSet<>();
    while (it.hasNext()) {
      TreeNode node = it.next();
      uids.add(node.getUUId());
      // AttributeStorage.clearCache(this, node.getUUId());
      clearAllCaches(this, node.getUUId());
    }
    cacheLocks.clear();
    cacheThreads.clear();
    cache.clear();
    Location.clearCache(this);
    // lr.clearCaches();
    String b = storage.getCacheStorage().startCacheTransaction("clear all caches",CacheStorage.COMMIT_ALL);
    try{
     // for (UUID uid : uids) storage.clearCachedCache(uid);
      storage.clearCaches();
    }
    finally{
      if(b==null)
        storage.getCacheStorage().commitCache();
    }
    ResManager.getManager().clearCache(this);
    ResManager.getManager().reOpen(this);
  }

  public static void clearAllCaches(TreeDB db, UUID uid) {
    //TODO handle lates
    if (db != null && db.cache.containsKey(uid)) {
      TreeNode node = db.cache.get(uid);
      if (node.getType().equals(Document.TYPE_NAME)) ((Document) node).clearCache();
      if (node.getTreeDB().equals(Location.TYPE_NAME)) ((Location) node).resetRelated();
    }
  }

//  public void replaceCache(UUID uuId, TreeNode rootNode) {
//    TreeNode old = cache.remove(uuId);
//    if (old != null) {
//      AttributeStorage.removeCache(uuId);
//      getLinkResolver().updateName(uuId, null, Version.getUnset());
//    }
//    cache.put(uuId, rootNode);
//    if (rootNode != null)
//      getLinkResolver().updateName(uuId, curNode.attrs().getAttribute(TreeNode.ATTR_NAME));
//  }

  public void removeCache(UUID uuId, UUID par, int deep) {
    if (deep <= 0) return;
    getLinkResolver().removeCache(uuId.toString());
    getCPLinkResolver().removeCache(getResName()+"::"+uuId.toString());
    Set<String> rem = attrRootRevCache.remove(uuId);
    if (rem != null)
      for (String s : rem) {
        attrRootCache.get(s).remove(uuId);
        if (attrRootCache.get(s).isEmpty()) attrRootCache.remove(s);
      }
    if (par != null) {
      TreeNode c = getCachedNodeOrVirt(uuId);
      if (c != null) c.clearShownChildren();
    }
    cache.remove(uuId);
    if (isVirtual) {
      TreeDB db = Instantiator.getBaseTree(this);
      if (db != null) db.removeCache(uuId, par, deep - 1);
    }
    for (TreeDB db : Instantiator.getVirtualTreeDBs(this)) db.removeCache(uuId, par, deep - 1);
    Location.clearCache(this);
    AttributeStorage.removeCache(this, uuId);
  }

  public void addRemovePool(TreeNode node) {
    rempool.add(node);
  }

  public Map<UUID, TreeNode> getCache() {
    return cache;
  }

  public void close() {
    // clearAllCaches(false);
    if (lr != null) lr.close();
    lr = null;
    if(ts!=null) ts.close();
    ts = null;
  }

  @Override
  public String toString() {
    return getName() + " " + super.toString();
  }

  public boolean inTransaction() {
    return storage instanceof TransactionStorage;
  }
  private static java.nio.file.Path cpPath = null; 
  static{
    cpPath = CoreUtils.getCacheDir(ResManager.getManager().getGlobalCacheLocation());
    gcs = new CacheStorage(null, cpPath,"crossproject","");
    glr = new LinkResolver(gcs);
  }
  public static CacheStorage getCPCacheStorage() {
    return gcs;
  }
  public static LinkResolver getCPLinkResolver() {
    return glr;
  }

  public TermsStorage getTermsStorage() {
    if(ts == null)
      ts = new TermsStorage(getCacheStorage(), this);
    return ts;
  }
  
}
