/*
 * Copyright 2015 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.unitesk.requality.core;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.net.URI;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.UUID;
import java.util.concurrent.CopyOnWriteArrayList;
import java.util.concurrent.locks.ReentrantReadWriteLock;
import java.util.concurrent.locks.ReentrantReadWriteLock.ReadLock;
import java.util.concurrent.locks.ReentrantReadWriteLock.WriteLock;
import com.unitesk.requality.core.attribute.Attribute;
import com.unitesk.requality.core.attribute.AttributeAvailability;
import com.unitesk.requality.core.attribute.AttributeDefinition;
import com.unitesk.requality.core.attribute.AttributeStorage;
import com.unitesk.requality.core.attribute.AttributeType;
import com.unitesk.requality.core.attribute.Flag;
import com.unitesk.requality.core.attribute.IReqURIResolver;
import com.unitesk.requality.core.attribute.ListValue;
import com.unitesk.requality.core.attribute.ReferenceValue;
import com.unitesk.requality.core.attribute.ReqURIResolveResult;
import com.unitesk.requality.core.attribute.VersionsStorage;
import com.unitesk.requality.core.attribute.VersionsStorage.Type;
import com.unitesk.requality.core.attribute.valueproviders.ConstValueProvider;
import com.unitesk.requality.core.attribute.valueproviders.HtmlValueProvider;
import com.unitesk.requality.core.attribute.valueproviders.LabelProvider;
import com.unitesk.requality.core.attribute.valueproviders.NameOrIdProvider;
import com.unitesk.requality.core.attribute.valueproviders.TemplValueProvider;
import com.unitesk.requality.core.attribute.valueproviders.TermsValueProvider;
import com.unitesk.requality.core.attribute.valueproviders.UserVisibleIdProvider;
import com.unitesk.requality.core.exceptions.CannotBeDeleted;
import com.unitesk.requality.core.instancer.Instantiator;
import com.unitesk.requality.core.instancer.TreeInstancer;
import com.unitesk.requality.core.operations.CopyPasteHelper;
import com.unitesk.requality.core.operations.PasteIntoHandler;
import com.unitesk.requality.core.propagators.ClonePropagator;
import com.unitesk.requality.core.propagators.IAttributeChangePropagator;
import com.unitesk.requality.core.propagators.LocationLabelPropagator;
import com.unitesk.requality.core.propagators.ParentAttrsPropagator;
import com.unitesk.requality.core.propagators.ReferenceRevertPropagator;
import com.unitesk.requality.nodetypes.NodeType;
import com.unitesk.requality.nodetypes.ReportSettings;
import com.unitesk.requality.nodetypes.Requirement;
import com.unitesk.requality.nodetypes.VirtualNode;
import com.unitesk.requality.tools.Index;
import com.unitesk.requality.tools.RequalityCLI;

/** Class that represents element in requirements catalog. */
public class TreeNode {
  public static ChangesSet<Attribute> getIdEvent(Attribute old, Attribute newa){
    return new ChangesSet<>(
      Collections.emptySet(),
      AttributeStorage.formSet(TreeNode.ATTR_ID),
      Collections.emptySet(),
      AttributeStorage.formAttributeMap(old),
      AttributeStorage.formAttributeMap(newa),
      AttributeStorage.formList(TreeNode.ATTR_ID));
  }
//
//  public static AttributeDe formIdAttribute(UUID uid, String id){
//    return new Attribute(Version.getZero(), uid, ATTR_ID, id);
//  }
  public static final String ATTR_TERMS = "_terms";
  public static final String ATTR_TERMS_REV = "_termsRev";
  public static final String ATTR_TYPE = "_type";
  public static final String ATTR_NAME = "_name";
  protected static String NumberRegex = "(-){0,1}\\d*";
  public static final String ATTR_REVERT_LINKS = "_revlinks";
  public static final String ATTR_PREDICATE = "_predicate";
  public static final String ATTR_FULL_PREDICATE = "_fullpredicate";
  public static final String ATTR_DESCRIPTION = "_description";
  public static final String ATTR_LOCATIONS_DESCRIPTION = "_ldescription";
  public static final String ATTR_GENERATORS = "_generators";
  public static final String NODE_GROUP = "com.unitesk.nodegroup.requality";
  // attribute for common undefined change over node
  public static final String ATTR_UNDEFINED = "_undefined";
  public static final String ATTR_ERROR = "_error";
  // public static final String ATTR_REVLINKS = "_%revlinks";
  public static final String ATTR_HIDDEN = VirtualNode.ATTR_HIDDEN_STATUS;
  public static final String ATTR_VIRT_REMOVED = "_virtRemoved";
  public static final String ATTR_IGNORE_NAME = "ignoreName";
  public static final String ATTR_NAMEORID = "_nameorid";
  public static final String ATTR_USERVISIBLEID = "_uservisibleid";
  public static final String ATTR_VIRT_STATUS = "_%virtstatus";
  public static final String ATTR_VIRT_CHILD_REMOVED = "_vchildremoved";
  public static final String ATTR_CLONE_PARENT = "_cloneparent";
  public static final String ATTR_ID = "_id";
  //public static final String ATTR_NODE_PATH = "node.resURL";
  public static final String ATTR_DOORS_ID = "ForeignID";
  public static final String ATTR_LINKS = "_links";
  public static final String ATTR_LABEL = "_label";
  private static final String REQ_URI_START = "requality://";
  private static Set<String> labelDeps = new HashSet<>();
  static {
    labelDeps.add(ATTR_NAMEORID);
    labelDeps.add(ATTR_ID);
  }
  
  protected boolean loaded = false;
  private UUID uuid;
  UUID vroot = null;
  private final TreeDB db;
  protected UUID cloneParent;
  //protected String id;
  UUID parent;
  
  private ReentrantReadWriteLock workChildLock = new ReentrantReadWriteLock();
  private WriteLock wl = workChildLock.writeLock();
  protected volatile CopyOnWriteArrayList<UUID> workChildren = new CopyOnWriteArrayList<>();
  //immutable cache of children uuids after visual struct modifiers to be shown to user
  //vistual struct modifiers are:
  //-hide by Predicate mechanics - hides node
  //-vnode collapse mechanics - adds vnode's children to shownChildren of its shownParent
  protected List<UUID> shownChildren = null;

  private boolean disposed = false;

  protected ReadLock readWCLock(){
    ReadLock rl = workChildLock.readLock();
    rl.lock();
    return rl;
  }
  protected void readWCUnlock(ReadLock r){
    r.unlock();
  }
  protected boolean writeWCLock(){
    if(wl.isHeldByCurrentThread())
      return false;
    wl.lock();
    return true;
  }
  protected void writeWCUnlock(){
    wl.unlock();
  }
  public String getForeignID() {
    return attrs().getStringAttributeValue(ATTR_DOORS_ID);
  }

  private AttributeStorage attributes;
  // link type(attr name)
  // private Map<TreeNode,Map<String,INodeChangeListener>>revlinks=null;


  protected boolean containsWorkChildUUID(UUID uid) {
    // if(outchildren==null)
    // getChildrenUUIds();
    ReadLock r = null;
    try{
      r = readWCLock();
      return workChildren.contains(uid);
    }
    finally{
      readWCUnlock(r);
    }
  }

  public String getMultiTypeLabel() {
    return getTypeLabel() + "s";
  }

  public final String getTypeLabel() {
    NodeType nt = db.getNodeType(getType());
    if (nt != null) return nt.getVisibleClassName();
    return "undefined";
  }

  protected void addChildUUID(UUID uid) {
    // getWorkChildrenUUIDs();
    if (uid == null) throw new RuntimeException("Attempt to add null child");
    if (uid.equals(getUUId())) {
      RLogger.logInfo("WARN! rec prevented " + getQualifiedId(true));
      throw new RuntimeException("Attempt to add self as child");
    }
    boolean wl = false;
    try{
      wl = writeWCLock();
      if (!workChildren.contains(uid)) {
        boolean b = workChildren.add(uid);
        if (b) {
          if (getParentUUIDs().contains(uid)) throw new RuntimeException("Recursion!");
          clearShownChildren();
          getVS().updateVersion(Type.STRUCT);
          VirtualNode.updateClonesChildren(this);
          // incVersion();
        }
      }
    }
    finally{
      if(wl)
        writeWCUnlock();
    }
  }

  private Set<UUID> getParentUUIDs() {
    Set<UUID> hs = new HashSet<>();
    hs.add(uuid);
    TreeNode par = getWorkParent();
    while (par != null) {
      hs.add(par.getUUId());
      par = par.getWorkParent();
    }
    return hs;
  }

  protected List<Object> getRemovedVirtualChilds() {
    List<Object> l = attrs().getListAttributeValue(ATTR_VIRT_CHILD_REMOVED);
    if (l == null) l = new LinkedList<>();
    return l;
  }

  protected void removeChildUUID(UUID uid) {
    // getWorkChildrenUUIDs();
    boolean b = false, w = false;
    try{
      w = writeWCLock();
      b = workChildren.remove(uid);
      // not vnode
    }finally{
      if(w)
        writeWCUnlock();
    }
    if (b) {
      shownChildren = null;
      getVS().updateVersion(Type.STRUCT);
    }
    // if(isVirtual()||getType().equals(VirtualNode.TYPE_NAME))
    // VirtualNode.updateClonesChildren(this);
  }

  public void setChildrenUUIDs(List<UUID> children2, boolean init) {
      shownChildren = null;
      boolean w = false;
      try{
        w = writeWCLock();
        Set<UUID> old = new HashSet<>();
        if (workChildren != null) {
          old.addAll(workChildren);
        }
        old.addAll(children2);
        old.removeAll(children2);
        boolean changes = setChildren(new ArrayList<>(children2), init);
        if (changes)
          if (!init) {
            VirtualNode.updateClonesChildren(this);
          }
      }
      finally{
        if(w)
          writeWCUnlock();
      }
  }

  public class TreeNodeChildIterator implements Iterator<TreeNode> {
    private int index;
    private boolean sh;

    public TreeNodeChildIterator(boolean shown) {
      sh = shown;
    }

    @Override
    public boolean hasNext() {
      return sh ? index < sizeShownChildren() : index < sizeWorkChildren();
    }

    @Override
    public TreeNode next() {
      return sh ? getShownChild(index++) : getWorkChild(index++);
    }

    @Override
    public void remove() {
      removeChild(sh ? getShownChild(index) : getWorkChild(index));
    }
  }

  /**
   * Elements can be grouped to support versions of storing formats for groups NODE_GROUP is related
   * to default group.
   *
   * @return identifier of node group
   */
  public String getGroup() {
    return NODE_GROUP;
  }

  public boolean isVirtual() {
    return cloneParent != null || getVirtState();
    // if(db == null)
    // return false;
    // return db.isVirtual();
  }

  private boolean getVirtState() {
    return attrs().getLocalStringAttributeValue(ATTR_VIRT_STATUS, null) != null;
  }

  /**
   * Method to get error messages appears during report generation
   *
   * @return error messages if any
   */
  public String getError() {
    //if (db.getStorage() instanceof TreeInstancer) db.updateNodeAttributes(this,,false);
    String s = attrs().getStringAttributeValue(ATTR_ERROR);
    if (s != null) return s;
    return "";
  }

  public TreeNode(TreeDB db, UUID uuid) {
    if (uuid == null) uuid = UUID.randomUUID();
    this.uuid = uuid;
    this.db = db;
    this.disposed = false;
//    if(isRoot())
//      addAPropogator(SortingPropogator.getOne());
//    addAPropogator(PredicatePropogator.getOne());
    addAPropogator(ClonePropagator.getOne());
    addAPropogator(ParentAttrsPropagator.getOne());
    addAPropogator(LocationLabelPropagator.getOne());
    addAPropogator(ReferenceRevertPropagator.getOne());
    // this.vs = VersionsStorage.initOrLoad(db, uuid);
    // String indexStr = attributes.getStringAttributeValue(
    // TreeDB.INDEX_ATTRIBUTE );
    // Attribute a = attrs().getAttribute( TreeDB.ORDER_ATTRIBUTE );
    // if(indexStr != null)
    // index = Index.create( indexStr );
  }

  /**
   * Method to set predicate that affect on node visibility in reports
   *
   * @param uuid2
   * @param pred predicate
   */
  public void setPredicate(String pred) {
    if (pred != null) {
      AttributeDefinition ad = new AttributeDefinition(getUUId(), ATTR_PREDICATE, new ConstValueProvider(AttributeType.STRING, pred));
//      Attribute a = attrs().createAttribute(ATTR_PREDICATE, pred);
  //    a.setAvailability(AttributeAvailability.SUBTREE);
      attrs().setAttribute(ad);
    } else attrs().removeAttribute(ATTR_PREDICATE);
  }

  /**
   * Method to get predicate that affect on node visibility in reports
   *
   * @return predicate
   */
  public String getPredicate() {
    return getPredicate(false);
  }

  public String getPredicate(boolean force) {
    // TODO handle versions
    // RLogger.logInfo(getQualifiedId()+" "+fullPredicate);
    String fullPredicate = null;
    StringBuilder sb = new StringBuilder();
    int addf = 0;
    //TODO proper predicate
    Attribute atrs = attrs().getLocalAttribute(ATTR_PREDICATE);
    String[] tolook = new String[3];
    if (atrs != null) {
      if (atrs.getOrigin().equals(getUUId())
          && atrs.getValue() != null
          && !atrs.getValue().toString().isEmpty())
        tolook[0] = attrs().getStringAttributeValue(atrs.getKey(), null);
      TreeNode par = getWorkParent();
      if (par != null) tolook[2] = par.getPredicate();
      TreeNode cp = getCloneParent();
      if (cp != null) tolook[1] = cp.getPredicate();
      for (int i = 0; i < 3; i++) {
        String sval = tolook[i];
        if (sval != null && !sval.isEmpty())
          if (sb.length() == 0) sb.append(sval);
          else {
            addf++;
            sb.append(")&&(");
            sb.append(sval);
            sb.append(")");
          }
      }
    }
    fullPredicate = sb.toString();
    sb.setLength(0);
    for (int i = 0; i < addf; i++) sb.append("(");
    sb.append(fullPredicate);
    return sb.toString();
  }

  public TreeNode getCloneParent() {
    UUID uid = getCloneParentUUId();
    if (uid == null || uid.equals(uuid)) {
      // TODO clonep?
      return null;
    }
    try {
      return db.getNodeExc(uid);
    } catch (NodeNotFoundException e) {
    }
    return null;
  }

  public UUID getCloneParentUUId() {
    if (cloneParent != null) return cloneParent;
    return null; // VirtualNode.getVirtualStorage(db).getCloneParent(uuid);
  }

  /**
   * Method to set type of element in form of string Types names are usually was in constants like
   * Requirement.TYPE_NAME
   *
   * @param type string name of type
   */
  public void setType(String type) {
    attrs().setConstAttribute(ATTR_TYPE, type, Flag.HIDDEN);
    //attrs().setAttribute(ATTR_TYPE, type);
  }

  /**
   * get UUID - unique identificator over element.
   *
   * @link java.util.UUID
   * @return uuid of element
   */
  public final UUID getUUId() {
    return uuid;
  }

  /**
   * Set identifier of element. Id is identical to name of corresponding file on storage. Needs to
   * be unique for siblings. Id also can form indexes that can be used in sorting.
   *
   * @param id new identifier of element.
   */
  public void setId(String id) {
    id = id.replace((char) 160, ' ');
    attrs().setAttribute(formIdAttribute(uuid, id));
  }

  /**
   * Get identifier of element. Id is identical to name of corresponding file on storage. Needs to
   * be unique for siblings. Id also can form indexes that can be used in sorting.
   *
   * @return current identifier of element.
   */
  public String getId() {
    // if(isVirtual()&&cloneParent==null){/
    // return VirtualNode.getVirtualStorage(db).getNodeId(uuid);
    // }
    // cannot be true for today
    Attribute old = attrs().getAttribute(ATTR_ID);
    if(old == null)
      return null;
    if(old.getValue() == null)
      return null;
    return old.getValue().toString();
    //return id;
  }

  public String getRawId() {
    return attrs().getAttribute(ATTR_ID).getValue().toString();
  }

  /** @return String name of node type */
  public String getType() {
    return attrs().getStringAttributeValue(ATTR_TYPE, "undefined");
  }

  /**
   * Set parent element
   *
   * @param parent new parent element
   */
  public void setParent(TreeNode parent) {
    if (isVirtual() && cloneParent == null) {
    } else {
      TreeNode par = getWorkParent();
      TreeNode targ = this;
      TreeNode newParent = parent;
      if (isVirtual()) {
        if (!parent.isVirtual()) { // TODO is it needed?
          TreeNode np = par;
          if (VirtualNode.TYPE_NAME.equals(np.getType())) {
            if (((VirtualNode) np).isCollapsed()) {
              if (np.getWorkChildrenUUIDs().size() == 1) {
                targ = np;
              }
            }
          }
          if (targ == this) {
            // masterDB = VirtualNode.getRootTree(db);
            CopyPasteHelper cph = CopyPasteHelper.getInstance();
            List<TreeNode> nodes = new ArrayList<>();
            List<NodeDesc> ns = cph.getNodes();
            nodes.add(this);
            cph.setNodes(nodes);
            cph.getNodes().get(0).removeAttribute(ATTR_VIRT_STATUS);
            PasteIntoHandler pih = new PasteIntoHandler();
            Object ret = pih.pasteNodes(newParent, null, true, true, true, true);
            if (ret instanceof String) throw new RuntimeException("Cannot change parent " + ret);
            TreeNode[] nodes2 = (TreeNode[]) ret;
            if (parent.isVirtual())
              (parent.getTreeDB().getStorage()).moveNodeTo(uuid, parent.getUUId());
            cph.setNodesDescs(ns);
            attrs().setConstAttribute(ATTR_VIRT_REMOVED, true, Flag.HIDDEN);
            //TODO del?
            // TODO error to show
            commitAttributes();
            if (nodes2 == null) return;
          }
        } else {
        }
      }
      db.moveNode(targ, parent);
    }
  }

  void fireGUINodeMoved(TreeNode oldParent, boolean internal) {

	if (guilistenerList != null)
	  synchronized (guilistenerList) {
		for (INodeChangeListener i : guilistenerList) i.moved(this, oldParent, internal);
	  }
  }

  void fireNodeMoved(TreeNode oldParent, boolean internal) {
    if (listenerList != null)
      synchronized (listenerList) {
        for (INodeChangeListener i : new ArrayList<>(listenerList))
          i.moved(this, oldParent, internal);
      }
  }

  /**
   * Get unique identifier of parent element. It does not cause parent element loading.
   *
   * @return unique identifier of parent element
   */
  public UUID getWorkParentUUID() {
    return parent;
  }
  public UUID getShownParentUUID() {
    try{
      return getShownParentUUIDExc(); // TODO error??
    }catch(Exception e)
    {
      e.printStackTrace();
    }
    return null;
  }
  public UUID getShownParentUUIDExc() {
    TreeNode node = getShownParentExc(); // TODO change?
    if (node != null) return node.getUUId();
    if (parent == null) return null;
    return null; // TODO error??
  }
  public void removeChild(TreeNode node) {
    removeChild(node,false);
  }

  /**
   * Remove child element from current
   *
   * @param node element to remove
   */
  public void removeChild(TreeNode node,boolean internal) {
    if(!db.getTreeLogic().canBeDeleted(node)&&!internal)
      throw new CannotBeDeleted(node);
    db.startChangeRow();
    try {
      //removeChildUUID(node.getUUId());

      // db.fireNodeGUIRemoved(node,NodeDesc.fromNode(node,false,false),
      // true);//TODO look into
      if (node.isVirtual() && !isVirtual()) {
        UUID uid = node.getVRoot();
        TreeDB rdb = Instantiator.getRootTree(this.db, TreeDB.DEFAULT_DEPTH);
        if (uid != null) {
          TreeDB db = Instantiator.getExistVirtualTreeDB(this.db, uid);
          if (db != null) db.removeNode(node);
          else rdb.removeNode(node);
        } else rdb.removeNode(node);
        // Instantiator.getVirtualTreeDB(db,Instantiator.DEFAULT_INSTANCE_NAME).removeNode(node);
      } else db.removeNode(node);
      // VirtualNode.updateClonesChildren(this);
    } finally {
      db.stopChangeRow();
    }
  }

  /**
   * Set text description for element. As an example may be description of requirements.
   *
   * @param description new description of element
   */
  public void setDescription(String description) {
    if (description == null || description.equals("")) attrs().removeAttribute(ATTR_DESCRIPTION);
    else{
      attrs().setAttribute(new AttributeDefinition(getUUId(), ATTR_DESCRIPTION, AttributeAvailability.LOCAL,
          new HtmlValueProvider(AttributeType.HTML, description),Flag.HIDDEN));
      //attrs().setAttribute(ATTR_DESCRIPTION, description);
    }
  }
  public void setDDescription(HtmlValueProvider hvp) {
    if (hvp == null) attrs().removeAttribute(ATTR_DESCRIPTION);
    else{
      attrs().setAttribute(new AttributeDefinition(getUUId(), ATTR_DESCRIPTION, AttributeAvailability.LOCAL,
          hvp,Flag.HIDDEN));
      //attrs().setAttribute(ATTR_DESCRIPTION, description);
    }
  }
  /**
   * Get current text description for element. As an example may be description of requirements.
   * This method returns description prepared for usage in reports. Picture paths are replaced to
   * corresponding paths in report folder.
   *
   * @param prefix optional prefix for description
   * @param postfix optional postfix for description
   * @param basePath path to report folder
   * @return description prepared to be used in report
   */
  public String getADescription(String basePath) {
    return getADescription("<div>", "</div>", basePath, null);
  }

  public String getADescription(String prefix, String postfix, String basePath, IReqURIResolver rslv) {
    String desc = getDescription(false, false);
    if (desc == null) return null;
    try {
      URI rp = getTreeDB().getStorage().getNodeResourcePath(getUUId());
      return resolveURI(rslv, RequalityCLI.getReportDescription(
          this, rp.toString(), desc, prefix, postfix, basePath));
    } catch (Exception e) {
      // TODO Auto-generated catch block
      e.printStackTrace();
    }
    return null;
  }

  protected String resolveURI(IReqURIResolver rslv, String descr) {
    if(rslv == null)
      return descr;
    StringBuffer ret = new StringBuffer();
    int im = descr.indexOf("<a"), lim = 0; // ,im2=descr.indexOf("<object");
    while (im > -1) {
      ret.append(descr.substring(lim, im));
      int end = descr.indexOf(">", im);
      if (end == -1) break; // not ended one
      int be = end-1;
      while(be>im && Character.isWhitespace(descr.charAt(be))){
        be--;
      }
      boolean nosec = descr.charAt(be) == '/';
      String val = "";
      int endatag = -1;
      if(!nosec){
        endatag = descr.indexOf("</a",end);
        val = descr.substring(end+1, endatag);
      }
      int hndl = im + 2;
      Map<String, Pair<Integer, String>> params = RequalityCLI.gatherParams(descr,hndl,end);
      Pair<Integer, String> par = null;
      for(Entry<String, Pair<Integer, String>> pp: params.entrySet()){
        String key = pp.getKey().toLowerCase(); 
        if(key.endsWith(":href")||key.equals("href"))
            par = pp.getValue();
      }
      //par = params.get("href");
      {
        char pth = 0;;
        byte state = 0;
        if(par!=null && par.getFirst()>-1){
          String href = par.getSecond().trim();
          if(href.startsWith(REQ_URI_START)){
            href = href.substring(REQ_URI_START.length());
            ReqURIResolveResult rs = rslv.resolve(this, href, val);
            if (rs != null) {
              ret.append(descr.substring(im, par.getFirst()));
              ret.append(rs.link);
              im = par.getFirst()+par.getSecond().length();
              if(nosec){
                ret.append(descr.substring(im, be));
                ret.append(">");
                ret.append(rs.value);
                ret.append("</a>");
                im = end+1;
              }else
              {
                ret.append(descr.substring(im, end+1));
                ret.append(rs.value);
                im = endatag;
              }
            } else {
              // ret.append(path);
            }
          }
        }
        
        lim = im;
      }
      im = descr.indexOf("<a",end);
    }
//    if(n!=null)
//      for (String s : toDel) {
//        // TODO delete or stay?
//  
//  //      n.deleteResource(s);
//      }
    // while (matcher.find()) {
    // if(matcher.group(1).equals("img") && matcher.group(2).equals("src")
    // ||matcher.group(1).equals("object") &&
    // matcher.group(2).equals("data")){
    // }
    // }
    if (lim > -1) ret.append(descr.substring(lim));
    return ret.toString();
  }

  public String getLocatonsDescription(String base) {
    return getLocatonsDescription("<div>", "</div>", base, " ", null);
  }

  public String getDescription(String prefix, String postfix, String basePath, IReqURIResolver reslv) {
    return getDescription(prefix, postfix, basePath, " ",reslv);
  }

  public String getDescription(String prefix, String postfix, String basePath, String separator,
      IReqURIResolver reslv) {
    String desc = getADescription(prefix, postfix, basePath, reslv);
    if (desc == null) {
      desc = getLocatonsDescription(prefix, postfix, basePath, separator, reslv);
      if (desc == null) return null;
      try {
        URI path = getTreeDB().getStorage().getNodeResourcePath(getUUId());
        return CoreUtils.htmlEncapsulate(prefix+
            RequalityCLI.getReportDescription(
                this, path != null ? path.toString() : null, desc, prefix, postfix, basePath)+postfix);
      } catch (Exception e) { // TODO Auto-generated catch block
        e.printStackTrace();
      }
      return null;
    }
    return CoreUtils.htmlEncapsulate(prefix+desc+postfix);
  }

  public String getLocatonsDescription(
      String prefix, String postfix, String basePath, String separator,
      IReqURIResolver rslv) {
    Attribute locs = attrs().getAttribute(Requirement.ATTR_LOCATIONS);
    if (locs != null && locs.getValue() != null) {
      List<?> res = ((ListValue) locs.getValue()).getValues();
      StringBuilder sb = new StringBuilder();
      Iterator<?> it = res.iterator();
      while (it.hasNext()) {
        ReferenceValue s = (ReferenceValue) it.next();
        if(s.hasErrorsOnResolve(attrs()))
          continue;
        TreeNode nd = s.getNodes(attrs())[0];
        if (nd != null) {
          sb.append(resolveURI(rslv,nd.getDescription(prefix, postfix, basePath, separator,rslv)));
          if (it.hasNext() && !sb.toString().endsWith(separator)) sb.append(separator);
        }
      }
      return sb.toString();
    }
    return null;
  }

  public String getLocatonsDescription(boolean process, String separator) {
    Attribute locs = attrs().getAttribute(Requirement.ATTR_LOCATIONS);
    if (locs != null && locs.getValue() != null) {
      List<?> res = ((ListValue) locs.getValue()).getValues();
      StringBuilder sb = new StringBuilder();
      Iterator<?> it = res.iterator();
      while (it.hasNext()) {
        ReferenceValue s = (ReferenceValue) it.next();
        if(s.hasErrorsOnResolve(attrs()))
          continue;
        TreeNode nd = s.getNodes(attrs())[0];
        if (nd != null) {
          sb.append(nd.getDescription(process, false));
          if (it.hasNext()) sb.append(separator);
        }
      }
      return sb.toString();
    }
    return null;
  }

  /**
   * Get current text description for element. As an example may be description of requirements.
   * This method returns description prepared for usage in reports. Picture paths are replaced to
   * corresponding paths in report folder.
   *
   * @param basePath path to report folder
   * @return description prepared to be used in report
   */
  public String getDescription(String basePath) {
    return getDescription(basePath, null);
  }
  public String getDescription(String basePath, IReqURIResolver rslv) {
    String result = getDescription("<div>", "</div>", basePath,rslv);
    return result;
  }

  /**
   * Return element description to be shown in uni editor and properties panels. Paths for
   * resources(for ex. images) are not processed
   *
   * @return element description
   */
  public String getDescription() {
    return getDescription(false, false);
  }

  /**
   * Return element description to be shown in uni editor and properties panels
   *
   * @param process process resource path flag
   * @param useLocs TODO
   * @return element description
   */
  public String getDescription(boolean process, boolean useLocs) {
    String desc = attrs().getStringAttributeValue(ATTR_DESCRIPTION);
    String ret = null;
    if (desc != null) {
      ret = desc;
    } else if (useLocs) ret = getLocatonsDescription(process, " ");
    if (process && ret != null) {
      ret = processResourcesPath(ret);
    }
    return ret;
  }

  public String getRawDescription() {
    return getRawDescription(false);
  }

  public String getRawDescription(boolean process) {
    Attribute desc = attrs().getAttribute(ATTR_DESCRIPTION);
    String ret = null;
    if (desc != null && desc.getDefinition().getRawValue()!=null && desc.getDefinition().getRawValue().getValue()!=null) {
      ret = desc.getDefinition().getRawValue().getValue().toString();
    }
    if (process && ret != null) {
      ret = processResourcesPath(ret);
    }
    return ret;
  }

  /**
   * Process paths in description, replace placeholder {node.resURL} to actual path to images
   *
   * @param desc description of element
   * @return description of element with correct path to resources of node
   */
  public String processResourcesPath(String desc) {
    URI resPath = getNodeResourcesPath();
    if (resPath != null) {
      int pos = desc.indexOf("{node.resURL}");
      while (pos > -1) {
        char q = '"';
        byte mode = 0;
        int apos = pos;
        while (mode == 0 && apos > 0) {
          apos--;
          char c = desc.charAt(apos);
          if (c == '"' || c == '\'') {
            q = c;
            mode = 1;
          }
        }
        if (mode == 0) {
          apos = -1;
          pos = desc.length();
        } else pos = desc.indexOf(q, pos);
        if (pos > -1) {
          String path =
              desc.substring(apos + 1, pos)
                  .replace("{node.resURL}", resPath.toString())
                  .replace("\\", "/");
          desc = desc.substring(0, apos + 1) + path + desc.substring(pos);
          pos = desc.indexOf("{node.resURL}");
        }
      }
      // result = desc.replace("", resPath);
    }
    return desc;
  }

  /**
   * Get tree database in which this element exists
   *
   * @return tree database
   */
  public TreeDB getTreeDB() {
    if (!db.isVirtual() && getVRoot() != null) return Instantiator.getExistVirtualTreeDB(db, getVRoot());
    return db;
  }

  /**
   * Set unique identifier of element
   *
   * @link java.util.UUID
   * @param uuid new unique identifier
   */
  void setUUId(UUID uuid) {
    this.uuid = uuid;
  }

  public TreeNode getShownParent() {
    try{
      return getShownParentExc();
    }
    catch(Exception e){//if cannot load then return null;
      RLogger.logInfo(CoreUtils.getTrace(e));
    }
    return null;
  }
  /**
   * Get parent of current element. Causes element loading if it is not in tree database.
   *
   * @return parent element
   */
  public TreeNode getShownParentExc() throws NodeNotFoundException {
    if (parent == null) {
      // no parent
      assert db == null
          || db.getRootNodeUUID() == null
          || db.getRootNodeUUID().equals(this.getUUId());
      // return null;
    }
    TreeNode node = null;
    if (db.isNodeExist(parent)
        && VirtualNode.TYPE_NAME.equals(db.getNodeType(parent))) {
      boolean find = false;
      Object clps = db.getAttributeValue(parent, VirtualNode.ATTR_COLLAPSEVNODE, false, false);
      if (clps == null || clps.toString().equals("true")) find = true;
      // List<Attribute>
      // hdn=db.getStorage().getAttributes(parent).get(VirtualNode.ATTR_COLLAPSEVNODE);

      // if(hdn!=null&&hdn.iterator().hasNext()&&
      // hdn.iterator().next().getValue().equals(true))
      // find = true;
      if (find) node = db.getNodeExc(parent).getShownParent();
      else node = db.getNodeExc(parent);
      // TODO do smth with code below
      // if(node.getCloneParent()==null)
      // node = Instantiator.getBaseTree(db)
      // .getNode(node.getUUId());
    } else if (parent != null) node = db.getNodeExc(parent);
    else return null;
    if (node!=null && VirtualNode.TYPE_NAME.equals(node.getType())
        && node.getCloneParent() == null
        && (node.isVirtual() || db.isVirtual())) {
      try {
        TreeNode nde2 = Instantiator.getRootTree(db).getNodeExc(node.getUUId());
        if (nde2 != null) return nde2;
      } catch (Exception e) {
        e.printStackTrace();
      }
    } 
    return node;
  }

  /**
   * Get parent of current element. Causes element loading if it is not in tree database.
   *
   * @return parent element
   */
  public TreeNode getWorkParent() {
    if (parent == null) {
      // no parent
      assert db == null
          || db.getRootNodeUUID() == null
          || db.getRootNodeUUID().equals(this.getUUId());
      // return null;
    }
    try {
      TreeNode node = null;
      try {
        if (parent != null) node = db.getNode(parent);
        else return null;

      } catch (RuntimeException re) {
        RLogger.logInfo("cannot load parent for "+uuid+" "+CoreUtils.getTrace(re));
      }
      return node;
    } catch (Exception e) {
      e.printStackTrace();
    }
    return null;
  }

  /**
   * Get children count for current element.
   *
   * @return children count
   */
  public int sizeShownChildren() {
    return sizeShownChildren(false);
  }

  public int sizeShownChildren(boolean onlyloaded) {
    if (shownChildren == null) getShownChildrenUUIDs(onlyloaded);
    if (shownChildren != null) return shownChildren.size();
    throw new RuntimeException("Shown children cannot be formed!");
    // return 0;
    // return getChildren().size();
  }

  public TreeNode getShownChild(int i) {
    return getShownChild(i, false); // TODO think
  }

  /**
   * Get child element with index i in children uuid list
   *
   * @param i index of child in uuid list
   * @param onlyloaded
   * @return child element
   */
  public TreeNode getShownChild(int i, boolean notcheck) {
    return getShownChild(i, notcheck, false);
  }

  public TreeNode getShownChild(int i, boolean notcheck, boolean onlyloaded){
    if (shownChildren == null) getShownChildrenUUIDs(onlyloaded);
    TreeNode node = null;
    if (shownChildren != null)
      if (onlyloaded) node = db.getCachedNodeOrVirt(shownChildren.get(i));else 
      try {
        node = db.getNode(shownChildren.get(i), TreeDB.DEFAULT_DEPTH, notcheck);
      } catch (NodeNotFoundException e){// TODO Auto-generated catch block
        RLogger.logInfo(CoreUtils.getTrace(e));
      }
    else throw new RuntimeException("Shown children cannot be formed!");
    // node =
    // db.getNode(getChildren().get(i),TreeDB.DEFAULT_DEPTH,notcheck);
    if (node != null && node.isVirtual() && !db.isVirtual()) {
      UUID uid = node.getVRoot();
      if (onlyloaded)
        return Instantiator.getExistVirtualTreeDB(db, uid).getCachedNodeOrVirt(node.getUUId());
      else return Instantiator.getExistVirtualTreeDB(db, uid).getNode(node.getUUId());
    }
    return node;
  }

  public TreeNode getRoot() {
    UUID vr = getVRoot();
    if (vr != null) {
      if (vr.equals(getUUId())) return this;
      return db.getNode(vr);
    }
    return db.getRootNode();
  }

  public boolean isVRoot() {
    // see treeDb.getnode
    UUID vr = getVRoot();
    if (vr != null) return uuid.equals(vr);
    return Instantiator.isVRoot(db, uuid, attrs().getLMap());
    // ITreeStorage storage = getTreeDB().getStorage();
    // UUID par = storage.getParent(uuid);
    // if(par!=null && storage.getNodeId(par,true).equals("Instances")){
    // par = storage.getParent(par);
    // if(par!=null&&storage.getRootNode().equals(par))
    // return true;
    // }
    // return false;
  }

  /**
   * Get path to element in form of identifier(id) of element and all its ancestors in form of
   * /RootId/.../ElementParentId/ElementId
   *
   * @return path to element
   */
  public String getQualifiedId() {
    return getQualifiedId(false);
  }
  public String getQualifiedId(boolean inner) {
    StringBuilder sb = new StringBuilder();
    getQualifiedId(sb, new HashSet<TreeNode>(), inner);
    return sb.toString();
  }

  private void getQualifiedId(StringBuilder sb, HashSet<TreeNode> hashSet, boolean inner) {
    hashSet.add(this);
    if (!inner && isDisposed()) return;
    //UUID paruid = getWorkParentUUID();
    TreeNode par = null;
    try{
       par = db.getNodeExc(parent);
    }
    catch(NodeNotFoundException e){
      //cutted subtree
      return;
    }
    if (hashSet.contains(par)) {
      RLogger.logInfo("recursion error!");
      return;
    }
    if (par == null || isVRoot()) {
      sb.append('/');
      return;
    }
    par.getQualifiedId(sb, hashSet, inner);

    /* append '/' if sb != "/" */
    if (sb.length() != 1) {
      sb.append('/');
    }
    sb.append(getId());
  }

  /**
   * Method to locate child with given id. Return null if it is not found.
   *
   * @param id id of child to find
   * @return child element
   */
  public TreeNode findWorkChild(String id) {
    id = id.replace((char) 160, ' ');
    if (id.startsWith("|:") && id.endsWith(":|")) { // child num
      int num = Integer.parseInt(id.substring(2, id.length() - 2));
      if (num < sizeWorkChildren()) {
        return getSortedWorkChildren()[num];
      }
      return null;
    }

    List<UUID> uids = getWorkChildrenUUIDs();
    for (UUID uid : uids) {
      try {
        TreeNode child = db.getNode(uid, TreeDB.DEFAULT_DEPTH, false);
        if (child.getRawId().equals(id)) {
          return child;
        }
        if (child.getId().toLowerCase().equals(id.toLowerCase())) {
          return child;
        }
        if (child.getUUId().toString().equals(id)) return child;
      } catch (Exception e) {
        //e.printStackTrace();
      }
    }
    return null;
  }

  /**
   * Get list of children UUIDs
   *
   * @return list of children UUIDs
   */
  public static String join(String r[], String d) {
    if (r.length == 0) return "";
    StringBuilder sb = new StringBuilder();
    int i;
    for (i = 0; i < r.length - 1; i++) sb.append(r[i] + d);
    return sb.toString() + r[i];
  }

  public List<UUID> getShownChildrenUUIDs() {
    return getShownChildrenUUIDs(false);
  }

  public List<UUID> getShownChildrenUUIDs(boolean onlyloaded) {
    // if(shownChildren!=null &&
    // !getVS().isParentVersionOutdated(Type.ATTRIBUTES))
    // return shownChildren;

    List<UUID> ch2 = new ArrayList<>();
    @SuppressWarnings("unused")
    boolean vchild = false;
    List<UUID> childs = new LinkedList<>(getWorkChildrenUUIDs(onlyloaded));
    for (UUID uid : childs) {
      if (uid == null) continue;
      if (getTreeDB().isNodeExist(uid)) {
        String type = getTreeDB().getNodeType(uid);
        if (type == null && !getTreeDB().isVirtual())
          for (TreeDB db : Instantiator.getVirtualTreeDBs(getTreeDB())) {
            type = db.getNodeType(uid);
            if (type != null) break;
          }
        if (type != null && VirtualNode.TYPE_NAME.equals(type)) {
          VirtualNode node = (VirtualNode) getTreeDB().getNode(uid);
          if (!node.isCollapsed()) ch2.add(uid);
          else {
            vchild = true;
            ch2.addAll(node.getWorkChildrenUUIDs(onlyloaded));
            ; // TODO mb shown here?
          }
        } else ch2.add(uid);
      }
    }
    if (
    /* (isVirtual()||vchild)&& */ ch2.size() > 0 && !isVRoot()) {
      ch2 = TreeDB.updateVisablility(getTreeDB(), ch2);
    }
    shownChildren = Collections.unmodifiableList(ch2); 
    return ch2;
  }


  public String getLabel() {
    Attribute a = attrs().getAttribute(ATTR_LABEL);
    if(a!=null && a.getValue()!=null)
      return a.getValue().toString();
    return null;
  }
  
  public Set<String> getLabelDeps(){
    if (db != null && !db.textIsSimple) {
      return Attribute.getHTMLNames();
    }
    return labelDeps;
  }
  
  public String formLabel(){
    if (db != null && !db.textIsSimple) {
      String desc = getDescription();
      if (desc != null && desc.length() > 5)
        return desc.substring(0, Math.min(desc.length(), 40)) + "\t[" + getId() + "]";
    }
    return getNameOrId();
  }
  
  @Override
  public String toString() {
    return getLabel();
  }

  /**
   * Create child element with given id and type "Node"
   *
   * @param id id of created element
   * @return new child element
   */
  public TreeNode createChild(String id) {
    if (findWorkChild(id) != null)
      throw new IllegalArgumentException("element with id '" + id + "' already exists");
    id = id.replace((char) 160, ' ');
    shownChildren = null;
    if (getType().equals("DBRoot")) {
      TreeNode node = db.createChildNode(uuid, id, "Node");
      return node;
    }
    TreeNode node = db.createChildNode(uuid, id, getType());
    // addChildUUID(node.getUUId());//TODO may be return
    return node;
  }

  public TreeNode createChild(String id, String type) {
    return createChild(id, type, TreeDB.EMPTY_TEMPLATE);
  }

  /**
   * Create child element with given id and type type
   *
   * @param id id of created element
   * @param templ
   * @type string type of element
   * @return new child element
   */
  public TreeNode createChild(String id, String type, String templ) {
    // TODO merge methods
    if (findWorkChild(id) != null)
      throw new IllegalArgumentException("element with id '" + id + "' already exists");
    id = id.replace((char) 160, ' ');
    shownChildren = null;
    TreeNode node = db.createChildNode(uuid, id, type, templ);
    return node;
  }

  public TreeNode createChild(UUID parent, UUID uid, String id, Map<String, AttributeDefinition> attrs) {
    if (findWorkChild(id) != null)
      throw new IllegalArgumentException("element with id '" + id + "' already exists");
    id = id.replace((char) 160, ' ');
    shownChildren = null;
    TreeNode node = db.createChildNode(uid, parent, id, attrs);
    return node;
  }

  public void delete() {
    delete(false);
  }
  /** Delete this element */
  public void delete(boolean internal) {
    if(!internal&&!db.getTreeLogic().canBeDeleted(this))
      throw new CannotBeDeleted(this);
    if (getWorkParent() != null) 
      getWorkParent().removeChild(this,internal);
    else {
      // Instantiator.getRootTree(db, TreeDB.DEFAULT_DEPTH);
      db.removeNode(this);
    }
  }

  // private void updateDirectCloneAttributes(Map<String, Attribute> adiff) {
  // if(uuid!=null && db!=null && db.getStorage()!=null &&
  // uuid!=db.getRootNodeUUID()&&!isVirtual()
  // &&!VirtualNode.getVirtualStorage(getTreeDB()).hasCloneParent(uuid)){
  // TreeNode clone = VirtualNode.getVirtualTreeDB(getTreeDB()).getNode(uuid);
  // Map<String,Attribute> attrs = clone.attributes;
  // for(String s : attrs.keySet()){
  // if(!attrs.containsKey(s)){
  // attrs.put(s, new LinkedList<Attribute>(attributes.get(s)));
  // }else
  // if(!attributes.containsKey(s))
  // attrs.remove(s);
  // else
  // attrs.put(s, new LinkedList<Attribute>(attributes.get(s)));
  // }
  // }
  // }
  //
  public boolean affectsPredicate(String what) {
    if (what.equals(ATTR_PREDICATE)) return true;
    String pred = getPredicate();
    if (pred != null && !pred.isEmpty()) {
      // for(String s:attributes.getRevertDependencies(ATTR_PREDICATE))
      if (pred.indexOf(what) > -1) return true;
    }
    return false;
  }

  public AttributeDefinition reactBeforeAttributeChange(String s, AttributeDefinition nlist) {
//    if (s.equals(ATTR_DESCRIPTION)) {
//      if (nlist != null) {
//        //TODO check
//        Object odesc = nlist.getRawValue().getValue(), raw = nlist.getRawValue().getValue();
//        // if(raw!=null){
//        // if desc is null then we also needs to delete all images
//        String desc = raw == null ? null : raw.toString();
//        desc =
//            RequalityCLI.updateImages(getNodeResourcesPath(), desc, this).replace((char) 160, ' ');
//        if (raw != null) {
//          try {
//            // TODO can be removed?
//            desc = desc.replace(this.getNodeResourcesPath().toString(), "{node.resURL}");
//          } catch (Exception e) { // unsop
//          }
//          return new AttributeDefinition(nlist.getOrigin(),
//              nlist.getKey(), new TemplValueProvider(AttributeType.HTML, desc));
//        }
//        // }
//      }
//    }
    return null;
  }

  public void reactAfterAttributeChange(
      String s, final Attribute attribute) {
    // attributes.put(s, attribute);
    // TODO activities on attribute change
    //if(s.equals(TreeNode.ATTR_ID)){
//      updateTreeNodes(this, TreeNode.ATTR_ID);
  //  }
    if (s.equals(TreeNode.ATTR_CLONE_PARENT)) {
      Object nw = null;
      if (attribute != null) nw = attribute.getDefinition().getRawValue().getValue();
      if (attribute != null && nw != null) cloneParent = UUID.fromString(nw.toString());
      return;
    }
    if (s.equals(ATTR_VIRT_STATUS)) {
      Object nw = null;
      if (attribute != null) nw = attribute.getValue();
      vroot = null;
      if (attribute != null && nw != null) vroot = UUID.fromString(nw.toString());
      return;
    }
    if (s.equals(ATTR_HIDDEN)) {
      TreeNode par = getShownParent();
      if (par != null) {
        // error(
        par.shownChildren = null;
        //par.getTreeDB().updateNodeChildren(par); // TODO check;
        // par = getParent();
        // if(par!=null)
        // par.setOutchildren(null);
      }
      return;
    }

    if (s.equals(ATTR_PREDICATE) || affectsPredicate(s)) {
      Attribute pred = attrs().getAttribute(ATTR_PREDICATE);
      if (pred == null || pred.getValue().equals(""))
        attrs().removeAttribute(ATTR_PREDICATE);
      //
      TreeNode nde = getShownParent();
      if (nde != null) {
        nde.shownChildren = null;
        //nde.getTreeDB().updateNodeChildren(nde);
      }
    }
      
    if (s.equals(ATTR_DESCRIPTION)) {
      Object nw = null;
      if (attribute != null) nw = attribute.getValue();
      if (attribute == null || (nw == null || nw.equals("")))
        attrs().removeAttribute(ATTR_DESCRIPTION);
      // if(desc.contains(""))
    }

    if (s.equals(TreeDB.INDEX_ATTRIBUTE))
      if (attribute != null) {
        Object nw = attribute.getValue();
        if (nw != null) this.index = (Index.create(nw.toString()));
        else this.index = Index.create(Integer.toString(Integer.MAX_VALUE));
      }

    // //Links core
    // updateLink(s, oldAttribute,attribute);
    // if(s.equals(ATTR_REVLINKS)){
    // revlinks = null;
    // getRevertLinks();
    // }
    if (s.equals(ATTR_NAME)) {
      getTreeDB().updateRequirementName(this);
      // fire related to name update actions
      // TODO update needed?
      //updateTreeNodes(this, TreeNode.ATTR_NAME);
    }
    // TODO move to db
    TreeNode templ = db.getTemplatesFolder();
    if (templ != null && getUUId().equals(templ.getUUId())) {
      if (s.equals(TreeDB.ACTIVE_TEMPLATES_ATTRIBUTE)) db.loadActiveTemplates();
    }
  }

  @Override
  public boolean equals(Object obj) {
    if (obj == null) return false;
    if (super.equals(obj)) return true;
    if (!(obj instanceof TreeNode)) return false;
    TreeNode node = (TreeNode) obj;
    if (!node.getUUId().equals(uuid)) return false;
    if (!getId().equals(node.getId())) return false;
    if (getWorkParentUUID() == null ^ node.getWorkParentUUID() == null
        || getWorkParentUUID() != null
            && node.getWorkParentUUID() != null
            && !getWorkParentUUID().equals(node.getWorkParentUUID())) return false;
    if (attrs().hashCode() != (node.attrs().hashCode()))
      // TODO rewrite
      return false;
    // for (String key : attrs().getAttributeKeys()) {
    // Object v1 =
    // attrs().getAttributeValue(key),v2=node.attrs().getAttributeValue(key);
    // if (v1==null ^ v2==null || !v1.equals(v2))
    // return false;
    // }
    return true;
  }

  /**
   * Get children elements in array sorted by default sort mechanism
   *
   * @return array of children elements
   */
  public TreeNode[] getSortedShownChildren() {
    return getSortedShownChildren(null, null, false);
  }

  /**
   * Get children elements of selected type in array sorted by default sort mechanism
   *
   * @param type required type of children
   * @return array of children elements of selected type
   */
  public TreeNode[] getSortedShownChildren(String type) {
    return getSortedShownChildren(null, type, false);
  }

  public TreeNode[] getSortedShownChildren(Comparator<TreeNode> c, String type) {
    return getSortedShownChildren(c, type, false);
  }

  /**
   * Get children elements of selected type in array sorted by order defined in given comparator
   *
   * @param c comparator that defines order of return elements
   * @param type required type of children
   * @param onlyloaded
   * @return array of children elements of selected type
   */
  public TreeNode[] getSortedShownChildren(Comparator<TreeNode> c, String type, boolean notcheck) {
    return getSortedShownChildren(c, type, notcheck, false);
  }

  public TreeNode[] getSortedShownChildren(
      Comparator<TreeNode> c, String type, boolean notcheck, boolean onlyloaded) {
    TreeNode[] r = getShownChildren(type, notcheck, onlyloaded);
    if (c == null) c = TreeNodeComparator.getComparator(Instantiator.getRootTree(db));
    Arrays.sort(r, c);
    return r;
  }

  public TreeNode[] getSortedWorkChildren() {
    return getSortedWorkChildren(null, null, false);
  }

  /**
   * Get children elements of selected type in array sorted by default sort mechanism
   *
   * @param type required type of children
   * @return array of children elements of selected type
   */
  public TreeNode[] getSortedWorkChildren(String type) {
    return getSortedWorkChildren(null, type, false);
  }

  public TreeNode[] getSortedWorkChildren(Comparator<TreeNode> c, String type, boolean notcheck) {
    return getSortedWorkChildren(c, type, notcheck, false, false);
  }

  public TreeNode[] getSortedWorkChildren(
      Comparator<TreeNode> c, String type, boolean notcheck, boolean onlyloaded, boolean ignoreCPUP) {
    TreeNode[] r = getWorkChildren(type, notcheck, onlyloaded, ignoreCPUP);
    if (c == null) c = TreeNodeComparator.getComparator(db);
    Arrays.sort(r, c);
    return r;
  }

  //
  // public TreeNode[] getRawChildren(String type,boolean onlyloaded){
  // ArrayList<TreeNode> result = new ArrayList<TreeNode>();
  // int num = getChildren().size();
  // for (int i = 0; i < num; i++) {
  // UUID chi = getChildren().get(i);
  // TreeNode ch = null;
  // if(VirtualNode.TYPE_NAME.equals(getType())||id.equals("Instances"))
  // ch = onlyloaded?db.getCachedNodeOrVirt(chi):db.getNode(chi);
  // else
  // ch = onlyloaded?db.getCachedNode(chi):db.getNode(chi);
  // if (ch!=null && (type == null || ch.getType().equals(type))) {
  // result.add(ch);
  // }
  // }
  // TreeNode[] r = (TreeNode[]) result.toArray(new TreeNode[0]);
  // return r;
  // }
  public TreeNode[] getShownChildren() {
    return getShownChildren(null, false);
  }

  public TreeNode[] getShownChildren(String type) {
    return getShownChildren(type, false);
  }

  public TreeNode[] getShownChildren(String type, boolean notcheck) {
    return getShownChildren(type, notcheck, false);
  }

  public TreeNode[] getShownChildren(String type, boolean notcheck, boolean onlyloaded) {
    ArrayList<TreeNode> result = new ArrayList<>();
    List<UUID> uids = getShownChildrenUUIDs(onlyloaded);
    //int num = sizeShownChildren(onlyloaded);
    for (int i = 0; i < uids.size(); i++) {
      TreeNode ch = null;
      try {
        ch = getNode(uids.get(i), notcheck, onlyloaded);
      } catch (RuntimeException ex) {
        RLogger.logInfo("Missing node! " + getQualifiedId(true) + " " + ch +"\n"+CoreUtils.getTrace(ex));
        shownChildren = null;
        //ex.printStackTrace();
      }
      if (ch != null && (type == null || type.equals(ch.getType()))) {
        result.add(ch);
      }
    }
    TreeNode[] r = result.toArray(new TreeNode[0]);
    return r;
  }

  public TreeNode[] getWorkChildren() {
    return getWorkChildren(null, false);
  }

  public TreeNode[] getWorkChildren(String type) {
    return getWorkChildren(type, false);
  }

  public TreeNode[] getWorkChildren(String type, boolean notcheck) {
    return getWorkChildren(type, notcheck, false, false);
  }

  public TreeNode[] getWorkChildren(String type, boolean notcheck, boolean onlyloaded, boolean ignoreCPUP) {
    return getWorkChildren(type,notcheck,onlyloaded,ignoreCPUP,true);
  }
  public TreeNode[] getWorkChildren(String type, boolean notcheck, boolean onlyloaded, boolean ignoreCPUP, boolean logNPE) {
    ArrayList<TreeNode> result = new ArrayList<>();
    List<UUID> uids = getWorkChildrenUUIDs(ignoreCPUP);
    //int num = sizeWorkChildren(onlyloaded);
    for (int i = 0; i < uids.size(); i++) {
      try{
        TreeNode ch = getNodeExc(uids.get(i),notcheck,onlyloaded);
        if (ch != null && (type == null || type.equals(ch.getType()))) {
          result.add(ch);
        }
      }catch(NodeNotFoundException nfe){
        if(logNPE)
          RLogger.logInfo(CoreUtils.getTrace(nfe));
      }
    }
    TreeNode[] r = result.toArray(new TreeNode[0]);
    return r;
  }

  /**
   * Get path to resources folder related to current element
   *
   * @return path to resources folder
   */
  public URI getNodeResourcesPath() {
    try {
      return db.getStorage().getNodeResourcePath(uuid);
    } catch (Exception e) {
      // TODO Auto-generated catch block
      e.printStackTrace();
    }
    return null;
  }

  public File getNodeResourcesFile() {
    URI path = getNodeResourcesPath();
    if (path == null) return null;
    return new File(path);
  }

  /**
   * Get intermediate representation of element for copying process.
   *
   * @return NodeDesc with information about current element
   */
  public NodeDesc toNodeDesc() {
    return toNodeDesc(true, true);
  }

  /**
   * Get intermediate representation of element for copying process.
   *
   * @param deep make intermediate representation of children
   * @return NodeDesc with information about current element(if deep is true then with all
   *     descendants)
   */
  public NodeDesc toNodeDesc(Boolean deep, boolean keepattrs) {
    return NodeDesc.fromNode(this, deep, keepattrs, false);
  }

  /**
   * Get content of resource with given name corresponding to this element
   *
   * @param name name of resource to get
   * @return content of resource in form of stream to read
   */
  public InputStream getResourceContent(String name) {
    if (name == null) return null;
    if (!db.getStorage().isResourceExists(uuid, name)) return null;
    return db.getStorage().getResourceContents(uuid, name);
  }

  /**
   * Get java file object with given name
   *
   * @param name name of resource
   * @return java file
   */
  public File getResource(String name) {
    if (name == null) return null;
    if (!db.getStorage().isResourceExists(uuid, name)) return null;
    return db.getStorage().getResource(uuid, name);
  }

  /**
   * List all available resources for current element
   *
   * @return list of resources names
   */
  public Set<String> listResources() {
    return db.getStorage().listResources(uuid);
  }

  /**
   * Set content of resource with given name on current node
   *
   * @param name name of resource
   * @param in content of resource in form of stream
   */
  public void setResourceContent(String name, InputStream in) {
    db.setResourceContent(this, name, in, null);
    try {
      in.close();
    } catch (IOException e) {
    }
  }

  /**
   * Set content of resource with given name on current node
   *
   * @param name name of resource
   * @param in content of resource in form of stream
   * @param data deprecated
   */
  public void setResourceContent(String name, InputStream in, Object data) {
    db.setResourceContent(this, name, in, data);
  }

  /**
   * Delete resource with given name from current element
   *
   * @param name name of resource to delete
   */
  public void deleteResource(String name) {
    db.deleteResource(this, name);
  }

  /**
   * Check if resource with given name exists on current element
   *
   * @param name name of resource to check
   * @return true if resource exists, false otherwise
   */
  public boolean isResourceExists(String name) {
    return db.isResourceExists(this, name);
  }

  protected Index index;

  public boolean isIndexUndefined() {
    return getIndex().isUndefined();
  }

  /**
   * Get index of element. Index is specific field that is used to define custom order of elements
   * in sorted list
   *
   * @return index of element
   */
  public Index getIndex() {
    if (index == null) {
      Attribute inda = attrs().getAttribute(TreeDB.INDEX_ATTRIBUTE);
      String indexStr = null;
      if (inda != null&&inda.getValue()!=null) indexStr = inda.getValue().toString();
      Attribute a = attrs().getAttribute(TreeDB.ORDER_ATTRIBUTE);
      if (a != null) {
        Object nw = a.getValue();
        if (nw instanceof String) {
          if (indexStr == null) indexStr = (String) nw; // old behaviour
        } else {
          if (indexStr == null) indexStr = ((Integer) nw).toString();
        }
      }
      if (indexStr != null) {
        index = Index.create(indexStr);
      } else {
        index = Index.getUndefined();
        ;
      }
    }
    return index;
  }

  public void removeIndex() {
    attrs().removeAttribute(TreeDB.INDEX_ATTRIBUTE);
    this.index = null;
  //  updateTreeNodes(this, TreeDB.INDEX_ATTRIBUTE);
  }

  /**
   * Set index of element. Index is specific field that is used to define custom order of elements
   * in sorted list
   *
   * @param index index of element
   */
  public void setIndex(Index index) {
    //TODO split into two methods
    UUID par = getWorkParentUUID();
    if (par == null || par.equals(db.getRootNode().getUUId()))
      return; // ignore index udpate for roots
    //Attribute indexAttr = attrs().createAttribute(TreeDB.INDEX_ATTRIBUTE, index.toString());
    attrs().setAttribute(new AttributeDefinition(uuid, TreeDB.INDEX_ATTRIBUTE, AttributeAvailability.LOCAL,
        new ConstValueProvider(AttributeType.STRING, index.toString()),Flag.HIDDEN));
    // TODO update needed?
    // updateTreeNodes(this, TreeDB.INDEX_ATTRIBUTE);
  }

  /**
   * Get disposed flag for element. It is true if element is deleted.
   *
   * @return true if element is disposed, false otherwise
   */
  public boolean isDisposed() {
    return disposed;
  }

  public void setNOTDisposed() {
    disposed = false;
  }

  /** Set disposed flag for element to true. It is usually do after element deletion. */
  public void setDisposed() {
    disposed = true;
  }

  /**
   * If element have 'name' attribute then return it, else return ParentUserFriendlyName/Id Tends to
   * be unique in project by agreement
   *
   * @param inner if true then attribute will be created and updated further
   * @return path to element
   */
  public String getUserVisibleName(boolean inner) {
    if(inner)
    {
      return attrs().getAttribute(ATTR_USERVISIBLEID).getGValue();
    }
    else{
      Attribute a = attrs().getLocalAttributes().get(ATTR_USERVISIBLEID);
      if(a!=null)
        return a.getGValue();
      else
        return getVirtualAttributeDefinition(ATTR_USERVISIBLEID).getValue(attrs()).toString();
    }
//
//      StringBuilder sb = new StringBuilder();
//    getUserVisibleName(inner, true, sb);
//    return sb.toString();
  }

  public String getFullName(boolean inner) {
    return Requirement.getFullName(this, getTreeDB(), inner);
  }

  public String getFullName() {
    return getFullName(false);
  }

  public String getUserVisibleName() {
    return getUserVisibleName(false);
  }

//  public void getUserVisibleName(boolean inner, boolean fc, StringBuilder sb) {
//  
//  }

  /**
   * Method to update Requirement id, name and corresponding Location nodes
   *
   * @param req requirement to update
   * @param what name of attribute to be updated(or "_%id" for when identifier is updating)
   */
//  public void updateTreeNodes(TreeNode req, String what) {
//    updateTreeNodes(req, what, false);
//  }

  /**
   * Update Requirement id, name and corresponding locations
   *
   * @param req requirement to update
   * @param what name of attribute to be updated(or "_%id" for when identifier is updating)
   * @param internal true if call source is internal
   */
//  protected void updateTreeNodes(TreeNode req, String what, boolean internal) {
//    TreeNode[] reqs = getActiveWorkChildren();
//
//    // TODO update needed?
//    for (TreeNode node : reqs) {
//      if (!node.getUUId().equals(req.getUUId())) node.updateTreeNodes(req, what, internal);
//    }
//    // if(!internal)
//    // db.fireNodeChange(req, true);
//    // req.fireNodeAttributeChange(what);
//    // db.fireNodeGUIChange(req, true);
//    // req.fireGUINodeAttributeChange(what);
//  }

  private List<INodeChangeListener> listenerList;
  private List<INodeChangeListener> guilistenerList;
  private List<IAttributeChangePropagator> propogators =
      new LinkedList<>();

  /**
   * Add Change Listener to be informed on changes of current node
   *
   * @param l change listener
   */
  public void addChangeListener(INodeChangeListener l) {
    if (listenerList == null) listenerList = new LinkedList<>();
    synchronized (listenerList) {
      if (!listenerList.contains(l)) listenerList.add(l);
    }
  }

  /**
   * Remove change listener from current node
   *
   * @param l change listener to be removed
   */
  public void removeChangeListener(INodeChangeListener l) {
    if (listenerList == null) return;
    synchronized (listenerList) {
      if (listenerList.contains(l)) listenerList.remove(l);
    }
  }
  

  public List<IAttributeChangePropagator> getAPropogators() {
    synchronized (propogators) {
      return propogators; 
    }
  }
  
  public void addAPropogator(IAttributeChangePropagator a){
    synchronized (propogators) {
      propogators.add(a); 
    }
  }
  public void removeAPropogator(IAttributeChangePropagator a){
    synchronized (propogators) {
      propogators.remove(a);
    }
  }

  /**
   * Add change listener to be notified after set of actual changes
   *
   * @param l change listener to add
   */
  public void addGuiChangeListener(INodeChangeListener l) {
	if (guilistenerList == null) guilistenerList = new LinkedList<>();
    synchronized (guilistenerList) {
      if (!guilistenerList.contains(l)) guilistenerList.add(l);
    }
  }

  /**
   * Remove change listener to be notified after set of actual changes
   *
   * @param l change listener to remove
   */
  public void removeGuiChangeListener(INodeChangeListener l) {
    if (guilistenerList == null) return;
    synchronized (guilistenerList) {
      if (guilistenerList.contains(l)) guilistenerList.remove(l);
    }
  }

  /** Fire event when node is deleted */
  protected void fireNodeDelete() {
    if (listenerList == null) return;
    INodeChangeListener[] listeners = null;
    synchronized (listenerList) {
      listeners = listenerList.toArray(new INodeChangeListener[0]);
    }
    int[] del = new int[listeners.length];
    int j = 0;
    for (int i = 0; i < listeners.length; i++) {
      INodeChangeListener k = listeners[i];
      try {
        k.deleted(this);
      } catch (Exception e) {
        e.printStackTrace();
        del[j++] = i;
      }
    }
    synchronized (listenerList) {
      for (int i = 0; i < j; i++) listenerList.remove(listeners[del[i]]);
    }
  }

  /**
   * Fire node change event when some attribute is changed
   *
   * @param changed attribute name
   */
  public void fireNodeAttributeChange(ChangesSet<Attribute> what) {
    if (listenerList == null) return;
    INodeChangeListener[] listeners = null;
    synchronized (listenerList) {
      listeners = listenerList.toArray(new INodeChangeListener[0]);
    }
    int[] del = new int[listeners.length];
    int j = 0;
    // for(String wh:what)
    for (int i = 0; i < listeners.length; i++) {
      INodeChangeListener k = listeners[i];
      try {
        k.attributesChange(this, what);
      } catch (Exception e) {
        // in the case of exception in
        // listener, delete it
        e.printStackTrace();
        del[j++] = i;
      }
    }
    synchronized (listenerList) {
      for (int i = 0; i < j; i++) listenerList.remove(listeners[del[i]]);
    }
  }

  /** Returns change listener that represents reaction on other nodes(used in linkto mechanics) */
  protected INodeChangeListener getChangeListener() {
    return null;
  }

  public boolean isNameDefined() {
    return attrs().getAttributeValue(ATTR_NAME) != null;
  }

  public String getNameOrId() {
    return attrs().getAttribute(ATTR_NAMEORID).getGValue();
  }


  public boolean setName(String name) {
    boolean changed = false;
    if (name == null || name.equals("")) {
      attrs().removeAttribute(ATTR_NAME);

    } else {
      name = name.replace((char) 160, ' ');
      AttributeDefinition newv = new AttributeDefinition(getUUId(), ATTR_NAME, AttributeAvailability.LOCAL,
          new TemplValueProvider(AttributeType.STRING,name),Flag.HIDDEN);
      //Attribute newv = attrs().createAttribute(ATTR_NAME, name);
      Attribute rem = attrs().setAttribute(newv);
      changed =
          rem == null
              || rem.getValue() == null
              || rem.getValue() != null && !rem.getValue().equals(newv.getRawValue().getValue());
    }
    return changed;
  }

  public String getName() {
    return attrs().getStringAttributeValue(ATTR_NAME);
  }

  /** Fire event when node is deleted */
  public void fireGUINodeDelete() {
	INodeChangeListener[] listeners = new INodeChangeListener[0];
	if (guilistenerList == null) return;
	  synchronized (guilistenerList) {	
	    listeners = guilistenerList.toArray(new INodeChangeListener[0]);
	  }
    int[] del = new int[listeners.length];
    int j = 0;
    for (int i = 0; i < listeners.length; i++) {
      INodeChangeListener k = listeners[i];
      try {
        k.deleted(this);
      } catch (Exception e) {
        e.printStackTrace();
        del[j++] = i;
      }
      // RLogger.logInfo(getType()+" "+getId()+" "+getName()+" p
      // "+(p==null?"null":p.getId()));
    }
    synchronized (guilistenerList) {
      for (int i = 0; i < j; i++) guilistenerList.remove(listeners[del[i]]);
    }
  }

  /**
   * File event when attribute was changed
   *
   * @param what name of changed attribute
   */
  public void fireGUINodeAttributeChange(ChangesSet<Attribute> what) {
	INodeChangeListener[] listeners = new INodeChangeListener[0];
	if (guilistenerList == null) return;
      synchronized (guilistenerList) {
	    listeners = guilistenerList.toArray(new INodeChangeListener[0]);
	  }
    int[] del = new int[listeners.length];
    int j = 0;
    // for(String wh : what)
    for (int i = 0; i < listeners.length; i++) {
      INodeChangeListener k = listeners[i];
      try {
        k.attributesChange(this, what);
      } catch (Exception e) {
        // in the case of exception in
        // listener, delete it
        e.printStackTrace();
        del[j++] = i;
      }
    }
    synchronized (guilistenerList) {
      for (int i = 0; i < j; i++) guilistenerList.remove(listeners[del[i]]);
    }
  }

  /** Get iterator for children */
  public Iterator<TreeNode> workIterator() {
    if (shownChildren == null) getShownChildrenUUIDs();
    return new TreeNodeChildIterator(false);
  }

  public Iterator<TreeNode> shownIterator() {
    if (shownChildren == null) getShownChildrenUUIDs();
    return new TreeNodeChildIterator(true);
  }

  public void fireNodeAttributeChange(
     ChangesSet<Attribute> attributeName, AttributeAvailability availability) {
    fireNodeAttributeChange(attributeName);
    // if(availability != AttributeAvailability.LOCAL)
    // {
    // for(TreeNode child:getUnsortedChildren())
    // child.fireNodeAttributeChange(attributeName,
    // availability == AttributeAvailability.DIRECT_CHILDREN ?
    // AttributeAvailability.LOCAL : availability);
    //
    // }
  }

  public void fireGUINodeAttributeChange(
      ChangesSet<Attribute> attributeName, AttributeAvailability availability) {
    fireGUINodeAttributeChange(attributeName);
    if (availability != AttributeAvailability.LOCAL) {
      for (TreeNode child : getWorkChildren())
        child.fireGUINodeAttributeChange(
            attributeName,
            availability == AttributeAvailability.DIRECT_CHILDREN
                ? AttributeAvailability.LOCAL
                : availability);
    }
  }

  public List<UUID> getActiveWorkChildrenUUIDs() {
    ReadLock r = null;
    try{
      r = readWCLock();
      return Collections.unmodifiableList(workChildren);
    }finally{
      readWCUnlock(r);
    }
  }

  public List<UUID> getWorkChildrenUUIDs() {
    return getWorkChildrenUUIDs(false);
  }

  public List<UUID> getWorkChildrenUUIDs(boolean ignoreCPUpdate) {
    try {
      if (!ignoreCPUpdate && getCloneParentUUId() != null) {
        if (getCloneParent() == null){
          return getActiveWorkChildrenUUIDs();
        }
        getCloneParent().getWorkChildrenUUIDs(false); // up cp children
        if (getVS().isCloneParentVersionOutdated(Type.STRUCT)) {
          // getVS().setCloneParentVersion(Type.STRUCT,
          // getCloneParent().getVS().getVersion(Type.STRUCT),true);
          UUID cloneParent = getCloneParentUUId();
          if (cloneParent != null) {
            boolean changes = false;
            Collection<UUID> result = new HashSet<>();
            Collection<UUID> stchi = new HashSet<>();
            Collection<UUID> stored = getActiveWorkChildrenUUIDs(); // getRepChildren(uuid,raw);
            if (stored != null) {
              result.addAll(stored);
              stchi.addAll(stored);
            }
            boolean w = false;
            boolean rl = false;
            if (cloneParent.equals(uuid)) {
              // workChildren = new ArrayList<UUID>(res);
            } else
              try {
                rl = getCloneParent().writeWCLock();
                w = writeWCLock();
                List<UUID> childs = getCloneParent().getWorkChildrenUUIDs();
                if (childs != null) {
                  Set<UUID> curCloneParent = new HashSet<>();
                  curCloneParent.addAll(childs);
                  Set<UUID> toAdd = new HashSet<>(curCloneParent);
                  Iterator<UUID> resit = result.iterator();
                  while (resit.hasNext()) {
                    UUID uid = resit.next();
                    UUID cp = getVirtualDB().getVStorage().getCloneParent(uid);
                    if (!curCloneParent.contains(cp)
                        || !Instantiator.getRootTree(db).isNodeExist(cp)) {
                      TreeNode node = Instantiator.getRootTree(db).getNode(uid);
                      resit.remove();
                      if (node != null) db.addRemovePool(node);
                    } else toAdd.remove(cp);
                    // else it is new added one
                    // TODO changes of type or attrs?
                  }
                  getVS().updateCloneParentVersion(Type.STRUCT);
                  for (UUID uid : toAdd) {
                    if (uid.equals(getUUId())) {
                      RLogger.logInfo("WARN! rec prevented " + getQualifiedId(true));
                      continue;
                    }
                    try {
                      TreeNode base = getTreeDB().getNode(uid);
                      if (base == null) continue;
                      UUID uid2 = UUID.randomUUID();
                      UUID cid =
                          getVirtualDB()
                              .getVStorage()
                              .cloneUUID(getTreeDB(), uid2, base, uuid, base.getId(), stchi, null);
                      if (cid == null) RLogger.logInfo("cloning error! " + uid2);
                      else {
                        result.add(cid);
                        stchi.add(cid);
                      }
                    } catch (Exception e) {
                      e.printStackTrace();
                    }
                  }
                  workChildren = new CopyOnWriteArrayList<>(result);
                  if (!toAdd.isEmpty() || childs.size() != result.size()) 
                    changes = true;
                  // if(changes){

                  // }
                  //getVS().updateCloneParentVersion(Type.STRUCT);
                }
              } catch (Exception e) {
                // TODO Auto-generated catch block
                e.printStackTrace();
              }
              finally{
                if(w)
                  writeWCUnlock();
                if(rl)
                  getCloneParent().writeWCUnlock();
              }
              if (changes) {
                getVS().updateVersion(Type.STRUCT);
              }
              ReadLock rl2 = null;
              try{
                rl2=readWCLock();
                return workChildren;
              }
              finally{
                readWCUnlock(rl2);
              }
          }
        }
      }
    } catch (Exception e) {
      e.printStackTrace();
    }
    return Collections.unmodifiableList(workChildren);
  }

  public int sizeWorkChildren() {
    return sizeWorkChildren(false);
  }

  public int sizeWorkChildren(boolean onlyloaded) {
    return getWorkChildrenUUIDs(onlyloaded).size();
  }

  public TreeNode getWorkChild(int i) {
    return getWorkChild(i, false); // TODO maybenot to check
  }

  public TreeNode getWorkChild(int i, boolean notcheck) {
    return getWorkChild(i, notcheck, false);
  }
  private TreeNode getNode(UUID child, boolean notcheck, boolean onlyloaded){
    try{
      return getNodeExc(child, notcheck, onlyloaded);
    } catch (NodeNotFoundException e) {
      e.printStackTrace();
    }
    return null;
  }
  private TreeNode getNodeExc(UUID child, boolean notcheck, boolean onlyloaded) throws NodeNotFoundException{
    TreeNode node = null;
      if (onlyloaded) node = db.getCachedNodeOrVirt(child);
      else node = db.getNode(child, TreeDB.DEFAULT_DEPTH, notcheck);
    if(node == null)
      return null;
    if (node.isVirtual() && !db.isVirtual()) {
      UUID uid = node.getVRoot();
      if (Instantiator.getExistVirtualTreeDB(db, uid) == null) Instantiator.getExistVirtualTreeDB(db, uid);
      if (onlyloaded)
        return Instantiator.getExistVirtualTreeDB(db, uid).getCachedNodeOrVirt(node.getUUId());
      else return Instantiator.getExistVirtualTreeDB(db, uid).getNode(node.getUUId());
    }
    return node;
  }
  public TreeNode getWorkChild(int i, boolean notcheck, boolean onlyloaded) {
    boolean wl = false;
    try{
      if(onlyloaded)
      {
        ReadLock r = null;
        try{
          r = readWCLock();
          UUID child = workChildren.get(i);
          return getNode(child, notcheck, onlyloaded);
        }finally{
          readWCUnlock(r);
        }
      }
      wl = writeWCLock();
      getWorkChildrenUUIDs();
      UUID child = workChildren.get(i);
      return getNode(child, notcheck, onlyloaded);
    }
    finally{
      if(wl)
        writeWCUnlock();
    }
  }

  public void dropTargetFor(TreeNode who) {
    who.setParent(this);
  }

  public boolean isRestricted() {
    return false;
  }

  public boolean isRoot() {
    if (db.isVirtual()) { // in virt
      TreeInstancer ti = db.getVStorage();
      if (ti != null) return db.getVStorage().getRootNode().equals(getUUId());
      // TODO other trees
    } else return db.getRootNodeUUID().equals(getUUId());
    return false;
  }

  public TreeNode findSingleWorkChildPlusName(String rt) {
    List<TreeNode> ret = findWorkChildPlusName(rt);
    if (ret == null || ret.isEmpty()) return null;
    return ret.iterator().next();
  }

  public List<TreeNode> findWorkChildPlusName(String rt) {
    rt = rt.replace((char) 160, ' ');
    List<TreeNode> ret = new ArrayList<>();
    if (rt.startsWith("|:") && rt.endsWith(":|")) { // child num
      int num = Integer.parseInt(rt.substring(2, rt.length() - 2));
      if (num < sizeWorkChildren()) {
        ret.add(getSortedWorkChildren()[num]);
      }
      return ret;
    }
    List<UUID> uids = getWorkChildrenUUIDs();
    for (UUID uid : uids) {
      TreeNode child = db.getNode(uid);
      if(child!=null && child.getId() == null)
        RLogger.logInfo("Child "+uid+" in "+getUUId()+" loading problem");
      else
      if (child != null
          && (child.getId().toLowerCase().equals(rt.toLowerCase())
              || child.getName() != null && child.getName().equals(rt)
              || child.getUUId().toString().equals(rt)
              || child.getRawId().equals(rt))) {
        ret.add(child);
      }
    }
    return ret;
  }

  public List<TreeNode> findShownChildsPlusName(String rt) {
    rt = rt.replace((char) 160, ' ');
    List<TreeNode> ret = new ArrayList<>();
    if (rt.startsWith("|:") && rt.endsWith(":|")) { // child num
      int num = Integer.parseInt(rt.substring(2, rt.length() - 2));
      if (num < sizeShownChildren()) {
        ret.add(getSortedShownChildren()[num]);
      }
      return ret;
    }
    List<UUID> uids = getShownChildrenUUIDs();
    for (UUID uid : uids) {
      TreeNode child = db.getNode(uid);
      if (child.getId().equals(rt)
          || child.getName() != null && child.getName().equals(rt)
          || child.getUUId().toString().equals(rt)) {
        ret.add(child);
      }
    }
    return ret;
  }

  private static Set<String> vanames = Collections.unmodifiableSet(
      new HashSet<String>() {
        /** */
        private static final long serialVersionUID = -1483522233796325954L;

        {
          {
            //add(ATTR_ID);
            add(ATTR_TERMS);
            add(ATTR_TERMS_REV);
            add(ATTR_LABEL);
            add(ATTR_NAMEORID);
            add(ATTR_USERVISIBLEID);
            add(ReportSettings.ATTR_GEN);
            //add(ATTR_NODE_PATH);
          }
        }
      });
  private static Set<String> altvanames = Collections.unmodifiableSet(
      new HashSet<String>() {
        /** */
        private static final long serialVersionUID = -1483522233796325954L;

        {
          {
            //add(ATTR_ID);
            add(ATTR_TERMS);
            add(ATTR_TERMS_REV);
            add(ATTR_LABEL);
            add(ATTR_NAMEORID);
            add(ATTR_USERVISIBLEID);
            add(ReportSettings.ATTR_GEN);
           // add("_cloneparentversion");
            //add(ATTR_NODE_PATH);
          }
        }
      });

  public Set<String> getVirtAttrsNames() {
    if (cloneParent == null) {
      return vanames;
    }
    return altvanames;
  }

  public static boolean isVirtName(TreeNode node, String key) {
//    if(key.startsWith(ATTR_NODE_PATH+"_")){
//      String ak = key.substring(ATTR_NODE_PATH.length()+1);
//      if(ak!=null && ak.length()>0)
//        return true;
//    }
    return altvanames.contains(key);
  }
  public static Set<String> getAllVirtAttrsNames(TreeNode node) {
    if(node!=null)
      return node.getVirtAttrsNames();
    return vanames;
  }  //private static AttributeDefinition idD, pathD;
  private Map<String,AttributeDefinition> vdefs = new HashMap<>();
  public AttributeDefinition getVirtualAttributeDefinition(String s) {
    if(s == null)
      return null;
    if(vdefs.containsKey(s))
      return vdefs.get(s);
    AttributeDefinition ad = null;
    switch (s) {
//      case ATTR_NODE_PATH:
//        {
//          try {
//            return new AttributeDefinition(getUUId(), ATTR_NODE_PATH, AttributeType.STRING,
//                AttributeAvailability.LOCAL,
//                NodeResPathVProvider.getOne(null),
//                Flag.HIDDEN|Flag.VIRTUAL);
//          } catch (Exception e) {
//          }
//        }
      case ATTR_NAMEORID:{
        ad = new AttributeDefinition(getUUId(), ATTR_NAMEORID, AttributeAvailability.LOCAL,
            NameOrIdProvider.getOne(),
            Flag.HIDDEN|Flag.VIRTUAL);
      }break;
      case ATTR_LABEL:
      {
        ad = new AttributeDefinition(getUUId(), ATTR_LABEL, AttributeAvailability.LOCAL,
            LabelProvider.getOne(),
            Flag.HIDDEN|Flag.VIRTUAL);
      }break;
      case ATTR_USERVISIBLEID:
      {
        ad = new AttributeDefinition(getUUId(), ATTR_USERVISIBLEID, AttributeAvailability.LOCAL,
            UserVisibleIdProvider.getOne(),
            Flag.HIDDEN|Flag.VIRTUAL);
      }break;
      case ATTR_TERMS:
      {
        ad = new AttributeDefinition(getUUId(), ATTR_TERMS, AttributeAvailability.LOCAL,
            TermsValueProvider.getOne(true),
            Flag.HIDDEN|Flag.VIRTUAL);
      }break;
      case ATTR_TERMS_REV:
      {
        ad = new AttributeDefinition(getUUId(), ATTR_TERMS, AttributeAvailability.LOCAL,
            TermsValueProvider.getOne(false),
            Flag.HIDDEN|Flag.VIRTUAL);
      }break;
      //case "_cloneparentversion"
    }
    if(ad!=null)
      vdefs.put(s, ad);
    return vdefs.get(s);
  }

  public AttributeStorage attrs() {
    if (attributes == null) attributes = AttributeStorage.loadOrCreate(this, db);
    return attributes;
  }

  public boolean canEditAttributesInTable() {
    return false;
  }

  /**
   * Save all attribute changes in storage. If it is not called after changes then all attribute
   * changes can be lost.
   */
  public void commitAttributes() {
    AttributeStorage as = attrs();
    if (as.hasChanges()) {
      as.commitAttributes();
    }
  }

  protected TreeDB getVirtualDB() {
    if (!isVirtual()) {
      UUID rmark = getVRoot();
      if (rmark != null) {
        return Instantiator.getExistVirtualTreeDB(db, UUID.fromString(rmark.toString()));
      } else return db; // Instantiator.getVirtualTreeDB(db,Instantiator.DEFAULT_INSTANCE_NAME);
    }
    return db;
  }

  protected boolean setChildren(List<UUID> storedChildren, boolean init) {
    boolean w = false;
    try{
      w = writeWCLock();
      boolean eqs = workChildren.equals(storedChildren);
      if (!eqs & !init) getVS().updateVersion(Type.STRUCT);
      if (storedChildren.contains(getUUId())) {
        RLogger.logInfo("WARN! rec prevented " + getQualifiedId(true));
        storedChildren.remove(getUUId());
      }
      this.workChildren = new CopyOnWriteArrayList<>(storedChildren);
      return eqs;
    }
    finally{
      if(w)
        writeWCUnlock();
    }
  }
  //
  // public void updateCPChildren() {
  // cPVersion = getCloneParent().getVersion().copy();
  // getTreeDB().updateNodeChildren(this);
  // }

  public VersionsStorage getVS() {
    return attrs().getVersionsStorage();
  }

  public UUID getVRoot() {
    if (!loaded && attrs() != null) {
      Attribute parentRoot = attrs().getAttribute(TreeNode.ATTR_VIRT_STATUS);
      if (parentRoot != null && parentRoot.getValue() != null) {
        String vroot = parentRoot.getValue().toString();
        return UUID.fromString(vroot);
      }
    }
    return vroot;
  }

  public void clearShownChildren() {
    shownChildren = null;
    getTreeDB().
      fireNodeChildrenOrderChange(this,false);
  }

  public String getRawName() {
    Attribute a = attrs().getAttribute(ATTR_NAME);
    if(a==null || a.getDefinition().getRawValue() == null)
      return null;
    Object o = a.getDefinition().getRawValue().getValue();
    if(o == null)
      return null;
    return o.toString();
  }

  // public void removeClone(UUID uuid2) {
  // if(uuid2==null)
  // return;
  //// Attribute clns = attrs().getAttribute(TreeInstancer.ATTR_CLONES);
  //// if(clns!=null){
  //// List<Object> vals = clns.getGValue(attrs());
  //// if(vals!=null && vals.contains(uuid2.toString())){
  //// vals.remove(uuid2.toString());
  //// attrs().setAttribute(TreeInstancer.ATTR_CLONES,
  //// new ListValue(AttributeType.STRING, vals));
  //// }
  //// }
  // }

  public TreeNode[] getActiveWorkChildren() {
    return getWorkChildren(null, false, true, true);
  }

  public boolean isHidden() {
    String s = "1";
    try {
      TreeInstancer ti = null;
      if (db.isVirtual()) ti = db.getVStorage();
      else {
        UUID vuid = getVRoot();
        if (vuid != null) ti = Instantiator.getExistVirtualTreeDB(db, vuid).getVStorage();
      }
      String pr = getPredicate();
      if (pr == null || pr.isEmpty()) s = "1";
      else if (ti == null) s = attrs().evaluatePredicateString(pr,AttributeStorage.formPredDeps(pr,attrs()));
      else s = ti.checkPredicate(getUUId(), pr);
    } catch (Exception e) {
      e.printStackTrace();
    }
    return (s.startsWith("0") || s.equals(false) || s.equals("false"))
        && !getPredicate().equals("false");
  }

  public List<Pair<String, TreeNode>> getRelations(IObjectFilter<String> filter) {
    List<Pair<String, TreeNode>> list = new LinkedList<>();
    for (String s : attrs().getLocalAttributeKeys()) // TODO all
    {
      if (filter == null || filter.filter(s)) {
        Attribute a = attrs().getAttribute(s);
        if (a != null && a.getType().equals(AttributeType.REFERENCE)) {
          List<TreeNode> linkeds = attrs().getLinkAttributeValue(a.getKey());
          if (linkeds != null)
            for (TreeNode linked : linkeds)
              list.add(new Pair<>(a.getKey(), linked));
        }
        if (a != null && (a.getListType() != null && a.getListType() == AttributeType.REFERENCE)) {
          try {
            List<TreeNode> nodes = attrs().getLinkAttributeValues(a.getKey());
            if (nodes != null)
              for (TreeNode node : nodes) list.add(new Pair<>(a.getKey(), node));
          } catch (Exception e) {
          }
        }
      }
    }
    return list;
  }

  public Set<TreeNode> getRelatedNodes(IObjectFilter<String> filter) {
    Set<TreeNode> nds = new HashSet<>();
    for (Pair<String, TreeNode> p : getRelations(filter)) {
      nds.add(p.getSecond());
    }
    return nds;
  }

  public void fireNodeResourceChange(String resName, Moment m, boolean inner) {
    if (listenerList != null)
      synchronized (listenerList) {
        for (INodeChangeListener l : listenerList) l.fireNodeResourceChange(this, resName, m, inner);
      }
    }

  public void fireGUINodeResourceChange(String resName, Moment m, boolean inner) {
	if (guilistenerList != null)
	  synchronized (guilistenerList) {
        for (INodeChangeListener l : guilistenerList)
          l.fireNodeResourceChange(this, resName, m, inner);
	  }
  }

//  public Version getNameVersion() {
//    Attribute a = attrs().getAttribute(ATTR_NAME);
//    if (a == null) return Version.getUnset();
//    return a.getVersion();
//  }

  public void revCacheChanged(TreeNode node, String s) {
    // TODO Auto-generated method stub

  }

  public boolean hasResource(String string) {
    InputStream is = getResourceContent(string);
    if (is != null) {
      try {
        is.close();
      } catch (IOException e) {
      }
      return true;
    }
    return false;
  }

  public static AttributeDefinition formIdAttribute(UUID uuid2, String nodeId) {
    //TODO change to proper
    return new AttributeDefinition(uuid2, ATTR_ID,AttributeAvailability.LOCAL,new ConstValueProvider(AttributeType.STRING, nodeId),
        Flag.HIDDEN);
  }

  public TreeNode getExposedNode() {
    return this;
  }

  public URI getNodeResourcesPath(String s) {
    URI rp = getNodeResourcesPath();
    if(rp!=null)
      return URI.create(rp.toString()+"/"+s);
    return null;
  }

  public boolean isSameTree(TreeNode node) {
    if(node == null)
      return false;
    return Instantiator.getRootTree(getTreeDB())==Instantiator.getRootTree(node.getTreeDB());
  }

  public boolean hasHeaderAttribute() {
    Attribute a = attrs().getAttribute("Вид");
    if(a!=null && a.getValue()!=null && a.getValue().toString().equals("Заголовок"))
      return true;
    return false;
  }
  public void setIgnoredName(String name) {
    setName(name);
    attrs().setAttribute(new AttributeDefinition(uuid, "ignoreName", 
        new ConstValueProvider(AttributeType.BOOL,true)));
  }
  public Map<String, Set<TreeNode>> getLinks(){
    Map<String, Set<TreeNode>> map1 =
        getTreeDB().getLinkResolver().getLinkedNodes(this);
    if(map1==null || map1.isEmpty())
      map1 = new HashMap<>();
    map1.putAll(getTreeDB().getTermsStorage().getDirectTerms(this));
    return map1;
  }
  public Set<TreeNode> getLinks(String s){
    return getLinks().get(s);
  }
  public Map<String, Set<TreeNode>> getRevertLinks(boolean b){
    Map<String, Set<TreeNode>> map1 =
        getTreeDB().getLinkResolver().getRevertLinks(this);
    if(map1==null || map1.isEmpty())
      map1 = new HashMap<>();
    if(b)
      map1.putAll(getTreeDB().getTermsStorage().getRevertTerms(this));
    return map1;
  }
  public Set<TreeNode> getRevertLinks(String s){
    return getRevertLinks(true).get(s);
  }
  public String[] getLinksOrErrorsKeys() {
    return getLinksOrErrors(true).keySet().toArray(new String[0]);
  }
  public ReferenceValue[] getLinksOrErrorsValues(String k) {
    return getLinksOrErrors(true).get(k);
  }
  public Map<String, ReferenceValue[]> getLinksOrErrors(boolean withterms) {
    Map<String, Set<ReferenceValue>>ret = null;
    if(withterms){
      Map<String,Set<TreeNode>>links = getTreeDB().getTermsStorage().getDirectTerms(this);
      for(String s:links.keySet()){
        if(ret == null)
          ret = new HashMap<>();
        ret.put(s,new HashSet<>());
        ret.get(s).add(new ReferenceValue(s, new HashSet<>(links.get(s)),attrs()));
      }
    }
    for(Entry<String, Attribute> en:attrs().getActiveAttributes().entrySet()){
      if(en.getValue().isReference()){
        ReferenceValue[] ta = getTreeDB().getLinkResolver().getRefValues(this, en.getKey());
        if(ta!=null && ta.length>0){
          if(ret == null)
            ret = new HashMap<>();
          if(!ret.containsKey(en.getKey()))
            ret.put(en.getKey(), new HashSet<>());
          for(ReferenceValue rv:ta)
            if(rv.hasErrorsOnResolve(attrs()))
              ret.get(en.getKey()).add(rv);
            else
              ret.get(en.getKey()).add(rv);
        }
      }
    }
    if(ret == null || ret.isEmpty())
      return Collections.emptyMap();
    Map<String,ReferenceValue[]> no = new HashMap<>();
    for(Entry<String, Set<ReferenceValue>> en:ret.entrySet())
      no.put(en.getKey(), en.getValue().toArray(new ReferenceValue[0]));
    return no;
  }
  public String getNodeVersion() {
    if(AttributesVersioningRule.isLegal(getType()))
      return attrs().getStringAttributeValue(AttributesVersioningRule.ATTR_NODE_VERSION,"1*");
    return AttributesVersioningRule.VALUE_ILLEGAL;
  }
}
