/*
 * Copyright 2015 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.unitesk.requality.core.attribute;

import java.net.URI;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;
import java.util.UUID;
import java.util.regex.Pattern;
import org.json.JSONArray;
import com.unitesk.requality.core.IAttributeFilter;
import com.unitesk.requality.core.SortingSettings.SortingMethod;
import com.unitesk.requality.core.TreeDB;
import com.unitesk.requality.core.TreeNode;
import com.unitesk.requality.core.instancer.Instantiator;
import com.unitesk.requality.nodetypes.Requirement;
import com.unitesk.requality.nodetypes.VirtualNode;
import com.unitesk.requality.tools.DeepFirstTreeWalker;

public class Attribute implements Comparable<Attribute> {
  public static final String ENUM_CHAR = "\u24BA";
  private static Set<String> statusNames = new HashSet<>(), hiddenReplace = new HashSet<>();

  private final TypedValue value;
  //private final Version version;
  private final AttributeDefinition adef;


  //Names of attributes to be replaced in raw value

  static {
    registerHTMLNames(TreeNode.ATTR_DESCRIPTION);
    registerStatusName(TreeNode.ATTR_TYPE);
  }
  public static Set<String> getStatusNames(){
    return statusNames;
  }
  public static Set<String> getHTMLNames(){
    return hiddenReplace;
  }
  public static void registerStatusName(String name) {
    statusNames.add(name);
  }

  public static void registerHTMLNames(String name) {
    hiddenReplace.add(name);
  }

  public boolean isVirtual(){
    return adef.isVirtual();
  }

  public static boolean affectsPosition(TreeNode node, String attr) {
    if (VirtualNode.ATTR_COLLAPSEVNODE.equals(attr)) return true;
    if (VirtualNode.ATTR_HIDDEN_STATUS.equals(attr)) return true;
    SortingMethod sm = node.getTreeDB().getSortingSettings().getMethod();
    if (sm == SortingMethod.BY_INDEX
        || sm == SortingMethod.BY_LOCATIONS
            && node.getTreeDB().getTreeLogic().usesIndexByDefault(node.getType()))
      if(attr == TreeDB.INDEX_ATTRIBUTE)
        return true;
    if (sm == SortingMethod.BY_NAMES) return attr.equals(TreeNode.ATTR_LABEL);
    return attr.equals(Requirement.ATTR_LOCATIONS) || attr.equals(TreeNode.ATTR_LABEL);
  }

  public AttributeType getType(){ 
    TypedValue tv = getTValue();
    if(tv == null)
      return AttributeType.STRING;//Default //TODO change?
    return tv.getType();
    //return adef.getType();
  }

  /***
   * Method used in calculations of node's general attributes version.
   * 
   */
  @Override
  public int hashCode() {
    int v = adef.hashCode();
    if(value!=null)
      v ^=  value.hashCode();
//    int v = origin.hashCode() ^ key.hashCode();
//    if (type != null) v = v ^ type.toString().hashCode();
//    if (version != null) v = v ^ Long.hashCode(version.getSInt());
//    else v = v ^ Long.hashCode(-1L);
//    if (generator != null) v = v ^ generator.hashCode(); // TODO version?
//    Object rawValue = getRawValue();
//    if (rawValue != null) v = v ^ rawValue.hashCode();
//    if (getListType() != null) v = v ^ getListType().toString().hashCode();
//    return v ^ getAvailability().toString().hashCode();
    return v;
  }

//  public Version getVersion() {
//    return version;
//  }

  public AttributeAvailability getAvailability() {
    return adef.getAvailability();
  }

  public boolean isMoreVisible(TreeDB db, Attribute a, UUID where, boolean samemore) {
    if (a == null) return true;
    if (where.equals(getOrigin())) {
      if (!where.equals(a.getOrigin())) return true;
    } else {
      if (where.equals(a.getOrigin())) return false;
    }
    if (adef.isGenerated()) {
      if (a.adef.isGenerated()) {
      } else return false;
    } else if (a.adef.isGenerated()) {
      return true;
    }
    return this.adef.isMoreAVVisibleThan(a.adef)
        || (this.adef.isSameAVVisibleWith(a.adef)&&this.isMasterMoreVisibleThan(
            db, a, this == a && samemore)); // )
  }
  // assume that all the rest are updated
 
  public Attribute(AttributeDefinition ad, AttributeStorage as){
    this.adef = ad;
    this.value = ad.getTValue(as);
  }

  public String getKey() {
    return adef.getKey();
  }
  
  public TypedValue getTValue(){
    //if(value == null)
      //value = adef.getValue(as);
    return value;
  }
  public <T> T getGValue(){
   // if(value == null)
     // value = adef.getValue(as);
    return (T) value.getValue();
  }

  private static List<Object> toList(ListValue o) {
    // TODO Auto-generated method stub
    List<Object> ret = new ArrayList<>();
    for (Object i : o.getValues())
      if (o.getType() == AttributeType.LIST) ret.add(toList((ListValue) i));
      else ret.add(i);
    return ret;
  }

  @Override
  public String toString() {
    StringBuilder sb = new StringBuilder();
    sb.append("[");
    sb.append(getAvailability() == AttributeAvailability.LOCAL
              ? "L"
              : getAvailability() == AttributeAvailability.DIRECT_CHILDREN ? "D" : "S");
    sb.append("]'");
    sb.append(getKey());
    sb.append("'");
    sb.append(":=");
    sb.append(value != null ? value.toString() : "null");
    sb.append("(");
    sb.append(super.toString());
    sb.append(")");
    //sb.append(getGenerator() == null ? "none" : getGenerator().toString());
    return sb.toString();
  }

  public static boolean equalsValues(Object value2, Object object) {
    if (value2 instanceof ListValue && object instanceof ListValue) {
      ListValue o1 = (ListValue) value2, o2 = (ListValue) object;
      if (o1.getValues().size() == o2.getValues().size() && o1.getType() == o2.getType()) {
        return Arrays.equals(o1.getValues().toArray(), o2.getValues().toArray());
      } else return false;
    }
    if (value2 instanceof AppliedEnumValue && object instanceof AppliedEnumValue) {
      AppliedEnumValue aev1 = (AppliedEnumValue) value2, aev2 = (AppliedEnumValue) object;
      return equalsValues(aev2.getValue(), aev1.getValue())
          && equalsValues(aev2.getComment(), aev1.getComment())
          && equalsValues(aev2.getBaseKey(), aev1.getBaseKey());
    }
    return value2 == null && object == null || value2 != null && value2.equals(object);
  }

  static final Pattern valid_inner_name_pattern =
      Pattern.compile("[_]{0,2}[%]{0,1}[A-Za-z0-9а-яА-Я_[$].,-]+[A-Za-z0-9а-яА-Я,._[$] -]*");
  static final Pattern valid_name_pattern =
      Pattern.compile("[A-Za-z0-9а-яА-Я,.-]+[A-Za-z0-9а-яА-Я,._[$] -]*");

  public static boolean checkNameIsValid(String name, boolean inner) {
    if (name == null) return false;
    return inner
        ? valid_inner_name_pattern.matcher(name).matches()
        : valid_name_pattern.matcher(name).matches();
  }

  @Override
  public int compareTo(Attribute o) {
    return this.getKey().compareTo(o.getKey());
  }
//  protected Attribute clone(AttributeStorage as) throws CloneNotSupportedException {
//    return new Attribute(adef, as);
//  }

  public UUID getOrigin() {
    return adef.getOrigin();
  }

  private boolean isMasterMoreVisibleThan(TreeDB db, Attribute a, boolean samemore) {
    if (getOrigin() == null || a.getOrigin() == null) return true;
    if (getOrigin().equals(a.getOrigin())) return samemore; // TODO check
    AttributeStorage as1 = AttributeStorage.getCachedStorage(db, getOrigin()),
        as2 = AttributeStorage.getCachedStorage(db, a.getOrigin());
    if (as1 != null && as2 != null) {
      String s1 = as1.getNode().getQualifiedId(), s2 = as2.getNode().getQualifiedId();
      int q1 = as1.getNode().getQualifiedId().length(),
          q2 = as2.getNode().getQualifiedId().length();
      if (q1 > q2) return true;
      return s1.equals(s2) ? samemore : false;
    }
    return false; // false for eqs by def
  }
  public boolean isList(){
    return getType() == AttributeType.LIST;
  }
  public static String valueToString(AttributeStorage as, AttributeType at, Object value) {
    StringBuilder sb = new StringBuilder();
    if (!checkValueByType(at, value)) return "INCORRECT!";
    EnumValue ev = null;
    if (value instanceof EnumValue) {
      ev = (EnumValue) value;
      value = ev.getValue();
    }
    if (at == AttributeType.LIST) {
      ListValue lv = (ListValue) value;
      AttributeType at2 = lv.getType();
      sb.append(at2);
      sb.append("=[");
      Iterator<Object> it = lv.iterator();
      while (it.hasNext()) {
        sb.append(valueToString(as, at2, it.next()));
        if (it.hasNext()) sb.append(",");
      }
      sb.append("]");
    } else if (at == AttributeType.ENUM_DEFINITION) {
      EnumDefinitionValue lv = (EnumDefinitionValue) value;
      AttributeType at2 = lv.getType();
      sb.append(at2);
      sb.append("=[");
      Iterator<EnumValue> it = lv.iterator();
      while (it.hasNext()) {
        sb.append(valueToString(as, at2, it.next().getValue()));
        if (it.hasNext()) sb.append(",");
      }
      sb.append("]");
    } else if (value != null){
      if(value instanceof IContextStringProvider)
        sb.append(((IContextStringProvider)value).toString(as));
      else
        sb.append(value.toString());
    }
    else sb.append("NULL");
    if (ev != null) {
      if (ev.getComment() != null) {
        sb.append("{");
        sb.append(ev.getComment());
        sb.append("}");
      }
    }
    return sb.toString();
  }
  public static boolean checkValueByType(AttributeType at, Object value2) {
    if (at == AttributeType.ENUM_DEFINITION) return value2 instanceof EnumDefinitionValue;
    if (at == AttributeType.ENUM) return value2 instanceof AppliedEnumValue;
    if (at == AttributeType.LIST) return value2 instanceof ListValue;
    if (at == AttributeType.BOOL) return value2 instanceof Boolean || value2 instanceof String;
    if (at == AttributeType.INT) return value2 instanceof Integer || value2 instanceof String;
    if (at == AttributeType.REFERENCE) return value2 instanceof ReferenceValue || value2 instanceof String;
    if (at == AttributeType.DATE) return value2 instanceof String; // TODO long?
    if (at == AttributeType.LIST) return checkListValue((ListValue) value2);
    // TODO additional checks
    return true;
  }

  public static Object rawValueToJSON(AttributeType at, Object value, String string) {
    StringBuffer sb2 = new StringBuffer();
    if (value instanceof EnumValue) {
      EnumValue ev = (EnumValue) value;
      value = ev.getValue();
    }
    if (at == AttributeType.LIST) {
      ListValue lv = (ListValue) value;
      AttributeType at2 = lv.getType();
      JSONArray joa = new JSONArray();
      // sb2.append("[");
      Iterator<Object> it = lv.iterator();
      while (it.hasNext()) {
        joa.put(rawValueToJSON(at2, it.next(), string));
      }
      return joa;
      // sb2.append("]");
    } else if (at == AttributeType.ENUM_DEFINITION) {
      EnumDefinitionValue lv = (EnumDefinitionValue) value;
      AttributeType at2 = lv.getType();
      JSONArray joa = new JSONArray();
      Iterator<EnumValue> it = lv.iterator();
      while (it.hasNext()) {
        joa.put(rawValueToJSON(at2, it.next().getValue(),string));
      }
      return joa;
    } else {
      if (value != null) {
        if(value instanceof String && string !=null)
          return ((String)value).replace("{node.resURL}", string);
        else
          return value;
      } else sb2.append("NULL");
    }
    return value;
  }

  public static Object resultValueToJSON(AttributeType at, Object value, URI uri) {
    if (at == AttributeType.LIST || at == AttributeType.ENUM_DEFINITION) {
      ArrayList<Object> lv = (ArrayList<Object>) value;
      JSONArray joa = new JSONArray();
      Iterator<Object> it = lv.iterator();
      while (it.hasNext()) {
        Object o = it.next();
        if (o instanceof List) joa.put(resultValueToJSON(at, o, uri));
        else {
          joa.put(o.toString().replace("{node.resURL}", uri.toString()));
        }
      }
      return joa;
    } else if (at == AttributeType.ENUM) {
      return ((EnumValue) value).getValue().toString().replace("{node.resURL}", uri.toString());
    } else {
      if (value != null) return (value.toString().replace("{node.resURL}", uri.toString()));
      else return ("NULL");
    }
  }

  /**
   * * Converts value of attribute type to another type
   *
   * @param from old type
   * @param to new type
   * @param newVal value to convert
   * @return new value if possible or null otherwise
   */
  public static Object convertValue(
      TreeNode node, AttributeType from, AttributeType to, Object newVal, String enumName) {
    if (from == to) return newVal;
    AttributeStorage as = node!=null?node.attrs():null;
    if (to == AttributeType.STRING){
      if(newVal instanceof IContextStringProvider)
        return ((IContextStringProvider)newVal).toString(as);
      return newVal.toString();
    }
    String newS = valueToString(as,from, newVal);
    if (from == AttributeType.LIST || from == AttributeType.ENUM_DEFINITION) newS = newS.replace("[", "").replace("]", "");
    if (to == AttributeType.BOOL) {
      try {
        return Boolean.parseBoolean(newS.trim().toLowerCase());
      } catch (Exception e) {
        e.printStackTrace();
        return false;
      }
    } else if (to == AttributeType.FLOAT) {
      try {
        return Double.parseDouble(newS.trim());
      } catch (Exception e) {
        e.printStackTrace();
        return 0.0d;
      }
    } else if (to == AttributeType.INT) {
      try {
        return Integer.parseInt(newS.trim());
      } catch (Exception e) {
        e.printStackTrace();
        return 0;
      }
    } else if (to == AttributeType.REFERENCE || to == AttributeType.DATE) {
      return newS;
    } else if (to == AttributeType.ENUM && enumName != null) {
      if(node == null)
        throw new UnsupportedOperationException("Enum conversion cannot be performed without node");
      Attribute edef = node.getTreeDB().getRootNode().attrs().getAttribute(enumName);
      TypedValue edf = edef.getTValue();
      if (edf != null&&edf.getType()==AttributeType.ENUM_DEFINITION){
        EnumDefinitionValue edv = edf.getGValue();
        if(edv!=null)
          for (EnumValue o : edv.getValues())
            if (o.getValue().toString().equals(newS)) return o.createApplied(edv.getType(), enumName);
        throw new RuntimeException(
            "Value "+(newS!=null?newS:"null")+" was not found in "+enumName+"!");
      }
    } else if (to == AttributeType.LIST) {
      List<Object> list = new ArrayList<>();
      list.add(newVal);
      return new ListValue(from, list);
    }
    throw new RuntimeException(
        "Unsopported operation - conversion from " + from + " to " + to + "!");
    // return null;
  }

  public static String[] getBuiltinTypes() {
    AttributeType[] ats = AttributeType.visibleValues();
    String[] ret = new String[ats.length - 1];
    for (int i = 0; i < ret.length; i++) ret[i] = ats[i].toString();
    return ret;
  }

  public static String[] getAvailableTypes(TreeDB db, IAttributeFilter filter) {
    TreeNode node = Instantiator.getRootTree(db).getRootNode();
    Set<String> names = new HashSet<>();
    for (Attribute a : node.attrs()) {
      // TODO was filter so needed?
      
      if (a.getDefinition().equals(AttributeType.ENUM_DEFINITION)){
        TypedValue tv = a.getTValue();
        if(filter.filter(a)
          && Attribute.checkValueByType(AttributeType.ENUM_DEFINITION, tv.getValue())
          && !((EnumDefinitionValue) tv.getValue()).getValues().isEmpty())
            names.add(a.getKey());
      }
    }
    String[] types = getBuiltinTypes();
    String[] ret = new String[types.length + names.size()];
    int i = 0;
    for (String at : types) ret[i++] = at;

    for (String s : names) ret[i++] = ENUM_CHAR + s;
    i = ret.length - 1;
    while (ret[i] == null) i--;
    return Arrays.copyOf(ret, i + 1);
  }



  public static String[] getAttributeValues(TreeDB db, String name) {

    DeepFirstTreeWalker walker = new DeepFirstTreeWalker(db.getRootNode(), false, true, false);
    Iterator<TreeNode> it = walker.iterator();
    Set<String> vals = new HashSet<>();
    while (it.hasNext()) {
      TreeNode node = it.next();
      Attribute a = node.attrs().getAttribute(name);
      if (a != null){
        TypedValue tv = a.getTValue();
        if(tv.getValue()!=null &&
          tv.getType() != AttributeType.LIST && 
          tv.getType() != AttributeType.RESOURCE &&
          a.getType() != AttributeType.ENUM_DEFINITION) {
            vals.add(
                tv.getType().name() + ": " + tv.getValue().toString());
        }
       
      }
    }
    return vals.toArray(new String[0]);
  }

  public boolean isLocal(UUID uid) {
    if (uid == null) return true;
    UUID or = getOrigin();
    return or == null || or.equals(uid);
  }

  public boolean isReference() {
    TypedValue tv = getTValue();
    return tv.getType().equals(AttributeType.REFERENCE)
        || tv.getType().equals(AttributeType.LIST) && 
            tv.getValue()!=null && 
            ((ListValue)tv.getValue()).getType().equals(AttributeType.REFERENCE);
  }

  public boolean strongEquals(Attribute attr) {
//    if (!attr.getVersion().equals(getVersion())) return false;
    return equals(attr);
  }


  private static boolean checkListValue(ListValue lv) {
    for (Object o : lv.getValues()) if (!checkValueByType(lv.getType(), o)) return false;
    return true;
  }
//
//  public void setVersion(Integer newv) {
//    version.setInt(newv);
//  }

  public AttributeDefinition getDefinition() {
    return adef;
  }
  public Object getValue() {
    TypedValue tv = getTValue();
    if(tv == null)
      return null;
    return tv.getValue();
  }
  public AttributeType getListType() {
    TypedValue tv = getTValue();
    if(tv.getType()==AttributeType.LIST){
      ListValue lv = tv.getGValue();
      if(lv!=null)
        return lv.getType();
    }
    if(tv.getType()==AttributeType.ENUM_DEFINITION){
      EnumDefinitionValue edv = tv.getGValue();
      if(edv!=null)
        return edv.getType();
    }
    return null;
  }
  public boolean isUserVisible() {
    return (getDefinition().getFlags()&Flag.HIDDEN)!=Flag.HIDDEN;
  }
  public static Object getJavaValue(AttributeType at, Object value) {
    switch(at){
      case LIST:{
        return toList((ListValue) value);
      }
      case ENUM:{
        AppliedEnumValue aev = (AppliedEnumValue)value;
        return getJavaValue(aev.getType(), aev.getValue());
      }
      case ENUM_DEFINITION:{
        throw new RuntimeException("Enum definition was not supported");
      }
      default:
        return value;
      
    }
  }
  public Object getJavaValue() {
    return getJavaValue(getType(), getValue());
  }
  public static TypedValue getTypedValue(Object object) {
    if(object instanceof TypedValue)
      return (TypedValue)object;
    return new TypedValue(AttributeType.detectType(object), object);
  }
}
