/*
 * Copyright 2015 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.unitesk.requality.documents;

import java.io.ByteArrayInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.io.PrintStream;
import java.io.Reader;
import java.io.StringWriter;
import java.io.Writer;
import java.nio.charset.Charset;
import java.util.HashSet;
import java.util.Set;
import org.htmlcleaner.BaseToken;
import org.htmlcleaner.CleanerProperties;
import org.htmlcleaner.HtmlCleaner;
import org.htmlcleaner.SimpleXmlSerializer;
import org.htmlcleaner.TagNode;
import org.htmlcleaner.XmlSerializer;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import com.unitesk.requality.core.CoreUtils;
import com.unitesk.requality.core.Path;
import com.unitesk.requality.core.TreeDB;
import com.unitesk.requality.core.TreeNode;
import com.unitesk.requality.nodetypes.Document;
import com.unitesk.requality.tools.dom.IXHTMLNode;
import com.unitesk.requality.tools.dom.XHTMLResourceCollector;

class XMLNNode implements IXHTMLNode {
  private Node cur;

  public XMLNNode(Node root) {
    cur = root;
  }

  @Override
  public String getName() {
    return cur.getNodeName();
  }

  @Override
  public IXHTMLNode[] getChildren() {
    NodeList nl = cur.getChildNodes();
    IXHTMLNode[] result = new XMLNNode[nl.getLength()];
    for (int i = 0; i < nl.getLength(); i++) {
      result[i] = new XMLNNode(nl.item(i));
    }
    return result;
  }

  @Override
  public String[] getAttributeKeys() {
    NamedNodeMap nl = cur.getAttributes();
    String[] result = new String[nl.getLength()];
    for (int i = 0; i < nl.getLength(); i++) {
      result[i] = nl.item(i).getNodeName();
    }
    return result;
  }

  @Override
  public String getAttributeValue(String key) {
    if (cur.getAttributes().getNamedItem(key) == null) {
      return null;
    }
    return cur.getAttributes().getNamedItem(key).getNodeValue();
  }

  @Override
  public void setAttribute(String key, String value) {
    if (cur.getAttributes().getNamedItem(key) == null) {
      return;
    }
    cur.getAttributes().getNamedItem(key).setNodeValue(value);
  }
}

public class DocumentImporter {

  private IDocumentProcessor processor = null;

  public void setProcessor(IDocumentProcessor processor) {
    this.processor = processor;
  }

  public void importDocument(
      String docId, File inFile, File outFile, TreeDB base, TreeNode node, String encoding, PrintStream pw)
      throws Exception {
    Reader reader = new InputStreamReader(new FileInputStream(inFile), Charset.forName(encoding));

    if (!outFile.getParentFile().exists()) outFile.getParentFile().mkdirs();

    try {
      StringWriter out = new StringWriter();
      cleanHtml(reader, encoding, out);
      out.close();
      reader.close();
      Document bdoc = null;
      if (base != null) 
        bdoc = (Document) base.getNode(docId, Path.QID);
      if (processor != null) {
        processor.process(docId, node, pw, null);
      } else {
        if(bdoc!=null)
          bdoc.setResourceContent(bdoc.getResourceName(), 
            new ByteArrayInputStream(out.toString().getBytes(Charset.forName("UTF-8"))));
        else{
          Writer writer = new OutputStreamWriter(new FileOutputStream(outFile), "UTF-8");
          try{
            writer.append(out.toString());
          }finally{
            writer.close();
          }
        }
      }
      //writer.close();
      IXHTMLNode rootNode = null;
      org.w3c.dom.Document fdoc = null;
      if (base != null) {
        if (bdoc != null) {
          bdoc.setMetaCharset("UTF-8");
          if (bdoc.getDOM() != null) {
            fdoc = bdoc.getDOM();
            rootNode = new XMLNNode(fdoc.getFirstChild());
          }
        }
      }
      final Document doc = bdoc;
      String srcHtmlFolder = inFile.getParent();
      File nodeResourceFolder = null;
      if (doc != null) {
        nodeResourceFolder = new File(doc.getHTMLPath()).getParentFile();
      }
      XHTMLResourceCollector docProc =
          new XHTMLResourceCollector(
              rootNode,
              srcHtmlFolder,
              nodeResourceFolder != null ? nodeResourceFolder.toURI().toString() : null);
      docProc.process();
      if(doc!=null)
        doc.saveDOM("", fdoc);
      pw.println("Import succeed");
      /*
       * NodeEditorInput input = new NodeEditorInput(doc); try {
       * IDE.openEditor
       * (PlatformUI.getWorkbench().getActiveWorkbenchWindow()
       * .getActivePage(), input, ReqMarker.ID); } catch
       * (PartInitException e) { e.printStackTrace(); }
       */
    } catch (Exception e) {
      pw.println(CoreUtils.getTrace(e));
      throw e;
    } finally {
      reader.close();
    }
  }

  public static void cleanHtml(Reader intHtml, String charset, Writer out) throws IOException {
    try {
      CleanerProperties props = createCleanerProperties();
      props.setCharset(charset);
      HtmlCleaner cleaner = new HtmlCleaner(props);
      TagNode root = cleaner.clean(intHtml);
      root.addAttribute("xmlns", "http://www.w3.org/1999/xhtml");
      if (root.hasAttribute("xmlns:xml")) {
        root.removeAttribute("xmlns:xml");
      }
      props = createCleanerProperties();
      props.setCharset(charset);
      recRemoveLiLI(root);
      XmlSerializer serializer = new SimpleXmlSerializer(props);
      root.serialize(serializer, out);
    } catch (Exception e) {
      e.printStackTrace();
    }
  }

  private static boolean recRemoveLiLI(TagNode root) {
    if(root.getName().equals("li"))
      if(root.getAllChildren().isEmpty())
        return true;
    if(root.getAllChildren().isEmpty())
      return false;
    Set<TagNode> trem = new HashSet<>();
    for(BaseToken node:root.getAllChildren())
      if(node instanceof TagNode)
        if(recRemoveLiLI((TagNode)node))
          trem.add((TagNode) node);
    for(TagNode node:trem)
      root.removeChild(node);
    return false;
  }

  private static CleanerProperties createCleanerProperties() {
    CleanerProperties properties = new CleanerProperties();
    properties.setAdvancedXmlEscape(true);
    properties.setTranslateSpecialEntities(true);
    properties.setRecognizeUnicodeChars(true);
    properties.setUseEmptyElementTags(false);
    properties.setNamespacesAware(true);
    properties.setUseCdataForScriptAndStyle(true);
    properties.setOmitDeprecatedTags(false);
    properties.setOmitUnknownTags(false);
    properties.setOmitDoctypeDeclaration(false);
    properties.setIgnoreQuestAndExclam(true);
    return properties;
  }
}
