/*
 * Copyright 2015 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.unitesk.requality.nodetypes;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStreamWriter;
import java.io.Writer;
import java.net.MalformedURLException;
import java.net.URI;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.ListIterator;
import java.util.Map;
import java.util.Set;
import java.util.UUID;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.OutputKeys;
import javax.xml.transform.Result;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerConfigurationException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;
import org.apache.commons.httpclient.URIException;
import org.apache.commons.httpclient.util.URIUtil;
import org.w3c.dom.Attr;
import org.w3c.dom.DocumentFragment;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.w3c.dom.ls.DOMImplementationLS;
import org.w3c.dom.ls.LSSerializer;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;
import com.unitesk.requality.core.ChangesSet;
import com.unitesk.requality.core.INodeChangeListener;
import com.unitesk.requality.core.Moment;
import com.unitesk.requality.core.Path;
import com.unitesk.requality.core.RLogger;
import com.unitesk.requality.core.TreeDB;
import com.unitesk.requality.core.TreeNode;
import com.unitesk.requality.core.TreeNodeComparator;
import com.unitesk.requality.core.attribute.Attribute;
import com.unitesk.requality.core.attribute.Flag;
import com.unitesk.requality.documents.DeepFirstXMLDOMWalker;
import com.unitesk.requality.documents.DeepFirstXMLDOMWalker.IgroreableIterator;
import com.unitesk.requality.tools.RequalityCLI;

public class Document extends TreeNode implements INodeChangeListener {
  public static final String TYPE_NAME = "Document";
  private static final String ATTRIBUTE_FILE = "_file";
  private static final String TYPE_ROOT_QID = "/Documents";
  public static final String ATTR_IS_UPDATING = "_isUpdating";
  public static final String ATTR_ALREADY_UPDATED = "_alreadyUpdated";
  public static final String ATTR_TEXT_VERSION = "_textVersion";
  private static final String LOCATION_CLASS_ID_PREFIX = "id_";
  private static final String ATTR_UTI = "_luidtoind";
  private Map<UUID, Integer> locationUUIdToIndex = null;

  //private Map<UUID,Integer> locationUUIdToIndex;
  private org.w3c.dom.Document xmlDoc;

  private static boolean verbose = false;
  private freemarker.ext.dom.NodeModel modelCache = null;

  @Override
  protected void addChildUUID(UUID uid) {
    //resetCache();
    super.addChildUUID(uid);
  };

  @Override
  public void removeChildUUID(UUID node) {
    super.removeChildUUID(node);

    //resetCache();
  };

  @Override
  public TreeNode findWorkChild(String id) {
    TreeNode node = super.findWorkChild(id);
    if (node != null) return node;
    try {
      return getTreeDB().getNodeExc(UUID.fromString(id));
    } catch (Throwable e) {
      //e.printStackTrace();
    }
    return null;
  }

  public Document() {
    super(null, null);
  }

  //Use it only on reports
  public freemarker.ext.dom.NodeModel getNodeModel() {
    if (modelCache == null)
      try {
        modelCache = freemarker.ext.dom.NodeModel.parse(new File(getHTMLPath()));
      } catch (SAXException e) {
        e.printStackTrace();
      } catch (IOException e) {
        e.printStackTrace();
      } catch (ParserConfigurationException e) {
        e.printStackTrace();
      }
    return modelCache;
  }

  private Element getElementChildIgnoreCase(Node node, String name) {
    for (int i = 0; i < node.getChildNodes().getLength(); i++) {
      if (node.getChildNodes().item(i).getNodeName().toLowerCase().equals(name.toLowerCase())) {
        return (Element) node.getChildNodes().item(i);
      }
    }
    return null;
  }

  public void setMetaCharset(String charset) {
    try {
      org.w3c.dom.Document dom = getDOM();
      if (dom == null) return;
      Node html = dom.getFirstChild();
      Node head = getElementChildIgnoreCase(dom.getFirstChild(), "head");
      if (head == null) {
        head = dom.createElement("head");
        html.appendChild(head);
      }
      Element meta = getElementChildIgnoreCase(head, "meta");
      if (meta == null) {
        meta = dom.createElement("meta");
        head.appendChild(meta);
      }
      meta.setAttribute("http-equiv", "Content-Type");
      meta.setAttribute("content", "text/html; charset=" + charset);
      saveDOM(null, dom);
    } catch (Exception e) {
      e.printStackTrace();
    }
  }

  public String getMetaCharset() {
    // TODO: check rfc for this meta
    String result = null;
    try {
      DeepFirstXMLDOMWalker dfw = new DeepFirstXMLDOMWalker(getDOM().getFirstChild());
      for (Node node : dfw) {
        if (node.getNodeName().toLowerCase().equals("meta")) {
          if (node.getAttributes().getNamedItem("http-equiv") == null) continue;
          if (!node.getAttributes()
              .getNamedItem("http-equiv")
              .getNodeValue()
              .toLowerCase()
              .equals("content-type")) continue;
          if (node.getAttributes().getNamedItem("content") == null) continue;
          String content = node.getAttributes().getNamedItem("content").getNodeValue();
          int i = content.indexOf("charset=");
          if (i < 0) return null;
          return content.substring(i + "charset=".length());
        }
      }
    } catch (Exception e) {
      // TODO Auto-generated catch block
      e.printStackTrace();
    }
    return result;
  }

  public static String getTypeRootQId() {
    return TYPE_ROOT_QID;
  }

  public Document(TreeDB db, UUID uuid) {
    super(db, uuid);
    setType(TYPE_NAME);
  }

  public void setResourceName(String name) {
    attrs().setConstAttribute(ATTRIBUTE_FILE, name, Flag.HIDDEN);
  }

  public String getResourceName() {
    return attrs().getStringAttributeValue(ATTRIBUTE_FILE);
  }

  // private static final String LOCATION_CLASS_ID_PREFIX =
  // "requality_text id_";

  public URI getHTMLPath() {
    URI path = null;
    try {
      path = getNodeResourcesPath();
    } catch (Exception e) {
      // TODO Auto-generated catch block
      e.printStackTrace();
    }
    if (path == null || getResourceName() == null) return null;
    //URI base = URI.create(path);
    try {
      return URI.create(path.toURL() + "/" + URIUtil.encodeQuery(getResourceName()));
    } catch (MalformedURLException e) {
      // TODO Auto-generated catch block
      e.printStackTrace();
    } catch (URIException e) {
      // TODO Auto-generated catch block
      e.printStackTrace();
    }
    return path;
  }

  public Map<UUID, Integer> getLocationIndexMap() {
    return getLocationUUIdToIndex(false);
  }
  //TODO refactor
  public Set<UUID> getXMLLocations() {
    Map<UUID, Integer> map = new HashMap<>();
    try {
      int cnt = 0;
      org.w3c.dom.Document doc = loadDOM();
      if (doc == null) return Collections.emptySet();
      DeepFirstXMLDOMWalker dfw = new DeepFirstXMLDOMWalker(doc.getFirstChild());
      for (Node node : dfw) {
        //try to handle not only location with span
        //but every that have mark in class
        //    if (node.getNodeName().equals("span")) {
        if (node.getAttributes() == null) continue;
        Node spanClass = node.getAttributes().getNamedItem("class");
        if (spanClass != null) {
          String classNames[] = spanClass.getNodeValue().split(" ");
          for (String className : classNames) {
            if (className.startsWith(LOCATION_CLASS_ID_PREFIX)) {
              String id = className.substring(LOCATION_CLASS_ID_PREFIX.length());
              UUID uuid = UUID.fromString(id);
              try {
                TreeNode lnode = findWorkChild(id);
                if (lnode == null) lnode = getTreeDB().getNodeExc(uuid);
                if (lnode != null) uuid = lnode.getUUId();
              } catch (RuntimeException re) {
              }
              if (!map.containsKey(uuid)) {
                map.put(uuid, cnt++);
              }
            }
          }
        }
        //  }
      }
      locationUUIdToIndex = map;
      saveLocationUUIDToIndex();
      //reloadLocationUUIdToIndex();
    } catch (Exception e) {
      e.printStackTrace();
    }
    return locationUUIdToIndex.keySet();
  }

  private Map<UUID, String> locDescrs = null;
  //Warning! Slow and untestes
  public Map<UUID, String> getLocationsDescriptions() {
    if (locDescrs != null) return locDescrs;
    try {
      int cnt = 0;
      org.w3c.dom.Document doc = loadDOM();
      if (doc == null) return null;
      Map<UUID, String> descs = new HashMap<>();
      DeepFirstXMLDOMWalker dfw = new DeepFirstXMLDOMWalker(doc.getFirstChild());
      IgroreableIterator it = (IgroreableIterator) dfw.iterator();
      DOMImplementationLS domImplementation = (DOMImplementationLS) doc.getImplementation();
      LSSerializer lsSerializer = domImplementation.createLSSerializer();
      lsSerializer.getDomConfig().setParameter("xml-declaration", false);

      while (it.hasNext()) {
        Node node = it.next();
        //try to handle not only location with span
        //but every that have mark in class
        //    if (node.getNodeName().equals("span")) {
        if (node.getAttributes() == null) continue;
        Node spanClass = node.getAttributes().getNamedItem("class");
        if (spanClass != null) {
          String classNames[] = spanClass.getNodeValue().split(" ");
          for (String className : classNames) {
            if (className.startsWith(LOCATION_CLASS_ID_PREFIX)) {
              String id = className.substring(LOCATION_CLASS_ID_PREFIX.length());
              UUID uuid = UUID.fromString(id);
              it.ignoreChildren(node);
              //TODO calculate desc
              StringBuffer sb = new StringBuffer();
              if (descs.containsKey(uuid)) sb.append(descs.get(uuid));
              NodeList nl = node.getChildNodes();
              for (int i = 0; i < nl.getLength(); i++) {
                if (i == 0 && nl.item(i).getNodeName().equals("a")) continue;
                if (nl.item(i) instanceof org.w3c.dom.Text) sb.append(nl.item(i).getTextContent());
                else sb.append(lsSerializer.writeToString(nl.item(i)));
              }
              descs.put(uuid, sb.toString());
              if (!locationUUIdToIndex.containsKey(uuid)) {
                locationUUIdToIndex.put(uuid, cnt++);
                saveLocationUUIDToIndex();
              }
            }
          }
        }
        //  }
      }
      locDescrs = descs;
      return descs;
    } catch (Exception e) {
      e.printStackTrace();
    }
    return null;
  }

  //    private void buildDescription(StringBuffer sb, Node item) {
  //    	for(int i = 0;i<nl.getLength();i++)
  //    		buildDescription(sb,nl.item(i));
  //
  //	}
  public int getLocationIndex(UUID locuid) {
    if (getLocationUUIdToIndex(false) == null) return -1;
    if (getLocationUUIdToIndex(false).containsKey(locuid))
      return getLocationUUIdToIndex(false).get(locuid);
    int num = sizeWorkChildren();
    for (int i = 0; i < num; i++) {
      TreeNode child = getWorkChild(i);
      if (child.getUUId().equals(locuid)) {
        UUID id = UUID.fromString(child.getId());
        if (getLocationIndexMap().containsKey(id)) return getLocationIndexMap().get(id);
      }
    }
    return -1;
  }

  public static class LocationRegion {
    public LocationRegion() {
      startNodeType = endNodeType = Node.ELEMENT_NODE;
    }

    public LocationRegion(String sX, int sT, int sOff, String eX, int eT, int eOff, String text) {
      startNodeXPath = sX;
      startNodeType = sT;
      startNodeOffset = sOff;
      endNodeXPath = eX;
      endNodeType = eT;
      endNodeOffset = eOff;
      this.text = text;
    }

    public boolean isStarted() {
      return startNodeXPath != null;
    }

    private String startNodeXPath;
    private int startNodeType;
    private int startNodeOffset;

    private String endNodeXPath;
    private int endNodeType;
    private int endNodeOffset;

    private String text = "";

    @Override
    public String toString() {
      String start = makeNodeString("start", startNodeXPath, startNodeType, startNodeOffset);
      if (endNodeXPath == null) {
        endNodeXPath = startNodeXPath;
        endNodeOffset = startNodeOffset + text.length();
      }
      String end = makeNodeString("end", endNodeXPath, endNodeType, endNodeOffset);
      return "{" + start + "," + end + ",\"text\":\"" + text + "\"}";
    }

    private String makeNodeString(String name, String xpath, int type, int offset) {
      StringBuilder sb = new StringBuilder();
      sb.append("\"");
      sb.append(name);
      sb.append("\":{\"node\":\"");
      sb.append(xpath);
      sb.append("\",\"nodeType\":");
      sb.append(type);
      sb.append(",\"offset\":");
      sb.append(offset);
      sb.append("}");
      return sb.toString();
    }

    public void text(String text) {
      if (!isText()) {
        startNodeType = endNodeType = Node.TEXT_NODE;
      }
      this.text += text;
    }

    public boolean isText() {
      return startNodeType == Node.TEXT_NODE;
    }

    public void xpath(String xpath) {
      if (!isStarted()) {
        startNodeXPath = xpath;
      } else {
        endNodeXPath = xpath;
      }
    }

    public void offset(int offset) {
      if (!isStarted()) {
        startNodeOffset = offset;
      } else {
        endNodeOffset = offset;
      }
    }
  }

  public String removeLocationRegion(final String uuid) {
    try {
      return removeLocationRegions(
          new HashSet<String>() {
            private static final long serialVersionUID = 6662684128925762858L;

            {
              {
                add(uuid);
              }
            }
          });
    } catch (Exception e) {
      e.printStackTrace();
    }
    return null;
  }

  private static boolean needDropNode(Node node, Set<String> uuids) {
    if (node.getNodeName().equals("a")) {
      if (node.getAttributes().getNamedItem("class") != null) {
        String cl = node.getAttributes().getNamedItem("class").getNodeValue();
        for (String uuid : uuids) {
          if (cl.contains("id_" + uuid) || cl.contains("requality_node")) {
            return true;
          }
          if (node.getAttributes().getNamedItem("id") != null) {
            String id = node.getAttributes().getNamedItem("id").getNodeValue();
            if (cl.contains("requality_id") && id.equals("id_" + uuid)) return true;
          }
        }
      }
    }
    //if(node.getNodeType() == 3 && node.getNodeValue().equals("\n"))
    //return true;
    //if(node.getNodeType()==3 && node.getNodeValue().equals("\n"))
    //	return true;
    if (node.getNodeName().equals("span")) {
      String className = node.getAttributes().getNamedItem("class").getNodeValue();
      for (String uuid : uuids) {
        if (className.startsWith("requality_link requality_link_" + uuid)) {
          return true;
        }
        if (className.startsWith("requality_link requality_link_in_" + uuid)) {
          return true;
        }
      }
    }

    return false;
  }

  public Location createLocation() {
    return createLocation(null);
  }

  public Location createLocation(UUID id) {
    if (id == null) id = UUID.randomUUID();
    Location l =
        (Location)
            getTreeDB().createChildNode(id, getUUId(), id.toString(), Location.TYPE_NAME, true);
    // saveDOM(); // moved to ReqMarker.addSelectionTo
    return l;
  }

  private class RegionDescription {
    public final String id;
    public final LocationRegion r;

    public RegionDescription(String uid, LocationRegion r) {
      this.r = r;
      this.id = uid;
    }
  }

  private void fillQueue(Node node, LinkedList<Node> lnodes, Set<String> map) {
    LocationDetectionResult ldres = isLocationNode(node);
    if (ldres != LocationDetectionResult.NOT_LOCATION) {
      String id = getUUID(node);
      if (map.contains(id)) {
        lnodes.addLast(node);
        //else
        //node.getParentNode().removeChild(node);
      }
    }
    NodeList cnl = node.getChildNodes();
    Node[] nodes = new Node[cnl.getLength()];
    for (int i = 0; i < cnl.getLength(); i++) nodes[i] = cnl.item(i);
    for (int i = 0; i < nodes.length; i++) fillQueue(nodes[i], lnodes, map);
  }
  // TODO: nested location support
  private void removeLocations(
      Set<String> map, Node rnode, List<RegionDescription> regions, final int parentOffset) {
    LinkedList<Node> lnodes = new LinkedList<>();
    //Map<Node, LocationDetectionResult> locNodes =
    //	new HashMap<Node,LocationDetectionResult>();
    fillQueue(rnode, lnodes, map);
    List<Object[]> selTexts = new LinkedList<>();

    for (Node node : lnodes) {
      LocationDetectionResult ldres = isLocationNode(node);
      String id = getUUID(node);
      Node next = node.getNextSibling(), par = node.getParentNode();
      //TODO check
      if (ldres == LocationDetectionResult.NOT_LOCATION) continue;
      if (ldres == LocationDetectionResult.SPAN_LOCATION) {
        NodeList nl = node.getChildNodes(); //TODO can be bug?
        if (nl != null) {
          Node[] nodes = new Node[nl.getLength()];
          for (int i = 0; i < nl.getLength(); i++) nodes[i] = nl.item(i);
          for (int i = 0; i < nodes.length; i++) {
            Node nde = nodes[i];
            if (needDropNode(nde, map)) {
              node.removeChild(nde);
              continue;
            }
            int start = 0, len = 0;
            String text = null;
            if (nde.getNodeType() == Node.TEXT_NODE) {
              len = nde.getNodeValue().length();
              text = nde.getNodeValue();
            }
            nde = par.insertBefore(nde, node);
            if (nde.getNodeValue() != null && nde.getNodeValue().equals("\n")) continue;
            if (nde.getNodeType() == Node.TEXT_NODE) {
              start = nde.getNodeValue().length() - len;
              Node prev = nde.getPreviousSibling();
              if (prev != null && prev.getNodeType() == Node.TEXT_NODE) {
                start += prev.getNodeValue().length();
                prev.setNodeValue(prev.getNodeValue() + nde.getNodeValue());
                par.removeChild(nde);
                nde = prev;
              }
              if (i == (nodes.length - 1)
                  && (next != null)
                  && next.getNodeType() == Node.TEXT_NODE) {
                nde.setNodeValue(nde.getNodeValue() + next.getNodeValue());
                par.removeChild(next);
                next = null;
              }
              selTexts.add(new Object[] {true, nde, start, start + len, id, text});
              //	    						if(!selTexts.keySet().contains(nde))
              //	    							if(nde.getNodeValue().replace("\n", "").length()>0){
              //	    								selTexts.put(nde, nde.getNodeValue().length());
              //	    								getuid.put(nde, uuid.toString());
            }
          }
        }
        node.getParentNode().removeChild(node);
      } else {
        String cl = node.getAttributes().getNamedItem("class").getNodeValue();
        for (String sid : map) cl = cl.replace("id_" + sid, "");
        if (cl.indexOf("id_") < 0) cl = cl.replace("requality_node", "");
        Attr a = node.getOwnerDocument().createAttribute("requality_node_active");
        a.setValue(cl);
        node.getAttributes().setNamedItem(a);
        selTexts.add(new Object[] {false, node, id});
      }
    }
    //StringBuffer jsData = new StringBuffer();
    ListIterator<Object[]> stit = selTexts.listIterator(selTexts.size());
    while (stit.hasPrevious()) {
      Object[] st = stit.previous();
      if ((boolean) st[0]) {
        //text
        Node n = (Node) st[1];
        LocationRegion r = new LocationRegion();
        Object[] o = getElementXpath(n);
        r.startNodeXPath = r.endNodeXPath = (String) o[0];
        r.startNodeType = r.endNodeType = (Short) o[1];
        r.startNodeOffset = (int) st[2];
        r.endNodeOffset = (int) st[3];
        r.text = st[5].toString();
        RLogger.logInfo(r.toString());
        String uid = (String) st[4];
        regions.add(new RegionDescription(uid, r));
        //TODO js?
        //if(!regions.containsKey(uid))
        //	regions.put(uid, new LinkedList<Document.LocationRegion>());
        //regions.get(uid).add(r);
      } else {
        Node n = (Node) st[1];
        LocationRegion r = new LocationRegion();
        Object[] o = getElementXpath(n);
        r.startNodeXPath = r.endNodeXPath = (String) o[0];
        r.startNodeType = r.endNodeType = (Short) o[1];
        RLogger.logInfo(r.toString());
        String uid = (String) st[2];
        regions.add(new RegionDescription(uid, r));
        //jsData.append(b)
      }
    }
  }

  public void saveDOM() {
    saveDOM(null);
  }

  public void saveDOM(String jsData) {
    saveDOM(jsData, null);
  }

  private static Transformer f = null;
  private static Transformer getFactory(){
    if(f!=null)
      return f;
    TransformerFactory tFactory =  TransformerFactory.newInstance();
    Transformer transformer;
    try {
      f = transformer = tFactory.newTransformer();
      transformer.setOutputProperty(OutputKeys.METHOD, "xml");
      transformer.setOutputProperty(OutputKeys.ENCODING, "UTF-8");
    } catch (TransformerConfigurationException e) {
      // TODO Auto-generated catch block
      e.printStackTrace();
    }
    return f;
  }
  
  public void saveDOM(String jsData, org.w3c.dom.Document doc) {
    if (doc == null)
      try {
        doc = getDOM();
      } catch (Exception e) {
        e.printStackTrace();
      }

    long t_start = System.currentTimeMillis();
    try {
      ByteArrayOutputStream newXml = new ByteArrayOutputStream();
      Writer out = new OutputStreamWriter(newXml, "UTF-8");
      Result result = new StreamResult(out);

      getFactory().transform(new DOMSource(doc), result);
      out.close();
      InputStream is = new ByteArrayInputStream(newXml.toByteArray());
      setResourceContent(getResourceName(), is, jsData);
      is.close();

    } catch (Exception e) {
      e.printStackTrace();
    }
    // TODO: optimize cache

    if (verbose) {
      long t_end = System.currentTimeMillis();
      RLogger.logInfo("time Document.saveDOM(" + jsData + ") = " + (t_end - t_start));
    }
  }

  private void saveLocationUUIDToIndex() {
    if (locationUUIdToIndex == null) getLocationIndexMap();
    Map<String, Object> ts = new HashMap<>();
    for (UUID s : locationUUIdToIndex.keySet())
      ts.put(s.toString(), locationUUIdToIndex.get(s).toString());
    getTreeDB().getCacheStorage().setCacheMap(getUUId().toString(), "locations", ATTR_UTI, ts);
  }

  public void reloadLocationUUIdToIndex() {
    locationUUIdToIndex = null;
    getXMLLocations();
  }

  public void saveDOMFromXMLCache() {
    long t_start = System.currentTimeMillis();
    ByteArrayOutputStream newXml = new ByteArrayOutputStream();
    try {
      org.w3c.dom.Document doc = getXmlDoc();
      Writer out = new OutputStreamWriter(newXml, "UTF-8");
      Result result = new StreamResult(out);
      Transformer t = TransformerFactory.newInstance().newTransformer();
      t.setOutputProperty(OutputKeys.METHOD,"xml");
      t.setOutputProperty(OutputKeys.ENCODING,"UTF-8");
      t.transform(new DOMSource(doc), result);
      out.close();
      InputStream is = new ByteArrayInputStream(newXml.toByteArray());
      setResourceContent(getResourceName(), is, null);
      is.close();
    } catch (Exception e) {
      e.printStackTrace();
    }
    // TODO: optimize cache

    if (verbose) {
      long t_end = System.currentTimeMillis();
      RLogger.logInfo("time Document.saveDOMFromXMLCache() = " + (t_end - t_start));
    }
  }

  private enum LocationDetectionResult {
    NOT_LOCATION,
    SINGLE_LOC_TAG,
    SPAN_LOCATION
  };

  private LocationDetectionResult isLocationNode(Node node) {
    //handling imgs?
    //if (node.getNodeName().equals("span")) {
    if (node.hasAttributes()) {
      Node spanClass = node.getAttributes().getNamedItem("class");
      if (spanClass == null) {
        return LocationDetectionResult.NOT_LOCATION;
      }
      String className = spanClass.getNodeValue();
      if (node.getNodeName().equals("span")) {
        if (className.contains(LOCATION_CLASS_ID_PREFIX) && className.contains("requality_text")) {
          return LocationDetectionResult.SPAN_LOCATION;
        }
      } else if (className.contains("requality_node") && className.contains("id_"))
        return LocationDetectionResult.SINGLE_LOC_TAG;
    }
    //}
    return LocationDetectionResult.NOT_LOCATION;
  }

  private String getUUID(Node locationNode) {
    try {
      String classNames[] =
          locationNode.getAttributes().getNamedItem("class").getNodeValue().split(" ");
      for (String className : classNames) {
        if (className.startsWith(LOCATION_CLASS_ID_PREFIX)) {
          return className.substring(LOCATION_CLASS_ID_PREFIX.length());
        }
      }
    } catch (Exception e) {
      e.printStackTrace();
      return null;
    }
    return null;
  }

  public Location getLocation(String locationId) {
    Location l = null;
    try {
      l = (Location) findWorkChild(locationId);
    } catch (RuntimeException re) {
    }
    if (l == null)
      try {
        l = (Location) getTreeDB().getNodeExc(UUID.fromString(locationId));
      } catch (RuntimeException e) {
      }
    return l;
  } 

  public org.w3c.dom.Document getDOM() throws Exception {
    //        if (getXmlDoc() == null) {
    //            setXmlDoc(loadDOM());
    //        }
    //    	setXmlDoc(loadDOM());
    //        return getXmlDoc();

    setXmlDoc(loadDOM());
    return getXmlDoc();
  }

  public static org.w3c.dom.Document loadDOM(InputStream is) throws Exception {
    long t_start = System.currentTimeMillis();
    DocumentBuilder docBuilder = DocumentBuilderFactory.newInstance().newDocumentBuilder();
    if (is == null) return null;
    org.w3c.dom.Document result = null;
    try {
      result = docBuilder.parse(is);
    } catch (SAXParseException spe) {

    }
    is.close();
    if (result != null) result.normalize();

    if (verbose) {
      long t_end = System.currentTimeMillis();
      RLogger.logInfo("time Document.loadDOM() = " + (t_end - t_start));
    }

    return result;
  }

  public org.w3c.dom.Document loadDOM() throws Exception {
    InputStream in = getResourceContent(getResourceName());
    try{
      return loadDOM(in);
    }
    finally{
      if(in!=null)
        in.close();
    }
  }

  private boolean containsLocationId(String[] classNames) {
    boolean requalityTextClass = false;
    boolean locationIdClass = false;

    for (String name : classNames) {
      if (name.equals("requality_text")) {
        requalityTextClass = true;
      }
      if (name.startsWith(LOCATION_CLASS_ID_PREFIX)) {
        locationIdClass = true;
      }
      if (requalityTextClass && locationIdClass) {
        return true;
      }
    }
    return requalityTextClass && locationIdClass;
  }

  private String getLocationIdClass(String[] classNames) {
    for (String name : classNames) {
      if (name.startsWith(LOCATION_CLASS_ID_PREFIX)) {
        return name;
      }
    }
    return null;
  }

  public String extractLocationText(String locUUID) {
    try {

      org.w3c.dom.Document d = getDOM();
      DeepFirstXMLDOMWalker dfw = new DeepFirstXMLDOMWalker(d.getFirstChild());

      DocumentFragment fragment = d.createDocumentFragment();
      for (Node node : dfw) {
        if (node.getNodeName().equals("span")) {
          Node spanClass = node.getAttributes().getNamedItem("class");
          if (spanClass != null) {
            String[] classNames = spanClass.getNodeValue().split(" ");

            if (containsLocationId(classNames)) {
              String locationIdClass = getLocationIdClass(classNames);
              UUID uuid =
                  UUID.fromString(locationIdClass.substring(LOCATION_CLASS_ID_PREFIX.length()));
              if (uuid.toString().equals(locUUID)) {
                for (int i = 0; i < node.getChildNodes().getLength(); i++) {
                  Node ch = node.getChildNodes().item(i);
                  if (ch.getNodeName().equalsIgnoreCase("a")
                      && ch.getAttributes().getNamedItem("name") != null
                      && ch.getAttributes().getNamedItem("name").getNodeValue().equals(locUUID)) {
                    continue;
                  }
                  fragment.appendChild(ch.cloneNode(true));
                }
              }
            }
          }
        }
      }
      ByteArrayOutputStream result = new ByteArrayOutputStream();
      TransformerFactory transFactory = TransformerFactory.newInstance();
      Transformer transformer = transFactory.newTransformer();
      transformer.setOutputProperty(OutputKeys.OMIT_XML_DECLARATION, "yes");
      transformer.setOutputProperty(OutputKeys.METHOD,"xml");
      transformer.setOutputProperty(OutputKeys.ENCODING,"UTF-8");
      transformer.transform(new DOMSource(fragment), new StreamResult(result));
      result.flush();
      return result.toString("UTF-8");

    } catch (Exception e) {
      e.printStackTrace();
    }
    return null;
  }

  public int sizeLocations() {
    return getLocations().size();
  }

  @Override
  public String toString() {
    return getId();
  }
  public void resetUTI() {
    locationUUIdToIndex = null;
    if (getTreeDB().getCacheStorage() != null) {
      //Map<String,String> sto = getTreeDB().getCacheStorage().getMap(getUUId(), ATTR_UTI);
      getTreeDB()
          .getCacheStorage()
          .setCacheMap(getUUId().toString(), "locations", ATTR_UTI, new HashMap<String, Object>());
      //            if(sto!=null){
      //                len = Integer.parseInt(var.toString());
      //                for(int i = 0;i<len;i++){
      //                    getTreeDB().getCacheStorage().setValue(getUUId(), ATTR_UTI+"_"+i+"_ind", null);
      //                    getTreeDB().getCacheStorage().setValue(getUUId(), ATTR_UTI+"_"+i+"_uid", null);
      //                }
      //            }
      //            getTreeDB().getCacheStorage().setValue(getUUId(), ATTR_UTI+"_len", null);
    }
  }
  public void resetCache() {
    // TODO: optimize cache
    if (verbose) {
      RLogger.logInfo("in resetCache");
    }
    
    int len = 0;
    resetUTI();
    modelCache = null;
    //        if(getUUId()!=getTreeDB().getStorage().getRootNode()&&
    //    		!isVirtual()&&!VirtualNode.getVirtualStorage(getTreeDB()).hasCloneParent(getUUId())){
    //    			TreeNode node = VirtualNode.getVirtualTreeDB(getTreeDB()).getNode(getUUId());
    //    			if (node.isVirtual())
    //    				((Document)node).resetCache();
    //        	
    xmlDoc = null;
    locDescrs = null;
    modelCache = null;
  }

  public org.w3c.dom.Document getXmlDoc() {
    return xmlDoc;
  }

  public void setXmlDoc(org.w3c.dom.Document xmlDoc) {
    this.xmlDoc = xmlDoc;
  }

  public String createLocationRangesDescriptor(String locId) {
    try {
      getDOM();
    } catch (Exception e) {
      e.printStackTrace();
    }
    return "";
  }

  // /////////////////////////
  // xpath stuff

  private static Set<Node> recLook(Node n) {
    Set<Node> list = new HashSet<>();
    NodeList nl = n.getChildNodes();

    for (int i = 0; i < nl.getLength(); i++)
      if (!list.contains(nl.item(i))) {
        list.add(nl.item(i));
        list.addAll(recLook(nl.item(i)));
      }
    return list;
  }

  public static Object[] getElementXpath(Node nodeNow) {
    Set<Node> allnodes = recLook(nodeNow.getOwnerDocument());
    Node[] aallNodes = allnodes.toArray(new Node[0]);
    String path = "";
    short type = Node.ELEMENT_NODE;
    if (nodeNow.getNodeType() == Node.TEXT_NODE) {
      type = Node.TEXT_NODE;
      nodeNow = nodeNow.getParentNode();
    }
    try {
      for (;
          nodeNow != null && nodeNow.getNodeType() == Node.ELEMENT_NODE;
          nodeNow = nodeNow.getParentNode()) {
        if (((Element) nodeNow).hasAttribute("id")) {
          short uniqueIdCount = 0;
          for (int i = 0; i < aallNodes.length; i++) {
            if (aallNodes[i].getAttributes() != null
                && aallNodes[i].getAttributes().getNamedItem("id") != null
                && aallNodes[i]
                    .getAttributes()
                    .getNamedItem("id")
                    .getNodeValue()
                    .equals(nodeNow.getAttributes().getNamedItem("id").getNodeValue()))
              uniqueIdCount++;
            if (uniqueIdCount > 1) break;
          }
          if (uniqueIdCount == 1) {
            path =
                "id(\\\""
                    + nodeNow.getAttributes().getNamedItem("id").getNodeValue()
                    + "\\\")"
                    + "/"
                    + path;
            path = path.substring(0, path.length() - 1);
            return new Object[] {path, type};
          } else {
            if (nodeNow.getLocalName() == null) RLogger.logInfo("null");
            path =
                "x:"
                    + nodeNow.getLocalName().toLowerCase()
                    + "[@id=\""
                    + nodeNow.getAttributes().getNamedItem("id").getNodeValue()
                    + "\"]"
                    + "/"
                    + path;
          }
        } else {
          int i;
          Node sib;
          for (i = 1, sib = nodeNow.getPreviousSibling();
              sib != null;
              sib = sib.getPreviousSibling()) {
            if (sib.getNodeName() != null && sib.getNodeName().equals(nodeNow.getNodeName())) i++;
          }
          path = "x:" + nodeNow.getNodeName().toLowerCase() + "[" + i + "]" + "/" + path;
        }
      }
      if (path.length() == 0) return null;
      else path = "/" + path.substring(0, path.length() - 1);
      return new Object[] {path, type};
    } catch (Exception ee) {
      ee.printStackTrace();
    }
    return null;
  }

  public void duplicateImages(String path) {
    RequalityCLI.duplicateImages(this, path);
  }

  public int compareTo(Document document) {
    String qid1 = getQualifiedId(), qid2 = getQualifiedId();
    int i = 0, li = 0;
    while (i >= 0 && qid1.subSequence(0, i).equals(qid2.subSequence(0, i)) && i < qid2.length()) {
      li = i;
      i = qid1.indexOf("/", i + 1);
    }
    int id1 = qid1.indexOf("/", li + 1), id2 = qid2.indexOf("/", li + 1);
    if (id1 == -1) id1 = qid1.length();
    if (id2 == -1) id2 = qid2.length();
    TreeNode n1 = getTreeDB().getNode(qid1.substring(0, id1), Path.QID),
        n2 = getTreeDB().getNode(qid2.substring(0, id2), Path.QID);
    return TreeNodeComparator.getComparator(n1.getTreeDB()).compare(n1, n2);
  }

  public void addUpdatedProblem() {
    attrs().setConstAttribute(ATTR_IS_UPDATING, "updating", Flag.HIDDEN);
    commitAttributes();
  }

  public void markDocumentAsUpdated() {
    attrs().setConstAttribute(ATTR_ALREADY_UPDATED, "already updated", Flag.HIDDEN);
    commitAttributes();
  }

  public String removeLocationRegions(Set<String> lsel) {
    try {
      //resetCache();
      List<RegionDescription> regs = new LinkedList<>();
      removeLocations(lsel, getDOM().getFirstChild(), regs, 0);

      StringBuilder sb = new StringBuilder();
      if (regs.isEmpty()) return null;
      sb.append("[");

      boolean ch = false;
      for (RegionDescription rd : regs) {
        ch = true;
        if (sb.length() > 1) sb.append(",");
        sb.append("{\"action\":\"delete\",\"uuid\":");
        sb.append("\"");
        sb.append(rd.id);
        sb.append("\",");
        sb.append("\"selection\":[");
        //for (int i = regions.size()-1;i>=0;i--) {
        //LocationRegion r = regions.get(i);
        sb.append(rd.r);
        sb.append("]}");
      }
      sb.append("]");
      String result = sb.toString();
      //saveDOM(result);
      //            for(String location:lsel){
      //            	//TODO error?
      //            	UUID uuid = UUID.fromString(location);
      //	            if(getLocationCache()!=null && getLocationCache().contains(uuid)){
      //	        		//getLocationUUIdToIndex(true, false).remove(uuid);
      //	        		ch = true;
      //	        	}
      //	            List<TreeNode> nodes = findWorkChildPlusName(location);
      //	            if(nodes!=null&&nodes.size()>0)
      //	            {
      //	            	TreeNode node = nodes.iterator().next();
      //	            	uuid = node.getUUId();
      //		            if(getLocationCache()!=null && getLocationCache().contains(uuid)){
      //		            	getLocationUUIdToIndex(true, false).remove(uuid);
      //		            	ch = true;
      //		        	}
      //	            }
      //            }
      if (ch) reloadLocationUUIdToIndex();
      //saveDOM(result,xmlDoc);
      // fireGUINodeAttributeChange(new HashSet<String>(){{{add(Requirement.ATTR_LOCATIONS);}}});//TODO needed?
      return result;
    } catch (Exception e) {
      e.printStackTrace();
    }
    //saveDOM(null,xmlDoc);
    return null;
  }

  public List<UUID> getLocations() {
    Map<UUID, Integer> locs = getLocationUUIdToIndex(false);
    if (locs != null) return new ArrayList<>(locs.keySet());
    return Collections.emptyList();
  }

  @Override
  public void attributeChange(TreeNode source, String what, Attribute old) {
    // TODO Auto-generated method stub

  }

  @Override
  public void deleted(TreeNode source) {
    // TODO Auto-generated method stub

  }

  @Override
  public void moved(TreeNode source, TreeNode oldParent, boolean internal) {
    //		if(source.getUUId().equals(getUUId())&&oldParent!=null){
    //			actMoved(oldParent.getQualifiedId()+"/"+getNameOrId());
    //		}
  }

  @Override
  public void attributesChange(TreeNode source, ChangesSet<Attribute> what) {
    // TODO Auto-generated method stub

  }

  @Override
  public void fireNodeResourceChange(TreeNode treeNode, String resName, Moment m, boolean inner) {
    if (m == Moment.AFTER
        && treeNode.getUUId().equals(getUUId())
        && equals(resName, getResourceName())) {
      getXMLLocations();
      reloadLocationUUIdToIndex();
      //saveAttributes();
    }
  }

  private boolean equals(String resName, String resourceName) {
    return resName == null && resourceName == null
        || resName != null && resName.equals(resourceName);
  }

  private Map<UUID, Integer> getLocationUUIdToIndex(boolean internal) {
    if (locationUUIdToIndex == null) {
      //			if(force && !internal)
      //			{
      //				getLocations(true);
      //				saveLocationUUIdToIndex();
      //				return locationUUIdToIndex;
      //			}
      //String len = null;
      Map<String, Object> st = null;
      Integer inited = (Integer) getTreeDB().getCacheStorage().getCacheValue(getUUId().toString(), "locations",  "locationsInited");
      if(inited == null || 1!=inited)
        if (getTreeDB().getCacheStorage() != null)
          st = getTreeDB().getCacheStorage().getCacheMap(getUUId().toString(), "locations", ATTR_UTI);
      if (st != null) {
        Map<UUID, Integer> map = new HashMap<>();
        //int l = Integer.parseInt(len);
        for (String s : st.keySet()) {
          String s2 =
              s; //getTreeDB().getCacheStorage().getValue(getUUId(), ATTR_UTI+"_"+(i)+"_ind");
          Object so1 =
              st.get(
                  s); //getTreeDB().getCacheStorage().getValue(getUUId(), ATTR_UTI+"_"+(i)+"_uid");
          String s1 = null;
          if(so1 != null)
            s1 = so1.toString();
          if (s1 != null && s2 != null && !s1.isEmpty() && !s2.isEmpty())
            map.put(UUID.fromString(s2), Integer.parseInt(s1));
        }
        if (!map.isEmpty() || getWorkChildren().length == 0) {
          //	setLocationUUIdToIndex(new HashMap<UUID,Integer>());
          locationUUIdToIndex = map;
        }
      } else {
        //if(!internal){
        getXMLLocations();
        //reloadLocationUUIdToIndex();
        //}
      }
      getTreeDB().getCacheStorage().setCacheValue(getUUId().toString(), "locations",  "locationsInited",1);
    }
    return locationUUIdToIndex;
  }

  public void clearCache() {
    resetCache();
  }
}
