/*
 * Copyright 2009-2011 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.unitesk.requality.report;

import java.io.File;
import java.io.FileOutputStream;
import java.io.OutputStreamWriter;
import java.io.Writer;
import java.net.URI;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.UUID;
import com.unitesk.requality.core.CoreUtils;
import com.unitesk.requality.core.RLogger;
import com.unitesk.requality.core.SharedConsts;
import com.unitesk.requality.core.TreeDB;
import com.unitesk.requality.core.TreeNode;
import com.unitesk.requality.core.api.API;
import com.unitesk.requality.core.instancer.Instantiator;
import com.unitesk.requality.nodetypes.Requirement;
import com.unitesk.requality.nodetypes.VirtualNode;
import com.unitesk.requality.tests.CoverageFromFile;
import com.unitesk.requality.tests.CoverageStorage;
import com.unitesk.requality.tests.ICoverageStorageSource;
import com.unitesk.requality.tools.RequalityCLI;
import freemarker.template.Configuration;
import freemarker.template.DefaultObjectWrapper;

public class ReportGenerator {
  private static class ReportGenHelper {
    Set<String> repFiles = new HashSet<>();
    Set<TreeDB> treeDBs = new HashSet<>();
    Template templ;
    Map<TemplateFile, Writer> twriters;
    String saveTo;
    Map<String, Object> userData;
    ClassLoader baseCL;

    public ReportGenHelper(
        Template templ,
        Map<TemplateFile, Writer> twriters,
        String saveTo,
        Map<String, Object> userData,
        ClassLoader baseCL) {
      this.templ = templ;
      this.twriters = twriters;
      this.saveTo = saveTo;
      this.userData = userData;
      this.baseCL = baseCL;
    }

    private void generate(Requirement rootReq) {
      try {
        ReportGenerator.generateReport(rootReq, templ, twriters, saveTo, userData, baseCL, this);
      } catch (Exception e) {
        RLogger.logInfo(CoreUtils.getTrace(e));
      }
    }

    private void registerFile(String file) {
      repFiles.add(file);
    }

    private Set<TreeDB> getGenerated() {
      return treeDBs;
    }

    private void markGenerated(TreeDB db) {
      treeDBs.add(db);
    }
  }

  private static boolean useStat;

  public static void clearOldImages(File f, String saveTo) {
    // TODO look
    if (!f.exists()) return;
    String fc = RequalityCLI.getFileContent(f);
    RequalityCLI.deleteOldImages(fc, null, null, saveTo);
  }

  private static void hideInstancers(TreeNode node, List<TreeNode> hidden) {
    // TreeDB root = VirtualNode.getRootTree(node.getTreeDB(),3);
    if (node == null) return;
    TreeDB root = node.getTreeDB(); // TODO root?
    if (!root.isVirtual()) {
      UUID uid = node.getVRoot();
      if (uid != null) root = Instantiator.getExistVirtualTreeDB(root, uid);
      else root = Instantiator.getRootTree(root); // Instantiator.getVirtualTreeDB(root,
      // Instantiator.DEFAULT_INSTANCE_NAME);
    }
    node = root.getNode(node.getUUId());
    if (node == null || node.getType() == null) 
      RLogger.logInfo("NF!");
    if (node!=null && node.getType().equals(VirtualNode.TYPE_NAME) && !(((VirtualNode) node).isCollapsed())) {
      ((VirtualNode) node).collapse();
      hidden.add(node);
      node.commitAttributes();
      //node.getTreeDB().updateShownNodeChildren(node.getWorkParent(), false);
    }
    if (node != null)
      for (UUID node2 : node.getWorkChildrenUUIDs())
        // if(!(virt.getVStorage()).isHidden(node2))
        hideInstancers(node.getTreeDB().getNode(node2), hidden);
  }

  public static void generateReport(
      Requirement rootReq,
      Template templ,
      Map<TemplateFile, Writer> twriters,
      String saveTo,
      Map<String, Object> userData,
      ClassLoader baseCL,
      ReportGenHelper reportGenHelper)
      throws Exception {
    if (rootReq == null) throw new RuntimeException("Root node is null!");
    List<TreeNode> hidden = new ArrayList<>();
    hideInstancers(rootReq, hidden);

    Configuration cfg = new Configuration();
    cfg.setObjectWrapper(new DefaultObjectWrapper());
    cfg.setDirectoryForTemplateLoading(new File(URI.create(templ.getReportPath())));
    for (TemplateFile t : templ.getTemplates()) {
      Writer out = twriters.get(t);

      try {
        Map<String, Object> data = new HashMap<>(userData);
        if (t.getDataClass().length() > 0) {
          IReportDataSource repObject = t.instanceDataObject(baseCL);
          data.putAll(repObject.getData());
          data.put("report", repObject);
        } else {
          data.put("report", new ReportTools());
        }
        data.put("reportLocation", saveTo);
        data.put("rootReq", rootReq);
        data.put("root", rootReq.getTreeDB().getRootNode());
        data.put("utils", new ReportUtils(rootReq.getTreeDB(), rootReq,t));

        CoverageStorage cs;
        if (data.containsKey("coverageStorageType")) {
          String type = (String) data.get("coverageStorageType"),
              setts = (String) data.get("coverageStorageSettings");
          ICoverageStorageSource ics = API.getCoverageSource(type);
          if (ics != null) {
            cs = CoverageStorage.getStorage(rootReq.getTreeDB(), ics);
            if (setts != null) cs.setSettings(setts.split(","));
          } else cs = CoverageStorage.getStorage(rootReq.getTreeDB(), new CoverageFromFile());
        } else {
          cs = CoverageStorage.getStorage(rootReq.getTreeDB(), new CoverageFromFile());
          if (data.containsKey("coverageFilePath"))
            cs.setSettings(new String[] {(String) data.get("coverageFilePath")});
        }
        data.put("coverage", cs);

        freemarker.template.Template ftl = cfg.getTemplate(t.getTemplatePath(), "UTF-8");
        // Statistics st = null;
        // if(cs.getSettings()!=null){
        // st = new
        // Statistics(rootReq.getTreeDB().getRootNode().getResourcesPath());
        // }
        Progress st =
            new Progress(rootReq.getTreeDB().getRootNode().getNodeResourcesPath(), saveTo);
        data.put("progress", st);
        ;

        if (useStat) {
          Class<? extends Object> c = SharedConsts.statManClass;
          if (c != null) {
            TreeDB db = rootReq.getTreeDB();
            if (db.isVirtual()) db = Instantiator.getRootTree(db);
            Object sm =
                c.getConstructor(CoverageStorage.class, TreeDB.class, URI.class)
                    .newInstance(cs, db, rootReq.getTreeDB().getRootNode().getNodeResourcesPath());
            data.put("statman", sm);
          }
        } else {
          StatisticManagerCLI sm = new StatisticManagerCLI(rootReq, cs);
          data.put("statman", sm);
        }
        data.put("userData", userData);
        data.put("coreUtils", new CoreUtils());
        data.put("generator", reportGenHelper);
        ftl.process(data, out);
        // TODO fix for report treeNode

        if (cs.getSettings(rootReq) != null)
          cs.saveCoverageToFile(
              saveTo
                  + "/"
                  + t.getOutputFileName(rootReq.getTreeDB())
                  + "_lastCov.xml");
      } catch (Exception e) {
        RLogger.logInfo(CoreUtils.getTrace(e));
        throw e;
      } finally {
        out.close();
        // for(TreeNode nd : hidden)
        // try{
        // if(nd.getType().equals(VirtualNode.TYPE_NAME)){
        // //TreeDB root = VirtualNode.getRootTree(nd.getTreeDB(),3);
        // //nd=root.getNode(nd.getUUId());
        // //TODO root?
        // ((VirtualNode)nd).unCollapse();
        // nd.saveAttributes();
        // rootReq.getTreeDB().updateNodeChildren(nd.getWorkParent(),
        // true);
        // }
        // }catch(Exception e){
        // e.printStackTrace();
        // }
      }
    }
  }

  public static void generateReport(
      Requirement rootReq,
      Template templ,
      String saveTo,
      Map<String, Object> userData,
      ClassLoader baseCL)
      throws Exception {
    Map<TemplateFile, Writer> map = new HashMap<>();

    try{
      for (TemplateFile t : templ.getTemplates()) {
  
        File outFile =
            new File(
                URI.create(
                    saveTo
                        + "/"
                        + t.getOutputFileName(rootReq.getTreeDB())));
        clearOldImages(outFile, saveTo + "/");
        outFile.getParentFile().mkdirs();
        FileOutputStream fos = new FileOutputStream(outFile);
        OutputStreamWriter osw = new OutputStreamWriter(fos, "UTF-8");
        map.put(t,osw);
      }
      generateReport(
        rootReq,
        templ,
        map,
        saveTo,
        userData,
        baseCL,
        new ReportGenHelper(templ, map, saveTo, userData, baseCL));
    }finally{
      for(Writer o: map.values())
        o.close();
    }
  }

  public static void generateReport(
      Requirement rootReq, Template templ, String saveTo, Map<String, Object> userData)
      throws Exception {
    useStat = false;
    generateReport(rootReq, templ, saveTo, userData, ClassLoader.getSystemClassLoader());
  }

  public static void generateReport(
      Requirement rootReq,
      Template templ,
      Map<TemplateFile, Writer> twriters,
      String saveTo,
      Map<String, Object> userData)
      throws Exception {
    useStat = true;
    ClassLoader cl = ClassLoader.getSystemClassLoader();
    generateReport(
        rootReq,
        templ,
        twriters,
        saveTo,
        userData,
        ClassLoader.getSystemClassLoader(),
        new ReportGenHelper(templ, twriters, saveTo, userData, cl));
  }
}
