/*
 * Copyright 2015 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.unitesk.requality.report;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.Reader;
import java.net.URI;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import javax.xml.namespace.QName;
import javax.xml.stream.XMLEventReader;
import javax.xml.stream.XMLInputFactory;
import javax.xml.stream.XMLStreamConstants;
import javax.xml.stream.events.StartElement;
import javax.xml.stream.events.XMLEvent;
import com.unitesk.requality.core.RLogger;

public class Template {
  private static final QName TEMPLATE = new QName("template");
  private static final QName FILE = new QName("file");
  private static final QName DIR = new QName("dir");
  private static final QName REPORT = new QName("report");
  private static final QName SOURCE = new QName("source");
  private static final QName LOCATION = new QName("location");
  private static final QName DATACLASS = new QName("class");
  private static final QName READONLY = new QName("read-only");
  private static final QName GUI = new QName("gui");
  private static final QName BUTTON = new QName("button");
  private static final QName LABEL = new QName("label");
  private static final QName COMMAND = new QName("command");
  private static final QName TEXT = new QName("text");
  private static final QName GLOBAL = new QName("global");
  public static final String DEFAULT_ID = "com.unitesk.requality.reports.requirements";

  private String reportPath;
  private List<TemplateFile> templates = new ArrayList<>();
  private String id;
  private String reportName;
  private boolean isGuiStarted;
  private long lastGuiIndex = 0;
  //id, map of params,including _type == [Button], _command = "command to call on click"
  private Map<String, Map<String, String>> guiMap = new HashMap<>();
  private boolean readonly = false;
  private boolean global = false;

  public Template(String reportPath) {
    this.reportPath = reportPath;
    try {
      load(reportPath);
    } catch (IOException e) {
      e.printStackTrace();
    }
  }

  public String getReportPath() {
    return reportPath;
  }

  public String load(String location) throws IOException {
    File file = new File(URI.create(reportPath + "/" + "template.xml"));
    if (!file.exists() || !file.isFile()) {
      throw new IOException("File " + file.getPath() + " does not exist.");
    }
    FileInputStream fis = new FileInputStream(file.getPath());
    InputStreamReader isr = new InputStreamReader(fis, "UTF-8");
    try{
      return load(isr);
    }
    finally{
      isr.close();
      fis.close();
    }
  }

  public String load(Reader reader) throws IOException {
    Map<String, Object> vars = new HashMap<>();
    String first = "";

    try {
      XMLInputFactory factory = XMLInputFactory.newInstance();
      XMLEventReader eReader = factory.createXMLEventReader(reader);

      while (eReader.hasNext()) {
        String res;
        XMLEvent event = eReader.nextEvent();
        switch (event.getEventType()) {
          case XMLStreamConstants.START_ELEMENT:
            res = processElem(event.asStartElement(), vars);
            if (first.length() == 0) {
              first = res;
            }
            break;
          case XMLStreamConstants.END_ELEMENT:
            if (isGuiStarted && event.asEndElement().getName().equals(GUI)) isGuiStarted = false;
          case XMLStreamConstants.CHARACTERS:
            break;
        }
      }
    } catch (Exception e) {
      e.printStackTrace();
    }
    return first;
  }

  public Map<String, Map<String, String>> getGUIElements() {
    return Collections.unmodifiableMap(guiMap);
  }

  private String processElem(StartElement elem, Map<String, Object> vars) throws IOException {
    javax.xml.stream.events.Attribute attr;
    attr = elem.getAttributeByName(SOURCE);
    String source = attr != null ? attr.getValue() : "";
    attr = elem.getAttributeByName(LOCATION);
    String location = attr != null ? attr.getValue() : "";
    attr = elem.getAttributeByName(DATACLASS);
    String className = attr != null ? attr.getValue() : "";

    QName name = elem.getName();
    // source = new File(folder, source).getAbsolutePath();
    if (name.equals(TEMPLATE)) {
      try {
        templates.add(new TemplateFile(reportPath, source, location, className, readonly, global));
      } catch (Exception e) {
        e.printStackTrace();
      }
      // processTemplate(source, vars, location);
    } else if (name.equals(FILE)) {
      copyFile(source, location);
    } else if (name.equals(DIR)) {
      copyDir(source, location);
    } else if (name.equals(REPORT)) {
      attr = elem.getAttributeByName(new QName("id"));
      if (attr == null) {
        throw new IOException("report must have id attribute");
      }
      id = attr.getValue();
      attr = elem.getAttributeByName(new QName("name"));
      reportName = attr == null ? id : attr.getValue();
      attr = elem.getAttributeByName(READONLY);
      readonly = attr != null ? attr.getValue().toLowerCase().equals("true") : false;
      attr = elem.getAttributeByName(GLOBAL);
      global = attr != null ? attr.getValue().toLowerCase().equals("true") : false;
    } else if (name.equals(GUI)) {
      guiMap = new HashMap<>();
      Map<String, String> settings = new HashMap<>();
      //TODO settings, for ex. layout
      guiMap.put("settings", settings);
      isGuiStarted = true;
    } else if (name.equals(BUTTON)) {
      attr = elem.getAttributeByName(COMMAND);
      lastGuiIndex++;
      HashMap<String, String> params = new HashMap<>();
      String id = name + "_" + lastGuiIndex;
      params.put("_type", name.getLocalPart());
      params.put("_command", attr != null ? attr.getValue() : "");
      attr = elem.getAttributeByName(TEXT);
      params.put("_text", attr != null ? attr.getValue() : id);
      guiMap.put(id, params);
    } else if (name.equals(LABEL)) {
      attr = elem.getAttributeByName(TEXT);
      lastGuiIndex++;
      HashMap<String, String> params = new HashMap<>();
      String id = name + "_" + lastGuiIndex;
      params.put("_type", name.getLocalPart());
      params.put("_text", attr != null ? attr.getValue() : id);
      guiMap.put(id, params);
    }
    return location;
  }

  /**
   * Copies catalogue into selected place.
   *
   * @param source path to catalogue which must be copied.
   * @param location path to where catalogue must be copied.
   * @return true if coping was successful, false - otherwise.
   */
  public boolean copyDir(String source, String location) {
    //TODO test it
    String fileSeparator = System.getProperty("file.separator");
    source = source.replace("/", fileSeparator);
    location = location.replace("\\", fileSeparator);
    boolean result = true;
    File loc = new File(location);
    if (!loc.exists()) {
      loc.mkdirs();
    }
    File dir = new File(source);
    if (dir.exists() && dir.isDirectory()) {
      File[] files = dir.listFiles();
      for (File file : files) {
        String filePath = file.getPath();
        String locPath = new File(location, file.getName()).getPath();

        if (file.isDirectory()) {
          copyDir(filePath, locPath);
        } else {
          copyFile(filePath, locPath);
        }
      }
    } else {
      System.err.println("Directory " + dir.getPath() + " does not exist.");
      result = false;
    }
    return result;
  }

  /**
   * Copies file into selected place.
   *
   * @param source path to file which must be copied.
   * @param location path to location where file must be copied.
   * @return true if coping successful, false - otherwise.
   */
  public boolean copyFile(String source, String location) {
    boolean result = true;

    try {
      String fileSeparator = System.getProperty("file.separator");
      source = source.replace("/", fileSeparator);
      location = location.replace("\\", fileSeparator);
      File inFile = new File(source);
      if (inFile.exists() && inFile.isFile()) {
        File outFile = new File(location);
        if (outFile.exists()) {
          System.err.println("File " + outFile.getPath() + " already exists.");
          outFile.delete();
        }
        if (!outFile.exists()) {
          RLogger.logInfo("Creating input stream from file: " + inFile.getPath());

          InputStream in = new FileInputStream(inFile);
          OutputStream out = new FileOutputStream(outFile);
          byte[] buf = new byte[1024];
          int len;
          while ((len = in.read(buf)) > 0) {
            out.write(buf, 0, len);
          }
          in.close();
          out.close();
        } else {
          System.err.println("Cannot delete file " + outFile.getPath() + ".");
        }
      } else {
        System.err.println("File " + inFile.getPath() + " does not exist.");
      }
    } catch (Exception e) {
      e.printStackTrace();
      result = false;
    }
    return result;
  }

  public static Template fromXml(String path, Reader reader) throws IOException {
    Template result = new Template(path);
    result.load(reader);
    return result;
  }

  public List<TemplateFile> getTemplates() {
    return templates;
  }

  public String getId() {
    return id;
  }

  public boolean getReadOnly() {
    return readonly;
  }

  public boolean getIsGlobal() {
    return global;
  }

  public String getName() {
    return reportName;
  }
}
