/*
 * Copyright 2009-2011 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.unitesk.requality.eclipse;

import java.io.File;
import java.io.IOException;
import java.net.URI;
import java.net.URISyntaxException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.SortedMap;
import java.util.SortedSet;
import java.util.TreeMap;
import java.util.TreeSet;
import org.eclipse.core.commands.AbstractHandler;
import org.eclipse.core.commands.Command;
import org.eclipse.core.commands.ExecutionEvent;
import org.eclipse.core.commands.ExecutionException;
import org.eclipse.core.commands.IHandler;
import org.eclipse.core.resources.IProject;
import org.eclipse.core.resources.IResourceChangeEvent;
import org.eclipse.core.resources.IResourceChangeListener;
import org.eclipse.core.resources.IWorkspace;
import org.eclipse.core.resources.IWorkspaceRoot;
import org.eclipse.core.resources.ResourcesPlugin;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.FileLocator;
import org.eclipse.core.runtime.IConfigurationElement;
import org.eclipse.core.runtime.IExtensionRegistry;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Platform;
import org.eclipse.core.runtime.Status;
import org.eclipse.jface.preference.IPreferenceStore;
import org.eclipse.jface.resource.ImageDescriptor;
import org.eclipse.jface.resource.ImageRegistry;
import org.eclipse.swt.widgets.Display;
import org.eclipse.ui.IPageListener;
import org.eclipse.ui.IPartListener2;
import org.eclipse.ui.IWindowListener;
import org.eclipse.ui.IWorkbenchPage;
import org.eclipse.ui.IWorkbenchPartReference;
import org.eclipse.ui.IWorkbenchWindow;
import org.eclipse.ui.PlatformUI;
import org.eclipse.ui.commands.ICommandService;
import org.eclipse.ui.handlers.IHandlerService;
import org.eclipse.ui.plugin.AbstractUIPlugin;
import org.eclipse.ui.progress.WorkbenchJob;
import org.osgi.framework.BundleContext;
import com.unitesk.requality.core.EventSource;
import com.unitesk.requality.core.IDeleteHandler;
import com.unitesk.requality.core.LoggerSource;
import com.unitesk.requality.core.RLogger;
import com.unitesk.requality.core.ResManager;
import com.unitesk.requality.core.SharedConsts;
import com.unitesk.requality.core.TreeDB;
import com.unitesk.requality.core.TreeNode;
import com.unitesk.requality.core.attribute.AttributeGenerator;
import com.unitesk.requality.core.descriptors.CoverageStorageSourceDescriptor;
import com.unitesk.requality.core.descriptors.DocumentProcessorDescriptor;
import com.unitesk.requality.core.descriptors.NodeTypeDescriptor;
import com.unitesk.requality.core.descriptors.TestGeneratorDescriptor;
import com.unitesk.requality.core.nodeiterators.IAttributeGenerator;
import com.unitesk.requality.core.selection.IUIStateChangeListener;
import com.unitesk.requality.core.selection.SelectionManager;
import com.unitesk.requality.documents.IDocumentProcessor;
import com.unitesk.requality.eclipse.api.EclipseAPI;
import com.unitesk.requality.eclipse.core.EclipseTracker;
import com.unitesk.requality.eclipse.cvs.CVSManager;
import com.unitesk.requality.eclipse.cvs.CVSSourcePlugin;
import com.unitesk.requality.eclipse.cvs.ICVSClient;
import com.unitesk.requality.eclipse.handlers.NewElementHandler;
import com.unitesk.requality.eclipse.natures.ReqCatalogue;
import com.unitesk.requality.eclipse.views.documents.DocumentProcessor;
import com.unitesk.requality.eclipse.views.properties.ProjectReferenceProperties;
import com.unitesk.requality.eclipse.wizards.DocWizard;
import com.unitesk.requality.report.IReportDataSource;
import com.unitesk.requality.report.ITemplatePathSource;
import com.unitesk.requality.report.StatisticManager;
import com.unitesk.requality.report.Template;
import com.unitesk.requality.report.TemplateFile;
import com.unitesk.requality.tests.ICoverageStorageSource;
import com.unitesk.requality.tests.ITestGenerator;
import com.unitesk.requality.tests.ITestGeneratorsProvider;
import com.unitesk.requality.tools.IMMagic;
import com.unitesk.requality.tools.Inkscape;
import com.unitesk.requality.tools.LibreCLI;

/** The activator class controls the plug-in life cycle */
public class Activator extends AbstractUIPlugin {

  // The plug-in ID
  public static final String PLUGIN_ID = "com.unitesk.requality";

  public static final String EXTENSION_NAME_DOC_PROCESSOR =
      "com.unitesk.requality.document.processor";
  public static final String EXTENSION_NAME_REFERENCE_ACR = "com.unitesk.requality.referencerename";
  public static final String EXTENSION_NAME_REPORT = "com.unitesk.requality.report";
  public static final String EXTENSION_NAME_TEST_GENERATORS = "com.unitesk.requality.testgenerator";
  public static final String EXTENSION_NAME_TEST_GENERATORS_PROVIDER =
      "com.unitesk.requality.testgeneratorprovider";
  public static final String EXTENSION_NAME_NODE_TYPES = "com.unitesk.requality.nodetype";
  public static final String EXTENSION_NAME_CHECKERS = "com.unitesk.requality.checker";
  public static final String EXTENSION_NAME_COVERAGE_SOURCES =
      "com.unitesk.requality.coveragesource";
  public static final String EXTENSION_NAME_ATTRIBUTE_GENERATORS =
      "com.unitesk.requality.attributegenerator";
  public static final String EXTENSION_NAME_HANDLERS = "org.eclipse.ui.handlers";

  // The shared instance
  private static Activator plugin;

  private CVSManager cvclManager;

  private void initAttributeGenerators() {
    AttributeGenerator.initGenMap(loadAttributeGeneratorsDescriptions());
  }

  private Map<String, List<ICVSClient>> supportMap = new HashMap<>();
  /** The constructor */
  public Activator() {}

  // //////////////////////////////////////////////////////////////////////////
  //
  // Bundle operations
  //
  // //////////////////////////////////////////////////////////////////////////
  private boolean loaded = false;;

  @Override
  public void stop(BundleContext context) throws Exception {
    for (TreeDB db : EclipseTracker.getTracker().getTrees()) {
      EclipseTracker.getTracker().getRepTask(db).halt();
    }
    super.stop(context);
  }
  private class ELogger extends RLogger{
    org.eclipse.e4.core.services.log.Logger logger;
    public ELogger(){
      logger = PlatformUI.getWorkbench().getService(org.eclipse.e4.core.services.log.Logger.class);
    }
    @Override
    public void lI(String s) {
      logger.info(s);
    };
    @Override
    public void lE(String s) {
      logger.error(s);
    };
    @Override
    public void lD(String s) {
      logger.debug(s);
    };
  }

  @Override
  public void start(BundleContext context) throws Exception {
    try {
      initTreesTracker();
      loaded = false;
      super.start(context);
      plugin = this;
      RLogger.loggerS = new LoggerSource(){
        ELogger el = null;
        @Override
        public RLogger getLogger() {
          if(el==null||el.logger==null)
            el=new ELogger();
          return el;
        };
      };
      initSharedConsts();
      initAPI();
      DocumentProcessor.eclipseGUI = true;
      loadReferenceAcronyms();
      initCVSClients();
      initTracker();
      initAttributeGenerators();
      initCLIs();
      ResourcesPlugin.getWorkspace().addResourceChangeListener(resourceChangeListener);
      loaded = true;
      SelectionManager.getOne()
          .addChangeListener(
              EventSource.GENERAL,
              new IUIStateChangeListener() {

                @Override
                public void newActiveNodes(List<TreeNode> nodes) {
                  updateDeleteInMenu();
                }
              });
      initPartListener();
      //startRPC();
    } catch (Exception e) {
      e.printStackTrace();
    }
  }

  public void initAPI() {
    EclipseAPI.init();
  }

  private void initPartListener() {
    WorkbenchJob wj = new WorkbenchJob("init pl"){
    
      @Override
      public IStatus runInUIThread(IProgressMonitor monitor) {
        getWorkbench().addWindowListener(new IWindowListener(){
          @Override
          public void windowOpened(IWorkbenchWindow window) {
            window.addPageListener(pl);
            for(IWorkbenchPage p:window.getPages())
              if(p!=null)
                p.addPartListener(
                  list);
          }
        
          @Override
          public void windowDeactivated(IWorkbenchWindow window) {
            // TODO Auto-generated method stub
            
          }
        
          @Override
          public void windowClosed(IWorkbenchWindow window) {
            window.removePageListener(pl);
          }
        
          @Override
          public void windowActivated(IWorkbenchWindow window) {
            // TODO Auto-generated method stub
            
          }
        });
        if(!getWorkbench().isStarting())
          for(IWorkbenchWindow w:getWorkbench().getWorkbenchWindows()){
            w.addPageListener(pl);
            for(IWorkbenchPage p:w.getPages())
              if(p!=null)
                p.addPartListener(
                  list);
          }
        return Status.OK_STATUS;
      }
    };
    wj.schedule();
  }
  IPageListener pl = new IPageListener(){
    
    @Override
    public void pageOpened(IWorkbenchPage page) {
      page.addPartListener(list);
    }
  
    @Override
    public void pageClosed(IWorkbenchPage page) {
      page.removePartListener(list);
    }
  
    @Override
    public void pageActivated(IWorkbenchPage page) {
      // TODO Auto-generated method stub
      
    }
  };
  IPartListener2 list = new IPartListener2() {

    @Override
    public void partVisible(IWorkbenchPartReference partRef) {
      // TODO Auto-generated method stub

    }

    @Override
    public void partOpened(IWorkbenchPartReference partRef) {
      // TODO Auto-generated method stub

    }

    @Override
    public void partInputChanged(IWorkbenchPartReference partRef) {
      // TODO Auto-generated method stub

    }

    @Override
    public void partHidden(IWorkbenchPartReference partRef) {
      // TODO Auto-generated method stub

    }

    @Override
    public void partDeactivated(IWorkbenchPartReference partRef) {
      // TODO Auto-generated method stub

    }

    @Override
    public void partClosed(IWorkbenchPartReference partRef) {
      // TODO Auto-generated method stub

    }

    @Override
    public void partBroughtToTop(IWorkbenchPartReference partRef) {
      // TODO Auto-generated method stub

    }

    @Override
    public void partActivated(IWorkbenchPartReference partRef) {
      if (partRef.getId().contains("RequalityExplorer"))
        SelectionManager.getOne().setSource(EventSource.EXPLORER);
      else if (partRef.getId().contains("Outline")||partRef.getId().contains("Editor")||partRef.getId().contains("Marker"))
        SelectionManager.getOne().setSource(EventSource.EDITOR);
      else if (partRef.getId().contains("RequalityLinksExplorer"))
        SelectionManager.getOne().setSource(EventSource.LCNF);
      else if (partRef.getId().contains("Properties"))
        SelectionManager.getOne().setSource(EventSource.PROPERTIES);
      else SelectionManager.getOne().setSource(EventSource.GENERAL);
      //RLogger.logInfo(partRef.getId()+" "+partRef.getPartName()+" "+partRef.getTitle());
    }
  };
  
  private void initTreesTracker() {
    ResManager.setTreesTracker(EclipseTracker.getTracker());
  }

  private void initSharedConsts() {
    SharedConsts.statManClass = StatisticManager.class;
    com.unitesk.requality.eclipse.handlers.location.CommonRemoveLocation.init(
        new com.unitesk.requality.eclipse.handlers.location.CommonRemoveLocation());
    com.unitesk.requality.core.operations.GenerateReportHandler.init(
        new com.unitesk.requality.eclipse.handlers.report.GenerateReportHandler());
  }

  private void initCLIs() {
    final IPreferenceStore pref = Activator.getDefault().getPreferenceStore();
    String path = pref.getString("com.unitesk.requality.librepath");
    LibreCLI.init(
        path,
        new Runnable() {
          public void run() {
            String val = LibreCLI.getDefault().getPath();
            pref.setValue("com.unitesk.requality.librepath", val);
            LibreCLI.getDefault().setPath(val);
          }
        });
    path = pref.getString("com.unitesk.requality.magick");
    IMMagic.init(
        path,
        new Runnable() {
          public void run() {
            String val = IMMagic.getDefault().getPath();
            pref.setValue("com.unitesk.requality.magick", val);
            IMMagic.getDefault().setPath(val);
          }
        });
    path = pref.getString("com.unitesk.requality.inkscape");
    Inkscape.init(
        path,
        new Runnable() {
          public void run() {
            String val = Inkscape.getDefault().getPath();
            pref.setValue("com.unitesk.requality.inkscape", val);
            Inkscape.getDefault().setPath(val);
          }
        });
  }
  //	private void startRPC() {
  //    	startRPC(9321,9321);
  //    }
  //    private void startRPC(int port,int start) {
  //    	try{
  //    		Endpoint end = Endpoint.publish("http://localhost:"+port+"/requalityService2",new ReqService(EclipseTracker.getTracker()));
  //    		end.isPublished();
  //    		end.getProperties();
  //    	     RLogger.logInfo("Initiated on "+port);
  //    	}catch(Exception be){
  //    		be.printStackTrace();
  //    		if(port<(start+4)){
  //    			startRPC(port+1,start);
  //    		}
  //    		else
  //    			be.printStackTrace();
  //    	}
  //	}

  public boolean isStarted() {
    return loaded;
  }

  private void initCVSClients() throws Exception {
    IConfigurationElement[] extns =
        Platform.getExtensionRegistry()
            .getConfigurationElementsFor("com.unitesk.requality.cvsprovider");
    for (IConfigurationElement e : extns) {
      try {
        Object o = e.createExecutableExtension("cvcloveransvc");
        if ((o != null) && (o instanceof CVSSourcePlugin)) {
          // список поддерживаемых TeamProvider (список строк, возвращаемых getID)
          List<String> tpList = ((CVSSourcePlugin) o).getSupportedTeamProvidersIDList();
          // побежим по списку поддерживаемых TeamProvider'ов
          Iterator<String> lIt = tpList.iterator();
          while (lIt.hasNext()) {
            String teamProviderID = lIt.next();
            // если этот провайдер уже кем-то поддерживается
            if (supportMap.containsKey(teamProviderID)) {
              // добавим
              List<ICVSClient> supp = supportMap.get(teamProviderID);
              supp.add(((CVSSourcePlugin) o).getInstanceOfSVCClient());
            } else {
              //... иначе начнём новый список поддрживающих
              List<ICVSClient> value = new ArrayList<>();
              value.add(((CVSSourcePlugin) o).getInstanceOfSVCClient());
              supportMap.put(teamProviderID, value);
            }
          }
        }
      } catch (Exception ez) {

      }
    }
  }

  public static Activator getDefault() {
    return plugin;
  }
  // //////////////////////////////////////////////////////////////////////////
  //
  // Active node tracking
  //
  // //////////////////////////////////////////////////////////////////////////

  private void updateDeleteInMenu() {
    IWorkbenchWindow window = PlatformUI.getWorkbench().getActiveWorkbenchWindow();
    if (window != null) {
      ICommandService commandService = window.getService(ICommandService.class);
      if (commandService != null) {
        commandService.refreshElements("org.eclipse.ui.edit.delete", null);
      }
    }
  }

  private void loadReferenceAcronyms() {
    Map<String, String> referenceAcronyms = new HashMap<>();
    IExtensionRegistry reg = Platform.getExtensionRegistry();
    IConfigurationElement[] configurationElements =
        reg.getConfigurationElementsFor(EXTENSION_NAME_REFERENCE_ACR);
    for (IConfigurationElement conf : configurationElements)
      referenceAcronyms.put(conf.getAttribute("Reference_name"), conf.getAttribute("Name_Acronym"));
    ProjectReferenceProperties.setGlobalRenames(referenceAcronyms);
  }

  private Set<NewElementHandler> newehandlers = null;

  public Set<NewElementHandler> getNewElementHandlers() {
    if (newehandlers == null) {
      newehandlers = new HashSet<>();
      IExtensionRegistry reg = Platform.getExtensionRegistry();
      IConfigurationElement[] configurationElements =
          reg.getConfigurationElementsFor(EXTENSION_NAME_HANDLERS);
      for (IConfigurationElement conf : configurationElements) {
        try {
          IHandler ih = (IHandler) conf.createExecutableExtension("class");
          if (ih instanceof NewElementHandler)
            newehandlers.add((NewElementHandler) conf.createExecutableExtension("class"));
        } catch (CoreException e) {
          // TODO Auto-generated catch block
          //	e.printStackTrace();
        }
      }
    }
    return newehandlers;
  }

  // //////////////////////////////////////////////////////////////////////////
  //
  // Report operations
  //
  // //////////////////////////////////////////////////////////////////////////

  private Map<String, Template> reportTemplates = new HashMap<>();
  /** Get report templates, includes template reloading */
  public Map<String, Template> getRenewedReportTemplates() {
    reportTemplates = new HashMap<>();
    return getReportTemplates();
  }

  public SortedMap<String, IDeleteHandler> getDeleteHandlers() {
    if (deleteHandlers == null) loadDeleteHandlers();
    return deleteHandlers;
  }

  public Map<String, Template> getReportTemplates() {
    loadReportTemplatesDescriptors();
    return Collections.unmodifiableMap(reportTemplates);
  }

  private SortedMap<String, IDeleteHandler> deleteHandlers;

  private void loadDeleteHandlers() {
    deleteHandlers = new TreeMap<>();
    IExtensionRegistry reg = Platform.getExtensionRegistry();
    IConfigurationElement[] configurationElements =
        reg.getConfigurationElementsFor(EXTENSION_NAME_HANDLERS);
    for (IConfigurationElement conf : configurationElements) {
      try {
        try {
          if (!conf.isValid()
              || conf.getAttribute("commandId") == null
              || !conf.getAttribute("commandId").equals("org.eclipse.ui.edit.delete")) continue;
          AbstractHandler ah = (AbstractHandler) conf.createExecutableExtension("class");
          if (ah != null && ah instanceof IDeleteHandler)
            deleteHandlers.put(conf.getAttribute("class"), (IDeleteHandler) ah);
        } catch (ClassCastException cce) {
          //so, this is bad handlers(
        }
      } catch (Exception e) {
        e.printStackTrace();
      }
    }
  }

  public IHandler[] getCommandHandlers(String command) {
    List<IHandler> handlers = new ArrayList<>();
    IExtensionRegistry reg = Platform.getExtensionRegistry();
    IConfigurationElement[] configurationElements =
        reg.getConfigurationElementsFor("org.eclipse.ui.handlers");
    for (IConfigurationElement conf : configurationElements) {
      if (conf.getAttribute("commandId").equals(command))
        try {
          handlers.add((IHandler) conf.createExecutableExtension("class"));
        } catch (CoreException e) {
          // TODO Auto-generated catch block
          e.printStackTrace();
        }
    }
    ICommandService cmdService =
        PlatformUI.getWorkbench().getActiveWorkbenchWindow().getService(ICommandService.class);
    Command comm = cmdService.getCommand(command);
    if (!comm.isDefined()) {
      return null;
    }
    return handlers.toArray(new IHandler[0]);
  }

  private void loadReportTemplatesDescriptors() {
    List<String> tplPaths = new ArrayList<>();
    try {
      tplPaths.add(Activator.getPath("templates"));
    } catch (IOException e) {
      e.printStackTrace();
    }
    IExtensionRegistry reg = Platform.getExtensionRegistry();
    IConfigurationElement[] configurationElements =
        reg.getConfigurationElementsFor(EXTENSION_NAME_REPORT);
    for (IConfigurationElement conf : configurationElements) {
      try {
        if (!conf.getName().equals("templatePath")) {
          continue;
        }
        ITemplatePathSource tplPath =
            (ITemplatePathSource) conf.createExecutableExtension("provider");
        String path = tplPath.getPath();
        if (path != null && new File(path).isDirectory()) {
          tplPaths.add(path);
        }
      } catch (Exception e) {
        e.printStackTrace();
      }
    }
    for (String path : tplPaths) {
      File tplPath = new File(path);
      for (String tplFolder : tplPath.list()) {
        String tpli = tplPath.toURI().toString();
        if (new File(URI.create(tpli + "/" + tplFolder + "/" + "template.xml")).exists()) {
          Template template = new Template(tpli + "/" + tplFolder);
          reportTemplates.put(template.getId(), template);
        }
      }
    }
    for (IConfigurationElement conf : configurationElements) {
      if (!conf.getName().equals("dataProvider")) {
        continue;
      }
      String tId = conf.getAttribute("templateId");
      String ftlName = conf.getAttribute("ftlName");
      if (reportTemplates.containsKey(tId)) {
        Template tpl = reportTemplates.get(tId);
        for (TemplateFile t : tpl.getTemplates()) {
          if (t.getTemplatePath().equals(ftlName)) {
            IReportDataSource dataObject;
            try {
              dataObject = (IReportDataSource) conf.createExecutableExtension("dataSource");
              t.setDataObject(dataObject);
            } catch (CoreException e) {
              e.printStackTrace();
            }
          }
        }
      }
    }
  }

  // ///////////////////////////////////////////////////////////////////////////
  //
  // Test generators operations
  //
  // ///////////////////////////////////////////////////////////////////////////

  private SortedSet<TestGeneratorDescriptor> generators = new TreeSet<>();

  public SortedSet<TestGeneratorDescriptor> getTestGenerators() {
    loadTestGeneratorsDescriptors();
    return Collections.unmodifiableSortedSet(generators);
  }

  public void loadTestGeneratorsDescriptors() {
    IExtensionRegistry reg = Platform.getExtensionRegistry();
    IConfigurationElement[] configurationElements =
        reg.getConfigurationElementsFor(EXTENSION_NAME_TEST_GENERATORS);
    for (IConfigurationElement conf : configurationElements) {
      try {
        ITestGenerator docProc = (ITestGenerator) conf.createExecutableExtension("class");
        if (docProc != null)
          generators.add(
              new TestGeneratorDescriptor(
                  conf.getAttribute("id"), conf.getAttribute("name"), docProc));
      } catch (Exception e) {
        e.printStackTrace();
      }
    }
    configurationElements =
        reg.getConfigurationElementsFor(EXTENSION_NAME_TEST_GENERATORS_PROVIDER);
    for (IConfigurationElement conf : configurationElements) {
      try {
        ITestGeneratorsProvider docProc =
            (ITestGeneratorsProvider) conf.createExecutableExtension("class");
        if (docProc != null) {
          TestGeneratorDescriptor[] desc = docProc.getDescriptors();
          if (desc != null) generators.addAll(Arrays.asList(desc));
        }
      } catch (Exception e) {
        e.printStackTrace();
      }
    }
  }

  // ///////////////////////////////////////////////////////////////////////////
  //
  // Processors operations
  //
  // ///////////////////////////////////////////////////////////////////////////

  private SortedSet<DocumentProcessorDescriptor> processors =
      new TreeSet<>();

  public SortedSet<DocumentProcessorDescriptor> getProcessors() {
    loadDocumentProcessorDescriptors();
    return Collections.unmodifiableSortedSet(processors);
  }

  public void loadDocumentProcessorDescriptors() {
    IExtensionRegistry reg = Platform.getExtensionRegistry();
    IConfigurationElement[] configurationElements =
        reg.getConfigurationElementsFor(EXTENSION_NAME_DOC_PROCESSOR);
    for (IConfigurationElement conf : configurationElements) {
      try {
        IDocumentProcessor docProc = (IDocumentProcessor) conf.createExecutableExtension("class");
        DocWizard wiz = null;
        try {
          wiz = (DocWizard) conf.createExecutableExtension("wizard");
        } catch (Exception e) {
          //TODO has no wizards
        }
        processors.add(
            new DocumentProcessorDescriptor(
                conf.getAttribute("id"), conf.getAttribute("name"), docProc, wiz));

      } catch (Exception e) {
        e.printStackTrace();
      }
    }
  }

  // //////////////////////////////////////////////////////////////////////////
  //
  // Resources
  //
  // //////////////////////////////////////////////////////////////////////////
  public static String getPath(String string) throws IOException {
    java.net.URL res =
        FileLocator.toFileURL(Activator.getDefault().getBundle().getResource(string));
    if (res == null) return null;
    return res.getFile();
  }

  public String getJSToolsPath() {
    String path = null;
    try {
      path = FileLocator.toFileURL(this.getBundle().getEntry("jstools")).getFile();
    } catch (IOException e) {
      e.printStackTrace();
    }
    return path;
  }

  /**
   * Returns an image descriptor for the image file at the given plug-in relative path
   *
   * @param path the path
   * @return the image descriptor
   */
  public static ImageDescriptor getImageDescriptor(String pluginId, String path) {
    return imageDescriptorFromPlugin(pluginId, path);
  }

  public static ImageDescriptor getImageDescriptor(String path) {
    return imageDescriptorFromPlugin(PLUGIN_ID, path);
  }

  @Override
  protected ImageRegistry createImageRegistry() {
    ImageRegistry imgreg = super.createImageRegistry();
    Map<String, NodeTypeDescriptor> types = RequalityPlugin.getAdditionalTypes();
    for (String n : types.keySet()) {
      NodeTypeDescriptor desc = RequalityPlugin.getAdditionalTypes().get(n);
      if (desc.getIcon() != null) {
        imgreg.put(
            "nodetype." + n,
            getImageDescriptor(
                desc.getPluginId(), RequalityPlugin.getAdditionalTypes().get(n).getIcon()));
      } else {
        imgreg.put("nodetype." + n, getImageDescriptor(desc.getPluginId(), "icons/unitesk.png"));
      }
    }
    java.io.File f = null;
    try {
      f =
          new File(
              FileLocator.toFileURL(Platform.getBundle(PLUGIN_ID).getEntry("icons/overlays"))
                  .toURI());
    } catch (URISyntaxException e1) {
      // TODO Auto-generated catch block
      e1.printStackTrace();
    } catch (IOException e1) {
      // TODO Auto-generated catch block
      e1.printStackTrace();
    }
    if (f != null && f.exists()) {
      File[] files = f.listFiles();
      for (File file : files) {
        try {
          imgreg.put(
              "icons/overlays/" + file.getName(),
              getImageDescriptor("icons/overlays/" + file.getName()));
        } catch (Exception e) {
          imgreg.dispose();
        }
      }
    }
    imgreg.put("nodetype.NodeType", getImageDescriptor("icons/nodetypes/NodeType.png"));
    imgreg.put("nodetype.default", getImageDescriptor("icons/unitesk.png"));
    imgreg.put("arrowup", getImageDescriptor("icons/uparrow.png"));
    imgreg.put("arrowdown", getImageDescriptor("icons/downarrow.png"));

    imgreg.put("green_dot", getImageDescriptor("icons/green_dot.png"));
    imgreg.put("red_dot", getImageDescriptor("icons/red_dot.png"));
    imgreg.put("yellow_dot", getImageDescriptor("icons/yellow_dot.png"));
    return imgreg;
  }

  // //////////////////////////////////////////////////////////////////////////
  //
  // Project tracking
  //
  // //////////////////////////////////////////////////////////////////////////

  public static EclipseTracker tracker = EclipseTracker.getTracker();

  private IResourceChangeListener resourceChangeListener =
      new IResourceChangeListener() {
        @Override
        public void resourceChanged(IResourceChangeEvent event) {
          //TODO may be replaced by sync?
//          Display.getDefault()
//              .asyncExec(
//                  new Runnable() {
//
//                    @Override
//                    public void run() {
                      if (event.getType() == IResourceChangeEvent.PRE_CLOSE) {
                        tracker.stopTracking((IProject) event.getResource(), false);
                        return;
                      }
                      if (event.getType() == IResourceChangeEvent.PRE_DELETE) {
                        tracker.stopTracking((IProject) event.getResource(), true);
                        return;
                      }
                      if (event.getType() == IResourceChangeEvent.POST_CHANGE) {
                        // Track new
                        IProject newProjects[] = getNewProjects();
                        if(newProjects.length>0)
                          Display.getDefault().asyncExec(new Runnable(){
                          
                            @Override
                            public void run() {
                              // TODO Auto-generated method stub
                              for (IProject proj : newProjects) {
                                tracker.trackProject(proj);
                              }
                            }
                          });
                      }
                      if (event.getType() == IResourceChangeEvent.PRE_BUILD) {
                        // Track new
                        IProject newProjects[] = getNewProjects();
                        
                        for (IProject proj : newProjects) {
                          try {
                            proj.refreshLocal(IProject.DEPTH_INFINITE, null);
                          } catch (CoreException e) {
                            // TODO Auto-generated catch block
                            e.printStackTrace();
                          }
                        }
                      }
                      // TODO Auto-generated method stub

                    }
//                  });
//        }
      };

  private HashMap<String, IAttributeGenerator> attributeGeneratorsMap;

  private static IProject[] getNewProjects() {
    EclipseTracker tracker = EclipseTracker.getTracker();
    Set<IProject> openProjects = new HashSet<>();

    for (TreeDB tree : tracker.getTrees())
      if (tracker.getProject(tree).isOpen()) {
        openProjects.add(tracker.getProject(tree));
      }
    Set<IProject> curProjects = new HashSet<>();
    for (IProject proj : ResourcesPlugin.getWorkspace().getRoot().getProjects()) {
      if (isInterestingProject(proj)) {
        curProjects.add(proj);
      }
    }

    Set<IProject> result = new HashSet<>(curProjects);
    result.removeAll(openProjects);
    return result.toArray(new IProject[0]);
  }

  private static boolean isInterestingProject(IProject proj) {
    try {
      if (!proj.isOpen()) {
        return false;
      }
      if (!proj.isNatureEnabled(ReqCatalogue.ID)) {
        return false;
      }
    } catch (CoreException e) {
      e.printStackTrace();
      return false;
    }
    return true;
  }

  private static void initTracker() throws CoreException {
    LinkedList<IProject> reqProjects = new LinkedList<>();
    for (IProject changedProject : ResourcesPlugin.getWorkspace().getRoot().getProjects()) {
      if (isInterestingProject(changedProject)) {
        reqProjects.add(changedProject);
      }
    }
    for (IProject changedProject : reqProjects) {
      if (!EclipseTracker.getTracker().isTracking(changedProject)) {
        EclipseTracker.getTracker().trackProject(changedProject);
      }
    }
  }
  //Method to get URL from file name. Created
  //here to be modified in future by using
  //bundles instead of file
  public static URI getUrl(String string) throws IOException {
    File f = new File(string);
    return getUrl(f.toURI());
  }

  public static URI getUrl(URI uri) throws IOException {
    try {
      return FileLocator.toFileURL(uri.toURL()).toURI();
    } catch (URISyntaxException e) {
      // TODO Auto-generated catch block
      e.printStackTrace();
    }
    return null;
  }

  private static Map<String, CoverageStorageSourceDescriptor> coverageInputMethods;

  private static Map<String, String> coverageIdToName;

  public static Map<String, CoverageStorageSourceDescriptor> getCoverageSources() {
    if (coverageInputMethods == null) loadCoverageInputMethods();
    return coverageInputMethods;
  }

  public static ICoverageStorageSource getCoverageSource(String type) {
    if (coverageInputMethods == null) loadCoverageInputMethods();
    if (coverageInputMethods.get(type) == null) return null;
    return coverageInputMethods.get(type).getStorageInput();
  }

  private static void loadCoverageInputMethods() {
    coverageInputMethods = new HashMap<>();
    coverageIdToName = new HashMap<>();
    IExtensionRegistry reg = Platform.getExtensionRegistry();
    IConfigurationElement[] configurationElements =
        reg.getConfigurationElementsFor(EXTENSION_NAME_COVERAGE_SOURCES);
    for (IConfigurationElement conf : configurationElements) {
      try {
        ICoverageStorageSource st =
            (ICoverageStorageSource) conf.createExecutableExtension("class");
        String id = conf.getAttribute("id");
        coverageInputMethods.put(
            id, new CoverageStorageSourceDescriptor(id, conf.getAttribute("name"), st));
        coverageIdToName.put(id, conf.getAttribute("name"));

      } catch (Exception e) {
        e.printStackTrace();
      }
    }
  }

  public Map<String, String> getCoverageSourceNames() {
    if (coverageIdToName == null) loadCoverageInputMethods();
    return coverageIdToName;
  }

  public static IProject getProjectByName(String split) {
    IWorkspace ws = ResourcesPlugin.getWorkspace();
    IWorkspaceRoot wr = ws.getRoot();
    IProject p = wr.getProject(split);
    return p;
  }

  public Map<String, IAttributeGenerator> loadAttributeGeneratorsDescriptions() {
    if (attributeGeneratorsMap == null) {
      attributeGeneratorsMap = new HashMap<>();
      IExtensionRegistry reg = Platform.getExtensionRegistry();
      IConfigurationElement[] configurationElements =
          reg.getConfigurationElementsFor(EXTENSION_NAME_ATTRIBUTE_GENERATORS);
      for (IConfigurationElement conf : configurationElements) {
        try {
          IAttributeGenerator st = (IAttributeGenerator) conf.createExecutableExtension("class");
          String id = conf.getAttribute("id");
          attributeGeneratorsMap.put(id, st);

        } catch (Exception e) {
          e.printStackTrace();
        }
      }
    }
    return attributeGeneratorsMap;
  }

  @SuppressWarnings("deprecation")
  public static void executeCommand(IHandler h, Map<?, ?> params) throws ExecutionException {
    IHandlerService hndlrService =
        PlatformUI.getWorkbench().getActiveWorkbenchWindow().getService(IHandlerService.class);
    h.execute(new ExecutionEvent(params, null, hndlrService.getCurrentState()));
  }

  private int maxFirstSubstring(String first, String second) {
    if (first == null || second == null || first.length() == 0 || second.length() == 0) {
      return 0;
    }
    int fl = first.length();
    int sl = second.length();
    int s = 0;
    for (s = 0; s < Math.min(sl, fl) && first.charAt(s) == second.charAt(s); s++) ;
    return s;
  }

  public String getProjectRelatedPath(String fname, IProject proj) {
    //TODO test
    String projPath = proj.getLocationURI().toString();
    int s = maxFirstSubstring(fname, projPath);
    if (s < 2) return fname;
    int pos = projPath.indexOf("/", s), upcount = projPath.length() > s ? 1 : 0;
    while (pos > -1) {
      upcount++;
      pos = projPath.indexOf("/", pos + 1);
    }
    StringBuilder sb = new StringBuilder("./");
    for (int i = 0; i < upcount; i++) sb.append("../");
    String end = fname.substring(s);
    if (sb.charAt(sb.length() - 1) == '/' && end.length() > 0 && end.charAt(0) == '/')
      end = end.substring(1);
    sb.append(end);
    return sb.toString();
  }

  public ICVSClient getSVCClientForTeamProvider(String id) {
    if (!supportMap.containsKey(id) || supportMap.get(id).size() == 0) return null;
    return supportMap.get(id).get(0);
  }

  public CVSManager getCVSManager() {
    if (cvclManager == null) cvclManager = new CVSManager();
    return cvclManager;
  }
}
