/*
 * Copyright 2009-2011 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.unitesk.requality.eclipse.core;

import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.PrintWriter;
import java.net.URI;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.UUID;
import java.util.concurrent.CopyOnWriteArrayList;
import org.eclipse.core.commands.operations.IOperationHistory;
import org.eclipse.core.commands.operations.IUndoContext;
import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IProject;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.resources.IWorkspace;
import org.eclipse.core.resources.IncrementalProjectBuilder;
import org.eclipse.core.resources.ResourcesPlugin;
import org.eclipse.core.resources.WorkspaceJob;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Path;
import org.eclipse.core.runtime.Status;
import org.eclipse.core.runtime.jobs.Job;
import org.eclipse.jface.dialogs.DialogSettings;
import org.eclipse.jface.dialogs.IDialogConstants;
import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.ui.PlatformUI;
import com.unitesk.requality.core.DefaultTreeChangeListener;
import com.unitesk.requality.core.IResTracker;
import com.unitesk.requality.core.ITreeStorage;
import com.unitesk.requality.core.RLogger;
import com.unitesk.requality.core.TreeDB;
import com.unitesk.requality.core.TreeNode;
import com.unitesk.requality.core.descriptors.NodeTypeDescriptor;
import com.unitesk.requality.core.instancer.Instantiator;
import com.unitesk.requality.core.instancer.TreeInstancer;
import com.unitesk.requality.core.transaction.ITransactionListener;
import com.unitesk.requality.core.transaction.TransactionStorage;
import com.unitesk.requality.eclipse.Activator;
import com.unitesk.requality.eclipse.RequalityPlugin;
import com.unitesk.requality.eclipse.natures.ReqCatalogue;
import com.unitesk.requality.eclipse.operations.TreeDBOperation;
import com.unitesk.requality.eclipse.storageformatversion.StoredFormatVersioning;
import com.unitesk.requality.eclipse.storageformatversion.VersionUpdateInfo;
import com.unitesk.requality.eclipse.ui.dialogs.BigMessageDialog;
import com.unitesk.requality.eclipse.views.properties.ProjectReferenceProperties;
import com.unitesk.requality.nodetypes.Document;
import com.unitesk.requality.nodetypes.Location;
import com.unitesk.requality.report.Template;
import com.unitesk.requality.report.TraceabilityDataSource;

public class EclipseTracker implements IResTracker {
  public enum TreeEvent {
    TREE_OPEN,
    TREE_CLOSE,
    TREE_DELETE
  }

  private Map<IProject, TreeDB> treeByProject = new HashMap<>();
  private Map<TreeDB, IProject> projectByTree = new HashMap<>();
  private Map<IProject, ResourceStorage> projectToStorage =
      new HashMap<>();
  private Map<TreeDB, TreeDBBackupRepeatingJob> smap =
      new HashMap<>();

  private static EclipseTracker tracker;
  private List<ITreeStateListener> listeners = new CopyOnWriteArrayList<>();

  public boolean addListener(ITreeStateListener l) {
    return listeners.add(l);
  }

  public boolean removeListener(ITreeStateListener l) {
    return listeners.remove(l);
  }

  private EclipseTracker() {}

  public static EclipseTracker getTracker() {
    if (tracker == null) tracker = new EclipseTracker();
    return tracker;
  }

  public synchronized void notifyListeners(TreeDB db, TreeEvent type) {
    for (ITreeStateListener listener : listeners) {
      switch (type) {
        case TREE_CLOSE:
          {
            TreeDBBackupRepeatingJob tbr = smap.get(db);
            if (tbr != null) tbr.singleAndStop(db);
            listener.close(db);
          }
          break;
        case TREE_OPEN:
          {
            TreeDBBackupRepeatingJob tbr = smap.get(db);
            if (tbr != null) tbr.start();
            listener.open(db);
          }
          break;
        case TREE_DELETE:
          {
            //	            	TreeDBBackupRepeatingJob tbr = smap.get(db);
            //	            	if(tbr!=null)
            //	            		tbr.singleAndStop();
            //db.clearAllCaches(false);
            listener.delete(projectByTree.get(db));
            TreeDBBackupRepeatingJob tbr = smap.remove(db);
            if(tbr!=null){
              try{
                tbr.halt();
              }catch(Exception e){
                e.printStackTrace();
              }
            }
          }
          break;
      }
    }
  }

  public IProject getProject(TreeDB db) {
    if (db == null) return null;
    IProject proj = projectByTree.get(db);
    if (proj == null && db.isVirtual()) {
      TreeDB root = Instantiator.getRootTree(db);
      if (root != db) return getProject(root);
    }
    return proj;
  }

  public TreeDB getTree(IProject proj) {
    return treeByProject.get(proj);
  }

  private void showError(final String title, final String message) {

    new Thread(
            new Runnable() {
              @Override
              public void run() {
                while (!PlatformUI.isWorkbenchRunning())
                  try {
                    this.wait(200);
                  } catch (Exception e) {
                  }
                PlatformUI.getWorkbench()
                    .getDisplay()
                    .asyncExec(
                        new Runnable() {
                          public void run() {
                            Shell sh =
                                PlatformUI.getWorkbench().getActiveWorkbenchWindow().getShell();
                            MessageDialog.openError(sh, title, message);
                          }
                        });
              }
            })
        .run();
  }

  private void showBigMessage(String string, String string2, String string3) {

    new Thread(
            new Runnable() {
              @Override
              public void run() {
                while (!PlatformUI.isWorkbenchRunning())
                  try {
                    this.wait(200);
                  } catch (Exception e) {
                  }
                PlatformUI.getWorkbench()
                    .getDisplay()
                    .asyncExec(
                        new Runnable() {
                          public void run() {
                            BigMessageDialog.openErr(
                                Display.getDefault().getActiveShell(),
                                string,
                                string2,
                                string3,
                                "",
                                new String[] {IDialogConstants.OK_LABEL});
                          }
                        });
              }
            })
        .run();
  }

  private void showMessage(final String title, final String message) {

    new Thread(
            new Runnable() {
              @Override
              public void run() {
                while (!PlatformUI.isWorkbenchRunning())
                  try {
                    this.wait(200);
                  } catch (Exception e) {
                  }
                PlatformUI.getWorkbench()
                    .getDisplay()
                    .asyncExec(
                        new Runnable() {
                          public void run() {
                            Shell sh =
                                PlatformUI.getWorkbench().getActiveWorkbenchWindow().getShell();
                            MessageDialog.openInformation(sh, title, message);
                          }
                        });
              }
            })
        .run();
  }

  private void showWarning(final String title, final String message) {

    new Thread(
            new Runnable() {
              @Override
              public void run() {
                while (!PlatformUI.isWorkbenchRunning())
                  try {
                    this.wait(200);
                  } catch (Exception e) {
                  }
                PlatformUI.getWorkbench()
                    .getDisplay()
                    .asyncExec(
                        new Runnable() {
                          public void run() {
                            Shell sh =
                                PlatformUI.getWorkbench().getActiveWorkbenchWindow().getShell();
                            MessageDialog.openWarning(sh, title, message);
                          }
                        });
              }
            })
        .run();
  }

  private void closeProject(final IProject project) {
    while (project.getWorkspace().isTreeLocked())
      try {
        this.wait(200);
      } catch (Exception e) {
      }
    Runnable r =
        new Runnable() {
          @Override
          public void run() {
            // TODO Auto-generated method stub
            try {
              project.close(null);
            } catch (CoreException e) {
              e.printStackTrace();
            }
          }
        };
    Display.getDefault().syncExec(r);
    //    	Job job = new Job("Close project Job") {
    //    		  @Override
    //    		  protected IStatus run(IProgressMonitor monitor) {
    //
    //  					PlatformUI.getWorkbench().getDisplay().asyncExec(new Runnable() {
    //  			    	    public void run() {
    //
    //  			    	    }
    //  					});
    //    		    return Status.OK_STATUS;
    //    		  }
    //    		};
    //    		job.setPriority(Job.DECORATE);
    //    		job.schedule(200);
  }

  private List<IProject> locked = new ArrayList<>();

  public boolean trackProject(final IProject project) {
    return trackProject(project, false);
  }

  public boolean trackProject(final IProject project, boolean ignoreversion) {
    try {
      try{
        project.refreshLocal(IResource.DEPTH_INFINITE, null);
      }catch(Exception e){
        e.printStackTrace();
      }
      if (!ignoreversion && locked.contains(project)) return false;
      if (!project.isOpen() || !project.isNatureEnabled(ReqCatalogue.ID)) {
        return false;
      } else {
        TreeDB db = getTree(project);
        if (db == null) {
          if (!locked.contains(project)) {
            List<VersionUpdateInfo> vilist = StoredFormatVersioning.collectUpdates(project);
            StringBuilder updates = new StringBuilder("");
            StringBuilder plugs = new StringBuilder("");
            String problems = "", p2 = null;
            int pl = 0;
            for (VersionUpdateInfo vi : vilist) {
              if (vi.getConverter() == null) {
                pl++;
                if (vi.getStatus().equals("plugin_not_found")) {
                  plugs.append("\n");
                  plugs.append(vi.getGroupId());
                } else p2 = vi.getNewVersion();
                problems = vi.getStatus();
              } else {
                updates.append("\n");
                updates.append(vi.toString());
              }
            }
            if (!ignoreversion && pl > 0) {
              locked.add(project);
              if (problems.equals("version_not_found")) {
                showError(
                    "Project Loading",
                    "Format of the project '"
                        + project.getName()
                        + "' requires newer version of Requality ("
                        + p2
                        + ").");
              } else if (problems.equals("plugin_not_found")) {
                pl--;
                plugs.append("\n");
                showWarning(
                    "Project Loading: '" + project.getName() + "'",
                    "Current set of plugins cannot handle the following elements: "
                        + plugs.toString()
                        + "Install additional plugins.");
              }
            }
            if (!ignoreversion && pl > 0) {
              closeProject(project);
              locked.remove(project);
              return false;
            } else {
              if (!ignoreversion && updates.length() > 0) {
                locked.add(project);
                try {
                  closeProject(project);
                } catch (Exception e) {
                  e.printStackTrace();
                }
                return showQuestion(
                    "Project Loading",
                    "Format of the project '"
                        + project.getName()
                        + "' is out of date, and installed Requality version can't process it. Do you want to update the project to proper format?",
                    new Runnable() {
                      @Override
                      public void run() {
                        try {
                          project.open(null);
                          List<VersionUpdateInfo> res =
                              StoredFormatVersioning.updateProject(project);
                          int rs = res.size();
                          StringBuffer err = new StringBuffer(), err2 = new StringBuffer();
                          if (res.size() > 0) {
                            StringBuffer sb = new StringBuffer(), st = new StringBuffer();
                            String problems = ""; //, pr2 = "";
                            int state = 0; //norm
                            for (VersionUpdateInfo vui : res) {
                              if (st.length() > 0) st.append("\n");
                              st.append(vui.getStatus());
                              if (vui.getConverter() == null) {
                                problems = vui.getStatus();
                                if(problems!=null){
                                  if (problems.equals("plugin_not_found")) {
                                    if (err.length() != 0) err.append("\n");
                                    err.append("Plugin '");
                                    err.append(vui.getGroupId());
                                    err.append("' is not found.");
                                    err.append("\n");
                                    err.append(vui.getGroupId());
                                    state = state | 1; //nf
                                    rs--;
                                  } else {
                                    String pr2 = vui.getNewVersion();
                                    if (problems.equals("version_not_found")) {
                                      if (err.length() != 0) err.append("\n");
                                      err.append("Project '");
                                      err.append(project.getName());
                                      err.append("' requires version ");
                                      err.append(pr2);
                                      err.append(" of plugin ");
                                      err.append(vui.getGroupId());
                                      state = state | 2; //nf vers
  
                                    } else if (!problems.isEmpty()) {
                                      if (err.length() != 0) err.append("\n");
                                      err.append("Conversion failed for '");
                                      err.append(project.getName());
                                      err.append("'");
                                      err2.append(problems);
  
                                      state = state | 4;
                                    }
                                  }
                                }
                              }
                            }
                            if (err.length() > 0)
                              showBigMessage("Project Loading", err.toString(), err2.toString());
                            else
                              showMessage(
                                  "Project Loading",
                                  "Conversion of '"
                                      + project.getName()
                                      + "' to the new format completed.\nINFO:\n"
                                      + st.toString());
                          }

                          locked.remove(project);
                          if (err.length() == 0) {
                            project.open(null);
                            trackProject(project);
                          }
                        } catch (Exception e) {
                          ByteArrayOutputStream baos = new ByteArrayOutputStream();
                          PrintWriter pw = new PrintWriter(baos);
                          e.printStackTrace(pw);
                          pw.close();
                          try {
                            baos.close();
                          } catch (IOException e1) {
                            // TODO Auto-generated catch block
                            e1.printStackTrace();
                          }
                          throw new RuntimeException(baos.toString());
                        }
                      }
                    },
                    new Runnable() {
                      public void run() {
                        try {
                          project.close(null);
                        } catch (CoreException e) {
                          e.printStackTrace();
                        }
                        locked.remove(project);
                      }
                    });
              }
            }
          }
          if (!project.isOpen()) {
            RLogger.logInfo("project is not opened");
            return false;
          }
          db = loadTreeDB(project,null,null);//TODO convert?
          bindBaseToProject(project, db);
          initClasses(db);
          db.addListener(
              new DefaultTreeChangeListener() {
                @Override
                public void created(TreeNode node) {
                  if (node.getGroup() != TreeNode.NODE_GROUP) {
                    IProject proj = EclipseTracker.getTracker().getProject(node.getTreeDB());
                    List<String> vers = StoredFormatVersioning.getFormatsVersions(proj);
                    String v = StoredFormatVersioning.getFullVersionByGroupId(node.getGroup());
                    if (!vers.contains(v)) {
                      vers.add(v);
                      StoredFormatVersioning.setFormatsVersions(proj, vers);
                    }
                  }
                }
              });
          db = getTree(project);
          try {
            Instantiator.removeTempInstances(db);
            Instantiator.loadDBs(db);
            db.getLinkResolver().initCaches();
            db.getTermsStorage().initCaches();
          } catch (Exception e) {
            e.printStackTrace();
          }
          ResourcesPlugin.getWorkspace()
              .addResourceChangeListener(
                  ((ResourceStorage) db.getStorage()).getResourceChangeListener());
          WorkspaceJob runBuilder =
              new WorkspaceJob("Check Requality Project: " + project.getName()) {
                @Override
                public IStatus runInWorkspace(IProgressMonitor monitor) throws CoreException {
                  try {
                    project.build(IncrementalProjectBuilder.FULL_BUILD, monitor);
                  } catch (CoreException e) {
                    e.printStackTrace();
                  }
                  return Status.OK_STATUS;
                }
              };
          runBuilder.setRule(ResourcesPlugin.getWorkspace().getRuleFactory().buildRule());
          runBuilder.schedule();
          TraceabilityDataSource.setRevertNames(
              db, ProjectReferenceProperties.getNamesAcronyms(db));
        }
        Location.clearCache(db);
        //db.clearAllCaches(true);
        //                	if(db.isLoaded()){
        //                		TreeDB vdb = Instantiator.getExistVirtualTreeDB(db,Instantiator.DEFAULT_INSTANCE_NAME,null,null);
        //                		if(vdb!=null)
        //                			vdb.getRootNode();;
        //                	}
        TreeDBBackupRepeatingJob savier = new TreeDBBackupRepeatingJob(db);
        savier.setPriority(Job.LONG);
        savier.schedule(TreeDBBackupRepeatingJob.DELAY);
        addListener(savier);
        smap.put(db, savier);
        notifyListeners(db, TreeEvent.TREE_OPEN);
      }
    } catch (Exception e) {
      ByteArrayOutputStream baos = new ByteArrayOutputStream();
      PrintWriter pw = new PrintWriter(baos);
      e.printStackTrace(pw);
      pw.close();
      try {
        baos.close();
      } catch (IOException e1) {
        // TODO Auto-generated catch block
        e1.printStackTrace();
      }
      throw new RuntimeException(baos.toString());
      //return false;
    }
    return true;
  }

  @Override
  public void initClasses(TreeDB db) {
    Map<String, NodeTypeDescriptor> nts = RequalityPlugin.getAdditionalTypes();
    for (String name : RequalityPlugin.getAdditionalTypes().keySet())
      db.addNodeType(name, nts.get(name).getClazz());
  }

  private boolean showQuestion(
      final String title, final String message, final Runnable runnable, final Runnable runnable2) {
    final Set<Boolean> resa = new HashSet<>();
    Display.getDefault()
        .syncExec(
            new Runnable() {
              public void run() {
                Shell sh = Display.getDefault().getActiveShell();
                if (MessageDialog.openConfirm(sh, title, message)) {
                  RLogger.logInfo("oksel");
                  ;
                  resa.add(true);
                  if (runnable != null) Display.getDefault().syncExec(runnable);
                } else {
                  RLogger.logInfo("cansel");
                  if (runnable2 != null) Display.getDefault().syncExec(runnable2);
                }
              }
            });
    return !resa.isEmpty();
  }

  private TreeDB loadTreeDB(final IProject project,TreeDB bdb,ResourceStorage rs) {
    if (!projectToStorage.containsKey(project)) {
      //        	try {
      //				project.refreshLocal(IResource.DEPTH_INFINITE, null);
      //			} catch (CoreException e) {
      //				e.printStackTrace();
      //			}
      if(rs == null)
        rs = new ResourceStorage(project);
      projectToStorage.put(project, rs);
    }
    ResourceStorage storage = projectToStorage.get(project);
    TreeDB db = bdb;
    if(bdb == null)
      db = TreeDB.load(storage);
    db.addTransactionListener(
        new ITransactionListener() {
          public IUndoContext getContext() {
            return PlatformUI.getWorkbench().getOperationSupport().getUndoContext();
          }

          @Override
          public void _dispose() {
            //IOperationHistory history = PlatformUI.getWorkbench()
            //        .getOperationSupport().getOperationHistory();
            //history.dispose(getContext(), true, true, true);
          }

          @Override
          public void execute(TransactionStorage tStorage) {
            TreeDBOperation operation = new TreeDBOperation(tStorage);

            operation.addContext(getContext());
            IOperationHistory history =
                PlatformUI.getWorkbench().getOperationSupport().getOperationHistory();
            history.add(operation);
          }

          @Override
          public void redo(TransactionStorage tStorage) {}

          @Override
          public void undo(TransactionStorage tStorage) {
            // TODO Auto-generated method stub

          }
        });

    db.setIsLoaded();
    return db;
  }

  public boolean stopTracking(IProject project, boolean deleted) {
    if (!isTracking(project)) {
      return false;
    }
    TreeDB db = getTree(project);
    if (db == null) return false;
    TreeNode node = db.getRootNode();
    if (node != null) Instantiator.removeInstance(db, node.getUUId());
    ITreeStorage ts = db.getStorage();
    //TODO error?
    if (ts instanceof TransactionStorage) ts = ((TransactionStorage) ts).getStorage();
    ResourceStorage storage = (ResourceStorage) ts;
    storage.forceNow = true; //avoid creating jobs
    try{
      ResourcesPlugin.getWorkspace()
          .removeResourceChangeListener(storage.getResourceChangeListener());
      projectToStorage.remove(project);
      if (!deleted) {
        notifyListeners(db, TreeEvent.TREE_CLOSE);
        unbindBaseFromProject(project, db);
        treeByProject.put(project, null);
      } else {
        notifyListeners(db, TreeEvent.TREE_DELETE);
        unbindBaseFromProject(project, db);
        db.getStorage().getCacheStorage().remove();
      }
      Instantiator.clearCache(db);
    }finally{
      storage.forceNow = false;
    }
    return true;
  }

  private void unbindBaseFromProject(IProject project, TreeDB db) {
    projectByTree.remove(db);
    treeByProject.remove(project);
  }

  private void bindBaseToProject(IProject project, TreeDB db) {
    projectByTree.put(db, project);
    treeByProject.put(project, db);
  }

  public boolean isTracking(IProject project) {
    boolean result = treeByProject.containsKey(project);
    return result;
  }

  public TreeDB[] getTrees() {
    return projectByTree.keySet().toArray(new TreeDB[0]);
  }

  public IProject[] getProjects() {
    return projectByTree.values().toArray(new IProject[0]);
  }

  private boolean isTrackable(IProject project) {
    try {
      return project.isNatureEnabled(ReqCatalogue.ID);
    } catch (Exception e) {
      e.printStackTrace();
      return false;
    }
  }

  public void setProjects(IProject[] projects) {
    HashSet<IProject> current = new HashSet<>(treeByProject.keySet());

    for (IProject project : projects) {

      try {
        if (isTrackable(project) && !isTracking(project)) {
          trackProject(project);
        } else if (!isTrackable(project) && isTracking(project)) {
          stopTracking(project, !project.exists());
        }
        current.remove(project);
      } catch (Exception e) {
        e.printStackTrace();
      }
    }

    for (IProject iProject : current) {
      stopTracking(iProject, iProject.exists());
    }
  }

  public static TreeNode getNodeByResource(IFile file) {
    return getNodeByResource(file, false);
  }
  public static TreeNode getNodeByResource(IFile file, boolean ignoreLock) {
    TreeDB db = getTracker().getTree(file.getProject());
    if (db == null) return null;
    ResourceStorage rs = getResourceStorage(db);
    UUID uuid = rs.getUUId(file, false);
    if (uuid == null) {
      String name = file.getParent().getName();
      if (name.endsWith("_resources")) {
        uuid =
            rs.getUUId(
                file.getParent()
                    .getParent()
                    .getFile(Path.fromOSString(name.substring(0, name.length() - 10) + ".json")));
      }
      if (uuid == null) return null;
    }
    try {
      return db.getNode(uuid);
    } catch (RuntimeException re) {
      //node not found
      return null;
    }
  }

  public static IResource getResourceByNode(TreeNode node) {
    //TODO check it
    TreeDB db = node.getTreeDB();
    UUID uid = node.getUUId();
    //    	if(node.isVirtual())
    //  		return null;
    ResourceStorage rs = getResourceStorage(db);
    if (rs != null) return rs.getResource(uid);

    return null;
  }

  public static ResourceStorage getResourceStorage(TreeDB db) {
    ResourceStorage rs = null;
    if (db == null) return null;
    if (db.inTransaction()) {
      if (((TransactionStorage) db.getStorage()).getStorage() instanceof ResourceStorage)
        rs = (ResourceStorage) ((TransactionStorage) db.getStorage()).getStorage();
    } else if (db.getStorage() instanceof ResourceStorage) {
      rs = (ResourceStorage) db.getStorage();
    } else if (db.getStorage() instanceof TreeInstancer) {
      TreeDB rdb = Instantiator.getRootTree(db, TreeDB.DEFAULT_DEPTH);
      ITreeStorage ts = null;
      if (rdb != db) ts = getResourceStorage(rdb);
      else ts = db.getStorage();
      if (ts instanceof ResourceStorage) rs = (ResourceStorage) ts;
    }
    return rs;
  }

  public void assignTreeDB(IProject project, ResourceStorage resourceStorage) {
    projectToStorage.put(project, resourceStorage);
  }

  public static IResource getMarkerResourceByNode(TreeNode node) {
    if (node.getType().equals(Document.TYPE_NAME)) {
      ResourceStorage rs = getResourceStorage(node.getTreeDB());
      if (rs != null) {
        String resname = ((Document) node).getResourceName();
        if (resname != null) return rs.getResourceFolder(node.getUUId()).getFile(resname);
      }
    }
    return getResourceByNode(node);
  }

  @Override
  public TreeDB[] getTreeDBs() {
    return getTrees();
  }

  @Override
  public void trackProject(String name, TreeDB db) {
    final IWorkspace workspace = ResourcesPlugin.getWorkspace();
    final IProject project = workspace.getRoot().getProject(name);
    locked.remove(project);
    if (db != null) {
      bindBaseToProject(project, db);
    }
    trackProject(project);
  }

  @Override
  public TreeDB createProject(String name) {
    final IWorkspace workspace = ResourcesPlugin.getWorkspace();
    final IProject project = workspace.getRoot().getProject(name);
    locked.add(project);
    if(!project.exists())
      try {
        project.create(null);
      } catch (CoreException e) {
        // TODO Auto-generated catch block
        e.printStackTrace();
      }
    TreeDB db = TreeDB.create(name, new ResourceStorage(project));
    db = loadTreeDB(project, db, (ResourceStorage) db.getStorage());
    initClasses(db);
    //db = EclipseTracker.getTracker().getTree(project);//TODO look into
    runCreate(project);
    return db;
  }

  private static TreeDB runCreate(IProject project) {
    List<String> vers = new ArrayList<>();
    vers.add(StoredFormatVersioning.getFullVersionByGroupId(TreeNode.NODE_GROUP));
    StoredFormatVersioning.setFormatsVersions(project, vers);
    TreeDB db = EclipseTracker.getTracker().getTree(project);
    return db;
  }

  @Override
  public void init(String... params) {}

  @Override
  public String getName(TreeDB db) {
    IProject p = getProject(db);
    if (p != null) return p.getName();
    return null;
  }

  @Override
  public TreeDB getTreeDB(String name) {
    for (TreeDB db : getTrees()) {
      IProject p = getProject(db);
      if (p != null && p.getName().equals(name)) return db;
    }
    return null;
  }

  @Override
  public void remove(String name) {
    try {
      getProject(getTreeDB(name)).delete(false, null);
    } catch (CoreException e) {
      // TODO Auto-generated catch block
      e.printStackTrace();
    }
  }

  @Override
  public String getLocation(TreeDB db) {
    return EclipseTracker.getTracker().getProject(db).getLocationURI().toString();
  }

  @Override
  public Map<String, Template> getReportTemplates() {
    return Activator.getDefault().getReportTemplates();
  }

  public TreeDBBackupRepeatingJob getRepTask(TreeDB db) {
    return smap.get(db);
  }

  @Override
  public void clearCache(TreeDB db) {
    IProject p = projectByTree.remove(db);
    if (p != null) {
      treeByProject.remove(p);
      projectToStorage.remove(p);
    }
    smap.remove(db);
  }

  @Override
  public void reOpen(TreeDB treeDB) {
    IProject p = getProject(treeDB); //TODO change to?
    try {
      if (p != null) {
        p.close(null);
        p.open(null);
      }
    } catch (Exception e) {
      e.printStackTrace();
    }
  }

  @Override
  public TreeDB trackTreeDB(URI path) {
    // TODO Auto-generated method stub
    return null;
  }

  @Override
  public TreeDB trackTreeDB(URI path, String name) {
    // TODO Auto-generated method stub
    return null;
  }

  @Override
  public java.nio.file.Path getRootLocation() {
    return ResourcesPlugin.getWorkspace().getRoot().getLocation().toFile().toPath();
  }

  @Override
  public java.nio.file.Path getGlobalCacheLocation() {
    DialogSettings ds = new DialogSettings("globCache");
    try {
      ds.load("globCache.ini");
    }
    catch(FileNotFoundException ffe){
      
    }
    catch (IOException e) {
      // TODO Auto-generated catch block
      e.printStackTrace();
    }
    if(ds.get("path")!=null){
      File path = new File(ds.get("path"));
      if(path.exists())
        return path.toPath();
    }
    return getRootLocation();
  }
}
