/*
 * Copyright 2015 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.unitesk.requality.eclipse.core;

import java.io.ByteArrayInputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.PrintWriter;
import java.io.StringWriter;
import java.io.UnsupportedEncodingException;
import java.net.URI;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Scanner;
import java.util.Set;
import java.util.Stack;
import java.util.UUID;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReentrantLock;
import java.util.concurrent.locks.ReentrantReadWriteLock;
import java.util.concurrent.locks.ReentrantReadWriteLock.ReadLock;
import java.util.concurrent.locks.ReentrantReadWriteLock.WriteLock;
import org.eclipse.compare.IStreamContentAccessor;
import org.eclipse.core.resources.IContainer;
import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IFolder;
import org.eclipse.core.resources.IMarker;
import org.eclipse.core.resources.IProject;
import org.eclipse.core.resources.IProjectDescription;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.resources.IResourceChangeListener;
import org.eclipse.core.resources.IResourceRuleFactory;
import org.eclipse.core.resources.IWorkspace;
import org.eclipse.core.resources.IWorkspaceRunnable;
import org.eclipse.core.resources.ResourcesPlugin;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IPath;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.Path;
import org.eclipse.core.runtime.jobs.ISchedulingRule;
import org.eclipse.core.runtime.jobs.Job;
import org.eclipse.core.runtime.jobs.MultiRule;
import org.json.JSONException;
import com.unitesk.requality.core.AbstractTreeStorage;
import com.unitesk.requality.core.CacheStorage;
import com.unitesk.requality.core.ChangesSet;
import com.unitesk.requality.core.CoreUtils;
import com.unitesk.requality.core.ITreeStorage;
import com.unitesk.requality.core.IUUIDsCached;
import com.unitesk.requality.core.Moment;
import com.unitesk.requality.core.NodeMapper;
import com.unitesk.requality.core.RLogger;
import com.unitesk.requality.core.TreeDB;
import com.unitesk.requality.core.TreeNode;
import com.unitesk.requality.core.attribute.Attribute;
import com.unitesk.requality.core.attribute.AttributeAvailability;
import com.unitesk.requality.core.attribute.AttributeDefinition;
import com.unitesk.requality.core.attribute.AttributeType;
import com.unitesk.requality.core.attribute.Flag;
import com.unitesk.requality.core.attribute.valueproviders.ConstValueProvider;
import com.unitesk.requality.core.attribute.valueproviders.HtmlValueProvider;
import com.unitesk.requality.eclipse.natures.ReqCatalogue;
import com.unitesk.requality.nodetypes.ErrorNode;

public class ResourceStorage extends AbstractTreeStorage implements IUUIDsCached {
  public static final String RESOURCE_POSTFIX = "_resources";
  public boolean forceNow = false;
  private static final String ROOT_NODE_FILENAME = "root.json";
  private Map<UUID,Map<String,AttributeDefinition>> activeAttributesD =
      new HashMap<>();
  private Map<UUID,Map<String,AttributeDefinition>> activeAttributes =
      new HashMap<>();
  private IProject project;
  private UUID rootNodeUUID = null;
  private ISchedulingRule activeRules = null;
  private Thread activeThread = null;

  private DeltaProcessor deltaProcessor = new DeltaProcessor(this);

  public IResourceChangeListener getResourceChangeListener() {
    return deltaProcessor;
  }

  public DeltaProcessor getDeltaProcessor() {
    return deltaProcessor;
  }

  public ResourceStorage(IProject project) {
    this.project = project;
    EclipseTracker.getTracker().assignTreeDB(project, this);
    cs = new CacheStorage(this,"");
    try {
      cs.getConnection();//additional init features
      project.getFile(CoreUtils.CACHE_FOLDER_NAME).refreshLocal(IResource.DEPTH_INFINITE, null);
    } catch (CoreException e) {
      e.printStackTrace();
    }
    getResourceMap();
  }

  public IFolder getResourceFolder(UUID uuid) {
    readLock();
    IFolder resFolder = null;
    try{
      String id = getNodeId(uuid);
      IFile file = getResource(uuid);
      if (file == null) {
        return null;
      } else {
        resFolder = file.getParent().getFolder(new Path(id + RESOURCE_POSTFIX));
        return resFolder;
      }
    }finally{
      boolean rl = true,wl = false;
      try{
//        if (resFolder!=null && !resFolder.isSynchronized(1)){
//          readUnlock();
//          rl = false;
//          writeLock();
//          wl = true;
//          try {
//            resFolder.refreshLocal(1, null);
//          } catch (CoreException e) {
//            e.printStackTrace();
//          }
//        }
      }finally{
        if(rl)
          readUnlock();
        if(wl)
          writeUnlock();
      }
    }
  }

  @Override
  public UUID getRootNode() {
    synchronized (this) {
      if (rootNodeUUID != null) {
        return rootNodeUUID;
      }
      
      writeLock();
      try{
        IFile file = project.getFile(ROOT_NODE_FILENAME);
        if (file.exists()) {
          //if (!resourceMap.inCache(file)) {
          rootNodeUUID = getUUId(file); //resourceMap.cachePut(file, uuid);
          //            }
          //            rootNodeUUID = resourceMap.cacheGet(file);
          return rootNodeUUID;
        }
        return null;
      }finally{
        writeUnlock();
      }
    }
  }

  @Override
  public String getNodeId(UUID uuid) {
    readLock();
    try{
      if (uuid == null) return null;
      if (uuid.equals(getRootNode())) {
        return "/";
      }
      IFile file = getResource(uuid);
      if (file == null) {
        return null; // node not found;
      }
      String jext = file.getFileExtension();
      String fileName = file.getName();
      return fileName.substring(0, fileName.length() - (jext.length() + 1));
    }
    finally{
      readUnlock();
    }
  }

  @Override
  public Map<String, AttributeDefinition> getStorageAttributes(UUID uuid) {
    readLock();
    try{
      if(activeAttributes.containsKey(uuid))
        return activeAttributes.get(uuid);
      IFile file = getResource(uuid);
      if (file == null) return new HashMap<>();
      String nodeSource = readFile(file);
      try {
        Map<String, AttributeDefinition> mp =  NodeMapper.getAttributes(uuid, nodeSource);
        activeAttributesD.put(uuid, mp);
        return mp;
      } catch (JSONException e) {
        StringWriter stackTrace = new StringWriter();
        e.printStackTrace(new PrintWriter(stackTrace));
        Map<String, AttributeDefinition> result = new HashMap<>();
  
        result.put(
            TreeNode.ATTR_DESCRIPTION,
            new AttributeDefinition(uuid, TreeNode.ATTR_DESCRIPTION, AttributeAvailability.LOCAL,
                new HtmlValueProvider(AttributeType.HTML, stackTrace.toString()),Flag.HIDDEN));
        result.put(
            TreeNode.ATTR_TYPE,
            new AttributeDefinition(uuid, TreeNode.ATTR_TYPE, AttributeAvailability.LOCAL,
                new ConstValueProvider(AttributeType.STRING, ErrorNode.TYPE_NAME),Flag.HIDDEN));
        activeAttributesD.put(uuid, result);
        return result;
      }
    }finally{
      readUnlock();
    }
  }

  @Override
  public void clearCaches() {
    writeLock();
    try{
      super.clearCaches();
      ResourceStorage rs = this;
      activeAttributes.clear();
      rs.clearAll();
      getCacheStorage().clearCached();
      rs.initRoot();
    }finally{
      writeUnlock();
    }
  }

  @Override
  public UUID getParent(UUID uuid) {
    readLock();
    try{
      if (getResource(uuid) == null) {
        return null; // node not found
      }
      return getParent(getResource(uuid));
    }finally
    {
      readUnlock();
    }
  }

  public void clearPersistent(UUID uuId) {
    getCacheStorage().clearCachedCache(uuId.toString());
    TreeDB.getCPCacheStorage().clearCachedCache(uuId.toString());
    IFile file = getResource(uuId);
    if (file == null) return;
    if (file.exists() && file.isAccessible()) {
      try {
        file.deleteMarkers(null, false, IResource.DEPTH_INFINITE);
      } catch (CoreException e) {
        e.printStackTrace();
      }
    }
  }

  public String getPersistent(UUID uid, String marker, String key) {
    IFile file = getResource(uid);
    if (file !=null && file.exists() && file.isAccessible()) {
      IMarker[] mrks;
      try {
        mrks = file.findMarkers("com.unitesk.requality." + marker, false, IResource.DEPTH_ZERO);
        for (IMarker mrk : mrks) {
          Object o = mrk.getAttribute(key);
          if (o != null) {
            String id = o.toString();
            if (id != null) return id;
          }
        }
      } catch (CoreException e) {
        // TODO Auto-generated catch block
        e.printStackTrace();
      }
    }
    return null;
  }

  public Map<String, String> getPersistents(UUID uid, String marker, Set<String> keys) {
    IFile file = getResource(uid);
    if (file == null) return null;
    if (file.exists() && file.isAccessible()) {
      IMarker[] mrks;
      try {
        mrks = file.findMarkers("com.unitesk.requality." + marker, false, IResource.DEPTH_ZERO);
        Map<String, String> ret = new HashMap<>();
        for (IMarker mrk : mrks) {
          for (String key : keys) {
            Object o = mrk.getAttribute(key);
            if (o != null) {
              String id = o.toString();
              ret.put(key, id);
            }
          }
        }
        return ret;
      } catch (CoreException e) {
        // TODO Auto-generated catch block
        e.printStackTrace();
      }
    }
    return null;
  }

  public void setPersistent(UUID uid, String marker, String key, String val) {
    IFile file = getResource(uid);
    if (file != null && file.exists() && file.isAccessible()) {
      try {
        //file.setPersistentProperty(new QualifiedName("com.unitesk.requality", key),val);
        IMarker[] ms =
            file.findMarkers("com.unitesk.requality." + marker, false, IResource.DEPTH_ZERO);
        IMarker m = null;
        if (ms == null || ms.length < 1) m = file.createMarker("com.unitesk.requality." + marker);
        else m = ms[0];
        while (!m.exists()) Thread.currentThread().join(100);
        m.setAttribute(key, val);
      } catch (CoreException ez) {
        ez.printStackTrace();
      } catch (InterruptedException ey) {
        ey.printStackTrace();
      } catch (Exception ex) {
        ex.printStackTrace();
      }
    }
  }

  public void setPersistents(UUID uid, String marker, Map<String, String> vals) {
    IFile file = getResource(uid);
    if (file != null && file.exists() && file.isAccessible()) {
      try {
        //file.setPersistentProperty(new QualifiedName("com.unitesk.requality", key),val);
        IMarker[] ms =
            file.findMarkers("com.unitesk.requality." + marker, false, IResource.DEPTH_ZERO);
        IMarker m = null;
        if (ms == null || ms.length < 1) m = file.createMarker("com.unitesk.requality." + marker);
        else {
          m = ms[0];
          Set<String> ks = new HashSet<>();
          if(m.getAttributes()!=null)
            ks = new HashSet<>(m.getAttributes().keySet());
          if (!ks.isEmpty()) {
            String[] allkeys = ks.toArray(new String[0]);
            Object[] allvals = m.getAttributes(allkeys);
            for (int i = 0; i < allkeys.length; i++) {
              String s = allkeys[i];
              if (s != null && !vals.containsKey(s)) vals.put(s, allvals[i].toString());
            }
          }
        }
        while (!m.exists()) Thread.currentThread().join(100);
        m.setAttributes(vals);
      } catch (CoreException ez) {
        ez.printStackTrace();
      } catch (InterruptedException ey) {
        ey.printStackTrace();
      } catch (Exception ex) {
        ex.printStackTrace();
      }
    }
  }

  public UUID getParent(IFile file) {
    readLock();
    try{
      if (file.getParent().getType() == IResource.PROJECT) {
        return null;
      }
      String folderName = file.getParent().getName();
      String parentId = folderName.substring(0, folderName.length());
  
      IFile parentResource;
      IContainer parentFolder = file.getParent().getParent();
      parentResource = parentFolder.getFile(new Path(parentId + ".json"));
      if (!parentResource.exists()) {
        if (getResourceMap().inCache(parentResource)) {
          return getResourceMap().cacheGet(parentResource);
        }
        return null;
      }
      return getUUId(parentResource);
    }
    finally{
      readUnlock();
    }
  }

  public boolean inCache(IFile f) {
    return getResourceMap().inCache(f);
  }

  public UUID getUUId(final IFile file) {
    return getUUId(file, true);
  }
  public UUID getUUId(final IFile file, boolean withcache) {
    String ext = file.getFileExtension();
    if (ext == null || !ext.contains("json")) return null;
    boolean hasp = false;
    if (!withcache || !getResourceMap().inCache(file)) {
      try{
        getResourceMap().writeLock();
        UUID uid = null;
        //              try {
        //                  String pp = file.getPersistentProperty(uuidname);
        //                  if(pp!=null){
        //                      uid = UUID.fromString(pp);
        //                  }
        //              } catch (CoreException e) {
        //                  // TODO Auto-generated catch block
        //                  e.printStackTrace();
        //              }
        
        uid = getFileUUID(file);
        if (uid != null) {
          getResourceMap().cachePut(file, uid, true, false);
          //if(!resourceMap.haveCache(file, uid)){
          //}
        }
        return uid;
      }finally{
        getResourceMap().writeUnlock();
      }
    } else {
      return getResourceMap().cacheGet(file);
    }
  }

  Job j = null;
  Stack<Object[]> toup = new Stack<>();

  public UUID getFileUUID(IFile file) {
    try {
      String content = readFile(file);
      if (content == null||content.isEmpty()){
        RLogger.logInfo("Empty file "+file.getFullPath().toString());
        return null;
      }
      return NodeMapper.getUUId(content);
    } catch (JSONException e) {
      RLogger.logError(CoreUtils.getTrace(e));
      return null;
    }
  }

  @Override
  public Collection<UUID> getChildren(UUID parentUUId) {
    Collection<UUID> result = new HashSet<>();
    IFile parentFile = getResource(parentUUId);
    if (parentFile == null) {
      return result; // node not found
    }

    String nodeId = getNodeId(parentUUId);
    if(nodeId == null)
      return result;
    if (nodeId.equals("/")) {
      nodeId = "root";
    }

    IFolder childrenFolder = parentFile.getParent().getFolder(Path.fromOSString(nodeId));

    if (!childrenFolder.exists()) {
      return result;
    }

    try {
      for (IResource res : childrenFolder.members()) {
        if (res.getType() == IResource.FILE && res.getName().endsWith(".json")) {
          if (res.exists()) {
            try {
              UUID childUUID = getUUId((IFile) res);
              if (childUUID != null) result.add(childUUID);
            } catch (Exception e) {
              e.printStackTrace(); //cannot load
            }
          }
        }
      }
    } catch (CoreException e) {
      e.printStackTrace();
    }
    return result;
  }

  @Override
  public boolean canLoad(UUID uuid) {
    return getResourceMap().inCache(uuid); //&&resourceMap.getResource(uuid).isAccessible();
  }

  public static String readFile(IFile file) {
    try {
      /*  if (!file.isSynchronized(IResource.DEPTH_ZERO)||!file.isAccessible()) {
          // TODO: add monitor
          file.refreshLocal(IResource.DEPTH_ZERO, null);
      }*/
      String fileExt = file.getFileExtension();
      if (fileExt == null || !fileExt.equals("json") || !file.exists()) {
        return null;
      }
      StringBuilder jsonsb = new StringBuilder();
      Scanner s = new Scanner(file.getContents(true), file.getCharset());
      while (s.hasNextLine()) {
        jsonsb.append(s.nextLine().trim());
        jsonsb.append("\n");
      }
      s.close();
      return jsonsb.toString();
    } catch (CoreException e) {
      e.printStackTrace();
    }
    return null;
  }
  public static String readFile(IStreamContentAccessor file) {
    try {
      /*  if (!file.isSynchronized(IResource.DEPTH_ZERO)||!file.isAccessible()) {
          // TODO: add monitor
          file.refreshLocal(IResource.DEPTH_ZERO, null);
      }*/
      StringBuilder jsonsb = new StringBuilder();
      Scanner s = new Scanner(file.getContents(), "UTF-8");
      while (s.hasNextLine()) {
        jsonsb.append(s.nextLine().trim());
        jsonsb.append("\n");
      }
      s.close();
      return jsonsb.toString();
    } catch (CoreException e) {
      e.printStackTrace();
    }
    return null;
  }
  @Override
  public URI getNodeResourcePath(UUID uuId) {
    IFolder folder = getResourceFolder(uuId);
    if (folder == null) return null;
    //TODO look into
    return getResourceFolder(uuId).getLocationURI();
  }

  public IProject getProject() {
    return project;
  }

  @Override
  public InputStream getResourceContents(UUID uuid, String resName) {
    try {
      // TODO why we lost sync here?
      IFolder resFolder = getResourceFolder(uuid);
      if (resFolder == null || !resFolder.exists()) return null;
      //if (!resFolder.isSynchronized(1)) {
      //}
      IFile f = resFolder.getFile(resName);
      if (f == null || !f.exists()) return null;
      return f.getContents();
    } catch (CoreException e) {
      e.printStackTrace();
    }
    return null;
  }

  @Override
  public boolean isResourceExists(UUID uuid, String resName) {
    if (resName == null) return false;
    IFolder resFolder = getResourceFolder(uuid);
    if (resFolder == null || !resFolder.exists()) return false;
    return resFolder.getFile(resName).exists();
  }

  private void recursiveResourceList(IFolder folder, String path, Set<String> result)
      throws CoreException {
    if (folder.getName().startsWith(".svn")) return;
    for (IResource res : folder.members()) {
      // use unix path separator
      String curPath = path + "/" + res.getName();
      if (res instanceof IFolder) {
        recursiveResourceList((IFolder) res, curPath, result);
      } else {
        result.add(curPath);
      }
    }
  }

  @Override
  public Set<String> listResources(UUID uuid) {
    Set<String> result = new HashSet<>();
    try {
      IFolder ifld = getResourceFolder(uuid);
      if (ifld != null && ifld.exists()) {
        recursiveResourceList(ifld, "", result);
      }
    } catch (CoreException e) {
      e.printStackTrace();
    }
    return result;
  }

  @Override
  public String getName() {
    return project.getName();
  }

  private IFile getFile(UUID parentUUId, String id, boolean create, IProgressMonitor monitor) {
    IFile file = null;
    if (parentUUId == null) {
      file = project.getFile(ROOT_NODE_FILENAME);
    } else {
      // get parent id
      IFolder childFolder = getParentFolder(parentUUId, id, null);
      //IWorkspace works=ResourcesPlugin.getWorkspace();
      //    if(works!=null)
      //    works.checkpoint(false);
      //                if(create)
      //                    if (!childFolder.exists()) {
      //                        try {
      //                            childFolder.create(true, true, monitor);
      //                        } catch (CoreException e) {
      //                            // TODO Auto-generated catch block
      //                            e.printStackTrace();
      //                        }
      //                    }
      if (childFolder == null) return null;
      file = childFolder.getFile(id + ".json");
    }
    return file;
  }

  private IFolder getParentFolder(UUID parentUUId, String id, IProgressMonitor monitor) {
    if (parentUUId == null) {
      return null; //root
    } else {
      IFile pfile = getFile(parentUUId);
//      // get parent id
//      String parentId = getNodeId(parentUUId, true);
      if (pfile == null) {
        RLogger.logInfo("parid null");
      } else {
//        if ("/".equals(parentId)) {
//          parentId = "root";
//        }
        IFolder childFolder;
        //IFile parentFile = getFile(parentUUId);
        String name = pfile.getName();
        if (pfile.getParent().getType() == IResource.PROJECT) {
          childFolder = ((IProject) pfile.getParent()).getFolder(name.substring(0,name.length()-5));
        } else {
          childFolder = ((IFolder) pfile.getParent()).getFolder(name.substring(0,name.length()-5));
        }
        //IWorkspace works=ResourcesPlugin.getWorkspace();
        //  if(works!=null)
        //  works.checkpoint(false);
        return childFolder;
      }
    }
    return null;
  }
  ReentrantReadWriteLock rwl = new ReentrantReadWriteLock();
  //final LockProxy<ReentrantLock> rl = new LockProxy<>(new ReentrantLock());
  private void writeLock(){
    rwl.writeLock().lock();
  }
  private void writeUnlock(){
    rwl.writeLock().unlock();
  }
  void readLock(){
    rwl.readLock().lock();
    //rl.lock();
  }
  void readUnlock(){
    rwl.readLock().unlock();
    //rl.unlock();
  }
  @Override
  public void createStorageNode(
      final UUID parentUUId, final UUID uuid, final String id, final Map<String, AttributeDefinition> attrs) {
    writeLock();
    try{
      final IFolder pfldr = getParentFolder(parentUUId, id, null);
      IWorkspaceRunnable createJob =
          new IWorkspaceRunnable() {
            @Override
            public void run(IProgressMonitor monitor) throws CoreException {
              try {
                  if (pfldr != null && !pfldr.exists())
                    try {
                      pfldr.create(true, true, monitor);
                    } catch (CoreException e) {
                      // TODO Auto-generated catch block
                      e.printStackTrace();
                    }
                } finally {
                }
              boolean wu = false;
              try {
                IFile file = getFile(parentUUId, id, true, monitor);
                getResourceMap().writeLock();
                wu = true;
                try{
                  getResourceMap().cachePut(file, uuid, false,true);
                  if (file == null) return;
                }
                finally{
                  if(wu){
                    getResourceMap().writeUnlock();
                    wu = false;
                  }
                }
                if (getDeltaTarget() != null)
                  getDeltaTarget().beforeCreated(uuid, parentUUId, attrs, true);
                //IFile file = getFile(parentUUId, id, true, monitor);
                getResourceMap().writeLock();
                wu = true;
                //getResourceMap().cachePut(file, uuid, false,true);
                //if (file == null) return;
                //without bottom line doc can have null id on import
                activeAttributesD.put(uuid, new HashMap<>(attrs));
                String jsonString = NodeMapper.toJSON(uuid, mapToList(attrs),true).toString(4);
                ByteArrayInputStream in = new ByteArrayInputStream(jsonString.getBytes("UTF-8"));
                //TODO wrong?
                //                  IWorkspace works=ResourcesPlugin.getWorkspace();
                //                  if(works!=null)
                //                      works.checkpoint(false);
                //s
                if (file.exists()) {
                  file.setContents(in, true, true, monitor);
                } else
                  try {
                    file.create(in, true, monitor);
                  } catch (CoreException re) {
                    try {
                      file.setContents(in, true, true, monitor);
                    } catch (CoreException re2) {
                      re2.printStackTrace();
                    }
                  }
               try {
                file = getFile(parentUUId, id, true, monitor);
                if (file == null) return;
                if (!file.getParent().exists()) return;
                if (file.exists())
                  if (!file.getCharset().toUpperCase().equals("UTF-8")) {
                    try {
                      file.refreshLocal(0, monitor);
                      file.setCharset("UTF-8", monitor);
                    } catch (CoreException ce) {
                      ce.printStackTrace();
                    }
                  }
              } finally {
              }
                //              resourceMap.toLook.add(file);
                getResourceMap().cachePut(file, uuid, true, true);
                if(wu){
                  wu = false;
                  getResourceMap().writeUnlock();
                }
  
                //тут поменяно местамиs
                //   synchronized (ResourceStorage.this) {
                // }
              } catch (UnsupportedEncodingException e) {
                e.printStackTrace();
              } catch (JSONException e) {
                e.printStackTrace();
              } catch (Exception e) {
                e.printStackTrace();
              } finally {
                if(wu){
                  wu = false;
                  getResourceMap().writeUnlock();
                }
                
              }
              IWorkspace works = ResourcesPlugin.getWorkspace();
              works.checkpoint(false);
            }
          };
  
      //strictRunWithoutDeltaProc(createJob, getFile(parentUUId,id,false,null).getParent());
      //strictRunWithoutDeltaProcWithReact(pfoldercreate, createJob,react, getFile(parentUUId,id,false,null).getParent());
      //works
      waitAccess(project);
  
      IWorkspace works = ResourcesPlugin.getWorkspace();
      IResourceRuleFactory ruleFactory = works.getRuleFactory();
      runWithoutDeltaProc(
          createJob,
          MultiRule.combine(
              ruleFactory.modifyRule(getFile(getRootNode())),
              ruleFactory.createRule(getFile(uuid).getParent())));
      runWithoutDeltaProc(new IWorkspaceRunnable(){
        @Override
        public void run(IProgressMonitor monitor) throws CoreException {
          try {
            if (getDeltaTarget() != null) getDeltaTarget().created(uuid, true);
          } finally {
            
          }
        }
      }, null);
    }
    finally{
      writeUnlock();
    }
  }

  @Override
  public void changeNodeId(final UUID uuid, final String id) {
    writeLock();
    try{
      IWorkspace works = ResourcesPlugin.getWorkspace();
      IResourceRuleFactory ruleFactory = works.getRuleFactory();
      IWorkspaceRunnable changeNodeId =
          new IWorkspaceRunnable() {
            @Override
            public void run(IProgressMonitor monitor) throws CoreException {
              try{
                String nodeId = getNodeId(uuid);
                if (id.equals(nodeId)) {
                  //strange
                  return;
                }
                IFile nodeFile = getResource(uuid);
                if (nodeFile == null || nodeFile.getParent() == null) {
                  return;
                }
                IPath newPath = nodeFile.getParent().getFullPath();
    
                newPath = newPath.addTrailingSeparator();
                IPath newFilePath = newPath.append(id + ".json");
                if (nodeFile.getFullPath().equals(newFilePath)) 
                  return;
                try {
                  nodeFile.move(newFilePath, true, monitor);
                } catch (Exception e) {
                  e.printStackTrace();
                }
                IFolder folder =
                    project.getFolder(
                        newPath
                            .append(nodeId + RESOURCE_POSTFIX)
                            .makeRelativeTo(project.getFullPath()));
                if (folder.exists()) {
                  folder.move(newPath.append(id + RESOURCE_POSTFIX), true, monitor);
                }
                folder =
                    project.getFolder(newPath.append(nodeId).makeRelativeTo(project.getFullPath()));
                if (folder.exists()) {
                  try {
    
                    folder.move(newPath.append(id), true, monitor);
    
                  } catch (Exception e) {
                    e.printStackTrace();
                  }
                }
                IFile newFile = project.getFile(newFilePath.makeRelativeTo(project.getFullPath()));
                //                if (!newFile.getCharset().equals("UTF-8")) {
                //                  newFile.setCharset("UTF-8", monitor);
                //                }
                getResourceMap().cacheReplace(nodeFile, newFile, uuid, project);
                getResourceMap()
                    .updatePath(
                        nodeFile.getParent().getLocation().append(nodeId),
                        newFile.getParent().getLocation().append(id),
                        project);
                
              IFile file = nodeFile;
              if (!file.getParent().exists()) return;
              if (file.exists())
                if (!file.getCharset().toUpperCase().equals("UTF-8")) {
                  try {
                    file.refreshLocal(0, monitor);
                    file.setCharset("UTF-8", monitor);
                  } catch (CoreException ce) {
                    ce.printStackTrace();
                  }
                }
                
                IWorkspace works = ResourcesPlugin.getWorkspace();
                works.checkpoint(false);
              }
              finally{
                
              }
              
            }
          };
      waitAccess(project);
      String old = getNodeId(uuid);
      runWithoutDeltaProc(
          changeNodeId,
          MultiRule.combine(
              ruleFactory.modifyRule(getFile(getRootNode())),
              ruleFactory.moveRule(getFile(uuid), getFile(uuid).getParent())));
    }finally{
      writeUnlock();
    }
    
  }

  @Override
  public void remove(final UUID uuid) {
    // TODO: remove with new delta processor
    Collection<UUID> children = getChildren(uuid);
    for (UUID child : children) {
      if (child != uuid) //fix some glitches
      remove(child);
    }
    writeLock();
    try{
      final IFile res = getResource(uuid);
      if(res==null || !res.exists())
        return;
      final IFolder fol = getParentFolder(getParent(uuid), getNodeId(uuid), null);
      IWorkspaceRunnable deleteJob =
          new IWorkspaceRunnable() {
            @Override
            public void run(IProgressMonitor monitor) throws CoreException {
              boolean wl = false; 
              activeAttributes.remove(uuid);
              String id = getNodeId(uuid);
              IContainer parent = res.getParent();
              IFolder children = parent.getFolder(new Path(id));
              IFolder resources = parent.getFolder(new Path(id + RESOURCE_POSTFIX));
              if (resources.exists()) {
                resources.delete(true, monitor);
              }
              if (children.exists()) {
                children.delete(true, monitor);
              }
              res.delete(true, monitor);
              if (fol != null
                  && fol.isAccessible()
                  && fol instanceof IFolder
                  && fol.members().length == 0) {
                fol.delete(true, monitor);
              }
            }
          };
      //if (res != null) { //res null - removed?
        //strictRunWithoutDeltaProcWithReact(deleteJob,rule, parrem,null, afterdel, null);
  
        IWorkspace works = ResourcesPlugin.getWorkspace();
        IResourceRuleFactory ruleFactory = works.getRuleFactory();
        waitAccess(project);
        ISchedulingRule mr =
            MultiRule.combine(
                ruleFactory.modifyRule(getFile(getRootNode())),
                MultiRule.combine(
                    ruleFactory.deleteRule(getResourceFolder(uuid)),
                    ruleFactory.deleteRule(getFile(uuid))));
        if (getParent(uuid) != null)
          mr = MultiRule.combine(getResource(getParent(uuid)).getParent(), mr);
        runWithoutDeltaProc(
            new IWorkspaceRunnable() {
              @Override
              public void run(IProgressMonitor monitor) throws CoreException {
                if (getDeltaTarget() != null) getDeltaTarget().deleted(uuid, Moment.BEFORE, true);
              }
            
            }, null);
        runWithoutDeltaProc(deleteJob, mr);

        runWithoutDeltaProc(
            new IWorkspaceRunnable() {
              @Override
              public void run(IProgressMonitor monitor) throws CoreException {
                if (getDeltaTarget() != null) getDeltaTarget().deleted(uuid, Moment.AFTER, true);

                getResourceMap().writeLock();
                boolean wl = true;
                try{
                  getResourceMap().cacheRemove(res, project);
                }
                finally{
                  if(wl)
                    getResourceMap().writeUnlock();
                }
                TreeDB.clearAllCaches(null, uuid);
                activeAttributes.remove(uuid);
                IWorkspace works = ResourcesPlugin.getWorkspace();
                works.checkpoint(false);
              }
            },
            null);
       //}
     }
     finally{
       writeUnlock();
     }
  }

  private static Collection<AttributeDefinition> mapToList(Map<String, AttributeDefinition> map) {
    if (map == null) return null;
    List<AttributeDefinition> attrs = new LinkedList<>();
    for (Entry<String, AttributeDefinition> s:map.entrySet())
    //attrs.addAll(map.values());
    {
      if (s.getValue()!=null) attrs.add(s.getValue());
    }
    return attrs;
  }

  @Override
  public void saveStorageAttributes(
      final UUID uuid, final ChangesSet<Attribute> changes) {
    writeLock();
    try{
      final IFolder parf = getParentFolder(getParent(uuid), getNodeId(uuid), null);
      //    if(uuid.equals(getRootNode()))
      //    return;
      //final Map<String, Attribute> oldAttrs = getAttributes(uuid);
      IWorkspaceRunnable saveAttributes =
          new IWorkspaceRunnable() {
            @Override
            public void run(IProgressMonitor monitor) {
              String jsonString;
              boolean wl = false;
              try {
//                 if (parf != null) {
//                  if (!parf.exists())
//                    try {
//                      parf.create(true, true, monitor);
//                      parf.refreshLocal(1, monitor);
//                    } catch (CoreException e) {
//                      // TODO Auto-generated catch block
//                      e.printStackTrace();
//                    }
//                  else parf.refreshLocal(1, monitor);
//                }
                IFile file = getResource(uuid);
                if (file == null) 
                  return;
                if(!file.isSynchronized(0))
                  file.refreshLocal(0, null);
                //  if(file.getFullPath().toOSString().endsWith("root.json"))
                //  return;
                getResourceMap().writeLock();
                wl = true;
                Map<String,AttributeDefinition> old = getStorageAttributes(uuid);
                if(old == null)
                  old = new HashMap<>();
                Map<String, Attribute> newA = changes.getNewMap();
                for(Entry<String, Attribute> en:newA.entrySet()){
                  Attribute a = en.getValue();
                  if(a == null || a.getValue() == null)
                    old.remove(en.getKey());
                  else
                    old.put(en.getKey(),a.getDefinition());
                }
                activeAttributes.put(uuid, new HashMap<>(old));
               
                jsonString = NodeMapper.toJSON(uuid, mapToList(old),true).toString(4);
  
                ByteArrayInputStream in = new ByteArrayInputStream(jsonString.getBytes("UTF-8"));
                //RLogger.logInfo(file.getFullPath());
                //                    if (!file.getCharset().equals("UTF-8")) {
                //                      file.setCharset("UTF-8", monitor);
                //                    }
                if (!file.exists()) file.create(in, IResource.FORCE, null);
                else file.setContents(in, IResource.FORCE, monitor);
                if (!file.equals(getResourceMap().getResource(uuid)))
                  //synchronized (ResourceStorage.this) {
                  getResourceMap().cachePut(file, uuid, true, true);
                file = getResource(uuid);
                if (file == null) return;
                if (!file.getParent().exists()) return;
//                if (file.exists())
//                  if (!file.getCharset().toUpperCase().equals("UTF-8")) {
//                    try {
//                      file.refreshLocal(0, monitor);
//                      file.setCharset("UTF-8", monitor);
//                    } catch (CoreException ce) {
//                      ce.printStackTrace();
//                    }
//                  }
                //}
  //              Map<String,Version> nva = new HashMap<String,Version>();
  //              for(String s:changes.getAddedChanged())
  //                nva.put(s, attributes.get(s).getVersion());
  //              saveVersions(uuid, changes, nva);
              } catch (JSONException e) {
                e.printStackTrace();
              } catch (CoreException e) {
                e.printStackTrace();
              } catch (UnsupportedEncodingException e) {
                e.printStackTrace();
              }
              finally{
                if(wl)
                    getResourceMap().writeUnlock();
                
              }
            }
          };
      //ISchedulingRule rule = project;
      waitAccess(project);
  
      IWorkspace works = ResourcesPlugin.getWorkspace();
      IResourceRuleFactory ruleFactory = works.getRuleFactory();
      waitAccess(project);
      ISchedulingRule rule =MultiRule.combine(
          ruleFactory.modifyRule(getFile(rootNodeUUID)), ruleFactory.modifyRule(getFile(uuid)));
      if(parf!=null)
        rule = MultiRule.combine(ruleFactory.modifyRule(parf),
            rule);
      runWithoutDeltaProc(
          saveAttributes, rule
          );
    }
    finally{
      writeUnlock();
    }
    //strictRunWithoutDeltaProcWithReact(pfoldercreate,saveAttributes,react, getFile(uuid).getParent());
    //strictRunWithoutDeltaProc(saveAttributes, getFile(uuid).getParent());
  }

  private IFile getFile(UUID uuid) {
    return getFile(getParent(uuid), getNodeId(uuid), false, null);
  }

  @Override
  public void moveNodeTo(final UUID curNodeUUId, final UUID newParentUUId) {
    writeLock();
    try{
      UUID oldParentUUID = getParent(curNodeUUId);
      String parentId = getNodeId(newParentUUId);
      if (parentId.equals("/")) {
        parentId = "root";
      }
      String nodeId = getNodeId(curNodeUUId);
      IFile toFile =
          getResource(newParentUUId)
              .getParent()
              .getFolder(Path.fromOSString(parentId))
              .getFile(Path.fromOSString(nodeId + ".json"));
      waitAccess(project);
  
      IWorkspaceRunnable moveJob =
          new MoveNodeJob(this, toFile, curNodeUUId, oldParentUUID, newParentUUId);
      IWorkspace works = ResourcesPlugin.getWorkspace();
      IResourceRuleFactory ruleFactory = works.getRuleFactory();
      runWithoutDeltaProc(
          moveJob,
          MultiRule.combine(
              ruleFactory.deleteRule(getFile(curNodeUUId).getParent()),
              ruleFactory.moveRule(getFile(curNodeUUId), toFile.getParent())));
      run(new IWorkspaceRunnable(){
      
        @Override
        public void run(IProgressMonitor monitor) throws CoreException {
          try {
            getDeltaTarget().moved(curNodeUUId, oldParentUUID, newParentUUId, Moment.AFTER, true);
          } catch (Exception e) {
            e.printStackTrace();
          }
          IWorkspace works = ResourcesPlugin.getWorkspace();
          works.checkpoint(false);
        }
      }, null);
     }finally{
       writeUnlock();
     }
  }

  private void recCreate(IContainer rt, IProgressMonitor monitor) throws CoreException {
    if(!rt.getProject().isAccessible())
      return;
    if (!rt.exists()) { 
      recCreate(rt.getParent(), monitor);
      if (rt instanceof IFolder) ((IFolder) rt).create(true, true, monitor);
    }
  }

  @Override
  public void setResourceContents(
      final UUID uuid, String resName, final InputStream in, Object data) {
    try {
      final String name = CoreUtils.URLEncode(resName);
      if (in.available() > 0) {
        writeLock(); 
        try{
          final IFolder folder = getResourceFolder(uuid);
          if (folder == null) return;
          IWorkspaceRunnable setResourceContents =
              new IWorkspaceRunnable() {
                @Override
                public void run(IProgressMonitor monitor) throws CoreException {
                  //long t = System.currentTimeMillis();
                  try
                  {
                    if (getDeltaTarget() != null) {
                      getDeltaTarget().resourceChanged(uuid, resName, Moment.BEFORE, true);
                    try {
                      //folder.refreshLocal(0, monitor);
                      if (!folder.exists()) {
                        recCreate(folder, monitor);
                        //folder.create(true, true, monitor);
                      }
                    } catch (CoreException ce) {
                      ce.printStackTrace();
                    }
                    //IFile file = folder.getFile(resName);
                    //while (folder.getProject().isAccessible()&&!folder.exists()) ;
                    java.io.File f =
                        new java.io.File(URI.create(folder.getLocationURI() + "/" + (name)));
                    //if(!f.exists()){
                    try {
                      f.getParentFile().mkdirs();
                      if (!f.exists()) f.createNewFile();
                      FileOutputStream fos = new FileOutputStream(f);
                      byte[] buf = new byte[32 * 1024];
                      //ReadableByteChannel f1 = Channels.newChannel(in);
                      //FileChannel f2 = fos.getChannel();
                      int len = 0;
                      while ((len = in.read(buf)) != -1) {
                        fos.write(buf, 0, len);
                      }
                      fos.close();
                    } catch (IOException e) {
                      // TODO Auto-generated catch block
                      e.printStackTrace();
                    }
                    
                    //folder.refreshLocal(1, null);
                    IFile file = folder.getFile(resName);
                    if(!file.isSynchronized(0))
                      file.refreshLocal(0, null);
                    if (!file.getParent().exists()) return;
//                    if (file.exists())
//                      if (!file.getCharset().toUpperCase().equals("UTF-8")) {
//                        try {
//                          file.refreshLocal(0, monitor);
//                          file.setCharset("UTF-8", monitor);
//                        } catch (CoreException ce) {
//                          ce.printStackTrace();
//                        }
//                      }
                  }
                }
                finally{
                  
                }
                  //                        }
                  //                            f..create(in, IResource.FORCE, null);
                  //                        else `
                  //f.setContents(in, IResource.FORCE, monitor);
                  //long d = System.currentTimeMillis()-t;
                  //RLogger.logInfo("time "+d+" "+resName);
              };
          };
          IWorkspace works = ResourcesPlugin.getWorkspace();
          runWithoutDeltaProc(setResourceContents, folder.getParent());
          run(new IWorkspaceRunnable(){
          
            @Override
            public void run(IProgressMonitor monitor) throws CoreException {

              if (getDeltaTarget() != null)
                getDeltaTarget().resourceChanged(uuid, resName, Moment.AFTER, true);
              IWorkspace works = ResourcesPlugin.getWorkspace();
              works.checkpoint(false);
            }
          }, null);
        }
        finally{
          writeUnlock();
        }
      }
    }
    catch (IOException e) {
      // TODO Auto-generated catch block
      e.printStackTrace();
    }
  }

  long time;

  @Override
  public void deleteResource(final UUID uuid, String name) {
    writeLock();
    try{
      name = CoreUtils.URLEncode(name);
      final String resName = name;
      IWorkspaceRunnable deleteResource =
          new IWorkspaceRunnable() {
            @Override
            public void run(IProgressMonitor monitor) throws CoreException {
              try{
                if (getDeltaTarget() != null)
                  getDeltaTarget().resourceChanged(uuid, resName, Moment.BEFORE, true);
                IFolder res_folder = getResourceFolder(uuid);
                if (res_folder != null && res_folder.exists()) {
                  res_folder.getFile(resName).delete(true, monitor);
                  if (res_folder.members().length == 0) {
                    res_folder.delete(true, monitor);
                  }
                }
              }finally{
                
              }
            }
          };
  
      IContainer rule = getResourceFolder(uuid);
      if (rule != null) rule = rule.getParent();
      waitAccess(project);
      IWorkspace works = ResourcesPlugin.getWorkspace();
      IResourceRuleFactory ruleFactory = works.getRuleFactory();
      IFolder rf = getResourceFolder(uuid);
      runWithoutDeltaProc(
          deleteResource,
          MultiRule.combine(
              ruleFactory.deleteRule(getFile(uuid)),
              rf==null?
                  ruleFactory.modifyRule(getFile(getRootNode()))
              :
                MultiRule.combine(
                  ruleFactory.deleteRule(getResourceFolder(uuid)),
                  ruleFactory.modifyRule(getFile(getRootNode())))));
      runWithoutDeltaProc(new IWorkspaceRunnable(){
      
        @Override
        public void run(IProgressMonitor monitor) throws CoreException {

          if (getDeltaTarget() != null)
            getDeltaTarget().resourceChanged(uuid, resName, Moment.AFTER, true);
          IWorkspace works = ResourcesPlugin.getWorkspace();
          works.checkpoint(false);
        }
      }, null);
    }
    finally{
      writeUnlock();
    }
  }

  private void waitAccess(IProject project) {
    Long l = System.currentTimeMillis();
    while (!project.isAccessible()) if (System.currentTimeMillis() - l > 1000) return;
  }

  @Override
  public void createStorage(String name) {
    writeLock();
    try{
      IWorkspaceRunnable createStorage =
          new IWorkspaceRunnable() {
            @Override
            public void run(IProgressMonitor monitor) throws CoreException {
              try{
                if (!project.exists()) project.create(monitor);
              
              }finally{
                //special case, nodes are created at bottom
              }
              project.open(monitor);
              IProjectDescription description = project.getDescription();
              String[] natures = description.getNatureIds();
              String[] newNatures = new String[natures.length + 1];
              System.arraycopy(natures, 0, newNatures, 0, natures.length);
              newNatures[natures.length] = ReqCatalogue.ID;
              description.setNatureIds(newNatures);
              project.setDescription(description, monitor);
              IWorkspace works = ResourcesPlugin.getWorkspace();
              works.checkpoint(false);
            }
          };
      IResource rule = ResourcesPlugin.getWorkspace().getRoot();
  
      IWorkspace works = ResourcesPlugin.getWorkspace();
      IResourceRuleFactory ruleFactory = works.getRuleFactory();
      waitAccess(project);
      runWithoutDeltaProc(createStorage, ruleFactory.createRule(rule));
    }
    finally{
      writeUnlock();
    }
  }

  @Override
  public void deleteStorage() {
  
    writeLock();
    try{
      IWorkspaceRunnable deleteStorage =
      
          new IWorkspaceRunnable() {
            @Override
            public void run(IProgressMonitor monitor) throws CoreException {
              try{
                project.delete(true, null);
                IWorkspace works = ResourcesPlugin.getWorkspace();
                works.checkpoint(false);
              }
              finally{
                
              }
            }
          };
      waitAccess(project);
      IWorkspace works = ResourcesPlugin.getWorkspace();
      IResourceRuleFactory ruleFactory = works.getRuleFactory();
      runWithoutDeltaProc(deleteStorage, ruleFactory.deleteRule(project));
    }
    finally{
      writeUnlock();
    }
  }

  private Map<IWorkspaceRunnable, ISchedulingRule> jbs =
      new HashMap<>();
  private LinkedList<IWorkspaceRunnable> jorder = new LinkedList<>();
  private void run(final IWorkspaceRunnable wjob, final ISchedulingRule rule){
    ISchedulingRule trule = rule,
        old = activeRules;
    if(activeRules!=null)
      if(rule != null)
        trule = MultiRule.combine(
          activeRules,
          rule
          );
      else
        trule = null;
  
    IWorkspace works = ResourcesPlugin.getWorkspace();

    try {
      try {
        activeRules = trule;
        deltaProcessor.deactivate();
        while (works.isTreeLocked()) ;
        if (forceNow) wjob.run(null);
        else works.run(wjob, trule, IWorkspace.AVOID_UPDATE, null);
      } finally {
        deltaProcessor.activate();
        if(old!=null)
          activeRules = old;
      }
    } catch (CoreException e) {
      // TODO Auto-generated catch block
      e.printStackTrace();
    }
  }
  private void runWithoutDeltaProc(
      final IWorkspaceRunnable wjob, final ISchedulingRule rule) {
    //works.checkpoint(false);
//    if (chRow > 0) {
//      //jbs.put(job,rule);
//      //jorder.addLast(job);
//      return;
//    }
    boolean sync = true;
    if(activeThread == null){
      activeThread = Thread.currentThread();
      sync = false;
    }else
      if(activeThread == Thread.currentThread())
        sync = false;
    if(sync){
      synchronized (activeThread) {
        run(wjob, rule);
      }
    }else
      run(wjob, rule);
  }
//
//  @Override
//  public int stopChangeRow() {
//    int i = super.stopChangeRow();
//    if (chRow == 0 && !jbs.isEmpty()) {
//      final Map<IWorkspaceRunnable, ISchedulingRule> jbs2 =
//          new HashMap<>(jbs);
//      jbs.clear();
//      IWorkspace works = ResourcesPlugin.getWorkspace();
//      deltaProcessor.deactivate();
//      works.checkpoint(false);
//      try {
//        works.run(
//            new IWorkspaceRunnable() {
//              @Override
//              public void run(IProgressMonitor monitor) throws CoreException {
//                for (IWorkspaceRunnable j : jorder) j.run(monitor);
//              }
//            },
//            jbs2.get(jorder.iterator().next()) /*test it*/,
//            IWorkspace.AVOID_UPDATE,
//            null);
//      } catch (CoreException e) {
//        // TODO Auto-generated catch block
//        e.printStackTrace();
//      } finally {
//        jorder.clear();
//        jbs.clear();
//        deltaProcessor.activate();
//      }
//    }
//    return i;
//  };

  // //////////////////////////////////////////////////////////////////////////
  //
  // cache related methods
  //
  // //////////////////////////////////////////////////////////////////////////
  private ResourceMap resourceMap = new ResourceMap(this);
  private boolean resinited = false;

  public ResourceMap getResourceMap() {
    if (resourceMap == null) {
      resourceMap = new ResourceMap(this);
    }
    if (!resinited && project != null) {
      resinited = true;
      resourceMap.initResMap(project);
    }
    return resourceMap;
  }

  public static class ResourceMap {
    private ReentrantReadWriteLock rwl = new ReentrantReadWriteLock();
    ReadLock rl = rwl.readLock();
    WriteLock wl = rwl.writeLock();
    private Map<IFile, UUID> resourceToNode = null;
    private Map<UUID, IFile> rresourceToNode = null;
    private ResourceStorage rs;
    protected boolean changed = false;

    public ResourceMap(ResourceStorage rs) {
      this.rs = rs;
    }

    public void writeUnlock() {
      wl.unlock();
    }

    public void writeLock() {
      wl.lock();
    }

    public void cacheRemove(IFile iFile, IProject project) {
      Map<IFile, UUID> rr = resourceToNode;
      if (rr.containsKey(iFile)) {;
        UUID uid = resourceToNode.remove(iFile);
        rresourceToNode.remove(uid);
        rs.getCacheStorage().removeByKey(rs.rootNodeUUID.toString(), "uuid", "_uuids",
          uid.toString());
          //changed = true;
        //changed = true;
      }
    }

    public boolean inCache(IResource resource) {
      
      rl.lock();
      Boolean b = rresourceToNode.get(resource) != null;
      rl.unlock();
      return b;
    }

    public boolean inCache(UUID uuid) {
      
      rl.lock();
      try{
        boolean b = rresourceToNode.get(uuid) != null;
        return b;
      }finally{
        rl.unlock();
      }
    }

    public void cachePut(IFile iFile, UUID uuid, boolean lock,boolean save) {
//      if (lock) {
//        wl.lock();
//      }
      try{
        resourceToNode.put(iFile, uuid);
        IFile f = rresourceToNode.put(uuid, iFile);
  
        changed = changed || f == null || !f.getFullPath().equals(iFile.getFullPath());
        UUID rid = rs.rootNodeUUID;//cannot use get, then it turns to inf loop
        if(save && rid!=null){
          String rids = rid.toString();
          Object o = rs.getCacheStorage().getCacheValue(rids, "uuid", "_uuids",
              uuid.toString());
          if(iFile!=null && !iFile.getProjectRelativePath().toPortableString().equals(o))
            rs.getCacheStorage().addCacheMapValue(rids, "uuid", "_uuids",
              uuid.toString(),
              iFile.getProjectRelativePath().toPortableString());
          else
            rs.getCacheStorage().removeByKey(rids, "uuid", "_uuids",
              uuid.toString());
        }
        else 
          changed = true;//on init we cannot getRoot, so save later
      }
      finally{
//        if (lock&&wl!=null) wl.unlock();
      }
    }

    public UUID cacheGet(IFile iFile) {
      
      rl.lock();
      UUID uid = resourceToNode.get(iFile);
      rl.unlock();
      return uid;
    }

    public void cacheReplace(IFile oldFile, IFile newFile, UUID uuid, IProject project) {
      if (inCache(uuid)) ;
      {
        
        wl.lock();
        try{
          resourceToNode.remove(oldFile);
          resourceToNode.put(newFile, uuid);
          rresourceToNode.put(uuid, newFile);
          changed = true;
        }
        catch(Exception e){
          e.printStackTrace();
        }finally{
          wl.unlock();
        }
      }
    }

    public void updatePath(IPath oldParentLocation, IPath newParentLocation, IProject project) {
      
      wl.lock();
      try{
        ArrayList<IResource> resources = new ArrayList<>(resourceToNode.keySet());
        for (IResource res : resources) {
          IPath curLocation = res.getLocation();
          if (oldParentLocation.isPrefixOf(curLocation)) {
            IFile newFile = null;
            if (newParentLocation != null) {
              IPath newLocation = curLocation.removeFirstSegments(oldParentLocation.segmentCount());
  
              newLocation = newParentLocation.append(newLocation);
              newFile = ResourcesPlugin.getWorkspace().getRoot().getFileForLocation(newLocation);
            }
            UUID uuid = resourceToNode.remove(res);
            if (newParentLocation != null) {
              rresourceToNode.put(uuid, newFile);
              resourceToNode.put(newFile, uuid);
            } else rresourceToNode.remove(uuid);
            changed = true;
          }
        }
      }
      finally{
        wl.unlock();
      }
    }

    public IFile getResource(UUID uuid) {
      return getResource(uuid,true);
    }
    public IFile getResource(UUID uuid,boolean lock) {
      if(lock)
        rl.lock();
      try{
        IFile f = rresourceToNode.get(uuid);
        //if(f == null)
        //f = (IFile) lookFor(uuid);
        try {
          if (f == null) return null;
        } catch (Exception e) {
          e.printStackTrace();
        }
        return f;
      }
      finally{
        if(lock)
          rl.unlock();
      }
      //return null;
    }

    private void recLook(IResource c, ResourceStorage rs, boolean lock) {
      if (!c.isAccessible()) return;
      if(lock)
        writeLock();
      try{
        if (c instanceof IFile) {
          //if(c.getName()!=null && c.getName().endsWith(".json")){
          IFile f = (IFile) c;
          UUID fuid = rs.getUUId(f, false); //UUID.fromString(uid);//rs.getUUId(f);
          if (fuid != null) {
            cachePut(f, fuid, lock,true);
            //added, nothing to write
          } else {
            RLogger.logInfo("ERROR: File has no uid! " + f);
          }
          //}
        } else if (c instanceof IContainer) {
          IContainer cnt = (IContainer) c;
          if (cnt.getName().startsWith(".")) return; //ignore hidden folders
          try {
            Map<String, IContainer> tohc = new HashMap<>();
            for (IResource f : cnt.members())
              if (f instanceof IFile) {
  
                if (!f.isAccessible()) continue;
                if (f.getName() != null && f.getName().endsWith(".json")) {
                  //recAdd(f,rs);
                  recLook(f, rs, false);
                  String cname = f.getName().substring(0, f.getName().length() - 5);
                  IContainer cn = tohc.remove(cname);
                  if (cn != null) {
                    recLook(cn, rs, false);
                    //recAdd(cn,rs);
                  } else tohc.put(cname, null);
                }
              } else if (f instanceof IContainer) {
                if (tohc.containsKey(f.getName())) recLook(f, rs, false);
                //recAdd(f,rs);
                else tohc.put(f.getName(), (IContainer) f);
              }
            //recAdd(f,rs);
          } catch (CoreException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
          }
        }
      }
      finally{
        if(lock)
          writeUnlock();
      }
    }

    private void initResMap(IProject project) {
      wl.lock();
      try{
        rresourceToNode = new HashMap<>();
        resourceToNode = new HashMap<>();
      }
      finally{
        wl.unlock();
      }
      initRoot(project);
      loadUUIDs(project);
    }

    private void initRoot(IProject project) {
      
      writeLock();
      try{
        IFile f = project.getFile("root.json");
        if (f.exists() && f.isAccessible()){
          UUID uid = rs.getUUId(f, false);
          cachePut(f, uid, false, true);
          rs.rootNodeUUID = uid;
        }
      }
      finally{
        writeUnlock();
      }
    }

    private void loadUUIDs(IProject p) {
      if(rs.getRootNode() == null)
        return;
      
      writeLock();
      try{
        Integer b = (Integer) rs.cs.getCacheValue(rs.getRootNode().toString(), "uuid", "_uuidsInited");
        if (b == null || b!=1) {
          initUUIDs(p, false);
        } else {
          Map<String, Object> kv = rs.cs.getCacheMap(rs.getRootNode().toString(), "uuid", "_uuids");
          if (kv != null)
            for (String key : kv.keySet()) {
              Object val = kv.get(key);
              if (val == null)
                RLogger.logInfo("null value for " + key + " on " + rs.getProject().getName());
              else {
                UUID uid = UUID.fromString(key);
                IFile f = rs.getProject().getFile(val.toString());
                if (f.exists() && f.isAccessible()) cachePut(f, uid, false, false);
              }
            }
        }
      }
      finally{
         writeUnlock();
      }
    }

    private void initUUIDs(IProject p, boolean lock) {
      if (lock) {
        wl.lock();
      }
      String stt = rs.getCacheStorage().startCacheTransaction("init uuids",CacheStorage.COMMIT_ALL);
      try{
        recLook(p, rs, false);
        saveUUIDs(false);
      }
      finally{
        if(stt == null)
          rs.getCacheStorage().commitCache();
        if (wl!=null && lock) wl.unlock();
      }
    }

    public boolean saveUUIDs(boolean lock) {
      if(rs.getRootNode() == null)
        return false;
      if (lock) wl.lock();
      try{
        Map<String, Object> uids = new HashMap<>();
        for (UUID uid : rresourceToNode.keySet()) {
          IFile f = rresourceToNode.get(uid);
          uids.put(uid.toString(), f.getProjectRelativePath().toPortableString());
        }
        rs.cs.setCacheMap(rs.getRootNode().toString(), "uuid", "_uuids", uids);
        rs.cs.setCacheValue(rs.getRootNode().toString(), "uuid", "_uuidsInited", true);
        changed = false;
      }
      finally{
        if (lock&&wl!=null) wl.unlock();
      }
      return true;
    }

    public void clearCache() {
      
      wl.lock();
      try{
        resourceToNode.clear();
        rresourceToNode.clear();
      }finally{
        wl.unlock();
      }
    }
  }
  public IFile getResource(UUID uuid) {
    return getResource(uuid, true);
  }
  public IFile getResource(UUID uuid,boolean lock) {
    return getResourceMap().getResource(uuid,lock);
  }

  private final Lock projectLock = new ReentrantLock();

  public boolean lockProject() {
    return projectLock.tryLock();
  }

  public void unLockProject() {
    projectLock.unlock();
  }

  @Override
  public ITreeStorage getBaseStorage() {
    // TODO Auto-generated method stub
    return null;
  }

  public void clearAll() {
    try {
      project.deleteMarkers(null, true, IResource.DEPTH_INFINITE);
      resourceMap.clearCache();
      resinited = false;
      getCacheStorage().remove();
      cs = new CacheStorage(this, "");
    } catch (Exception e) {
      e.printStackTrace();
    }
  }
  @Override
  public void clearCachedCache(UUID uuId) {
    boolean b1=false, b2=false;
    try{
      writeLock();
      b1 = true;
      super.clearCachedCache(uuId);
      getVersionsStorage().remove(uuId);
      getResourceMap().writeLock();
      b2 = true;
      IFile res = getResource(uuId, false);
      if (res != null) {
        getResourceMap().cacheRemove(res, project);
      }
      activeAttributes.remove(uuId);
      clearPersistent(uuId);
    }finally{
      if(b1)
        writeUnlock();
      if(b2)
        getResourceMap().writeUnlock();
    }
    
    //resourceMap.clearCache();

    //resinited = false;
    //resourceMap = null;
  }
//
//  @Override
//  public String getCacheVar(UUID uuid, String section, String var) {
//    return cs.getCacheValue(uuid, section, var);
//    //return getPersistent(uuid, section, var);
//  }
//
//  @Override
//  public void setCacheVars(UUID uid, String section, Map<String, Map<String,Object>> ts) {
//    Map<String, Map<String,Object>> oldAttributes = getCacheVars(uid, section, ts.keySet());
//    HashSet<String> allS = new HashSet<>();
//    allS.addAll(ts.keySet());
//    if (oldAttributes != null) allS.addAll(oldAttributes.keySet());
//    //for(String s:allS)
//    if (getDeltaTarget() != null)
//      getDeltaTarget()
//          .cacheChanged(
//              uid,
//              section,
//              allS,
//              oldAttributes != null ? oldAttributes : null,
//              ts,
//              Moment.BEFORE,
//              true);
//    cs.setCacheMaps(uid, section, ts);
//    //setPersistents(uid, section, ts);
//    //for(String s:allS)
//    if (getDeltaTarget() != null)
//      getDeltaTarget()
//          .cacheChanged(
//              uid,
//              section,
//              allS,
//              oldAttributes != null ? oldAttributes : null,
//              ts,
//              Moment.AFTER,
//              true);
//  }
//
//  @Override
//  public Map<String, Map<String,Object>> getCacheVars(UUID uid, String section, Set<String> keySet) {
//    return cs.getCacheMaps(uid, section, keySet);
//  }
//
//  @Override
//  public void setCacheVar(UUID uuid, String section, String var, String value) {
//    String old = getCacheVar(uuid, section, var);
//    Set<String> s = new HashSet<>();
//    s.add(var);
//    Map<String, Map<String,Object>> ov = new HashMap<>(), nv = new HashMap<>();
//    if(old!=null){
//      ov.put(var, new HashMap<>());
//      ov.get(var).put("", old);
//    }
//    nv.put(var, new HashMap<>());
//    nv.get(var).put("", value);
//    if (getDeltaTarget() != null)
//      getDeltaTarget().cacheChanged(uuid, section, s, ov, nv, Moment.BEFORE, true);
//    
//    if (getDeltaTarget() != null)
//      getDeltaTarget().cacheChanged(uuid, section, s, ov, nv, Moment.AFTER, true);
//  }

  @Override
  public boolean saveUUIDs() {
    return getResourceMap().saveUUIDs(true);
  }

  public void initRoot() {
    getResourceMap().initRoot(project);
  }

  @Override
  public boolean areUUIDsChanged() {
    return getResourceMap().changed;
  }

  @Override
  public File getResource(UUID uuid, String resName) {
    IFolder resFolder = getResourceFolder(uuid);
    if (resFolder == null || !resFolder.exists()) return null;
    //if (!resFolder.isSynchronized(1)) {
//    try {
//      resFolder.refreshLocal(1, null);
//    } catch (Exception e) {
//      e.printStackTrace();
//    }
    //}
    IFile f = resFolder.getFile(resName);
    if (f == null || !f.exists()) return null;
//    if (!f.isSynchronized(0))
//      try {
//        f.refreshLocal(0, null);
//      } catch (CoreException e) {
//        e.printStackTrace();
//      }
    return new java.io.File(f.getLocationURI());
  }

  @Override
  public java.nio.file.Path getLocation() {
    return project.getLocation().toFile().toPath();
  }
}
