package com.unitesk.requality.eclipse.cvs;

import java.io.File;
import java.util.ArrayList;

import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IProject;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.runtime.IProgressMonitor;

import com.unitesk.requality.core.FileSpaceTimePoint;

public interface ICVSClient {
  /**
   * Получить идентификатор клиента (должен быть уникальным среди всех клиентов)
   *
   * @return идентификатор клиента
   */
  public String getID();

  /**
   * Полчить содержимое файла заданной ревизии
   *
   * @param filePathInRepo -- путь до файла относительно крневой папки репозитория
   * @param rev -- ревизия, искомого файла
   * @param monitor -- прогресс-монитор, в котором нужно отображать пройцесс выполнения операции
   *     (может принимать значение null)
   * @return -- содержимой файла
   * @throws Exception в случае, возниконвения проблем с соединенийм или репозиторием (например,
   *     файл данной ревизии отсутствует, так как носит другое имя)
   */
  public String getFileContentFromRepoS(String filePathInRepo, String rev, IProgressMonitor monitor)
      throws Exception;

  /**
   * Получить содержимое файла заданной ревизии с указанием стержневой ревизии
   *
   * @param filePathInRepo -- путь до файла относительно крневой папки репозитория
   * @param pegRev -- стержневая ревизия
   * @param revOfIntrst -- ревизия, искомого файла
   * @param monitor -- прогресс-монитор, в котором нужно отображать пройцесс выполнения операции
   *     (может принимать значение null)
   * @return -- содержимой файла
   * @throws Exceptionв в случае, возниконвения проблем с соединенийм или репозиторием
   */
  public String getFileContentFromRepoS(
      String filePathInRepo, String pegRev, String revOfIntrst, IProgressMonitor monitor)
      throws Exception;

  /**
   * Получить список ревизий для заданного файла (ресурса) проекта
   *
   * @param componentFile -- файл (ресурс) проекта, для которого нужно ролучить список ревизий
   * @param monitor -- прогресс-монитор, в котором нужно отображать пройцесс выполнения операции
   *     (может принимать значение null)
   * @return список "пространсвенно-временных" точек файла (тройки путь до файла-ревизия-метка
   *     времени создания/изменения файла)
   * @throws Exception в случае, возниконвения проблем с соединенийм или репозиторием
   */
  public ArrayList<FileSpaceTimePoint> getResourceRevisions(
      IResource componentFile, IProgressMonitor monitor) throws Exception;

  /**
   * Получить список ревизий для заданного файла (ресурса) проекта, начиная с заданной ревизии
   *
   * @param componentFile -- файл (ресурс) проекта, для которого нужно ролучить список ревизий
   * @param startRev -- ревизия, с которой нужно начать
   * @param monitor -- прогресс-монитор, в котором нужно отображать пройцесс выполнения операции
   *     (может принимать значение null)
   * @return список "пространсвенно-временных" точек файла (тройки путь до файла-ревизия-метка
   *     времени создания/изменения файла)
   * @throws Exception в случае, возниконвения проблем с соединенийм или репозиторием
   */
  public ArrayList<FileSpaceTimePoint> getResourceRevisions(
      IResource componentFile, String startRev, IProgressMonitor monitor) throws Exception;

  /**
   * Связать SVC клиента с заданныйм проектом
   *
   * @param proj -- проект, с которым нужно связать SVC клиента
   * @throws Exception в случае, возниконвения проблем с соединенийм или репозиторием
   */
  public void init(IProject proj) throws Exception;

  /**
   * Узнать ревизию локальной копии файла
   *
   * @param localFile -- путь к файлу проекта, для которого нужно узнать, в какой ревизии он
   *     находится
   * @return ревизию, в которой находится заданный локальный файл
   * @throws Exception в случае, возниконвения проблем с соединенийм или репозиторием
   */
  public String getLocalCopyRevision(IResource localFile) throws Exception;

  /**
   * Накатить изменения в файле в репозиторий
   *
   * @param file -- файл, чьи изменения нужно накатить
   * @param commitMessage -- сообщение, которое нужно ассойиировать с накатываемыми изменениями
   * @param monitor -- прогресс-монитор, в котором нужно отображать пройцесс выполнения операции
   *     (может принимать значение null)
   * @throws Exception в случае, возниконвения проблем с соединенийм или репозиторием (например,
   *     локальный файл находится в более ранней ревизии, чем его копия в репозитории, или файл
   *     вообще был удалён из репозитория)
   */
  public void commit(IFile file, String commitMessage, IProgressMonitor monitor) throws Exception;

  /**
   * Получить УРС репозитория
   *
   * @return -- УРС репозитория
   */
  public String getRepoURL();

  /**
   * Получить относительный путь в репозитории, и номер его последней ревизии для локального файла
   *
   * @param componentFile -- локальный файл, для которого нужно получить информацию
   * @return массив строк с информацией; в ячейке с индексом 0 находится строка, хранящая
   *     относительный путь файла в репозитории, в ячейке 1 -- номер ревизии
   * @throws Exception в случае, возниконвения проблем с соединенийм или репозиторием (например,
   *     несоответствие транспортных протоколов, или УРС репозитория)
   */
  public String[] getRelativePathAndLatestRevForAFile(IFile componentFile) throws Exception;
  /**
   * Исключить из рассмотрания клиента репозитория (Team provider) директорию .reqcsv
   *
   * @throws Exception В случае возникновения проблем с репозиторием или попыткой выполнить эту
   *     операцию до init
   */
  public void excludeDotReqCSVDir() throws Exception;

  File getContainerFromRepoS(String filePathInRepo, String rev, IProgressMonitor monitor)
      throws Exception;
}
