/*
 * Copyright 2015 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.unitesk.requality.eclipse.handlers.document;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;
import java.util.TreeSet;
import java.util.UUID;
import org.eclipse.core.commands.AbstractHandler;
import org.eclipse.core.commands.ExecutionEvent;
import org.eclipse.core.commands.ExecutionException;
import org.eclipse.core.commands.operations.IUndoContext;
import org.eclipse.ui.PlatformUI;
import com.unitesk.requality.core.IDeleteHandler;
import com.unitesk.requality.core.Path;
import com.unitesk.requality.core.TreeDB;
import com.unitesk.requality.core.TreeNode;
import com.unitesk.requality.core.attribute.ReferenceValue;
import com.unitesk.requality.eclipse.ui.dialogs.BigMessageDialog;
import com.unitesk.requality.nodetypes.Comment;
import com.unitesk.requality.nodetypes.DocFolder;
import com.unitesk.requality.nodetypes.Document;
import com.unitesk.requality.nodetypes.Location;
import com.unitesk.requality.nodetypes.Requirement;
import com.unitesk.requality.nodetypes.TextNode;
import com.unitesk.requality.tools.DeepFirstTreeWalker;

public class DeleteDocumentHandler extends AbstractHandler implements IDeleteHandler {
  private static final String
      WARN_LOCATIONS = "Selected Document (one or more) contains marked Locations.",
      WARN_DOCUMENTS = "Document and Document Folder deletion cannot be undone.";
  private boolean isDocUpdating = false;

  public String titleOverride(List<TreeNode> nodes) {
    for (TreeNode node : nodes) if (!DocFolder.TYPE_NAME.equals(node.getType())) return null;
    return "Document Folder";
  }

  public void recUpdateProcessed(TreeNode node, List<UUID> processed) {
    processed.add(node.getUUId());
    for (UUID uid : node.getWorkChildrenUUIDs()) {
      TreeNode nde = node.getTreeDB().getNode(uid);
      recUpdateProcessed(nde, processed);
    }
  }

  //start transaction and commit, please!
  private void recDelNode(TreeNode node, List<UUID> processed) {
    if (processed.contains(node.getUUId())) return;
    if (node instanceof Document) {
      deleteDocument((Document) node, processed);
    }

    if (node instanceof DocFolder) {
      while (node.sizeWorkChildren() > 0) {
        recDelNode(node.getWorkChild(0), processed);
      }
      node.delete();
    }
  }

  List<Requirement> reqs_to_delete;
  //start transaction and commit, please!
  private void deleteDocument(Document doc, List<UUID> processed) {
    if (doc.attrs().getAttribute(Document.ATTR_IS_UPDATING) != null) {
      isDocUpdating = true;
      doc.attrs().removeAttribute(Document.ATTR_IS_UPDATING);
      doc.commitAttributes();
    }
    removeLocationsFromRequirement(doc, processed);

    TreeDB db = doc.getTreeDB();
    DocFolder folder = (DocFolder) db.getNode(Document.getTypeRootQId(), Path.QID);

    if (isDocUpdating) {
      Document anotherDoc = findStarredDoc(folder, db);
      if (anotherDoc != null) {
        anotherDoc.attrs().removeAttribute(Document.ATTR_IS_UPDATING);
        anotherDoc.commitAttributes();
      }
    }
    doc.delete();
  }

  public Document findStarredDoc(DocFolder folder, TreeDB db) {
    Iterator<TreeNode> it = folder.workIterator();
    while (it.hasNext()) {
      TreeNode next = it.next();
      if (next instanceof DocFolder) {
        return findStarredDoc((DocFolder) next, db);
      } else if (next instanceof Document) {
        Document doc = (Document) next;
        if (doc.attrs().getAttribute(Document.ATTR_IS_UPDATING) != null) {
          return doc;
        }
      }
    }
    return null;
  }

  private boolean folderHasDocumentWithLocation(DocFolder folder) {
    if (folder.sizeWorkChildren() == 0) {
      return false;
    }

    for (int i = 0; i < folder.sizeWorkChildren(); i++) {
      TreeNode node = folder.getWorkChild(i);
      if (node instanceof DocFolder) {
        if (folderHasDocumentWithLocation((DocFolder) node)) {
          return true;
        }
      }
      if (node instanceof Document) {
        Document doc = (Document) node;
        if (doc.sizeLocations() > 0) {
          return true;
        }
      }
    }
    return false;
  }

  //start transaction and commit, please!
  private void removeLocationsFromRequirement(Document doc, List<UUID> processed) {
    TreeDB db = doc.getTreeDB();
    Set<UUID> locQIds = new HashSet<>();
    for (int i = 0; i < doc.sizeWorkChildren(); i++) {
      locQIds.add(doc.getWorkChild(i).getUUId());
    }

    if (isDocUpdating) {
      DeepFirstTreeWalker dftw = new DeepFirstTreeWalker(db.getRootNode(), true, true, false);
      Iterator<TreeNode> it = dftw.iterator();
      while (it.hasNext()) {
        TreeNode tn = it.next();
        if (tn.getType().equals(Requirement.TYPE_NAME)
            || TextNode.TYPE_NAME.equals(tn.getTreeDB())) {
          Requirement req = (Requirement) tn;
          if (req.attrs().getAttribute(Requirement.ATTR_TRANSF_STATUS) != null) {
            req.attrs().removeAttribute(Requirement.ATTR_TRANSF_STATUS);
            req.commitAttributes();
          }
        }
      }
    }

    DeepFirstTreeWalker dftw =
        new DeepFirstTreeWalker(doc.getTreeDB().getNode(Requirement.getTypeRootQId(), Path.QID));
    Iterator<TreeNode> it = dftw.iterator();
    ArrayList<Location> removeLocations = new ArrayList<>();
    while (it.hasNext()) {
      TreeNode node = it.next();
      if (processed.contains(node.getUUId())) continue;
      if (!node.isDisposed()
          && (Requirement.TYPE_NAME.equals(node.getType())
              || TextNode.TYPE_NAME.equals(node.getTreeDB()))) {
        Collection<ReferenceValue> locs = ((Requirement) node).getLocationQIds();
        boolean process = false;
        if (locs != null && !locs.isEmpty()) {
          for (ReferenceValue s : locs){
            if(s.hasErrorsOnResolve(node.attrs()))
              continue;
            UUID[] uids = s.getNodeUUIDs(node.attrs()); 
            if (locQIds.contains(uids[0])) { //TODO can be wrong
              process = true;
              removeLocations.add((Location)node.getTreeDB().getNode(uids[0]));
            }
          }
          Requirement req = (Requirement) node;
          if (!removeLocations.isEmpty()) {
            for (Location loc : removeLocations) {
              req.removeLocation(loc);
              processed.add(loc.getUUId());
            }
          }
          if (process) {
            if (((req.getWorkChildrenUUIDs().size() - req.getWorkChildren(Comment.TYPE_NAME).length)
                    == 0)
                && !reqs_to_delete.contains(req)
                && req.getLocationQIds().size() == 0
                && req.getTreeDB().getTreeLogic().canBeDeleted(req)) reqs_to_delete.add(req);
            req.commitAttributes();
          }
        }
      }
    }
  }

  @Override
  public boolean isSuitableFor(TreeNode tn) {
    return tn.getType().equals(Document.TYPE_NAME) || tn.getType().equals(DocFolder.TYPE_NAME);
  }

  @Override
  public void process(List<TreeNode> nodes, List<UUID> processed, boolean multi)
      throws ExecutionException {
    int todelete = 0;
    for (TreeNode node : nodes) {
      if (processed.contains(node.getUUId())) continue;
      if (!isSuitableFor(node) || !node.getTreeDB().getTreeLogic().canBeDeleted(node)) continue;
      todelete++;
    }
    if (todelete == 0) {
      return;
    }
    reqs_to_delete = new LinkedList<>();
    for (TreeNode node : nodes) {
      if (processed.contains(node.getUUId())) continue;
      if (!node.getTreeDB().getTreeLogic().canBeDeleted(node)) continue;
      if (node instanceof DocFolder) {
        List<UUID> toprocess = new ArrayList<>();
        recUpdateProcessed(node, toprocess);
        recDelNode(node, processed);
        processed.addAll(toprocess);
      }
      if (node instanceof Document) {

        // TODO: document delete undo
        // node.getTreeDB().startTransaction("Delete Document");
        node.getTreeDB().startChangeRow();
        try {
          List<UUID> toprocess = new ArrayList<>();
          recUpdateProcessed(node, toprocess);
          deleteDocument((Document) node, processed);
          processed.addAll(toprocess);
        } finally {
          node.getTreeDB().stopChangeRow();
        }
        // node.getTreeDB().commit();
      }
    }
    IUndoContext context = PlatformUI.getWorkbench().getOperationSupport().getUndoContext();
    PlatformUI.getWorkbench()
        .getOperationSupport()
        .getOperationHistory()
        .dispose(context, true, true, true);
    //TODO message
    if (reqs_to_delete.size() > 0
        && !(reqs_to_delete.get(0).getTreeDB().inTransaction())) {
      String strList = "";
      for (Requirement req : reqs_to_delete) strList += req.getUserVisibleName(true) + "\n";
      if (strList.length() > 0) strList = strList.substring(0, strList.length() - 1);
      if (BigMessageDialog.open(
          PlatformUI.getWorkbench().getActiveWorkbenchWindow().getShell(),
          "Delete Requirement" + ((reqs_to_delete.size() > 1) ? "s" : ""),
          "Requirement" + (reqs_to_delete.size() == 1 ? "" : "s"),
          strList,
          ((reqs_to_delete.size() > 1) ? " has" : " have") + " no locations, delete?")) {
        TreeDB db = reqs_to_delete.get(0).getTreeDB();
        if (db.startTransaction(
            "Delete Requirement" + (reqs_to_delete.size() > 1 ? "s" : ""), true)) {
          for (Requirement req : reqs_to_delete) {
            if (!processed.contains(req.getUUId())) {
              recUpdateProcessed(req, processed);
              req.delete();
            }
          }
          db.commit();
        }
      }
    }
  }

  @Override
  public String[] getWarnings(List<TreeNode> nodes) {
    Set<String> warns = new TreeSet<>();
    for (TreeNode node : nodes) {
      if (DocFolder.TYPE_NAME.equals(node.getType())) {
        warns.add(WARN_DOCUMENTS);
        if (folderHasDocumentWithLocation((DocFolder) node)) warns.add(WARN_LOCATIONS);
      }
      if (Document.TYPE_NAME.equals(node.getType())) {
        warns.add(WARN_DOCUMENTS);
        if (((Document) node).sizeLocations() > 0) warns.add(WARN_LOCATIONS);
      }
    }
    return warns.toArray(new String[0]);
  }

  @Override
  public Object execute(ExecutionEvent event) throws ExecutionException {
    return null;
  }
}
