/*
 * Copyright 2015 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.unitesk.requality.eclipse.handlers.report;

import java.io.File;
import java.io.IOException;
import java.net.URI;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.List;
import java.util.regex.Matcher;
import org.apache.commons.httpclient.util.URIUtil;
import org.eclipse.core.commands.AbstractHandler;
import org.eclipse.core.commands.ExecutionEvent;
import org.eclipse.core.commands.ExecutionException;
import org.eclipse.core.commands.IHandler;
import org.eclipse.core.resources.IContainer;
import org.eclipse.core.resources.IFile;
import org.eclipse.core.runtime.IPath;
import org.eclipse.core.runtime.Path;
import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.swt.widgets.DirectoryDialog;
import org.eclipse.ui.PlatformUI;
import com.unitesk.requality.core.CoreUtils;
import com.unitesk.requality.core.TreeNode;
import com.unitesk.requality.core.selection.SelectionManager;
import com.unitesk.requality.eclipse.Activator;
import com.unitesk.requality.eclipse.core.EclipseTracker;
import com.unitesk.requality.nodetypes.ReportData;
import com.unitesk.requality.report.Template;
import com.unitesk.requality.tools.RequalityCLI;

// TODO report undo
public class ExportReportHandler extends AbstractHandler implements IHandler {
  private static String DIALOG_TITLE = "Report Export",
      DIALOG_MESSAGE = "Select folder for report(s) to be exported to";

  public static IContainer getReportFolder(TreeNode node) {
    IFile nodeFile = (IFile) EclipseTracker.getResourceByNode(node);
    IContainer res = nodeFile.getParent().getFolder(Path.fromOSString(node.getId() + "_resources"));
    return res;
  }

  @Override
  public Object execute(ExecutionEvent event) throws ExecutionException {
    List<TreeNode> nodes = SelectionManager.getOne().getSelectedNodes();
    if (nodes == null) return null;
    boolean manyNodes = nodes.size() != 1;
    DirectoryDialog dd =
        new DirectoryDialog(PlatformUI.getWorkbench().getActiveWorkbenchWindow().getShell());
    dd.setText(DIALOG_TITLE);
    dd.setMessage(DIALOG_MESSAGE);
    String exportDir = dd.open();
    if (exportDir == null) return null;
    File f = new File(exportDir);
    if (!f.exists()) {
      MessageDialog.openError(
          PlatformUI.getWorkbench().getActiveWorkbenchWindow().getShell(),
          "Report export",
          "Target folder does not exists.");
      return null;
    }
    for (TreeNode node : nodes) {
      if (node.attrs().getAttribute(ReportData.ATTR_LAST_GENERATED) == null) return null;
      if (node.getType().equals(ReportData.TYPE_NAME)) {
        if (!((ReportData) node).reportExists()) return null;
      }
    }
    for (TreeNode node : nodes) {
      ReportData report = null;
      if (node instanceof ReportData) {
        report = (ReportData) node;
        exportReport(report, manyNodes ? node.getId() : null, f);
      }
    }
    return null;
  }

  private static SimpleDateFormat ft = new SimpleDateFormat("yyyy.MM.dd HH-mm-ss");

  public static void exportReport(ReportData node, String folder, File exportFolder) {
    URI href = null;
    try {
      URI savePath = node.getSaveDir();
      String filename = node.attrs().getStringAttributeValue(ReportData.ATTR_LAST_GENERATED);
      href = URI.create(savePath.toString() + "/" + URIUtil.encodeQuery(filename));
      Date dNow = new Date();
      File repfolder =
          new File(
              exportFolder.getAbsolutePath()
                  + IPath.SEPARATOR
                  + node.getId()
                  + "-"
                  + ft.format(dNow));
      ;
      try {
        RequalityCLI.copyfile(
            null,
            href,
            URI.create(repfolder.toURI().toURL() + "/" + URIUtil.encodeQuery(filename)),
            true);
        String content = RequalityCLI.getFileContent(new File(href));
        boolean changes = false;
        Matcher imgs = RequalityCLI.getImagePattern().matcher(content);//TODO mb remove
        while (imgs.find()) {
          if (imgs.group(1).equals("img") && imgs.group(2).equals("src")
              || imgs.group(1).equals("object") && imgs.group(2).equals("data")) {
            String src = CoreUtils.URLEncode(imgs.group(4));
            if (src.startsWith("data:image")) continue;
            try {
              File f2 = new File(repfolder.getAbsolutePath() + IPath.SEPARATOR + src);
              File f =
                  RequalityCLI.copyfile(
                      null,
                      new URI(savePath + "/" + URIUtil.encodeQuery(src)),
                      new URI(repfolder.toURI().toURL() + "/" + URIUtil.encodeQuery(src)),
                      true);
              if (!f.getName().equals(f2.getName())) {
                content =
                    content.replace(
                        imgs.group(0), imgs.group(0).replace(f2.getName(), f.getName()));
                changes = true;
              }
            } catch (IOException ioe) {
              //image cannot be copied
            }
          }
        }
        if (changes) RequalityCLI.setFileContent(new File(href), content);
      } catch (IOException ioe) {
        MessageDialog.openError(
            PlatformUI.getWorkbench().getActiveWorkbenchWindow().getShell(),
            "Report export",
            "File access error appears. Export is aborted.");
      }

    } catch (Exception e) {
      e.printStackTrace();
    }
  }

  public static Template getTemplate(String tid) {
    return Activator.getDefault().getReportTemplates().get(tid);
  }

  @Override
  public boolean isEnabled() {
    List<TreeNode> selected = SelectionManager.getOne().getSelectedNodes();
    if (selected == null || selected.size() == 0) return false;
    for (TreeNode node : selected) {
      if (node.attrs().getAttribute(ReportData.ATTR_LAST_GENERATED) == null) return false;
      if (node.getType().equals(ReportData.TYPE_NAME)) {
        if (!((ReportData) node).reportExists()) return false;
      }
    }
    return super.isEnabled();
  }
}
