/*
 * Copyright 2015 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.unitesk.requality.eclipse.ui.cnf;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.SortedMap;
import java.util.TreeMap;
import java.util.UUID;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Status;
import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.jface.util.LocalSelectionTransfer;
import org.eclipse.jface.viewers.ITreeSelection;
import org.eclipse.swt.dnd.DropTargetEvent;
import org.eclipse.swt.dnd.TransferData;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.ui.PlatformUI;
import org.eclipse.ui.navigator.CommonDropAdapter;
import org.eclipse.ui.navigator.CommonDropAdapterAssistant;
import com.unitesk.requality.core.ITreeLogic.IdContainsWrongChars;
import com.unitesk.requality.core.ITreeLogic.IdEmpty;
import com.unitesk.requality.core.ITreeLogic.PastingIntoLeafException;
import com.unitesk.requality.core.ITreeLogic.SameIdExistsInChildren;
import com.unitesk.requality.core.ITreeLogic.WrongProjectException;
import com.unitesk.requality.core.SortingSettings.SortingMethod;
import com.unitesk.requality.core.TreeDB;
import com.unitesk.requality.core.TreeLogic;
import com.unitesk.requality.core.TreeNode;
import com.unitesk.requality.core.instancer.Instantiator;
import com.unitesk.requality.core.nodeiterators.ReuseRequalityChildIterator;
import com.unitesk.requality.eclipse.Activator;
import com.unitesk.requality.nodetypes.Comment;
import com.unitesk.requality.nodetypes.Requirement;
import com.unitesk.requality.nodetypes.TestPurpose;
import com.unitesk.requality.nodetypes.TextNode;
import com.unitesk.requality.nodetypes.VirtualNode;
import com.unitesk.requality.tools.DeepFirstTreeWalker;
import com.unitesk.requality.tools.Index;
import com.unitesk.requality.tools.RequalityCLI;

public class CNFDropAssistant extends CommonDropAdapterAssistant {
  private List<TreeNode> srcNodes;

  public CNFDropAssistant() {
    // TODO Auto-generated constructor stub
  }

  @Override
  public IStatus validateDrop(Object target, int operation, TransferData transferType) {
    if (!isSupportedType(transferType)) {
      return Status.CANCEL_STATUS;
    }
    if (target instanceof TreeNode) {
      LocalSelectionTransfer transfer = LocalSelectionTransfer.getTransfer();
      ITreeSelection selection = (ITreeSelection) transfer.getSelection();
      if (selection == null) {
        return Status.CANCEL_STATUS;
      }
      srcNodes = new ArrayList<>();
      Object[] select = selection.toArray();
      for (Object sel : select) {
        if (sel instanceof TreeNode) {
          srcNodes.add((TreeNode) sel);
        } else {
          return Status.CANCEL_STATUS;
        }
      }
      if (select.length == 1 && select[0] == target) return Status.CANCEL_STATUS;
      return Status.OK_STATUS;
    }
    return Status.CANCEL_STATUS;
  }

  private static String[] allowedAlt =
      new String[] {
        Requirement.TYPE_NAME, TextNode.TYPE_NAME, Comment.TYPE_NAME, TestPurpose.TYPE_NAME
      };

  @Override
  public IStatus handleDrop(
      CommonDropAdapter aDropAdapter, DropTargetEvent aDropTargetEvent, Object aTarget) {
    if (srcNodes == null) {
      return Status.CANCEL_STATUS;
    }
    if (!(aTarget instanceof TreeNode)) {
      return Status.CANCEL_STATUS;
    }
    List<TreeNode> workList = new ArrayList<>();
    SortedMap<UUID, String> quids = new TreeMap<>();
    TreeDB db = srcNodes.get(0).getTreeDB();
    for (TreeNode node : srcNodes) {
      quids.put(node.getUUId(), node.getQualifiedId());
      //it is not possible to move nodes over different treedbs
      if (node.getTreeDB() != db && (node.getTreeDB() != Instantiator.getRootTree(db)))
        return Status.CANCEL_STATUS;
    }
    TreeNode basetarget = (TreeNode) aTarget, target = basetarget;
    TreeNode lastNode = null;
    SortedMap<String, Integer> cnt = new TreeMap<>();
    boolean alt = RequalityCNF.altPressed;
    for (int ii = 0; ii < srcNodes.size(); ii++) {
      TreeNode node = srcNodes.get(ii);
      if (alt) {
        boolean validSt = false;
        for (String s : allowedAlt) if (s.equals(node.getType())) validSt = true;
        if (!validSt) return Status.CANCEL_STATUS;
      }
      boolean add = true;
      for (int j = 0; j < ii; j++) {
        TreeNode n2 = srcNodes.get(j);
        if (n2 != node
            && (quids.get(node.getUUId()) + "/").contains((quids.get(n2.getUUId())) + "/")) {
          add = false;
          break;
        }
      }
      if (add) {
        //TODO check
        if (node.isVirtual() && VirtualNode.TYPE_NAME.equals(node.getWorkParent().getType())) {
          if (((VirtualNode) node.getWorkParent()).isCollapsed()
              && node.getWorkParent().getWorkChildrenUUIDs().size() == 1)
            node = node.getWorkParent();
        }
        if (!node.isVirtual())
          node = Instantiator.getRootTree(node.getTreeDB()).getNode(node.getUUId());
        workList.add(node);
        if (!cnt.containsKey(node.getType())) cnt.put(node.getType(), 0);
        cnt.put(node.getType(), cnt.get(node.getType()) + 1);
      }
      TreeLogic l = node.getTreeDB().getTreeLogic();
      if ((!l.canBeChildOf(node.getType(), basetarget.getType())
              && l.canBeChildOf(node.getType(), basetarget.getWorkParent()))
          || RequalityCNF.shiftPressed) {
        lastNode = basetarget;
        target = basetarget.getWorkParent();
      }
    }
    if (lastNode == null && target.sizeShownChildren() > 0)
      lastNode = target.getSortedShownChildren()[target.sizeShownChildren() - 1];
    if (workList.size() == 0) return Status.CANCEL_STATUS;
    TreeLogic l = workList.get(0).getTreeDB().getTreeLogic();
    String generalssting = workList.size() > 1 ? "s" : "";
    String op = "Move";
    if (alt) op = "Reuse";
    for (TreeNode srcNode : workList) {
      boolean adds = cnt.get(srcNode.getType()) > 1;
      try {
        if (!l.canBeCopiedToExc(srcNode, target, true)) {
          MessageDialog.openError(
              PlatformUI.getWorkbench().getActiveWorkbenchWindow().getShell(),
              op + " Element" + generalssting,
              (adds ? srcNode.getMultiTypeLabel() : srcNode.getTypeLabel())
                  + " cannot be "
                  + op.toLowerCase()
                  + "d to '"
                  + target.getUserVisibleName(true)
                  + "'");
          return Status.CANCEL_STATUS;
        }
      } catch (PastingIntoLeafException e) {
        MessageDialog.openError(
            PlatformUI.getWorkbench().getActiveWorkbenchWindow().getShell(),
            op + " Element" + generalssting,
            "Can't "
                + op.toLowerCase()
                + " "
                + (adds ? srcNode.getMultiTypeLabel() : srcNode.getTypeLabel())
                + " to "
                + target.getTypeLabel()
                + " with "
                + (srcNode.getType().equals(TestPurpose.TYPE_NAME) ? "non-" : "")
                + "Test Purpose child");
        return Status.CANCEL_STATUS;
      } catch (WrongProjectException e) {
        MessageDialog.openError(
            PlatformUI.getWorkbench().getActiveWorkbenchWindow().getShell(),
            op + " Element" + generalssting,
            op.replace("e", "").toLowerCase() + "ing elements between projects is prohibited");
        return Status.CANCEL_STATUS;
      } catch (SameIdExistsInChildren e) {
        //not possible here
      } catch (IdEmpty e) {
        MessageDialog.openError(
            PlatformUI.getWorkbench().getActiveWorkbenchWindow().getShell(),
            op + " Element" + generalssting,
            "Element id is empty");
        return Status.CANCEL_STATUS;
      } catch (IdContainsWrongChars e) {
        MessageDialog.openError(
            PlatformUI.getWorkbench().getActiveWorkbenchWindow().getShell(),
            op + " Element" + generalssting,
            "Element id contains wrong characters :" + e.getWrongCharacter());
        return Status.CANCEL_STATUS;
      }
      for (TreeNode node : new DeepFirstTreeWalker(srcNode, true, true, false)) {
        if (node.getQualifiedId().equals(target.getQualifiedId())) {
          return Status.CANCEL_STATUS;
        }
      }
      if (target.getTreeDB().getTreeLogic().checkVNodeRecursion(srcNode, target)) {
        //target.getTreeDB().rollback();
        MessageDialog.openError(
            getShell(),
            op + " Element" + generalssting,
            "Some of selected elements contains Virtual Node that points to target or its parents. So "
                + op.toLowerCase().substring(0, op.length() - 1)
                + "ing it will cause recursion and will be prohibited. ");
        return Status.CANCEL_STATUS;
      }
    }
    if (!alt) {
      String fromNames = "";
      List<Integer> renIndex = new ArrayList<>();
      int ccc = 0;
      for (int i = 0; i < workList.size(); i++)
        if (!l.isChildIdValid(target, workList.get(i), workList.get(i).getId())) {
          renIndex.add(i);
          if (!workList
              .get(i)
              .getTreeDB()
              .getTreeLogic()
              .isAutoRenamable(workList.get(i).getType())) {
            fromNames += workList.get(i).getId() + ", ";
            ccc++;
          }
        }
      String purposes = "", adds = ccc > 1 ? "s" : "";
      byte purpcnt = 0;
      if (!fromNames.equals("")) {
        purposes +=
            "\n- Element"
                + adds
                + " with id"
                + adds
                + " "
                + fromNames.substring(0, fromNames.length() - 2)
                + " already exist"
                + (adds.equals("") ? "s" : "")
                + " in "
                + target.getUserVisibleName(true)
                + ".";
        purpcnt++;
      }
      boolean askrename = true;
      for (int i = 0; i < workList.size() && askrename; i++) {
        if (l.isAutoRenamable(workList.get(i).getType())) continue;
        for (int j = i + 1; j < workList.size(); j++)
          if (workList.get(i).getId().equals(workList.get(j).getId())) {
            if (l.isAutoRenamable(workList.get(i).getType())) continue;
            purposes += "\n- Selection contains elements with same ids.";
            purpcnt++;
            ccc += 2;
            askrename = false;
            break;
          }
      }
      Shell shell = Activator.getDefault().getWorkbench().getActiveWorkbenchWindow().getShell();
      if (!purposes.equals(""))
        if (!MessageDialog.openConfirm(
            shell,
            op + " Element" + generalssting,
            op
                + "d element"
                + (generalssting.equals("") ? " has" : generalssting + " have")
                + " id conflict"
                + (purpcnt > 1 ? "s" : "")
                + ":\n"
                + purposes
                + "\n"
                + "\nConflicting element"
                + adds
                + " will be renamed. If you still want to proceed click OK, otherwise click Cancel.")) {
          return Status.CANCEL_STATUS;
        }
    }
    TreeDB tdb = workList.get(0).getTreeDB();
    if (alt) {
      //TODO look
      //tdb = Instantiator.getVirtualTreeDB(tdb,Instantiator.DEFAULT_INSTANCE_NAME);
    }
    if (tdb.startTransaction(op + " Element" + generalssting, true)) {
      try{
        boolean flaglast = false;
        if (target.sizeShownChildren() > 0
            && lastNode == target.getSortedShownChildren()[target.sizeShownChildren() - 1])
          flaglast = true;
        if (target != lastNode) {
          TreeNode[] sorted = target.getSortedWorkChildren();
          for (int j = 0; j < sorted.length; j++) {
            Index lastInd;
            if (j > 0) lastInd = Index.create(Integer.toString(j + 1));
            else lastInd = Index.createOne();
            sorted[j].setIndex(lastInd);
            sorted[j].commitAttributes();
          }
        }
        for (int i = 0; i < workList.size(); i++) {
          String possibleId = "";
          boolean isIdValid = true;
          if (!alt) {
            isIdValid =
                target
                    .getTreeDB()
                    .getTreeLogic()
                    .isChildIdValid(target, workList.get(i), workList.get(i).getId());
            if (!isIdValid) {
              Iterable<TreeNode> it = null;
              String tp = "";
              if (workList.get(i).getType().equals(Comment.TYPE_NAME)) {
                tp = Comment.TYPE_NAME + " ";
                it = Arrays.asList(target.getWorkChildren());
              }
              possibleId = RequalityCLI.getNextId(it, target, workList.get(i).getType(), tp);
              workList
                  .get(i)
                  .setId(
                      RequalityCLI.getNextId(
                          Arrays.asList(workList.get(i).getWorkParent().getWorkChildren()),
                          target,
                          workList.get(i).getType(),
                          ""));
              workList.get(i).commitAttributes();
            }
          }
          if (!alt
              && !target.getTreeDB().getTreeLogic().canBeChildOf(workList.get(i), target, false)) {
            target.getTreeDB().rollback();
            MessageDialog.openError(
                getShell(),
                op + " Element" + generalssting,
                "Selected elements cannot be " + op.toLowerCase() + "d together. ");
            return Status.CANCEL_STATUS;
          }
  
          if (alt) {
            VirtualNode vn =
                (VirtualNode)
                    target.createChild(
                        RequalityCLI.getNextId(target, VirtualNode.TYPE_NAME, null),
                        VirtualNode.TYPE_NAME);
            try {
              vn.setTarget(workList.get(i));
              vn.setChildIterator(new ReuseRequalityChildIterator(null, vn.getUUId(), "0"));
              vn.collapse();
              vn.commitAttributes(); //TODO dont updated
              TreeNode child = vn.getWorkChild(0);
              //child.setName(workList.get(i).getNameOrId());
              //child.saveAttributes();
              if (target != lastNode) {
                child.setIndex(
                    Index.getNextIndex(
                        vn.getWorkChild(0).getType(),
                        target,
                        lastNode,
                        vn.getWorkChild(0).getType().equals("TestStep")));
                vn.setIndex(child.getIndex());
                child.commitAttributes();
                vn.commitAttributes();
              } else if (target.getTreeDB().getTreeLogic().usesIndexByDefault(child.getType())
                  || target.getTreeDB().getSortingSettings().getMethod() == SortingMethod.BY_INDEX) {
                vn.getWorkChild(0)
                    .setIndex(
                        Index.getNextIndex(
                            child.getType(), target, lastNode, child.getType().equals("TestStep")));
                vn.setIndex(child.getIndex());
                vn.getWorkChild(0).commitAttributes();
                vn.commitAttributes();
              }
              if (flaglast && target.sizeWorkChildren() > 0) {
                lastNode = target.getSortedWorkChildren()[target.sizeWorkChildren() - 1];
              } else {
                lastNode = child;
              }
            } catch (Exception e) {
              e.printStackTrace();
            }
          } else {
            if (workList.get(i).getWorkParent() != target) workList.get(i).setParent(target);
            if (target != lastNode) {
              Index ind =
                  Index.getNextIndex(
                      workList.get(i).getType(),
                      target,
                      lastNode,
                      workList.get(i).getType().equals("TestStep"));
              if (ind != Index.getUndefined()
                      && (db.getSortingSettings().getMethod() == SortingMethod.BY_INDEX)
                  || db.getTreeLogic().usesIndexByDefault(workList.get(i).getType())) {
                workList.get(i).setIndex(ind);
                workList.get(i).commitAttributes();
              }
            } else if (target.getTreeDB().getTreeLogic().usesIndexByDefault(workList.get(i).getType())
                || target.getTreeDB().getSortingSettings().getMethod() == SortingMethod.BY_INDEX) {
              workList
                  .get(i)
                  .setIndex(
                      Index.getNextIndex(
                          workList.get(i).getType(),
                          target,
                          lastNode,
                          workList.get(i).getType().equals("TestStep")));
              workList.get(i).commitAttributes();
            }
            if (flaglast && target.sizeWorkChildren() > 0) {
              lastNode = target.getShownChildren()[target.sizeShownChildren() - 1];
            } else {
              lastNode = workList.get(i);
            }
          }
          if (!isIdValid){
            workList.get(i).setId(possibleId);
            workList.get(i).commitAttributes();
          }
        }
        tdb.commit();
        return Status.OK_STATUS;
      }
      catch(Exception e){
        if (tdb.inTransaction()) tdb.rollback();
      }
    }
    return Status.CANCEL_STATUS;
  }
}
