/*
 * Copyright 2014-2015 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.unitesk.requality.eclipse.views.documents;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.regex.Pattern;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import com.unitesk.requality.core.RLogger;

/** Class for representing document as plain text */
public class DocStringRepresentation {

  protected Document doc;
  protected StringBuilder strRep = new StringBuilder();
  protected List<SpanInfo> spans = new ArrayList<>();
  protected List<NodeLocation> tagsList = new ArrayList<>();
  protected Map<Node, SpanStartEnd> tagsMap = new HashMap<>();
  protected List<Integer> tagStartToTagNumber = new ArrayList<>();

  protected Set<String> addedSpanIds = new HashSet<>();
  int transferredSpansCounter = 0;
  int foundButNotTransferredCounter = 0;
  int foundInFewNodesCounter = 0;
  int transferredFromNextDOMTags = 0;

  public Map<Node, Node> debugRemovedNodes = new HashMap<>();
  private boolean verbose = false;

  public DocStringRepresentation(final Document doc) {
    this.doc = doc;
    buildString(doc, true);
    if (verbose) {
      RLogger.logInfo("spans size = " + spans.size());
    }
    Collections.sort(tagsList);
    for (int i = 0; i < tagsList.size(); ++i) {
      tagStartToTagNumber.add(tagsList.get(i).locationPosition.start);
    }
  }

  public void buildString(final Node currentNode, boolean removeNewLines) {
    int start = strRep.length();
    boolean appendNewLine = false;

    if (currentNode.getNodeType() == Node.TEXT_NODE) {
      String text = currentNode.getNodeValue();
      append(text, removeNewLines);
    } else {
      Pattern BLOCK_ELEMENT_RE = Pattern.compile("p|h[0-9]|br|div|tr|td", Pattern.CASE_INSENSITIVE);
      String nodeName = currentNode.getNodeName();
      if (currentNode.getNodeType() == Node.ELEMENT_NODE
          && BLOCK_ELEMENT_RE.matcher(nodeName).matches()) {
        removeNewLines = true;

        appendNewLine = true;
      }

      NodeList nl = currentNode.getChildNodes();

      for (int i = 0; i < nl.getLength(); i++) {
        buildString(nl.item(i), removeNewLines);
      }

      if (currentNode.getNodeName().compareTo("span") == 0) {
        Node clazz = currentNode.getAttributes().getNamedItem("class");
        if (clazz != null) {
          String spanId = clazz.getNodeValue();
          if (spanId.startsWith("requality_text ")) {
            if (strRep.substring(start).length() > 0) {
              spans.add(
                  new SpanInfo(
                      new SpanStartEnd(start, strRep.length()), spanId, strRep.substring(start)));
              if (verbose) {
                RLogger.logInfo("added span <" + strRep.substring(start) + ">");
              }
            }
          }
        }
      }
      if (appendNewLine) {
        appendNewLine();
      }
    }

    SpanStartEnd newLoc = new SpanStartEnd(start, strRep.length());
    NodeLocation nodeLoc = new NodeLocation(newLoc, currentNode);
    tagsList.add(nodeLoc);
    tagsMap.put(currentNode, newLoc);
  }

  private void append(String text, boolean removeNewLines) {
    if (removeNewLines) {
      //			text = text.replaceAll("\r?\n", " ").replaceAll("\t|\\s{2,}", " ")/*.trim()*/;
      //			text = text.replaceAll("\r?\n", " ").replaceAll("\t", " ").replaceAll("\\s{2,}", " "); //.replaceAll("\\s[,\\.]", "!");
      text = getRawString(text);
      text = text.replaceAll("\\s[\\.]", ".").replaceAll("\\s[,]", ",");
      //			text = text.replaceAll("\r?\n", " ").replaceAll("\t", " ").replaceAll("\\s{2,}", " ").replaceAll("\\s,", ",");
    }
    if (strRep.length() > 0 && (lastChar() == ' ' /* || lastChar() == '\n'*/)) {
      text = text.replaceFirst("^\\s+", "");
    }

    if (strRep.length() > 0 && (lastChar() == ' ' /* || lastChar() == '\n'*/)) {
      //			text = text.replaceFirst("^[,\\.]", "");
      if (text.startsWith(",") || text.startsWith(".") || text.startsWith(";")) {
        strRep.deleteCharAt(strRep.length() - 1);
      }
    }
    strRep.append(text);
  }

  private void appendNewLine() {
    //		if (strRep.length() > 0 && (lastChar() == ' '/* || lastChar() == '\n'*/) ) {
    //			strRep.setLength(strRep.length()-1);
    //		}
    strRep.append("\r\n");
  }

  private int getTagNumber(int positionStart) {

    int low = 0;
    int last = tagStartToTagNumber.size() - 1;
    int high = last;

    while (low <= high) {
      int mid = low + (high - low) / 2;
      int midVal = tagStartToTagNumber.get(mid);

      int next = (mid < last) ? tagStartToTagNumber.get(mid + 1) : tagStartToTagNumber.get(last);
      if ((midVal <= positionStart) && (positionStart <= next)) {
        low = mid;
        break;
      }

      if (positionStart < midVal) {
        high = mid - 1;
      } else if (positionStart > midVal) {
        low = mid + 1;
      }
    }

    // moving to first DOM node with the same start
    if (low > last) {
      low = last;
    }
    while ((low > 0) && (tagStartToTagNumber.get(low).equals(tagStartToTagNumber.get(low - 1)))) {
      --low;
    }

    return low;
  }

  private char lastChar() {
    return strRep.charAt(strRep.length() - 1);
  }

  public String getCleanString() {
    return strRep.toString();
  }

  public List<SpanInfo> getSpans() {
    return spans;
  }

  public List<NodeLocation> getTags() {
    return tagsList;
  }

  public boolean DocModify(
      final SpanToTransfer spanToInsertPositionInNewFile, final String spanId) {
    Node node = null;
    String spanText =
        strRep
            .toString()
            .substring(
                spanToInsertPositionInNewFile.getLocation().getStart(),
                spanToInsertPositionInNewFile.getLocation().getEnd());

    if (spanText.trim().length() == 0) {
      return false;
    }

    int domTagNumber = getTagNumber(spanToInsertPositionInNewFile.getLocation().start);

    node = tagsList.get(domTagNumber).getNode();
    if (node == null) {
      return false;
    }

    boolean result = false;
    if (node.getNodeValue() != null
        && (node.getNodeValue().contains(spanText)
            || getRawString(node.getNodeValue()).contains(spanText))) {
      if (!node.hasChildNodes() && node.getNodeName().compareTo("#text") == 0) {
        result = updateForSpanNodes(spanToInsertPositionInNewFile, spanId, node, spanText);
        return result;
      }
      return false;

    } else {
      // TODO: append few tags from tagList and update few nodes
      RLogger.logInfo("in else block foundButNotTransferredCounter");
      ++foundButNotTransferredCounter;
      int offset = 1;
      if ((domTagNumber + offset) >= tagsList.size()) {
        return false;
      }
      Node nextNode = tagsList.get(domTagNumber + offset).getNode();

      String strToTransferSpan = "";

      int counter = 0;

      //        	while (!strToTransferSpan.replaceAll("[\\r\\n\\s]+", " ").contains(spanText)) {
      while (!getRawString(strToTransferSpan).contains(spanText)) {
        if (verbose) {
          RLogger.logInfo(
              "strToTransferSpan = <"
                  + strToTransferSpan.replaceAll("[\\r\\n\\s]+", "@")
                  + "> === span = "
                  + spanText);
        }

        ++counter;
        if ((domTagNumber + offset) >= tagsList.size()) {
          return false;
        }
        nextNode = tagsList.get(domTagNumber + offset).getNode();
        if (nextNode.getNodeValue() != null) {
          if (!nextNode.getNodeValue().equals("\n")) {
            strToTransferSpan += nextNode.getNodeValue();
          }
        }
        ++offset;

        if (counter > 25) {
          if (verbose) {
            RLogger.logInfo("Text has not been found in 25 appended nodes");
          }
          return false;
        }
      }
      foundInFewNodesCounter++;
      RLogger.logInfo("in foundInFewNodesCounter");

      // understanding how many nodes we should change
      // currently working if span is only in one node
      int numberOfNodesToUpdate = 0;
      String str = "";
      for (int i = offset - 1; i >= 0; --i) {
        Node last = tagsList.get(domTagNumber + i).getNode();
        str = last.getNodeValue() + str;
        ++numberOfNodesToUpdate;
        //        		if (str.contains(spanText) || str.replaceAll("[\\r\\n\\s]+", " ").contains(spanText)) {
        if (str.contains(spanText) || getRawString(str).contains(spanText)) {
          break;
        }
      }

      if (numberOfNodesToUpdate == 1) {
        nextNode = tagsList.get(domTagNumber + offset - 1).getNode();
        result = updateForSpanNodes(spanToInsertPositionInNewFile, spanId, nextNode, spanText);
        if (result) {
          ++transferredFromNextDOMTags;
        }

      } else {

        RLogger.logInfo("TODO: update few DOM nodes " + numberOfNodesToUpdate);
        // TODO update few nodes in DOM
      }

      return result;
    }
  }

  private boolean updateForSpanNodes(
      final SpanToTransfer locationPositionInNewFile,
      final String spanId,
      Node node,
      String spanText) {

    Node currentNode = node;

    if (verbose) {
      RLogger.logInfo("Node = " + currentNode);
    }

    while (debugRemovedNodes.keySet().contains(currentNode)) {
      currentNode = debugRemovedNodes.get(currentNode);
    }

    if (verbose) {
      RLogger.logInfo("currentNode = " + currentNode);
      RLogger.logInfo("spanText = <" + spanText + ">");
    }

    if (currentNode == null) {
      if (verbose) {
        RLogger.logInfo(
            "in updateForSpanNodes for span <"
                + spanText
                + ">, in if block with currentNode is null");
      }
      // if here - than the text of next span is contains in one of the previous, so currently we find wrong location
      return false;
    }

    SpanStartEnd locationOfNodeInDoc = tagsMap.get(currentNode);

    if (locationOfNodeInDoc == null) {
      if (verbose) {
        RLogger.logInfo("returning false as tagsMap doesn't contain currentNode");
      }
      return false;
    }

    if (!((locationPositionInNewFile.getLocation().getStart() >= locationOfNodeInDoc.getStart())
        || (locationOfNodeInDoc.getEnd() >= locationPositionInNewFile.getLocation().getEnd()))) {
      return false;
    }

    SpanStartEnd occurencePosition =
        getOccurence(locationPositionInNewFile, locationOfNodeInDoc, currentNode, spanText);
    if (occurencePosition == null) {
      return false;
    }

    Node left = null;
    try {
      left =
          doc.createTextNode(
              currentNode.getTextContent().substring(0, occurencePosition.getStart()));
    } catch (StringIndexOutOfBoundsException e) {
      left = null;
    }

    Node right = null;
    try {
      right =
          doc.createTextNode(currentNode.getTextContent().substring(occurencePosition.getEnd()));
    } catch (StringIndexOutOfBoundsException e) {
      right = null;
    }

    Element newNode = doc.createElement("span");
    newNode.setAttribute("class", spanId);
    Node newTextNode = doc.createTextNode(spanText);
    newNode.appendChild(newTextNode);

    if (verbose) {
      RLogger.logInfo("appended child " + newTextNode);
    }

    Node parent = currentNode.getParentNode();

    if (left != null) {
      parent.insertBefore(left, currentNode);
    }

    parent.insertBefore(newNode, currentNode);
    ++transferredSpansCounter;
    boolean isFirstSpanOccurence = addedSpanIds.add(spanId);
    if (isFirstSpanOccurence) {
      // adding tag <a>
      Element tagA = doc.createElement("a");

      tagA.setAttribute("class", "requality_id");
      tagA.setAttribute("id", spanId.substring(15));
      tagA.setAttribute("name", spanId.substring(18));

      Node emptyTextNode = doc.createTextNode("");
      tagA.appendChild(emptyTextNode);
      newNode.insertBefore(tagA, newTextNode);
    }

    if (right != null) {
      parent.insertBefore(right, currentNode);

      debugRemovedNodes.put(currentNode, right);

      int newEnd = locationOfNodeInDoc.getEnd();
      
      SpanStartEnd modifiedLoc = new SpanStartEnd(newEnd - right.getTextContent().length(), newEnd);
  
      // modify location in tagsMap
      tagsMap.put(right, modifiedLoc);
      if (verbose) {
        RLogger.logInfo("Changing tagsMap in:");
        RLogger.logInfo("        [" + right + "] -> [" + modifiedLoc + "]");
      }
    }

    return true;
  }

  private SpanStartEnd getOccurence(
      SpanToTransfer locationPositionInNewFile,
      SpanStartEnd locationOfNodeInDoc,
      Node currentNode,
      String spanText) {
    int startOccurenceIndex =
        locationPositionInNewFile.getLocation().getStart() - locationOfNodeInDoc.getStart();
    int endOccurenceIndex =
        locationPositionInNewFile.getLocation().getEnd() - locationOfNodeInDoc.getStart();

    if (startOccurenceIndex < 0 || endOccurenceIndex < 0) {
      if (verbose) {
        RLogger.logInfo(
            "Error while finding indexes of spanText in DOM node text, ignore this span");
      }
      return null;
    }

    try {
      String spanAsTextSegment =
          currentNode.getTextContent().substring(startOccurenceIndex, endOccurenceIndex);
      if (verbose) {
        RLogger.logInfo("spanAsTextSegment = <" + spanAsTextSegment + ">");
        RLogger.logInfo("spanText = <" + spanText + ">");
      }

      //        	if (! (spanAsTextSegment.replaceAll("[\\r\\n\\s]+", " ")).equals(spanText)) {
      if (!getRawString(spanAsTextSegment).equals(spanText)) {
        if (verbose) {
          RLogger.logInfo("spanText and spanAsTextSegment are not equals");
        }

        String nodeTextContentWithoutMultipleSpaces =
            currentNode.getTextContent().replaceAll("\\s{2,}", " ");

        if (nodeTextContentWithoutMultipleSpaces.contains(spanText)) {
          // we should modify start end end indexes for update in DOM of new document
          if (verbose) {
            RLogger.logInfo("Trying to transfer this requirement");
          }

          // TODO do we need this check?
          int fromIndex = nodeTextContentWithoutMultipleSpaces.indexOf(spanText);
          if (fromIndex == -1) {
            return null;
          }
          // check for another occurrence  - if found - it is ambiguous, ignore it
          if (fromIndex < nodeTextContentWithoutMultipleSpaces.length()
              && nodeTextContentWithoutMultipleSpaces.indexOf(spanText, fromIndex + 1) != -1) {
            if (verbose) {
              RLogger.logInfo("Ambiguous occurence");
            }
            return null;
          }

          String firstSpanTextWord = spanText.split("\\s")[0];
          startOccurenceIndex = currentNode.getTextContent().indexOf(firstSpanTextWord, fromIndex);
          fromIndex += spanText.length();

          String[] wordsArray = spanText.split("\\s");
          int len = wordsArray[wordsArray.length - 1].length();
          String lastSpanTextOccur = spanText.substring(spanText.length() - len);
          endOccurenceIndex =
              currentNode.getTextContent().indexOf(lastSpanTextOccur, fromIndex - len)
                  + lastSpanTextOccur.length();

          String newSpanString =
              currentNode.getTextContent().substring(startOccurenceIndex, endOccurenceIndex);
          if (verbose) {
            RLogger.logInfo("newSpanString = <" + newSpanString + ">");
          }
          int spanTextLen = spanText.length();
          int newSpanStringLen = newSpanString.length();

          //    				if (! newSpanString.replaceAll("[\\r\\n\\s]+", " ").equals(spanText) ||
          if (!getRawString(newSpanString).equals(spanText)
              || !newSpanString.substring(0, 1).equals(spanText.substring(0, 1))
              || !newSpanString
                  .substring(newSpanStringLen - 1, newSpanStringLen)
                  .equals(spanText.substring(spanTextLen - 1, spanTextLen))) {
            //    					if (! newSpanString.replaceAll("[\\r\\n\\s]+", " ").equals(spanText)) {\
            if (!getRawString(newSpanString).equals(spanText)) {
              if (verbose) {
                RLogger.logInfo("Case 1");
              }
            }
            if (!newSpanString.substring(0, 1).equals(spanText.substring(0, 1))) {
              if (verbose) {
                RLogger.logInfo("Case 2");
              }
            }
            if (!newSpanString
                .substring(newSpanStringLen - 1, newSpanStringLen)
                .equals(spanText.substring(spanTextLen - 1, spanTextLen))) {
              if (verbose) {
                RLogger.logInfo(
                    "newSpanString.substring(newSpanStringLen - 1, newSpanStringLen) = "
                        + newSpanString.substring(newSpanStringLen - 1, newSpanStringLen));
                RLogger.logInfo(
                    "spanText.substring(spanTextLen-1, spanTextLen) = "
                        + spanText.substring(spanTextLen - 1, spanTextLen));
              }
            }

            if (verbose) {
              RLogger.logInfo("Failed in requirement transfer");
            }
            return null;
          }

        } else {
          if (verbose) {
            RLogger.logInfo(
                "Something went wrong with extracting span text from DOM node, "
                    + "span texts are totally not equal, ignore this span");
          }

          return null;
        }
      }

    } catch (StringIndexOutOfBoundsException e) {
      if (verbose) {
        RLogger.logInfo(
            "Something went wrong with extracting span text from DOM node, ignore this span");
      }
      return null;
    }

    return new SpanStartEnd(startOccurenceIndex, endOccurenceIndex);
  }

  void deleteRemovedNodes() {
    for (Node node : debugRemovedNodes.keySet()) {
      Node parent = node.getParentNode();
      if (parent != null) {
        parent.removeChild(node);
      }
    }
  }

  private String getRawString(String text) {
    return text.replaceAll("[\\r\\t\\n\\s]+", " ").replaceAll("\\s{2,}", " ");
  }
}
