/*
 * Copyright 2011-2015 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.unitesk.requality.eclipse.views.documents;

import java.io.BufferedInputStream;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.io.PrintStream;
import java.io.Reader;
import java.io.Writer;
import java.util.HashMap;
import java.util.Map;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Status;
import org.eclipse.jface.dialogs.DialogSettings;
import org.eclipse.jface.dialogs.ErrorDialog;
import org.eclipse.jface.dialogs.IDialogConstants;
import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.jface.wizard.WizardDialog;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.ui.IWorkbench;
import org.eclipse.ui.PlatformUI;
import org.json.JSONException;
import com.unitesk.requality.core.Path;
import com.unitesk.requality.core.RLogger;
import com.unitesk.requality.core.TreeDB;
import com.unitesk.requality.core.TreeNode;
import com.unitesk.requality.documents.IDocumentProcessor;
import com.unitesk.requality.documents.IProgressMonitor;
import com.unitesk.requality.eclipse.RequalityPlugin;
import com.unitesk.requality.eclipse.core.ReqSettings;
import com.unitesk.requality.eclipse.ui.dialogs.BigMessageDialog;
import com.unitesk.requality.eclipse.ui.dialogs.SelectDocDialog;
import com.unitesk.requality.eclipse.wizards.DocWizard;
import com.unitesk.requality.nodetypes.Document;

// TODO: headless document processing support

public abstract class DocumentProcessor implements IDocumentProcessor {
  public static boolean eclipseGUI = false;
  protected String encoding = "UTF-8";
  //    protected boolean showSuperMessageBox = true;
  protected boolean showMessages = true;
  protected String messageBoxTitle = null;
  protected String messageBoxMessage = null;
  protected DocWizard wiz = null;

  public void setShowMessages(boolean v) {
    showMessages = v;
  }

  public void toggleShowMessages() {
    showMessages = !showMessages;
  }

  @Override
  public int[] getCounterMax() {
    return new int[] {1000, 0, 400};
  };

  @Override
  public void setEncoding(String encoding) {
    this.encoding = encoding;
  }

  int[] progress = new int[] {0, 0, 0};
  private IProgressMonitor monitor;

  private int sumCountersMax() {
    int[] maxs = getCounterMax();
    int sum = 0;
    for (int i = 0; i < 3; i++) sum += maxs[i];
    return sum;
  }

  protected IProgressMonitor getMonitor() {
    return monitor;
  }

  private int lastTargetWorked = 0;

  private void updateProgressMonitor() {
    int s = 0;
    for (int i = 0; i < 3; i++) s += progress[i];
    if (monitor != null) monitor.worked(s - lastTargetWorked);
    lastTargetWorked = s;
  }

  @Override
  public void setCounterValues(int lc, int pc, int sc) {
    progress[0] = lc;
    progress[1] = pc;
    progress[2] = sc;
    updateProgressMonitor();
  }

  @Override
  public int[] getCounterValues() {
    return progress;
  }

  protected void cancelError(String t1, String t2, Status st) {
    if (restoretempl && templ != null)
      for (String s : templ.keySet()) doc.getTreeDB().setActiveTemplate(s, null);
    showErrorBox(t1, t2, st);
  }

  protected void cancelMessage(String s1, String s2) {
    if (restoretempl && templ != null)
      for (String s : templ.keySet()) doc.getTreeDB().setActiveTemplate(s, null);
    showMessageBox(s1, s2);
  }

  @Override
  public void process(String docQId, TreeNode rootNode,PrintStream ps, IProgressMonitor monitor)
      throws IOException {
    //Shell parent = PlatformUI.getWorkbench().getActiveWorkbenchWindow()
    //      .getShell();

    progress = new int[] {0, 0, 0};
    lastTargetWorked = 0;
    TreeDB db = rootNode.getTreeDB();
    if (db.inTransaction()) {
      cancelError(
          "Document Processor Cancelled",
          "Document Processor Cancelled: Project catalog is used by other process",
          null);
      return;
    }
    if (db.startTransaction("Document Processor", true)) {
      try {
        Document doc = (Document) db.getNode(docQId, Path.QID);
        String htmlName = doc.getResourceName();
        MonitoredInputStream inStream =
            new MonitoredInputStream(doc.getResourceContent(htmlName), 0);
        //inStream = new BufferedInputStream(doc.getResourceContent(htmlName));
        this.monitor = monitor;
        Reader reader = new InputStreamReader(inStream, encoding);
        if (monitor != null)
          monitor.beginTask("Document '" + docQId + "' processing", sumCountersMax());
        ByteArrayOutputStream byteOut = new ByteArrayOutputStream();
        Writer writer = new OutputStreamWriter(byteOut, encoding);
        if (!isCanceled() && process(docQId, reader, writer, rootNode, ps)) {
          writer.close();
          reader.close();
          inStream.close();
          if (isCanceled()) {
            writer.close();
            reader.close();
            db.rollback();
            cancelMessage("Document Processor", "Document processing canceled");
            return;
          }
          MonitoredInputStream outIn =
              new MonitoredInputStream(new ByteArrayInputStream(byteOut.toByteArray()), 2);
          doc.setResourceContent(htmlName, outIn, new ProcessorData());
          outIn.close();
          if (isCanceled()) {
            writer.close();
            reader.close();
            db.rollback();
            cancelMessage("Document Processor", "Document processing canceled");
            return;
          }
          int[] cv = getCounterValues();
          cv[2] = getCounterMax()[2];
          setCounterValues(cv[0], cv[1], cv[2]);
          db.commit();

          if (messageBoxTitle != null && messageBoxMessage != null) {
            showMessageBox(messageBoxTitle, messageBoxMessage);
          } else {
            showMessageBox("Document Processor", "Document processing finished");
          }
        } else {
          writer.close();
          reader.close();
          db.rollback();
          if (isCanceled()) cancelMessage("Document Processor", "Document processing is cancelled");
        }

      } catch (final Exception e) {
        e.printStackTrace();
        db.rollback();
        cancelError(
            "Document Processor Error",
            "Document Processor Error: " + e.getLocalizedMessage(),
            new Status(IStatus.ERROR, RequalityPlugin.ID, 0, e.getMessage(), null));
      }
    } else {
      cancelError(
          "Document Processor Error",
          "Some actions are already being performed",
          new Status(IStatus.ERROR, RequalityPlugin.ID, 0, "transaction started", null));
    }
    if (restoretempl && templ != null)
      for (String s : templ.keySet()) doc.getTreeDB().setActiveTemplate(s, null);
  }

  protected boolean isCanceled() {
    if (monitor == null) return false;
    return monitor.isCanceled();
  }

  private void setCounterValues(int[] vals) {
    setCounterValues(vals[0], vals[1], vals[2]);
  }

  private class MonitoredInputStream extends BufferedInputStream {
    int max;
    int param;

    public MonitoredInputStream(InputStream in, int param) {
      super(in);
      this.param = param;
      try {
        max = available();
      } catch (IOException e) {
        e.printStackTrace();
      }
    }

    @Override
    public synchronized int read(byte[] b, int off, int len) throws IOException {
      int r = super.read(b, off, len);
      int mx = getCounterMax()[param];
      int[] cv = getCounterValues();
      cv[param] = max == 0 ? 0 : (mx - available() * mx / max);
      setCounterValues(cv);
      return r;
    };
  }

  protected abstract boolean process(String docId, Reader reader, Writer writer, TreeNode rootNode, PrintStream ps)
      throws IOException;

  private boolean restoretempl = false;
  private Map<String, String> templ;
  private Document doc;

  @Override
  public boolean configure(Document doc, Map<String,Object> params, PrintStream pw) {
    Map<String, String> templ = doc.getTreeDB().getActiveTemplates();
    boolean hastempl = false;
    for (String s : templ.keySet()) {
      if (templ.get(s) != null && !templ.get(s).isEmpty()) {
        hastempl = true;
        this.doc = doc;
        break;
      }
    }
    if (hastempl)
      if (!showQuestionBox(
          "Several templates are active",
          "WARNING. One or more templates are active for current project",
          "Please press ok if you want it to stay active.",
          "Or press cancel if you want to disable it for current document processing.")) {
        for (String s : templ.keySet()) doc.getTreeDB().setActiveTemplate(s, null);
        restoretempl = true;
      }
    return true;
  }

  protected void showBIGMessageBox(final String messageBoxTitle, final String messageBoxMessage) {
    if (!eclipseGUI || !showMessages) {
      RLogger.logInfo(messageBoxTitle + ": " + messageBoxMessage);
    } else {
      IWorkbench w = PlatformUI.getWorkbench();
      w.getDisplay()
          .syncExec(
              new Runnable() {
                public void run() {
                  Shell activeShell =
                      PlatformUI.getWorkbench().getActiveWorkbenchWindow().getShell();
                  BigMessageDialog.open(
                      activeShell,
                      messageBoxTitle,
                      "",
                      messageBoxMessage,
                      "",
                      new String[] {IDialogConstants.OK_LABEL});
                }
              });
    }
  }

  protected void showMessageBox(final String messageBoxTitle, final String messageBoxMessage) {
    if (!eclipseGUI || !showMessages) {
      RLogger.logInfo(messageBoxTitle + ": " + messageBoxMessage);
    } else {
      IWorkbench w = PlatformUI.getWorkbench();
      w.getDisplay()
          .syncExec(
              new Runnable() {
                public void run() {
                  Shell activeShell =
                      PlatformUI.getWorkbench().getActiveWorkbenchWindow().getShell();
                  MessageDialog.openInformation(activeShell, messageBoxTitle, messageBoxMessage);
                }
              });
    }
  }

  protected void showErrorBox(final String title, final String message, final Status status) {
    if (!eclipseGUI || !showMessages || !showMessages) {
      RLogger.logInfo(title + ": " + message);
    } else {
      IWorkbench w = PlatformUI.getWorkbench();
      w.getDisplay()
          .asyncExec(
              new Runnable() {
                public void run() {
                  Shell activeShell =
                      PlatformUI.getWorkbench().getActiveWorkbenchWindow().getShell();
                  ErrorDialog.openError(
                      activeShell, title, message, status == null ? Status.OK_STATUS : status);
                }
              });
    }
  }

  private static Boolean lastquestionboxvalue;

  protected Boolean showQuestionBox(
      final String title, final String label1, final String message, final String label2) {
    lastquestionboxvalue = false;
    if (eclipseGUI) {
      IWorkbench w = PlatformUI.getWorkbench();
      w.getDisplay()
          .syncExec(
              new Runnable() {

                @Override
                public void run() {
                  Shell shell = PlatformUI.getWorkbench().getActiveWorkbenchWindow().getShell();
                  lastquestionboxvalue =
                      com.unitesk.requality.eclipse.ui.dialogs.BigMessageDialog.open(
                          shell, title, label1, message, label2);
                }
              });
    } else {
      //TODO ask for keypress
      lastquestionboxvalue = false;
      RLogger.logInfo(title + ": " + label1 + " " + message + " " + label2);
    }
    return lastquestionboxvalue;
  }

  private Document tempprefferedDoc;

  protected Document showSelectAnotherDocumentDialog(final Document doc, Document prefferedDoc) {
    return showSelectAnotherDocumentDialog(doc, prefferedDoc, false);
  }

  protected Document showSelectAnotherDocumentDialog(
      final Document doc, Document prefferedDoc, boolean forcePreffered) {
    tempprefferedDoc = null;
    if (!eclipseGUI || !showMessages || !showMessages || forcePreffered) {
      return prefferedDoc;
    } else {
      IWorkbench w = PlatformUI.getWorkbench();
      w.getDisplay()
          .syncExec(
              new Runnable() {
                @Override
                public void run() {
                  Shell shell = PlatformUI.getWorkbench().getActiveWorkbenchWindow().getShell();
                  SelectDocDialog dialog =
                      new SelectDocDialog(shell, doc.getTreeDB()) {
                        @Override
                        public boolean canSelect() {
                          if (!super.canSelect()) return false;
                          if (!selectedNode.getType().equals(Document.TYPE_NAME)) return false;
                          if (selectedNode.getQualifiedId().equals(doc.getQualifiedId()))
                            return false;
                          return true;
                        }
                      };
                  if (dialog.open() == SelectDocDialog.OK
                      && dialog.getResult() instanceof Document) {
                    tempprefferedDoc = (Document) dialog.getResult();
                  }
                }
              });
    }
    return tempprefferedDoc;
  }

  protected DialogSettings showWizard(String name) {
    DialogSettings ds = new DialogSettings(name);

    return ds;
  }

  @Override
  public void setWizard(Object w) {
    wiz = (DocWizard) w;
  }

  public Map<String, Object> showWizard(final Document docm) {
    if (wiz == null) return null;
    try {
      final DocWizard dw = wiz.getClass().newInstance();
      if (!eclipseGUI) return null; //TODO console
      IWorkbench w = PlatformUI.getWorkbench();
      final Map<String, Object> ret = new HashMap<>();
      w.getDisplay()
          .syncExec(
              new Runnable() {
                @Override
                public void run() {
                  Shell shell = PlatformUI.getWorkbench().getActiveWorkbenchWindow().getShell();
                  ReqSettings ds = new ReqSettings(getClass().getCanonicalName());
                  try {
                    ds.load(DocumentProcessor.this.getClass().getCanonicalName(), docm.getTreeDB());
                  } 
                  catch(FileNotFoundException fnfe){//normal case
                    Map<String, Object> def = wiz.getDefaultSettings();
                    for (String s : def.keySet()) ds.put(s, def.get(s));
                  }
                  catch (IOException | JSONException e) {
                    Map<String, Object> def = wiz.getDefaultSettings();
                    for (String s : def.keySet()) ds.put(s, def.get(s));
                    RLogger.logError(e);
                  }
                  if(ds.getMap().isEmpty())
                    ds.setMap(wiz.getDefaultSettings());
                  
                  dw.setReqSettings(ds);
                  WizardDialog wd = new WizardDialog(shell, dw);
                  long k = wd.open();
                  if (k == WizardDialog.OK) {
                    for (String s : ds.getKeys()) ret.put(s, ds.get(s));
                    try {
                      ds.save(DocumentProcessor.this.getClass().getCanonicalName());
                    } catch (IOException e) {
                      e.printStackTrace();
                    } catch (JSONException e) {
                      e.printStackTrace();
                    }
                  }
                }
              });
      return ret;
    } catch (InstantiationException e) {
    } catch (IllegalAccessException e) {
    }
    return null;
  }
}
