/*
 * Copyright 2015 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.unitesk.requality.eclipse.views.documents;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.io.PrintStream;
import java.io.PrintWriter;
import java.io.Reader;
import java.io.UnsupportedEncodingException;
import java.io.Writer;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.UUID;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.OutputKeys;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerConfigurationException;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.TransformerFactoryConfigurationError;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;
import org.eclipse.ui.PartInitException;
import org.eclipse.ui.PlatformUI;
import org.xml.sax.SAXException;
import com.unitesk.requality.core.Path;
import com.unitesk.requality.core.RLogger;
import com.unitesk.requality.core.TreeDB;
import com.unitesk.requality.core.TreeNode;
import com.unitesk.requality.core.attribute.Attribute;
import com.unitesk.requality.core.attribute.Flag;
import com.unitesk.requality.core.attribute.TypedValue;
import com.unitesk.requality.eclipse.views.documents.DiffProcessor.Diff;
import com.unitesk.requality.eclipse.views.documents.DiffProcessor.Operation;
import com.unitesk.requality.nodetypes.Document;
import com.unitesk.requality.nodetypes.Location;
import com.unitesk.requality.nodetypes.NodeType;
import com.unitesk.requality.nodetypes.Requirement;
import com.unitesk.requality.nodetypes.ToDoItem;
import com.unitesk.requality.tools.RequalityCLI;

/**
 * Handles the action in the context menu. Transfer requirements from old document to new document
 */
public class ProcessNewSpecificationVersion extends DocumentProcessor {

  private static final Object PREFFERED_DOC = "prefferedDoc";
  private static final Object FORCE = "force";
  public boolean verbose = false;
  public boolean writeDebugFiles = false;

  private Document oldDocument = null;
  private Document newDocument = null;
  private TreeDB db;

  private int diffFoundSpanCounter = 0;
  private int transferedLocCounter = 0;

  private DocStringRepresentation oldDocRep = null;
  private DocStringRepresentation newDocRep = null;

  private int outsideEqualCount = 0;
  private int partiallyOutsideEqual = 0;

  private int completeReqsCount = 0;

  private int incompleteReqsCount = 0;
  private int failedReqsCount = 0;
  private boolean showMessageBoxWithStatistics = true;

  public ToDoList toDoList;

  private Map<String, Integer> reqIdsToLocCounter = new HashMap<>();
  private Map<String, Integer> reqIdsToTransferredSpanCounter = new HashMap<>();

  public void unsetShowMessageBoxWithStats() {
    showMessageBoxWithStatistics = false;
  }

  public void setShowMessageBoxWithStats() {
    showMessageBoxWithStatistics = true;
  }

  @Override
  public boolean configure(final Document doc, Map<String, Object> params, PrintStream pw) {
    if (!super.configure(doc, params, pw)) {
      return false;
    }
    TypedValue frs = (TypedValue) params.get(FORCE);
    return configure(doc, (Document)params.get(PREFFERED_DOC), 
        frs!=null? frs.getGValue():false);
  }

  public boolean configure(Document doc, Document prefferedDoc, boolean force) {
    this.db = doc.getTreeDB();
    if (doc.attrs().getAttribute(Document.ATTR_ALREADY_UPDATED) != null) {
      super.showMessageBox(
          "Document Processor",
          "This document was already updated. " + "Documents cannot be updated more than once.");
      return false;
    }

    oldDocument = showSelectAnotherDocumentDialog(doc, prefferedDoc, force);
    return oldDocument != null;
  }

  /** This method preferred to default. Can be used in console mode. */
  public boolean configure(Document doc, Document prefferedDoc) {
    return configure(doc, prefferedDoc, false);
  }

  @Override
  protected boolean process(String docId, final Reader in, final Writer out, TreeNode rootNode, PrintStream ps) {

    newDocument = (Document) db.getNode(docId, Path.QID);

    if (isCanceled()) return false;

    boolean b = processReqsTransfer(out);
    ps.append("Processing completed");
    return b;
  }

  /**
   * Method transfers locations from old document to new document. Can be used in console mode Isn't
   * uses in eclipse plug-in mode
   */
  public void transferRequirements(Document oldDoc, Document newDoc) {

    newDocument = newDoc;
    oldDocument = oldDoc;

    this.db = oldDocument.getTreeDB();

    ByteArrayOutputStream byteOut = new ByteArrayOutputStream();
    Writer out = null;
    try {
      out = new OutputStreamWriter(byteOut, encoding);
      processReqsTransfer(out);
      out.close();
      String htmlName = newDocument.getResourceName();

      if (verbose) {
        RLogger.logInfo("htmlName = " + htmlName);
      }

      ByteArrayInputStream outIn = new ByteArrayInputStream(byteOut.toByteArray());
      newDocument.setResourceContent(htmlName, outIn, new ProcessorData());
      outIn.close();

    } catch (UnsupportedEncodingException e) {
      e.printStackTrace();
    } catch (IOException e) {
      e.printStackTrace();
    }
  }

  private boolean processReqsTransfer(final Writer out)
      throws TransformerFactoryConfigurationError {

    transferedLocCounter = 0;
    diffFoundSpanCounter = 0;

    completeReqsCount = 0;
    incompleteReqsCount = 0;
    failedReqsCount = 0;

    try {
      if (isCanceled()) {
        return false;
      }

      newDocRep = new DocStringRepresentation(newDocument.getDOM());

      if (isCanceled()) {
        return false;
      }

      oldDocRep = new DocStringRepresentation(oldDocument.getDOM());

      if (writeDebugFiles) {
        printCleanStringsToFiles();
      }

      if (isCanceled()) {
        return false;
      }

      Map<SpanInfo, List<SpanToTransfer>> locationsInNewFile = processDiff();

      if (isCanceled()) {
        return false;
      }

      boolean status = updateReqAttributes(locationsInNewFile);

      if (!status || isCanceled()) {
        return false;
      }

      status = setStatuses();

      if (!status || isCanceled()) {
        return false;
      }

      newDocRep.deleteRemovedNodes();
      status = createToDoLinks();

      if (verbose) {
        RLogger.logInfo("ToDo list items created");
      }

      if (!status) {
        return false;
      }

      if (isCanceled()) {
        return false;
      }

      newDocument.setXmlDoc(newDocRep.doc);
      newDocument.saveDOMFromXMLCache();

      if (isCanceled()) return false;

      transform(out);

      if (isCanceled()) return false;

      oldDocument.addUpdatedProblem();
      newDocument.addUpdatedProblem();

      newDocument.setXmlDoc(newDocRep.doc);
      newDocument.saveDOMFromXMLCache();

      newDocument.markDocumentAsUpdated();

      if (verbose) {
        RLogger.logInfo("newDocument.saveDOMFromXMLCache() is done");
      }

      PlatformUI.getWorkbench()
          .getDisplay()
          .syncExec(
              new Runnable() {
                public void run() {
                  if (verbose) {
                    RLogger.logInfo("In run() for ToDoList view");
                  }

                  try {
                    ToDoList viewPart;
                    viewPart =
                        (ToDoList)
                            PlatformUI.getWorkbench()
                                .getActiveWorkbenchWindow()
                                .getActivePage()
                                .showView("com.unitesk.requality.views.ToDoList");
                    viewPart.refresh(db);
                  } catch (PartInitException e) {
                    e.printStackTrace();
                  }
                }
              });

      generateTextForMessageBox();

      return true;

    } catch (TransformerConfigurationException e) {
      e.printStackTrace();
    } catch (TransformerException e) {
      e.printStackTrace();
    } catch (ParserConfigurationException e) {
      e.printStackTrace();
    } catch (IOException e) {
      e.printStackTrace();
    } catch (SAXException e) {
      e.printStackTrace();
    } catch (Exception e) {
      e.printStackTrace();
    }

    return false;
  }

  private void printCleanStringsToFiles() throws FileNotFoundException {
    PrintWriter oldCleanString = new PrintWriter(new File("oldCleanStr.txt"));
    //    		oldCleanString.print(oldDocRep.getCleanString().replaceAll("\n", "@"));
    oldCleanString.print(oldDocRep.getCleanString());
    oldCleanString.close();

    PrintWriter oldTagsFile = new PrintWriter(new File("oldTags.txt"));
    for (NodeLocation nodeLoc : oldDocRep.getTags()) oldTagsFile.println(nodeLoc);
    oldTagsFile.close();

    PrintWriter newCleanString = new PrintWriter(new File("newCleanStr.txt"));
    //	    	newCleanString.print(newDocRep.getCleanString().replaceAll("\n", "@"));
    newCleanString.print(newDocRep.getCleanString());
    newCleanString.close();
    PrintWriter newTagsFile = new PrintWriter(new File("newTags.txt"));
    for (NodeLocation nodeLoc : newDocRep.getTags()) newTagsFile.println(nodeLoc);
    newTagsFile.close();
  }

  private boolean updateReqAttributes(Map<SpanInfo, List<SpanToTransfer>> spansInNewFile) {

    for (SpanInfo spanLoc : oldDocRep.getSpans()) {
      if (verbose) {
        RLogger.logInfo("spanLoc = " + spanLoc.spanId + " " + spanLoc.spanText);
      }

      String id = spanLoc.getSpanId().substring(18);
      if (reqIdsToLocCounter.keySet().contains(id)) {
        int oldCount = reqIdsToLocCounter.get(id);
        reqIdsToLocCounter.put(id, oldCount + 1);
      } else {
        reqIdsToLocCounter.put(id, 1);
      }
    }

    if (isCanceled()) {
      return false;
    }

    Map<String, UUID> idToUUID = new HashMap<>();
    for (SpanInfo spanLoc : oldDocRep.getSpans()) {

      if (verbose) {
        RLogger.logInfo("spanLoc = " + spanLoc);
      }

      if (isCanceled()) {
        return false;
      }

      List<SpanToTransfer> spansToTransfer = spansInNewFile.get(spanLoc);
      if (spansToTransfer != null) {
        // currently there is max one element in locsToTransfer, this cycle is for
        // preposition of different variants to user
        for (SpanToTransfer spanToTransfer : spansToTransfer) {
          ++diffFoundSpanCounter;
          boolean forUpdate = false;
          if (!idToUUID.containsKey(spanLoc.getSpanId())) {
            idToUUID.put(spanLoc.getSpanId(), UUID.randomUUID());
          }
          UUID spanID = idToUUID.get(spanLoc.getSpanId());
          if (spanToTransfer.getSpanText().equals(spanLoc.getText())) {
            forUpdate =
                newDocRep.DocModify(spanToTransfer, "requality_text id_" + spanID.toString());
          } else {
            if (verbose) {
              RLogger.logInfo("Different span texts to modify. Continue.");
            }
          }

          if (forUpdate) {
            // adding locations in new doc for current requirement
            ++transferedLocCounter;
            String sspanID = spanID.toString();
            String oldspanID = spanLoc.getSpanId().substring(18);
            Location oldLoc = oldDocument.getLocation(oldspanID);

            if (reqIdsToTransferredSpanCounter.keySet().contains(oldspanID.toString())) {
              int count = reqIdsToTransferredSpanCounter.get(oldspanID.toString());
              reqIdsToTransferredSpanCounter.put(oldspanID.toString(), count + 1);
            } else {
              reqIdsToTransferredSpanCounter.put(oldspanID.toString(), 1);
            }

            Location loc;

            if (newDocument.getLocation(sspanID) == null) {
              loc = newDocument.createLocation(spanID);
              loc.commitAttributes();
              if (oldLoc != null) {
                Requirement[] reqs = oldLoc.getRequirements();
                for (Requirement req : reqs) {
                  req.addLocation(loc.getUUId());
                  req.commitAttributes();
                }
              }
            }
          }
        }
      }
    }

    return true;
  }

  private boolean setStatuses() {

    if (verbose) {
      RLogger.logInfo("reqIdsToTransferredLocCounter:");
      for (String str : reqIdsToTransferredSpanCounter.keySet()) {
        RLogger.logInfo(str + " " + reqIdsToTransferredSpanCounter.get(str));
      }
    }

    Set<Requirement> requirementsFromOldDoc = new HashSet<>();

    if (verbose) {
      RLogger.logInfo("in setStatuses ");
    }

    for (UUID uuid : oldDocument.getLocations()) {

      if (verbose) {
        RLogger.logInfo("uuid =  " + uuid);
      }

      Location location = oldDocument.getLocation(uuid.toString());

      if (location == null) {
        // TODO add checkers before update processor
        continue;
      }

      if (verbose) {
        RLogger.logInfo("location = " + location);
      }

      for (Requirement req : location.getRequirements()) {
        requirementsFromOldDoc.add(req);
      }
    }

    if (isCanceled()) {
      return false;
    }

    for (Requirement req : requirementsFromOldDoc) {
      // uncomment if we don't want to mark structure requirements without locations with statuses overlays
      //    			if (req.getLocations().isEmpty()) {
      //    			  continue;
      //    			}

      if (req != null) {
        List<Location> reqLocations = req.getLocations();
        int numberOfTransferredSpans = 0;
        int numberOfInitialSpans = 0;

        int numberOfLocationsInOldDoc = 0;
        int numberOfTransferredLocations = 0;

        for (Location loc : reqLocations) {

          Integer numberOfSpansInOldDoc = 0;
          Integer numberOfSpansInNewDoc = 0;

          if (loc.getDocument().equals(oldDocument)) {
            numberOfLocationsInOldDoc++;
            if (reqIdsToTransferredSpanCounter.keySet().contains(loc.getId())) {
              numberOfSpansInNewDoc = reqIdsToTransferredSpanCounter.get(loc.getId());
              numberOfTransferredSpans += numberOfSpansInNewDoc;
            }
            if (reqIdsToLocCounter.get(loc.getId()) != null)
              numberOfSpansInOldDoc = reqIdsToLocCounter.get(loc.getId());

            if (numberOfSpansInOldDoc == numberOfSpansInNewDoc) {
              numberOfTransferredLocations++;
            }
            numberOfInitialSpans += numberOfSpansInOldDoc;
          }
        }

        Requirement.TransferStatus statusOfCurrentReq = null;

        if (numberOfLocationsInOldDoc > 0) {
          if (numberOfTransferredLocations == numberOfLocationsInOldDoc) {
            statusOfCurrentReq = Requirement.TransferStatus.COMPLETE;
            completeReqsCount++;
          } else {
            if ((numberOfTransferredSpans > 0)
                && (numberOfTransferredSpans < numberOfInitialSpans)) {
              statusOfCurrentReq = Requirement.TransferStatus.INCOMPLETE;
              incompleteReqsCount++;
            }
            if (numberOfTransferredSpans == 0) {
              statusOfCurrentReq = Requirement.TransferStatus.FAILED;
              failedReqsCount++;
            }
          }
        }
        if (statusOfCurrentReq != null) {
          req.setTransferProblem(statusOfCurrentReq);
        }

        if (verbose) {
          RLogger.logInfo("req = " + req);
          RLogger.logInfo("numberOfTransferredSpans = " + numberOfTransferredSpans);
          RLogger.logInfo("numberOfInitialSpans = " + numberOfInitialSpans);
          RLogger.logInfo("number of transferred locations = " + numberOfTransferredLocations);
          RLogger.logInfo("number of initial locations = " + numberOfLocationsInOldDoc);
          RLogger.logInfo("status = " + statusOfCurrentReq);
        }
      }
    }

    //		 uncomment if we want to get the worst status from all children

    //		 for (TreeNode node : getOrderFromLeaves(db.getRootNode())) {
    //
    //			 if (verbose) {
    //				 RLogger.logInfo("node = " + node);
    //			 }
    //
    //			 if(isCanceled()) {
    //					return false;
    //			 }
    //
    //			 if (node instanceof Requirement) {
    //				 Requirement req = (Requirement) node;
    //				 Requirement.TransferStatus worstStatus = null;
    //
    //				 Attribute reqAttr = req.getAttribute(Requirement.ATTR_TRANSF_STATUS);
    //				 if (reqAttr != null) {
    //					 worstStatus = Requirement.TransferStatus.valueOf(reqAttr.getRawValue().toString());
    //				 }
    //
    //				 if (node.sizeChildren() > 0) {
    //					 TreeNode[] children = node.getSortedChildren();
    //					 for (TreeNode child : children) {
    //						 Attribute attr = child.getAttribute(Requirement.ATTR_TRANSF_STATUS);
    //						 if (attr != null) {
    //							 Requirement.TransferStatus childStatus = Requirement.TransferStatus.valueOf(attr.getRawValue().toString());
    //							 if (childStatus != null) {
    //								 if (worstStatus == null ) {
    //									 worstStatus = childStatus;
    //								 } else if (!childStatus.isBetterThan(worstStatus)) {
    //									 worstStatus = childStatus;
    //								 }
    //							 }
    //						 }
    //					 }
    //				 }
    //
    //				 if (worstStatus != null) {
    //					 req.setTransferProblem(worstStatus);
    //				 }
    //			 }
    //		 }
    return true;
  }

  private boolean createToDoLinks() {

    TreeNode toDoFolder = db.getNode("/ToDoLinks", Path.QID);
    if (toDoFolder == null) {
      toDoFolder = db.getRootNode().createChild("ToDoLinks", NodeType.TYPE_NAME);
    }

    Set<Requirement> reqsWithToDoItems = new HashSet<>();
    for (TreeNode node : getOrderFromLeaves(db.getNode(toDoFolder.getUUId()))) {
      if (node instanceof ToDoItem) {
        Object attr_req_link =
            UUID.fromString(node.attrs().getStringAttributeValue(ToDoItem.ATTR_REQ_LINK));
        if (attr_req_link instanceof java.util.UUID) {
          UUID reqUUID = (UUID) attr_req_link;
          TreeNode taskParent = db.getNode(reqUUID);
          if (taskParent != null) {
            if (taskParent.getType().equals(Requirement.TYPE_NAME)) {
              reqsWithToDoItems.add((Requirement) taskParent);
            }
          }
        }
      }
    }

    UUID root = db.getNode(Requirement.getTypeRootQId(), Path.QID).getUUId();
    for (TreeNode node : getOrderFromLeaves(db.getNode(Requirement.getTypeRootQId(), Path.QID))) {

      if (node.getUUId().equals(root)) {
        continue;
      }

      if (verbose) {
        RLogger.logInfo("node = " + node);
      }

      if (isCanceled()) {
        return false;
      }

      if (node instanceof Requirement) {
        Requirement req = (Requirement) node;

        if (reqsWithToDoItems.contains(req)) {
          continue;
        }

        if (req.getLocations().isEmpty()) {
          continue;
        }

        Requirement.TransferStatus status = null;

        Attribute reqAttr = req.attrs().getAttribute(Requirement.ATTR_TRANSF_STATUS);
        if (reqAttr != null) {
          status = Requirement.TransferStatus.valueOf(reqAttr.getValue().toString());
        }

        if (status != null) {
          TreeNode toDoItem =
              toDoFolder.createChild(
                  RequalityCLI.getNextId(toDoFolder, ToDoItem.TYPE_NAME, "link_"),
                  ToDoItem.TYPE_NAME);
          Task.Action action;
          switch (status) {
            case INCOMPLETE:
              action = Task.Action.ADD_LOCATIONS;
              break;
            case COMPLETE:
              action = Task.Action.VERIFY;
              break;
            case FAILED:
              action = Task.Action.FIND_LOCATIONS;
              break;
            default:
              action = Task.Action.FIND_LOCATIONS;
              break;
          }

          toDoItem.attrs().setConstAttribute(ToDoItem.ATTR_REQ_LINK, req.getUUId().toString(),0);
          toDoItem.attrs().setConstAttribute(ToDoItem.ATTR_TRANSF_ACTION, action.toString(),Flag.HIDDEN);
          toDoItem.attrs().setConstAttribute(ToDoItem.OLD_DOC, oldDocument.getQualifiedId(),Flag.HIDDEN);
          toDoItem.attrs().setConstAttribute(ToDoItem.NEW_DOC, newDocument.getQualifiedId(),Flag.HIDDEN);
          toDoItem.commitAttributes();
        }
      }
    }
    if (verbose) {
      RLogger.logInfo("Before exiting ToDoLinks creation");
    }
    return true;
  }

  private List<TreeNode> getOrderFromLeaves(TreeNode rootNode) {
    List<TreeNode> order = new ArrayList<>();
    bfs(rootNode, order);
    Collections.reverse(order);
    return order;
  }

  private void bfs(TreeNode node, List<TreeNode> queue) {
    queue.add(node);
    for (TreeNode child : node.getWorkChildren()) {
      bfs(child, queue);
    }
  }

  private Map<SpanInfo, List<SpanToTransfer>> processDiff() {

    Map<SpanInfo, List<SpanToTransfer>> locationsInNewFile =
        new HashMap<>();

    DiffProcessor differ = new DiffProcessor();
    LinkedList<Diff> diffResult;

    diffResult = differ.diff_main(oldDocRep.getCleanString(), newDocRep.getCleanString(), true);

    int indexInDiffList = 0;

    for (SpanInfo spanLoc : oldDocRep.getSpans()) {
      String spanText = spanLoc.getText();
      int spanBeginIndex = spanLoc.getSpanPosition().getStart();
      Diff df = diffResult.get(indexInDiffList);
      boolean isSpanStartInsideDiffBlock =
          (spanBeginIndex >= (df.getOriginalOffset() - (df.getText().length())))
              && (spanBeginIndex <= df.getOriginalOffset());
      while (!isSpanStartInsideDiffBlock) {
        ++indexInDiffList;
        df = diffResult.get(indexInDiffList);
        isSpanStartInsideDiffBlock =
            (spanBeginIndex >= (df.getOriginalOffset() - (df.getText().length())))
                && (spanBeginIndex <= df.getOriginalOffset());
      }
      List<SpanToTransfer> locationsToTransfer = new ArrayList<>();
      Diff currentDiff = df;
      if (currentDiff.getOperation().equals(Operation.EQUAL)) {
        int spanTextLen = spanText.length();
        if ((spanBeginIndex + spanTextLen) <= (currentDiff.getOriginalOffset())) {
          // the span is all inside EQUAL diff
          int startIndexInNewDoc =
              currentDiff.getReplacementOffset() + spanBeginIndex - currentDiff.getOriginalOffset();
          SpanToTransfer locToTransfer =
              new SpanToTransfer(startIndexInNewDoc, startIndexInNewDoc + spanTextLen, spanText);
          locationsToTransfer.add(locToTransfer);
          locationsInNewFile.put(spanLoc, locationsToTransfer);

        } else {
          // TODO upgrade for case of partially in equal block
          ++partiallyOutsideEqual;
        }
      } else {
        ++outsideEqualCount;
      }
    }

    return locationsInNewFile;
  }

  public int getOutsideEqualCount() {
    return outsideEqualCount;
  }

  public int getPartiallyOutsideEqual() {
    return partiallyOutsideEqual;
  }

  private void generateTextForMessageBox() {

    long totalReqCounter = completeReqsCount + incompleteReqsCount + failedReqsCount;
    long transferredCounter = completeReqsCount + incompleteReqsCount;
    String message = "Document processing finished.\n";

    if (totalReqCounter == 0) {
      message += "There is no extracted requirements in marked document";
    } else {

      if (transferredCounter == 0) {
        message +=
            "None of the requirements have been transferred from "
                + totalReqCounter
                + " requirements.";
      } else {
        message += "Transferred " + transferredCounter + " requirement";
        if (transferredCounter > 1) {
          message += "s";
        }
        message += " from " + totalReqCounter + ".";

        if (incompleteReqsCount > 0) {
          message += "\nWhere for " + incompleteReqsCount + " requirement";
          if (incompleteReqsCount > 1) {
            message += "s";
          }
          message += " not all text fragments have been transferred.";
        }
      }
    }

    super.messageBoxMessage = message;
    super.messageBoxTitle = "Document Processor";

    if (verbose) {
      RLogger.logInfo("From MessageBox: " + messageBoxMessage);

      RLogger.logInfo("Stats:");
      RLogger.logInfo("number of spans: " + oldDocRep.getSpans().size());
      RLogger.logInfo("diff found: " + diffFoundSpanCounter + " spans to transfer");
      RLogger.logInfo("number of transferred spans: " + transferedLocCounter);
      RLogger.logInfo("outsideEqualCount: " + outsideEqualCount);
      RLogger.logInfo("partiallyOutsideEqual: " + partiallyOutsideEqual);

      RLogger.logInfo("Partially transferred number of reqs : " + incompleteReqsCount);
      RLogger.logInfo("Totally transferred number of reqs: " + completeReqsCount);
      RLogger.logInfo("Not transferred number of reqs: " + failedReqsCount);
      RLogger.logInfo("Total number of reqs in initial doc: " + totalReqCounter);
    }
  }

  private void transform(final Writer out)
      throws Exception, TransformerFactoryConfigurationError, TransformerConfigurationException,
          TransformerException {
    DOMSource domSource = new DOMSource(newDocument.getDOM());

    StreamResult result = new StreamResult(out);
    TransformerFactory tf = TransformerFactory.newInstance();
    Transformer transformer;

    transformer = tf.newTransformer();
    transformer.setOutputProperty(OutputKeys.METHOD, "xml");
    transformer.setOutputProperty(OutputKeys.ENCODING, "UTF-8");
    transformer.transform(domSource, result);
  }

  public int getTransferedLocCounter() {
    return transferedLocCounter;
  }

  public int getDiffFoundSpanCounter() {
    return diffFoundSpanCounter;
  }

  public int getTotalSpansCounter() {
    return oldDocRep.getSpans().size();
  }

  public int getCompleteReqsCount() {
    return completeReqsCount;
  }

  public int getIncompleteReqsCount() {
    return incompleteReqsCount;
  }

  public int getFailedReqsCount() {
    return failedReqsCount;
  }

  public int getTotalReqsCount() {
    return (completeReqsCount + incompleteReqsCount + failedReqsCount);
  }
}
