/*
 * Copyright 2015 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.unitesk.requality.eclipse.views.properties;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.UUID;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.jface.action.MenuManager;
import org.eclipse.jface.viewers.DoubleClickEvent;
import org.eclipse.jface.viewers.IDoubleClickListener;
import org.eclipse.jface.viewers.ISelection;
import org.eclipse.jface.viewers.IStructuredSelection;
import org.eclipse.jface.viewers.ITreeContentProvider;
import org.eclipse.jface.viewers.TreePath;
import org.eclipse.jface.viewers.TreeSelection;
import org.eclipse.jface.viewers.TreeViewer;
import org.eclipse.jface.viewers.Viewer;
import org.eclipse.swt.SWT;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.ui.IEditorReference;
import org.eclipse.ui.IViewPart;
import org.eclipse.ui.IWorkbenchPartSite;
import org.eclipse.ui.PartInitException;
import org.eclipse.ui.PlatformUI;
import org.eclipse.ui.ide.IDE;
import org.eclipse.ui.model.WorkbenchLabelProvider;
import org.eclipse.ui.views.properties.tabbed.TabbedPropertySheetPage;
import com.unitesk.requality.core.ChangesSet;
import com.unitesk.requality.core.DefaultTreeChangeListener;
import com.unitesk.requality.core.ITreeChangeListener;
import com.unitesk.requality.core.Moment;
import com.unitesk.requality.core.TreeNode;
import com.unitesk.requality.core.attribute.Attribute;
import com.unitesk.requality.eclipse.RequalityPlugin;
import com.unitesk.requality.eclipse.editors.NodeEditorInput;
import com.unitesk.requality.eclipse.editors.browser.ReqMarker;
import com.unitesk.requality.eclipse.handlers.location.ILocationTab;
import com.unitesk.requality.eclipse.ui.cnf.RequalitySorter;
import com.unitesk.requality.eclipse.ui.editors.UniEditor;
import com.unitesk.requality.nodetypes.Document;
import com.unitesk.requality.nodetypes.Location;
import com.unitesk.requality.nodetypes.Requirement;

public class DocLocationPropertySection extends TreeNodePropertiesTab implements ILocationTab {

  private TreeViewer locations;
  private Document doc;

  public DocLocationPropertySection() {}

  @Override
  public void createControls(Composite parent, TabbedPropertySheetPage aTabbedPropertySheetPage) {
    super.createControls(parent, aTabbedPropertySheetPage);
    parent.setLayout(new GridLayout());
    parent.setLayoutData(new GridData(GridData.FILL_BOTH));

    locations = new TreeViewer(parent, SWT.BORDER | SWT.H_SCROLL | SWT.V_SCROLL | SWT.MULTI);
    locations.getControl().setLayoutData(new GridData(GridData.FILL_BOTH));
    locations.setLabelProvider(WorkbenchLabelProvider.getDecoratingWorkbenchLabelProvider());
    locations.setComparator(new RequalitySorter());
    locations.setContentProvider(
        new ITreeContentProvider() {
          @Override
          public void inputChanged(Viewer viewer, Object oldInput, Object newInput) {}

          @Override
          public void dispose() {}

          @Override
          public boolean hasChildren(Object element) {
            TreeNode node = (TreeNode) element;
            if (node instanceof Location) {
              return ((Location) node).getRequirements().length > 0;
            }
            return node.sizeWorkChildren() > 0;
          }

          @Override
          public Object getParent(Object element) {
            return ((TreeNode) element).getWorkParent();
          }

          @Override
          public Object[] getElements(Object inputElement) {
            return (Object[]) inputElement;
          }

          @Override
          public Object[] getChildren(Object parentElement) {
            TreeNode node = (TreeNode) parentElement;
            ArrayList<Object> result = new ArrayList<>();
            if (node.getType().equals(Location.TYPE_NAME)) {
              Location l = (Location) node;
              if (doc.equals(l.getWorkParent()))
                for (Requirement r : l.getRequirements(true)) if (r != null) result.add(r);
            }
            return result.toArray();
          }
        });
    locations.addDoubleClickListener(
        new IDoubleClickListener() {

          @Override
          public void doubleClick(DoubleClickEvent event) {
            IStructuredSelection selection = (IStructuredSelection) locations.getSelection();
            NodeEditorInput input;
            if (selection.getFirstElement() instanceof Requirement) {
              Requirement doc = (Requirement) selection.getFirstElement();
              input = new NodeEditorInput(doc);
              try {
                IDE.openEditor(
                    PlatformUI.getWorkbench().getActiveWorkbenchWindow().getActivePage(),
                    input,
                    UniEditor.ID,
                    false);
              } catch (PartInitException e) {
                e.printStackTrace();
              }
              return;
            } else if (selection.getFirstElement() instanceof Location) {
              Location loc = (Location) selection.getFirstElement();
              input = new NodeEditorInput(loc.getWorkParent());
              input.setData(loc);
            } else {
              return;
            }

            try {
              String docId = input.getStorage().getNode().getQualifiedId();
              IEditorReference ref = RequalityPlugin.findMarkEditorForDocument(docId);
              if (ref != null && ref.getEditor(true) != null) {
                ReqMarker ed = (ReqMarker) ref.getEditor(true);
                ed.setTargetLocation((Location) input.getData(), true);
              }
              IDE.openEditor(
                  PlatformUI.getWorkbench().getActiveWorkbenchWindow().getActivePage(),
                  input,
                  ReqMarker.ID,
                  false);
            } catch (PartInitException e) {
              e.printStackTrace();
            } catch (CoreException e) {
              e.printStackTrace();
            }
          }
        });
    IViewPart view =
        PlatformUI.getWorkbench()
            .getActiveWorkbenchWindow()
            .getActivePage()
            .findView("org.eclipse.ui.views.PropertySheet");
    if (view == null) return;
    IWorkbenchPartSite site = view.getSite();
    MenuManager mgr = new MenuManager();
    mgr.setRemoveAllWhenShown(true);
    site.registerContextMenu(mgr, site.getSelectionProvider());
    locations.getControl().setMenu(mgr.createContextMenu(locations.getControl()));
  }

  public Document getDocument() {
    return doc;
  }

  @Override
  public void aboutToBeHidden() {
    if (doc != null) doc.getTreeDB().removeGUIListener(tcl);
  };

  public ITreeChangeListener tcl =
      new DefaultTreeChangeListener() {
        @Override
        public void deleted(final TreeNode nodeDesc, boolean internal, Moment m) {
          if(m == Moment.BEFORE)
            return;
          if (isDisposed) {
            if (doc != null) doc.getTreeDB().removeListener(tcl);
            return;
          }
          boolean update = false;
          if (nodeDesc.getType().equals(Location.TYPE_NAME)
              && doc.getLocations().contains(nodeDesc.getUUId())) update = true;
          if (nodeDesc.getType().equals(Requirement.TYPE_NAME)) {
            List<UUID> ids = nodeDesc.getActiveWorkChildrenUUIDs();
            for (UUID uid : ids) if (doc.getLocations().contains(uid)) update = true;
          }
          if (update)
            PlatformUI.getWorkbench()
                .getDisplay()
                .asyncExec(
                    new Runnable() {
                      public void run() {
                        if (isDisposed) {
                          if (doc != null) doc.getTreeDB().removeListener(tcl);
                          return;
                        }
                        locations.refresh();
                      }
                    });
        }

        @Override
        public void changed(
            TreeNode node, ChangesSet<Attribute> params, boolean internal) {
          if (isDisposed) {
            if (doc != null) doc.getTreeDB().removeListener(tcl);
            return;
          }
          if (node.getType().equals(Location.TYPE_NAME)
              && doc.getLocations().contains(node.getUUId())) updateNode(node);
          if (node.getType().equals(Requirement.TYPE_NAME)) {
            Requirement r = (Requirement) node;
            for (Location l : r.getLocations())
              if (doc.getLocations().contains(l.getUUId())) updateNode(l);
          }
        }

        private void updateNode(final TreeNode node) {
          TreeNode[] toup = null;
          if (node.getType().equals(Location.TYPE_NAME)
              && doc.getLocations().contains(node.getUUId()))
            toup = new TreeNode[]{ node };
          if (node.getType().equals(Requirement.TYPE_NAME)) {
            Requirement r = (Requirement) node;
            List<Location> locs = r.getLocations();
            toup = new TreeNode[locs.size()];
            int i = 0;
            for (Location l : locs)
              if (doc.getLocations().contains(l.getUUId())) toup[i++] = l;
          }
          if(toup!=null && toup.length>0){
            
            TreeNode[] ftoup = toup;
            PlatformUI.getWorkbench()
              .getDisplay()
              .asyncExec(
                  new Runnable() {
                    public void run() {
                      if (isDisposed) {
                        if (doc != null) doc.getTreeDB().removeListener(tcl);
                        return;
                      }
                      for(TreeNode node:ftoup)
                        locations.refresh(node,true);
                    }
                  });
          }
        }
      };

  public void refreshLocations() {
    setTreeNode(doc);
  }

  @Override
  public void dispose() {
    super.dispose();
  }

  @Override
  protected void setTreeNode(TreeNode node) {
    if (tcl != null && doc != null) doc.getTreeDB().removeGUIListener(tcl);
    doc = toDocument(node);
    if (doc == null || locations == null || locations.getControl().isDisposed()) {
      return;
    }
    Set<Location> recs = new HashSet<>();
    for (UUID l : doc.getLocations()) {
      Location loc = doc.getLocation(l.toString());
      if (loc != null && !recs.contains(loc)) recs.add(loc);
    }
    locations.setInput(recs.toArray());
    locations.expandAll();
    if (tcl != null && doc != null) doc.getTreeDB().addGUIListener(tcl);
  }

  protected Document toDocument(TreeNode node) {
    if (node instanceof Document) {
      return (Document) node;
    }

    if (node instanceof Location) {
      Location location = (Location) node;
      Document[] docs = location.getDocuments();
      if (docs.length > 0) {
        return docs[0];
      }
    }
    return null;
  }

  @Override
  public Map<Location, List<Requirement>> getLocationsAndRequirements() {
    Map<Location, List<Requirement>> map = new HashMap<>();
    ISelection sel = locations.getSelection();
    if (!locations.getControl().isFocusControl()
        || doc == null
        || !(sel instanceof IStructuredSelection)) {
      return null;
    }
    TreeSelection tsel = ((TreeSelection) sel);
    for (TreePath tp : tsel.getPaths()) {
      if ((tp.getFirstSegment() instanceof Location)) {
        if ((tp.getLastSegment() instanceof Requirement)) {
          Location l = (Location) tp.getFirstSegment();
          Requirement r = (Requirement) tp.getLastSegment();
          if (!map.containsKey(l)) map.put(l, new ArrayList<Requirement>());
          if (!map.get(l).contains(r)) map.get(l).add(r);
        } else {
          Location l = (Location) tp.getFirstSegment();
          if (!map.containsKey(l)) map.put(l, new ArrayList<Requirement>());
          Requirement[] reqs = l.getRequirements();
          for (Requirement r : reqs) if (!map.get(l).contains(r)) map.get(l).add(r);
        }
      }
    }
    return map.size() > 0 ? map : null;
  }

  @Override
  protected boolean affectedByAttribute(String a) {
    return TreeNode.ATTR_ID.equals(a)
        || TreeNode.ATTR_NAME.equals(a)
        || Requirement.ATTR_LOCATIONS.equals(a);
  }
  private static Set<String> ids = new HashSet<>();
  static{
    ids.add(TreeNode.ATTR_ID);
    ids.add(TreeNode.ATTR_NAME);
    ids.add(Requirement.ATTR_LOCATIONS);
  }
  @Override
  public boolean requireUpdate(TreeNode source, ChangesSet<Attribute> what) {
    for(String s:what.getAll())
      if(ids.contains(s))//TODO iterator check
        return true;
    return false;
  };
}
