/*
 * Copyright 2015 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.unitesk.requality.eclipse.views.properties;

import java.util.Collections;
import org.eclipse.core.commands.operations.IOperationHistoryListener;
import org.eclipse.core.commands.operations.OperationHistoryEvent;
import org.eclipse.core.resources.IProject;
import org.eclipse.jface.viewers.ISelection;
import org.eclipse.jface.viewers.IStructuredSelection;
import org.eclipse.swt.events.FocusAdapter;
import org.eclipse.swt.events.FocusEvent;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Text;
import org.eclipse.ui.IWorkbenchPart;
import org.eclipse.ui.PlatformUI;
import org.eclipse.ui.views.properties.tabbed.AbstractPropertySection;
import org.eclipse.ui.views.properties.tabbed.TabbedPropertySheetPage;
import com.unitesk.requality.core.ChangesSet;
import com.unitesk.requality.core.CoreUtils;
import com.unitesk.requality.core.INodeChangeListener;
import com.unitesk.requality.core.ITreeNodeProvider;
import com.unitesk.requality.core.Moment;
import com.unitesk.requality.core.TreeDB;
import com.unitesk.requality.core.TreeNode;
import com.unitesk.requality.core.attribute.Attribute;
import com.unitesk.requality.core.attribute.AttributeStorage;
import com.unitesk.requality.eclipse.core.EclipseTracker;
import com.unitesk.requality.eclipse.core.ITreeStateListener;
import com.unitesk.requality.nodetypes.Location;
import com.unitesk.requality.nodetypes.Requirement;

public abstract class TreeNodePropertiesTab extends AbstractPropertySection {
  protected TreeNode node;
  boolean isDisposed = true;
  ElementDescriptionPanel descr;
  
  INodeChangeListener nodeChangeListener = new PropertyTabNodeChangeListener();
  private DescriptionFocusAdapter standardDescAdapter;

  protected FocusAdapter getDescriptionFocusAdapter(ElementDescriptionPanel descr, Text adesc, TreeNode node) {
    if (standardDescAdapter == null) standardDescAdapter = new DescriptionFocusAdapter();
    standardDescAdapter.setViewer(adesc, toRequirementOrNode(node));
    this.descr = descr;
    return standardDescAdapter;
  }

  private class DescriptionFocusAdapter extends FocusAdapter {
    private Text adesc;
    private TreeNode node;

    public void setViewer(Text adesc2, TreeNode node2) {
      adesc = adesc2;
      node = node2;
    }

    public void focusLost(FocusEvent e) {
      super.focusLost(e);
      if (updating) return;
      //if new attribute text is "", then remove this attribute rather than set it to "",
      //because otherwise we get modified .json-files with empty attribute added,
      //when simply navigating nodes in Requality Explorer and Properties View,
      //and this is affecting .svn
      if (node == null) {
        return;
      }
      String oldDescr = node.getRawDescription();
      if (oldDescr == null) {
        oldDescr = "";
      }
      String newDescr = "";
      if (adesc.getText() == null) return;
      newDescr = adesc.getText();
      newDescr = CoreUtils.htmlEncapsulate(newDescr);
      if (!descr.isDirty()||oldDescr.equals(newDescr)) {
        return;
      }
      if (node.getTreeDB().startTransaction("Update " + node.getTypeLabel(), true)) {
        node.setDescription(newDescr);
        node.commitAttributes();
        node.getTreeDB().commit();
      } else adesc.setText(node.getDescription());
      //TODO else cancel
    }

    private boolean isDirty() {
      return descr.isDirty();
    }

    public void setTreeNode(TreeNode node2) {
      node = node2;
    }
  }

  private class PropertyTabNodeChangeListener implements INodeChangeListener {

    @Override
    public void attributeChange(TreeNode source, final String what,Attribute old) {
      
      attributesChange(
          source,
          new ChangesSet<>(Collections.emptySet(),AttributeStorage.formSet(what),Collections.emptySet(),
              AttributeStorage.formAttributeMap(old),
              AttributeStorage.formAttributeMap(node.attrs().getAttribute(what)),AttributeStorage.formList(what)));
    }

    @Override
    public void deleted(TreeNode source) {
      Display d = Display.getCurrent();
      if (d == null) {
        d = Display.getDefault();
      }
      updating = true;
      if (node == null || (source.getQualifiedId(true) + "/").contains(node.getQualifiedId(true) + "/")) 
        d.asyncExec(
          new Runnable() {
            @Override
            public void run() {
              {
                node = null;
                if (page.getControl().isDisposed()) {
                  return;
                }
                setTreeNode(null);
              }
            }
          });
    }

    @Override
    public void moved(TreeNode source, TreeNode oldParent, boolean internal) {
      // TODO Auto-generated method stub

    }

    @Override
    public void attributesChange(TreeNode source, ChangesSet<Attribute> what) {
      if (node == null) setTreeNodeSafe(source);
      else if (TreeNodePropertiesTab.this.requireUpdate(source,what)) {
        boolean af = false;
        for (String s : what.getAll()) if (affectedByAttribute(s)) af = true;
        if (af) setTreeNodeSafe(source);
      }
    }

    @Override
    public void fireNodeResourceChange(TreeNode treeNode, String resName, Moment m, boolean inner) {
      // TODO Auto-generated method stub

    }
  }

  ITreeStateListener treeStateListener =
      new ITreeStateListener() {
        @Override
        public void open(TreeDB db) {
          setTreeNodeSafe(node);
        }

        @Override
        public void close(TreeDB db) {
          setTreeNodeSafe(node);
        }

        @Override
        public void delete(IProject proj) {}
      };
  private TabbedPropertySheetPage page;

  @Override
  public void createControls(Composite parent, TabbedPropertySheetPage aTabbedPropertySheetPage) {
    page = aTabbedPropertySheetPage;
    super.createControls(parent, aTabbedPropertySheetPage);
    isDisposed = false;
  }

  public abstract boolean requireUpdate(TreeNode source, ChangesSet<Attribute> what);
  
  @Override
  public boolean shouldUseExtraSpace() {
    return true;
  }

  public TreeNodePropertiesTab() {
    super();
    EclipseTracker.getTracker().addListener(treeStateListener);
  }

  @Override
  public void dispose() {
    isDisposed = true;
    if (node != null) node.removeGuiChangeListener(nodeChangeListener);
    node = null;
    EclipseTracker.getTracker().removeListener(treeStateListener);
    super.dispose();
  }

  protected boolean updating = false;

  protected boolean checkNodeChanges(final TreeNode newNode){
    if(newNode!=null&&node==null || node!=newNode)
      return true;
    return false;
  }
  private void setTreeNodeSafe(final TreeNode newNode) {
    if (node != null) node.removeGuiChangeListener(nodeChangeListener);
    Display d = Display.getCurrent();
    if (d == null) {
      d = Display.getDefault();
    }
    updating = true;
    if(newNode != null && !newNode.isDisposed())
      d.asyncExec(
        new Runnable() {
          @Override
          public void run() {
            //	TreeNodePropertiesTab.this.node = newNode;
            if (!isDisposed) {
              TreeNodePropertiesTab.this.node = newNode;
              if (node == null) {
                updating = false;
                return;
              }
              if (standardDescAdapter != null)
                standardDescAdapter.setTreeNode(toRequirementOrNode(newNode));
              if (isDisposed) {
                updating = false;
                return;
              }
              setTreeNode(newNode);
              updating = false;
            }
          }
        });
    if (newNode != null) newNode.addGuiChangeListener(nodeChangeListener);
  }

  protected abstract void setTreeNode(TreeNode node);

  @Override
  public void setInput(IWorkbenchPart part, ISelection selection) {
    super.setInput(part, selection);
    TreeNode node = TreeNodePropertiesTab.toTreeNode(selection);
    if (node != null) {

      PlatformUI.getWorkbench().getOperationSupport().getOperationHistory().addOperationHistoryListener(new IOperationHistoryListener(){
      
        @Override
        public void historyNotification(OperationHistoryEvent event) {
          refresh();
        }
      });
      setTreeNodeSafe(node);
    }
  }

  protected static TreeNode toRequirementOrNode(TreeNode newNode) {
    TreeNode node = toRequirement(newNode);
    if (node != null) return node;
    return newNode;
  }

  protected static Requirement toRequirement(TreeNode node) {
    if (node instanceof Requirement) {
      return (Requirement) node;
    }

    if (node instanceof Location) {
      Location location = (Location) node;
      Requirement[] reqs = location.getRequirements(true);
      if (reqs == null) return null;
      if (reqs.length > 0) {
        return reqs[0];
      }
    }
    return null;
  }

  public static TreeNode toTreeNode(ISelection sel) {
    if (sel == null) return null;
    Object o = ((IStructuredSelection) sel).getFirstElement();
    if (o instanceof ITreeNodeProvider) return ((ITreeNodeProvider)o).getNode();
    if (o instanceof TreeNode) return (TreeNode) o;
    if (o instanceof IProject) {
      TreeDB db = EclipseTracker.getTracker().getTree((IProject) o);
      if (db != null) return db.getRootNode();
    }
    return null;
  }

  protected boolean affectedByAttribute(String a) {
    return false;
  }
}
